package org.artifactory.federation;

import org.apache.commons.lang3.NotImplementedException;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.descriptor.authentication.AuthenticationToken;
import org.artifactory.descriptor.repo.FederatedMember;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.federation.status.ActionStatus;
import org.artifactory.repo.RepoDetails;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.spring.EntitlementsEventListener;

import javax.annotation.Nonnull;
import java.util.Collection;
import java.util.List;
import java.util.Optional;
import java.util.Set;

/**
 * @author Alexei Vainshtein
 */
public interface FederatedRepoService extends EntitlementsEventListener, ContextCreationListener {

    /**
     * Sync the config to members upon the action required.
     */
    default void syncConfigToMembers(ActionStatus action,
            FederatedRepoDescriptor newDescriptor, FederatedRepoDescriptor oldDescriptor) {

    }
    /**
     * Fetch from each remote url the existed federated repos.
     *
     * @param urls - The remote urls to check
     * @return a list that contains a json with url and federation repos of that url.
     */
    List<FederationReposResponse> getReposFromRemote(List<String> urls, String proxyName);

    /**
     * @return a list of all federated repositories and the info per repo
     */
    List<RepoDetails> getRepos();

    /**
     * @return the token that associated with remote url and the key provided
     */
    String getToken(String remoteUrl, String key);

    /**
     * Validates the given repository
     *
     * @param key                      - repo name
     * @param clientBaseUrl            - base url of the requesting server
     * @param timeFromRequestingServer - the current time of the server that is requesting the validation.
     */
    void validate(String key, String clientBaseUrl, long timeFromRequestingServer, String repoType, String sourceRepo);

    /**
     * Checks if the license provided and subscription (if AOL) are enterprise and above.
     * {@link ForbiddenException} is thrown if the license/subscription are not enterprise and above.
     */
    void validateFederationLicenseInstalled();

    /**
     * Validates remote url
     *
     * @param fullUrl - The remote url including the repository to validate
     */
    void validateRemoteServer(String fullUrl, String proxyName, RepoType repoType, String sourceRepoKey);

    /**
     * Migrates local repository to a federated repository
     *
     * @param repoKey - The repository name that need to be migrated.
     */
    default void migrate(String repoKey) {

    }

    default void triggerFullSync(String repoKey, String mirrorUrl) {

    }

    default void configSync(String repoKey) {

    }

    /**
     * Returns whether federated repo is ready to use
     */
    default boolean isActive(String repoKey) {
        return false;
    }

    /**
     * Returns a json that include all the metadata from all the package types provider + the basic config
     * @param types - All the package types that metadata need to be calculated for
     * @return a json with all the metadata info
     */
    default String getMetadata(Collection<String> types) {
        return StringUtils.EMPTY;
    }

    default void invalidateMetadata(String packageType, String memberUrl, String proxyKey) {}

    void addInconsistentRepoMember(String memberUrl, String repoKey);

    Set<String> getInconsistentRepos();

    void removeInconsistentRepoMember(String memberUrl);

    /**
     * Remove an inconsistent repo from the list by repo key.
     * @param repoKey - The repo key to remove from the list
     */
    default void removeInconsistentRepoByRepoKey(String repoKey) {

    }
    void removeRepoFromRemoteMembers(FederatedRepoDescriptor federatedDescriptor);

    void saveAuthenticationMasterToken(String tokenId, String tokenType, String scope,
            String baseUrl, String exchangeUrl, String accessToken, Long expiresIn);

    Collection<AuthenticationToken> getPairedAuthenticationTokens();

    Optional<AuthenticationToken> getPairedAuthenticationToken(String baseUrl);

    void deleteAuthenticationToken(String baseUrl);

    FullSyncStatus getFullSyncStatus(@Nonnull String mirrorRemoteUrl);

    void putFullSyncStatus(@Nonnull String mirrorRemoteUrl, @Nonnull FullSyncStatus syncStatus);

    void invalidateFullSyncStatus(@Nonnull String mirrorRemoteUrl);

    /**
     * checks if the build info repo is federated.
     *
     * @return return true is the build info repo is federated otherwise returns false
     */
    boolean isProjectBuildInfoRepoFederated(String projectKey);

    /**
     * Validate there aren't any member that is disabled
     * @param newRepoDescriptor - The descriptor to get the members from.
     * If a disabled member is found, a {@link FederatedRemoteMembersUpdateException thrown}
     */
    default void validateNoDisabledFederationMember(FederatedRepoDescriptor newRepoDescriptor) {
    }

    /**
     * Checks if status has changed of the local member
     * @param repoKey - The repo key of the local member
     * @param newMembers - The new members
     * @param oldMembers - The old members
     * @return true if changed, false otherwise
     */
    default boolean isStatusChanged(String repoKey, List<FederatedMember> newMembers, Set<FederatedMember> oldMembers) {
        return false;
    }

    default Collection<String> findConflicts(String repoKey) {
        throw new NotImplementedException("Method not implemented");
    }

    default void deRegister(String repoKey, String remoteUrl) {}

    default void setDisabledRepoTypes(String repoTypes) {}

    default boolean isNegotiationAllowedForRepoType(String repoType) {return true;}

    default void enableRepos(List<String> repoTypes){}

    /**
     * Put metadata that came from the origin url in the local instance
     *
     * @param metadata      the newer metadata
     * @param originUrl     the origin member that sends the metadata
     * @param updateMembers a boolean that represents if the other members of repo type need to be updated as well
     */
    default void putMirrorMetadata(String metadata, String originUrl, boolean updateMembers){}
}
