package org.artifactory.hqc.aop;

import com.google.common.annotations.VisibleForTesting;
import lombok.extern.slf4j.Slf4j;
import org.aopalliance.intercept.MethodInterceptor;
import org.aopalliance.intercept.MethodInvocation;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.hqc.annotation.HeavyQueryCache;
import org.artifactory.hqc.config.HeavyQueryCacheConfig;
import org.artifactory.hqc.enums.HeavyQueryCacheType;
import org.artifactory.hqc.service.HeavyQueryCacheService;
import org.springframework.stereotype.Component;

import java.util.Arrays;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Callable;
import java.util.stream.Collectors;

/**
 * @author Andrey Tashlitsky
 */
@Slf4j
@Component
public class HeavyQueryCacheAdvice implements MethodInterceptor {

    private static final int NAME_MAX_LENGTH = 255;
    private final Map<String, Integer> properties = new HashMap<>();

    @Override
    public Object invoke(MethodInvocation methodInvocation) throws Throwable {
        log.debug("Inside HeavyQueryCacheAdvice, invoking: {}", methodInvocation.getMethod());
        HeavyQueryCache annotation = methodInvocation.getMethod().getAnnotation(HeavyQueryCache.class);
        String name = getName(methodInvocation, annotation.hqcType());
        log.debug("The calculated name is: '{}'", name);
        return ContextHelper.get().beanForType(HeavyQueryCacheService.class)
                .get(
                        getCallable(methodInvocation),
                        methodInvocation.getMethod().getReturnType(),
                        name,
                        annotation.hqcType(),
                        paramsToConfig(annotation)
                );
    }

    /**
     * Calculate a string that later will be used as part of the cache key. The 'name' string will contain all the
     * wrapped method argument toString. If the string is too long then calculates the hashcode of the arguments list.
     *
     * @param methodInvocation Wrapped method.
     * @param hqcType          Query type.
     * @return A unique string of this method invocation.
     */
    @VisibleForTesting
    String getName(MethodInvocation methodInvocation, HeavyQueryCacheType hqcType) {
        String name = Arrays.stream(methodInvocation.getArguments())
                .map(Object::toString)
                .collect(Collectors.joining("_"));
        if (name.length() > 0 && hqcType.toString().length() + name.length() + 1 > NAME_MAX_LENGTH) {
            log.debug("Name is too long: {}, allowed length is {}", name.length(), NAME_MAX_LENGTH);
            name = String.valueOf(Arrays.hashCode(methodInvocation.getArguments()));
        }
        return name;
    }

    /**
     * Wraps the method invocation in a callable.
     */
    @VisibleForTesting
    <T> Callable<T> getCallable(MethodInvocation methodInvocation) {
        return () -> {
            try {
                return (T) methodInvocation.proceed();
            } catch (Throwable e) {
                throw new RuntimeException(e);
            }
        };
    }

    /**
     * Converts {@link HeavyQueryCache} values to {@link HeavyQueryCacheConfig} object.
     */
    @VisibleForTesting
    HeavyQueryCacheConfig paramsToConfig(HeavyQueryCache annotation) {
        String property = "artifactory." + annotation.hqcType().toString().replace('_', '.');
        String minCallableDurationMillisProperty = property + ".min.callable.duration.millis";
        String maxBlobSizeKbProperty = property + ".max.blob.size.kb";
        String factorProperty = property + ".expiry.factor";
        String upperBoundProperty = property + ".expiry.upper.bound";
        String lowerBoundProperty = property + ".expiry.lower.bound";

        HeavyQueryCacheConfig.HeavyQueryCacheConfigBuilder builder = HeavyQueryCacheConfig.builder();
        if (annotation.minCallableDurationMillis() != -1) {
            int minCallableDurationMillis =
                    getProperty(minCallableDurationMillisProperty, annotation.minCallableDurationMillis());
            builder.minCallableDurationMillis(minCallableDurationMillis);
        }
        if (annotation.maxBlobSizeKb() != -1) {
            int maxBlobSizeKb = getProperty(maxBlobSizeKbProperty, annotation.maxBlobSizeKb());
            builder.maxBlobSizeKb(maxBlobSizeKb);
        }
        if (annotation.expiryFactor() != -1) {
            int expiryFactor = getProperty(factorProperty, annotation.expiryFactor());
            builder.expiryFactor(expiryFactor);
        }
        if (annotation.expiryUpperBound() != -1) {
            int expiryUpperBound = getProperty(upperBoundProperty, annotation.expiryUpperBound());
            builder.expiryUpperBound(expiryUpperBound);
        }
        if (annotation.expiryLowerBound() != -1) {
            int expiryLowerBound = getProperty(lowerBoundProperty, annotation.expiryLowerBound());
            builder.expiryLowerBound(expiryLowerBound);
        }
        HeavyQueryCacheConfig hqcConfig = builder.build();
        log.debug("Converted annotation {} to hqcConfig {}", annotation, hqcConfig);
        return hqcConfig;
    }

    /**
     * Reads a value from a given property.
     *
     * @param property     Given property name.
     * @param defaultValue Default value to return in case the property is empty.
     * @return Value from property if present, otherwise returns the default value.
     */
    private int getProperty(String property, int defaultValue) {
        return properties.computeIfAbsent(property, k -> {
            if (System.getProperty(property) != null) {
                int value = Integer.parseInt(System.getProperty(property));
                log.trace("Found '{}={}', overriding value", property, value);
                return value;
            }
            return defaultValue;
        });
    }
}
