package org.artifactory.hqc.service;

import lombok.NonNull;
import org.artifactory.common.ConstantValues;
import org.artifactory.hqc.config.HeavyQueryCacheConfig;
import org.artifactory.hqc.enums.HeavyQueryCacheType;

import javax.annotation.Nullable;
import java.util.concurrent.Callable;

/**
 * @author Tomer Slutsky
 * @author Andrey Tashlitsky
 */
public interface HeavyQueryCacheService {

    /**
     * Tries to return an entry from the cache table, and uses the callback if it fails.
     * An entry wouldn't be used even if it exists and the expiry_time of the entry passes.
     * If no cached entry was returned, adds it to the cache table if it is allowed by the following conditions:
     * Slower than {@link ConstantValues#hqcMinCallableDurationMillis} and smaller than
     * {@link ConstantValues#hqcMaxBlobSizeKb} in size.
     *
     * @param callback  The supplied callback to happen if the entry doesn't exist in the cache.
     * @param clazz     The class which will be used to deserialize the BLOB.
     * @param name      A name to use as part of the entry key.
     * @param hqcType   Entry type, name of the feature that is wrapped by HQC mechanism.
     * @param hqcConfig The configuration object containing values for caching.
     * @return Cached query response if exists. Otherwise, calls the callable and returns its result. null can be
     * returned in case the callable returned null or if {@link InterruptedException} was caught.
     */
    @Nullable
    <T> T get(@NonNull Callable<T> callback, @NonNull Class<T> clazz, @NonNull String name,
            @NonNull HeavyQueryCacheType hqcType, @NonNull HeavyQueryCacheConfig hqcConfig);

    /**
     * Runs cleanup of expired entries in the heavy_query_cache DB table.
     *
     * @return The amount of deleted entries.
     */
    int runCleanup();
}
