package org.artifactory.ph;

import com.google.common.collect.Maps;
import com.google.common.collect.SetMultimap;
import lombok.NonNull;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.JerseyArtifactoryResponse;
import org.artifactory.api.request.UploadService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageNotFoundException;
import org.artifactory.core.api.ph.exception.PackageRuntimeException;
import org.artifactory.core.api.ph.service.PackageUploadService;
import org.artifactory.core.api.ph.virtual.resolution.PackageUploadContext;
import org.artifactory.factory.InfoFactoryHolder;
import org.artifactory.fs.MutableFileInfo;
import org.artifactory.md.Properties;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.SaveResourceContext;
import org.artifactory.repo.StoringRepo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.service.deploy.ArtifactoryDeployRequest;
import org.artifactory.repo.service.deploy.ArtifactoryDeployRequestBuilder;
import org.artifactory.resource.FileResource;
import org.artifactory.rest.exception.RestException;

import javax.annotation.Nonnull;
import javax.ws.rs.core.HttpHeaders;
import javax.ws.rs.core.Response;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Map;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_INTERNAL_SERVER_ERROR;
import static org.artifactory.api.request.ArtifactoryHeaders.CHECKSUM_MD5;
import static org.artifactory.api.request.ArtifactoryHeaders.CHECKSUM_SHA1;
import static org.artifactory.api.request.ArtifactoryHeaders.CHECKSUM_SHA256;

@Slf4j
public class PackageUploadServiceImpl implements PackageUploadService {

    private final InternalRepositoryService internalRepositoryService;
    private final UploadService uploadService;


    public PackageUploadServiceImpl(InternalRepositoryService internalRepositoryService,
            UploadService uploadService) {
        this.internalRepositoryService = internalRepositoryService;
        this.uploadService = uploadService;
    }

    @Override
    public void upload(String repoKey, String path, InputStream inputStream) {
        LocalRepo<?, ?> localRepo = internalRepositoryService.localFederatedOrCachedRepositoryByKey(repoKey);
        if (localRepo != null && localRepo.isCache()) {
            try {
                uploadToVirtualOrRemoteCache(repoKey, path, inputStream);
                return;
            } catch (PackageException e) {
                throw new PackageRuntimeException(e.getStatus(), e.getPublicReason(), e.getInternalReason(), e);
            }
        }
        RepoPath repoPath = RepoPathFactory.create(repoKey, path);
        internalRepositoryService.deploy(repoPath, inputStream);
    }

    @Override
    public BaseResource uploadToVirtualOrRemoteCache(String repoKey, String path, InputStream inputStream)
            throws PackageException {
        StoringRepo<?, ?> storingRepo = internalRepositoryService.storingRepositoryByKey(repoKey);
        if (storingRepo == null) {
            throw new PackageNotFoundException("repository not found");
        }
        if (!storingRepo.isCache() && storingRepo.isReal()) {
            throw new PackageException(SC_BAD_REQUEST, "Bad request",
                    "Bad request: internal uploads allowed only to remote-cache and virtual repos");
        }
        try {
            MutableFileInfo fileInfo = InfoFactoryHolder.get().createFileInfo(RepoPathFactory.create(repoKey, path));
            fileInfo.createTrustedChecksums();
            FileResource repoResource = new FileResource(fileInfo);
            SaveResourceContext saveContext = new SaveResourceContext.Builder(repoResource, inputStream).build();
            return internalRepositoryService.saveResource(storingRepo, saveContext);
        } catch (Exception e) {
            throw new PackageException(SC_INTERNAL_SERVER_ERROR, "internal server error", e.getMessage(), e);
        }
    }

    @Override
    public Response upload(PackageUploadContext context) throws PackageException {
        JerseyArtifactoryResponse res = new JerseyArtifactoryResponse();
        ArtifactoryDeployRequest req = buildArtifactoryRequest(context);
        try {
            uploadService.upload(req, res);
            return res.build();
        } catch (RestException e) {
            throw new PackageException(e.getStatusCode(), e.getMessage(), e.getMessage());
        } catch (IOException | RepoRejectException e) {
            log.error(e.getMessage(), e);
            throw new PackageException(SC_BAD_REQUEST, e.getMessage(), e.getMessage());
        }
    }

    @NonNull
    @Override
    public File getTempUploadDir() {
        return ArtifactoryHome.get().getTempUploadDir();
    }

    @Nonnull
    private ArtifactoryDeployRequest buildArtifactoryRequest(PackageUploadContext context) {
        Properties properties = parseUploadProperties(context);
        ArtifactoryDeployRequest req =
                new ArtifactoryDeployRequestBuilder(RepoPathFactory.create(context.getRepoKey(), context.getPath()))
                        .inputStream(context.getContent())
                        .properties(properties)
                        .contentLength(contentLength(context))
                        .build();

        Map<String, String> headers = Maps.newHashMap();
        headers.putAll(context.getRequestHeaders());
        if (StringUtils.isNotBlank(context.getSha1())) {
            headers.put(CHECKSUM_SHA1, context.getSha1());
        }
        if (StringUtils.isNotBlank(context.getSha256())) {
            headers.put(CHECKSUM_SHA256, context.getSha256());
        }
        if (StringUtils.isNotBlank(context.getMd5())) {
            headers.put(CHECKSUM_MD5, context.getMd5());
        }
        req.addHeaders(headers);
        return req;
    }

    private long contentLength(PackageUploadContext context) {
        if (context.getContentLength() > 0) {
            return context.getContentLength();
        }
        Map<String, String> headers = context.getRequestHeaders();
        String headerValue = headers.entrySet()
                .stream()
                .filter(entry -> HttpHeaders.CONTENT_LENGTH.equalsIgnoreCase(entry.getKey()))
                .findFirst().map(Map.Entry::getValue)
                .orElse("-1");
        try {
            return Long.parseLong(headerValue);
        } catch (NumberFormatException e) {
            log.warn("Content-Length header value is not a number: '{}', path: {}", headerValue, context.getPath());
        }
        return -1; //unspecified
    }


    private Properties parseUploadProperties(PackageUploadContext context) {
        SetMultimap<String, String> contextAttributes = context.getAttributes();
        Properties properties = (Properties) InfoFactoryHolder.get().createProperties();
        contextAttributes.forEach(properties::put);
        return properties;
    }


}
