package org.artifactory.repo.cache;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.base.Joiner;
import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.ImmutableCollection;
import com.google.common.collect.ImmutableList;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.concurrent.ArtifactoryExecutor;
import org.artifactory.concurrent.ArtifactoryExecutorFactory;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.RepoConfigCopyFactory;
import org.artifactory.model.RepoConfigPropagationModel;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoConfigReload;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ThreadLocalRandom;
import java.util.concurrent.TimeUnit;
import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collectors;

/**
 * Repository config cache. Based on the {@link LoadingCache}.
 * Caches all repo configs from the app start and provides a way to update the cache depending on repo
 * create/update/delete operations.
 *
 * @author vovam
 */
@Slf4j
public class RepoConfigCache {
    private final Supplier<Map<String, Long>> getAllRepoKeysFunction;
    private final LoadingCache<String, RepoConfig> cache;
    private final ArtifactoryExecutor artifactoryExecutor;
    private final Supplier<Map<String, RepoConfig>> getAllRepoConfigsFunction;
    private final Function<String, RepoConfig> getRepoConfigByKeyFunction;
    private final Function<Set<String>, Map<String, RepoConfig>> getRepoConfigsByKeysFunction;
    private final BiFunction<RepoConfigPropagationModel, RepoConfigCache, Void> reloadRepoConfigurationsFunction;
    private volatile ConcurrentHashMap<String, Long> allRepoKeysAndModificationDates;
    private final String identifier = RandomStringUtils.randomAlphabetic(4);

    public RepoConfigCache(long syncUpPeriodMillis, long repoConfigExpirationMillis, int batchOperationSize,
            double optimizationRatio, Supplier<Map<String, RepoConfig>> getAllRepoConfigsFunction,
            Function<String, RepoConfig> getRepoConfigByKeyFunction,
            Function<Set<String>, Map<String, RepoConfig>> getRepoConfigsByKeysFunction,
            BiFunction<RepoConfigPropagationModel, RepoConfigCache, Void> reloadRepoConfigurationsFunction,
            Supplier<Map<String, Long>> getAllRepoKeysFunction
    ) {

        if (syncUpPeriodMillis <= 0) {
            throw new IllegalArgumentException("Repo config cache: sync repo keys period can't be negative or zero");
        }
        this.getAllRepoKeysFunction = getAllRepoKeysFunction;
        this.getAllRepoConfigsFunction = getAllRepoConfigsFunction;
        this.getRepoConfigByKeyFunction = getRepoConfigByKeyFunction;
        this.getRepoConfigsByKeysFunction = getRepoConfigsByKeysFunction;
        this.reloadRepoConfigurationsFunction = reloadRepoConfigurationsFunction;

        log.debug("Initializing repo config cache instance {}. Getting all repo configs from DB...", identifier);
        Map<String, RepoConfig> allRepoConfigs = getAllRepoConfigsFunction.get();
        log.debug("Finishing getting all repo configs");

        this.allRepoKeysAndModificationDates = new ConcurrentHashMap<>(allRepoConfigs.size());
        allRepoConfigs.values().forEach(r -> this.allRepoKeysAndModificationDates.put(r.getKey(), r.getModification()));

        CacheLoader<String, RepoConfig> cacheLoader =
                new RepoConfigCacheLoader(batchOperationSize, optimizationRatio, this::countAll,
                        getAllRepoConfigsFunction, getRepoConfigByKeyFunction, getRepoConfigsByKeysFunction);
        this.cache = CacheBuilder.newBuilder()
                .initialCapacity(allRepoConfigs.size())
                .expireAfterWrite(repoConfigExpirationMillis, TimeUnit.MILLISECONDS)
                .build(cacheLoader);
        this.cache.putAll(allRepoConfigs);

        ScheduledExecutorService scheduledExecutorService =
                Executors.newSingleThreadScheduledExecutor(new BasicThreadFactory.Builder()
                        .namingPattern("repo-conf-cache-" + identifier)
                        .daemon(true)
                        .uncaughtExceptionHandler(
                                (t, e) -> log.error("Can't sync repo config keys while updating repo config cache", e))
                        .build());
        this.artifactoryExecutor =
                ArtifactoryExecutorFactory.newArtifactoryExecutor(scheduledExecutorService, ContextHelper.get(), false);

        long randomizedSyncUpPeriodMillis = ThreadLocalRandom.current().nextLong(0, syncUpPeriodMillis);

        ((ScheduledExecutorService) this.artifactoryExecutor.getExecutor()).scheduleAtFixedRate(
                this::syncRepoConfigKeys, randomizedSyncUpPeriodMillis, syncUpPeriodMillis,
                TimeUnit.MILLISECONDS);
        log.debug("Finishing repo config cache initialization");
    }

    /**
     * Check is repo config present by key.
     */
    public boolean existByKey(String repoKey) {
        log.trace("Checking does repo config by key: {} exist in repo cache...", repoKey);
        boolean exist = allRepoKeysAndModificationDates.keySet().contains(repoKey);
        log.trace("Finishing checking repo config by key: {} in repo cache. Exists: {}", repoKey, exist);
        return exist;
    }

    /**
     * Return a cached repo config by key.
     * NOTE: If loading cache can't load a repo config by the given key because of not-synced state -
     * as a fallback action this repo config will be loaded from DB directly and will not be cached.
     *
     * @throws IllegalStateException - error occurs if loading cache can't load data because of an internal error
     */
    public RepoConfig getByKey(String repoKey) {
        try {
            log.trace("Getting repo config by key: {} from repo cache...", repoKey);
            // to avoid repeated requests to a DB check the repo keys set first
            if (allRepoKeysAndModificationDates.containsKey(repoKey)) {
                log.trace("Repo config with key: {} is present in cache", repoKey);
                return RepoConfigCopyFactory.copy(cache.get(repoKey));
            }
            return null;
        } catch (ExecutionException e) {
            throw new IllegalStateException("Can not get repo by key from cache. Repo key: " + repoKey, e);
        } catch (CacheLoader.InvalidCacheLoadException e) {
            log.debug("Can't get repo by key from loading cache. Executing a fallback query to DB...");
            return getRepoConfigByKeyFunction.apply(repoKey);
        } finally {
            log.trace("Finishing getting repo config by key: {} from repo cache", repoKey);
        }
    }

    /**
     * Return cached repo configs by set of keys.
     * NOTE: If loading cache can't load repo configs by given keys because of not-synced state -
     * as a fallback action these repo configs will be loaded from DB directly and will not be cached.
     *
     * @throws IllegalStateException - error occurs if loading cache can't load data because of an internal error
     */
    public ImmutableCollection<RepoConfig> getRepositoryConfigsByKeys(Set<String> repoKeys) {
        try {
            if (log.isTraceEnabled()) {
                log.trace("Getting repo configs by keys: [{}] from repo cache...", String.join(",", repoKeys));
            }
            return cache.getAll(repoKeys).values().stream().map(RepoConfigCopyFactory::copy)
                    .collect(ImmutableList.toImmutableList());
        } catch (CacheLoader.InvalidCacheLoadException e) {
            log.debug("Executing a fallback query to DB...");
            Collection<RepoConfig> repoConfigs = getRepoConfigsByKeysFunction.apply(repoKeys).values();
            return new ImmutableList.Builder<RepoConfig>().addAll(repoConfigs).build();
        } catch (ExecutionException e) {
            String errorMessage = "Can't load repo config by keys: [" + Joiner.on(", ").join(repoKeys) + "]";
            log.error(errorMessage);
            throw new IllegalStateException(errorMessage, e);
        } finally {
            if (log.isTraceEnabled()) {
                log.trace("Finishing getting repo configs by keys: [{}] from repo cache",
                        Joiner.on(",").join(repoKeys));
            }
        }
    }

    /**
     * Return all cached data
     * If loading cache can't load repo configs by keys that are present in cached keys set because of not-synced
     * state - as a fallback action all repo configs will be loaded from DB directly and will not be cached.
     *
     * @throws IllegalStateException - error occurs if loading cache can't load data because of an internal error
     */
    public Collection<RepoConfig> getAll() {
        try {
            log.trace("Getting all repo config from repo cache...");
            return cache.getAll(allRepoKeysAndModificationDates.keySet()).values().stream()
                    .map(RepoConfigCopyFactory::copy)
                    .collect(Collectors.toList());
        } catch (ExecutionException e) {
            throw new IllegalStateException("Can not get all repo configs from cache", e);
        } catch (CacheLoader.InvalidCacheLoadException e) {
            log.debug("Executing a fallback query to DB...");
            return getAllRepoConfigsFunction.get().values();
        } finally {
            log.trace("Finishing getting all repo config from repo cache");
        }
    }

    /**
     * Return count of all cached data
     */
    public int countAll() {
        log.trace("Getting all repo config count from repo cache...");
        int size = allRepoKeysAndModificationDates.size();
        log.trace("Finishing getting all repo config count from repo cache. Count: {}", size);
        return size;
    }

    /**
     * Add only repo keys. if a new repo was added - a cache will load it lazily.
     * It decreases number of race conditions, however increases number of calls to DB.
     */
    public void addRepoConfigKeys(Map<String, Long> keysAndDates) {
        if (log.isTraceEnabled()) {
            log.trace("Adding repo keys: [{}] to cache...", String.join(",", keysAndDates.keySet()));
        }
        allRepoKeysAndModificationDates.putAll(keysAndDates);
        log.trace("Finishing adding repo keys to cache");
    }

    public void updateRepoConfigKeys(Set<String> keys) {
        if (log.isTraceEnabled()) {
            log.trace("Updating repo keys: [{}] in cache...", String.join(",", keys));
        }
        cache.invalidateAll(keys);
        log.trace("Finishing updating repo keys in cache");
    }

    /**
     * Delete repo configs from cache by keys
     */
    public void deleteRepoConfigsByKeys(Collection<String> keys) {
        if (log.isTraceEnabled()) {
            log.trace("Deleting repo config keys: [{}] from repo cache...", String.join(",", keys));
        }
        // it is ok to not sync it since repo keys like a gate - will not get value from cache even it is cached
        // because key is not in the repo keys set
        keys.stream().forEach(allRepoKeysAndModificationDates::remove);
        cache.invalidateAll(keys);
        if (log.isTraceEnabled()) {
            log.trace("Finishing repo config deletion by keys");
        }
    }

    public void addAndUpdateRepoConfigKeys(Map<String, Long> keysAndDates) {
        if (log.isTraceEnabled()) {
            log.trace("Adding and updating repo keys: [{}] to cache...", String.join(",", keysAndDates.keySet()));
        }
        allRepoKeysAndModificationDates.putAll(keysAndDates);
        cache.invalidateAll(keysAndDates.keySet());
        log.trace("Finishing adding and updating repo keys to cache");
    }

    /**
     * Return set of all cached repo keys
     */
    public Set<String> getAllRepoKeys() {
        log.trace("Getting all repo config keys from repo cache...");
        Set<String> keys = Set.copyOf(allRepoKeysAndModificationDates.keySet());
        if (log.isTraceEnabled()) {
            log.trace("Finishing getting all repo config keys from repo cache. Keys: [{}]", String.join(",", keys));
        }
        return keys;
    }

    /**
     * Fill repo config keys set and remove cached repo configs if they were deleted.
     * DO NOT CALL THIS METHOD OUT OF THE SINGLE THREAD SCHEDULER because it is not synchronized.
     */
    @VisibleForTesting
    void syncRepoConfigKeys() {
        log.debug("Repo cache instance: {}. Sync repo config keys. Getting all repo config keys from db...",
                identifier);
        Map<String, Long> dbRepoKeysAndDates = getAllRepoKeysFunction.get();
        if (log.isDebugEnabled()) {
            StringBuilder messageBuilder = new StringBuilder("Repo cache instance: ");
            messageBuilder.append(identifier);
            messageBuilder.append(". Finishing getting all repo config keys from db");
            if (log.isTraceEnabled()) {
                messageBuilder
                        .append(". Repo config keys: [")
                        .append(Joiner.on(",").join(dbRepoKeysAndDates.keySet()))
                        .append("]");
            }
            log.debug(messageBuilder.toString());
        }

        ConcurrentHashMap<String, Long> dbRepoKeysConcurrentHashMap =
                new ConcurrentHashMap<>(dbRepoKeysAndDates.size());
        dbRepoKeysConcurrentHashMap.putAll(dbRepoKeysAndDates);

        if (log.isDebugEnabled()) {
            StringBuilder messageBuilder = new StringBuilder("Repo cache instance: ")
                    .append(identifier)
                    .append(". Repo keys are updated");
            if (log.isTraceEnabled()) {
                messageBuilder = messageBuilder.append(". Prev state: [")
                        .append(Joiner.on(", ").join(allRepoKeysAndModificationDates.keySet()))
                        .append("]. New state: [")
                        .append(Joiner.on(", ").join(dbRepoKeysConcurrentHashMap.keySet()))
                        .append("]");
            }
            log.debug(messageBuilder.toString());
        }

        RepoConfigPropagationModel propagationModel = calculateCacheDiff(dbRepoKeysAndDates);
        if (propagationModel == null) {
            return;
        }
        reloadRepoConfigurationsFunction.apply(propagationModel, this);
    }

    @Nullable
    public RepoConfigPropagationModel calculateCacheDiff(Map<String, Long> dbRepoKeysAndDates) {
        Set<RepoConfigReload> repoConfigDelete = new HashSet<>();
        Set<RepoConfigReload> repoConfigCreateOrUpdate = new HashSet<>();
        collectRepoConfigUpdateOrDelete(dbRepoKeysAndDates, allRepoKeysAndModificationDates, repoConfigDelete,
                repoConfigCreateOrUpdate);
        collectRepoConfigCreate(dbRepoKeysAndDates.keySet(), allRepoKeysAndModificationDates,
                repoConfigCreateOrUpdate);
        if (repoConfigCreateOrUpdate.isEmpty() && repoConfigDelete.isEmpty()) {
            return null;
        }
        return new RepoConfigPropagationModel(repoConfigCreateOrUpdate, repoConfigDelete, false);
    }

    private void collectRepoConfigCreate(Set<String> dbRepoKeys, ConcurrentHashMap<String, Long> oldKeysAndDates,
            Set<RepoConfigReload> repoConfigCreateOrUpdate) {
        Set<String> newRepoKeys = new HashSet<>(dbRepoKeys);
        newRepoKeys.removeAll(oldKeysAndDates.keySet());
        newRepoKeys.forEach(newRepoKey -> {
            RepoConfig newRepoConfig = getRepoConfigByKeyFunction.apply(newRepoKey);
            if (newRepoConfig == null) {
                log.warn("Repo {} does not exist in DB", newRepoKey);
                return;
            }
            repoConfigCreateOrUpdate.add(
                    new RepoConfigReload(newRepoKey, newRepoConfig.getRepoType(), newRepoConfig.getPackageType(),
                            RepoConfigChange.ConfigAction.REPO_CREATE, newRepoConfig.getModification()));
            log.debug("Repo {} is not cached, but exists in the DB. Cache will be reloaded", newRepoKey);
        });
    }

    private void collectRepoConfigUpdateOrDelete(Map<String, Long> dbRepoKeysAndDates,
            ConcurrentHashMap<String, Long> oldKeysAndDates, Set<RepoConfigReload> repoConfigDelete,
            Set<RepoConfigReload> repoConfigCreateOrUpdate) {
        oldKeysAndDates.forEach((key, modification) -> {
            RepoConfig oldRepoConfig = getByKey(key);
            if (oldRepoConfig == null) {
                log.warn("Repo {} does not exist", key);
                return;
            }
            if (dbRepoKeysAndDates.containsKey(key)) {
                Long dbModificationDate = dbRepoKeysAndDates.get(key);
                if (dbModificationDate == null) {
                    log.warn("Modification date for repo {} should not be null", key);
                    return;
                }
                if (!dbModificationDate.equals(modification)) {
                    repoConfigCreateOrUpdate.add(
                            new RepoConfigReload(key, oldRepoConfig.getRepoType(), oldRepoConfig.getPackageType(),
                                    RepoConfigChange.ConfigAction.REPO_UPDATE, dbModificationDate));
                    log.debug("Modification date for repo {} is changed. Cache will be reloaded", key);
                }
            } else {
                repoConfigDelete.add(
                        new RepoConfigReload(key, oldRepoConfig.getRepoType(), oldRepoConfig.getPackageType(),
                                RepoConfigChange.ConfigAction.REPO_DELETE, modification));
                log.debug("Repo {} does not exist in DB but still exists in cache. Cache will be reloaded", key);
            }
        });
    }

    /**
     * Terminate a thread scheduler that syncs repo keys on a time basis.
     */
    public void destroy() {
        try {
            if (!artifactoryExecutor.isShutdown()) {
                log.trace("Destroying repo config cache...");
                this.artifactoryExecutor.shutdown();
                log.trace("Repo config cache has been successfully destroyed");
            }
        } catch (Throwable th) {
            log.warn("Issue occurs while destroying repo config cache. Reason: {}", th.getMessage());
        }
    }
}
