package org.artifactory.repo.confighandler.manager;

import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.model.RepoConfig;
import org.artifactory.repo.ReferencedConfigChange;
import org.artifactory.repo.confighandler.ConfigChangeEventContext;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.RepoConfigChangeReloadableListener;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;

/**
 * @author Mor Merhav
 * Standard strategy to be used by
 * {@link org.artifactory.repo.confighandler.manager.ConfigChangeEventHandlerManagerImpl}.
 * Performs update via {@link org.artifactory.api.config.RepositoryConfigService} and reload via
 * {@link org.artifactory.api.repo.RepositoryService} and notifies listeners
 * {@link org.artifactory.spring.RepoConfigChangeReloadableListener}
 *
 * @see org.artifactory.repo.confighandler.manager.ConfigChangeEventHandlerManagerImpl
 * @see org.artifactory.api.config.RepositoryConfigService
 * @see org.artifactory.api.repo.RepositoryService
 * @see org.artifactory.spring.RepoConfigChangeReloadableListener
 */
@SuppressWarnings("ClassCanBeRecord")
class StandardUpdateOrReloadStrategy implements UpdateOrReloadStrategy {

    private final InternalRepositoryService repositoryService;
    private final RepositoryConfigService repositoryConfigService;
    private final Collection<RepoConfigChangeReloadableListener> repoConfigChangeReloadableListeners;

    StandardUpdateOrReloadStrategy(InternalRepositoryService repositoryService,
            RepositoryConfigService repositoryConfigService,
            Collection<RepoConfigChangeReloadableListener> repoConfigChangeReloadableListeners) {
        this.repositoryService = repositoryService;
        this.repositoryConfigService = repositoryConfigService;
        this.repoConfigChangeReloadableListeners = repoConfigChangeReloadableListeners;
    }

    @Override
    public void updateOrReload(ConfigChangeEventContext context) {
        updateRepoConfigsIfNeeded(context);
        reloadRepoIfNeeded(context);
    }

    /**
     * Update repositories after a config descriptor change
     */
    private void updateRepoConfigsIfNeeded(ConfigChangeEventContext configChangeEventContext) {
        Map<String, RepoConfig> reposToUpdate = configChangeEventContext.getReposToUpdate();
        if (reposToUpdate != null && !reposToUpdate.isEmpty()) {
            repositoryConfigService.bulkUpsertRepositoryConfigs(new ArrayList<>(reposToUpdate.values()), true);
        }
    }

    /**
     * Reload repositories after a config descriptor change which does not require repo-config changes
     */
    private void reloadRepoIfNeeded(ConfigChangeEventContext configChangeEventContext) {
        if (configChangeEventContext.getReposToReload().isEmpty()) {
            return;
        }
        List<ReferencedConfigChange> referencedConfigChanges = configChangeEventContext.getReposToReload().entrySet()
                .stream()
                .map(entry -> new ReferencedConfigChange(entry.getKey(), entry.getValue()))
                .toList();

        // Reload the repositories first
        repositoryService.reloadOnReferencedConfigChange(referencedConfigChanges);
        repoConfigChangeReloadableListeners.stream()
                .filter(listener -> !(listener instanceof InternalRepositoryService))
                .forEach(listener -> listener.reloadOnReferencedConfigChange(referencedConfigChanges));
    }
}
