/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.repo.interceptor.config;

import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.config.ConfigurationException;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.interceptor.RepoConfigChangeInterceptor;
import org.artifactory.repo.system.SystemRepos;
import org.artifactory.repo.system.suppliers.BuildInfoRepo;
import org.jetbrains.annotations.NotNull;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_CREATE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_DELETE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_UPDATE;
import static org.artifactory.repo.RepoDetailsType.FEDERATED;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL;
import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;

/**
 * @author Fred Simon
 */
@Component
public class BuildInfoRepoConfigInterceptor implements RepoConfigChangeInterceptor {
    private static final Logger log = LoggerFactory.getLogger(BuildInfoRepoConfigInterceptor.class);
    private static final String ERR_DEL_BUILD_REPO = "Deletion of BuildInfo repository is not allowed";
    private RepositoryConfigService repositoryConfigService;
    private CentralConfigService centralConfigService;
    private boolean deleteAllMode;

    @Autowired
    public void setCentralConfigService(CentralConfigService centralConfigService) {
        this.centralConfigService = centralConfigService;
    }

    @Autowired
    public void setRepositoryConfigService(RepositoryConfigService repositoryConfigService) {
        this.repositoryConfigService = repositoryConfigService;
    }

    @Override
    public void setDeleteAllMode(boolean deleteAllMode) {
        this.deleteAllMode = deleteAllMode;
    }

    /**
     * Asserting the following rules about BuildInfo repository:
     * 1. No more than 1 local default BuildInfo Repo
     * 2. No remote or virtual BuildInfo Repo
     * 3. BuildInfo repository must exist. In case it doesn't exist, we add it with default name
     * {@link SystemRepos#BUILDINFO_REPO_KEY}
     */
    @Override
    public void onBeforeSave(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges) {
        List<RepoConfigChange> localAndFederatedRepoChanges =
                Stream.of(repoConfigChanges.getOrDefault(LOCAL, List.of()),
                                repoConfigChanges.getOrDefault(FEDERATED, List.of()))
                        .flatMap(Collection::stream)
                        .collect(Collectors.toList());
        assertNoMoreThanOneDefaultBuildInfo(localAndFederatedRepoChanges);
        assertNoBuildInfoInRemoteVirtual(repoConfigChanges);
        addBuildInfoRepoIfNeeded(repoConfigChanges, localAndFederatedRepoChanges);
    }

    private void assertNoMoreThanOneDefaultBuildInfo(List<RepoConfigChange> localAndFederatedRepoChanges) {
        boolean buildInfoRepoExists = localAndFederatedRepoChanges.stream()
                .filter(repoConfigChange -> REPO_CREATE.equals(repoConfigChange.getConfigAction()) ||
                        REPO_UPDATE.equals(repoConfigChange.getConfigAction()))
                .map(RepoConfigChange::getNewRepoConfig)
                .filter(repo -> RepoType.BuildInfo.equals(repo.getPackageType()))
                .filter(repo -> repo.getKey().startsWith(BUILDINFO_REPO_KEY))
                .count() > 1;
        if (buildInfoRepoExists) {
            throw new ConfigurationException("Can't have more than one " + RepoType.BuildInfo.getType() +
                    " repository.");
        }
    }

    private void assertNoBuildInfoInRemoteVirtual(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges) {
        List<RepoConfigChange> remoteRepoConfigChanges = repoConfigChanges.getOrDefault(REMOTE, List.of());
        List<RepoConfigChange> virtualRepoConfigChanges = repoConfigChanges.getOrDefault(VIRTUAL, List.of());
        if (Stream.of(remoteRepoConfigChanges, virtualRepoConfigChanges)
                .flatMap(Collection::stream)
                .filter(repoConfigChange -> REPO_CREATE.equals(repoConfigChange.getConfigAction()) ||
                        REPO_UPDATE.equals(repoConfigChange.getConfigAction()))
                .map(RepoConfigChange::getNewRepoConfig)
                .anyMatch(repo -> RepoType.BuildInfo.equals(repo.getPackageType()))) {
            throw new ConfigurationException(
                    "Configuring a Remote/virtual " + RepoType.BuildInfo.getType() + " repository is not allowed");
        }
    }

    private void addBuildInfoRepoIfNeeded(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges,
            List<RepoConfigChange> localAndFederatedRepoChanges) {
        List<RepoConfig> existingBuildInfoRepos =
                repositoryConfigService.findRepositoriesByPackageType(RepoType.BuildInfo);
        // build info repo does not exist in storage
        if (existingBuildInfoRepos.isEmpty()) {
            boolean isBuildInfoRepoCreated = localAndFederatedRepoChanges.stream()
                    .filter(repoConfigChange -> REPO_CREATE.equals(repoConfigChange.getConfigAction()) ||
                            REPO_UPDATE.equals(repoConfigChange.getConfigAction()))
                    .map(RepoConfigChange::getNewRepoConfig)
                    .anyMatch(repo -> RepoType.BuildInfo.equals(repo.getPackageType()));
            // build info isn't being created
            if (!isBuildInfoRepoCreated) {
                addDefaultBuildInfoRepo(repoConfigChanges, localAndFederatedRepoChanges);
            }
        }

        if (deleteAllMode) {
            return;
        }

        List<RepoConfigChange> buildInfoDeletions = repoConfigChanges.getOrDefault(LOCAL, List.of()).stream()
                .filter(repoConfigChange -> REPO_DELETE.equals(repoConfigChange.getConfigAction()))
                .filter(repo -> RepoType.BuildInfo.equals(repo.getOldRepoConfig().getPackageType()))
                .collect(Collectors.toList());
        // build info repo is being deleted
        if (!buildInfoDeletions.isEmpty()) {
            log.warn(ERR_DEL_BUILD_REPO);
            updateOldBuildInfoToDefault(repoConfigChanges, buildInfoDeletions);
        }
    }

    private void addDefaultBuildInfoRepo(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges,
            List<RepoConfigChange> localAndFederatedRepoChanges) {
        String repoKey = BUILDINFO_REPO_KEY;
        List<RepoConfigChange> localRepoChanges = repoConfigChanges.getOrDefault(LOCAL, new ArrayList<>());

        // 5 fallbacks allowed for saving the repo
        for (int i = 0; i <= 5; i++) {
            repoKey = (i > 0) ? BUILDINFO_REPO_KEY + "-" + i : BUILDINFO_REPO_KEY;

            if (repositoryConfigService.exists(repoKey) ||
                    isRepoConfigChangesContainsRepo(localAndFederatedRepoChanges, repoKey)) {
                log.warn("'{}' is already taken as repo name for {} repo", repoKey, RepoType.BuildInfo.getType());
            } else {
                LocalRepoConfig buildInfoRepo = getDefaultBuildInfoRepo(centralConfigService.getDescriptor(), repoKey);
                localRepoChanges.add(new RepoConfigChange(repoKey, REPO_CREATE, buildInfoRepo, null));
                repoConfigChanges.put(LOCAL, localRepoChanges);
                log.info("Added {} repository under {}", RepoType.BuildInfo.getType(), repoKey);
                break;
            }
        }
        assertBuildInfoCreated(localRepoChanges, repoKey);
    }

    @NotNull
    private LocalRepoConfig getDefaultBuildInfoRepo(CentralConfigDescriptor descriptor, String repoKey) {
        return new BuildInfoRepo().buildRepoConfig(descriptor, repoKey);
    }

    private void updateOldBuildInfoToDefault(Map<RepoDetailsType, List<RepoConfigChange>> repoConfigChanges,
            List<RepoConfigChange> buildInfoDeletions) {
        RepoConfigChange buildInfoDeletion = buildInfoDeletions.get(0);
        buildInfoDeletions.forEach(deletion -> repoConfigChanges.get(LOCAL).remove(deletion));
        String oldBuildInfoKey = buildInfoDeletion.getRepoKey();
        LocalRepoConfig defaultBuildInfoRepo = getDefaultBuildInfoRepo(centralConfigService.getDescriptor(),
                buildInfoDeletion.getRepoKey());
        List<RepoConfigChange> localRepoChanges = repoConfigChanges.getOrDefault(LOCAL, new ArrayList<>());
        localRepoChanges.add(new RepoConfigChange(oldBuildInfoKey, REPO_UPDATE, defaultBuildInfoRepo,
                buildInfoDeletion.getOldRepoConfig()));
        repoConfigChanges.put(LOCAL, localRepoChanges);
    }

    private boolean isRepoConfigChangesContainsRepo(List<RepoConfigChange> localRepoChanges, String repoKey) {
        return localRepoChanges.stream()
                .filter(c -> !Objects.equals(c.getConfigAction(), REPO_DELETE))
                .anyMatch(c -> Objects.equals(c.getRepoKey(), repoKey));
    }

    private void assertBuildInfoCreated(List<RepoConfigChange> localRepoChanges, String repoKey) {
        if (localRepoChanges.stream().noneMatch(repoConfigChange -> repoKey.equals(repoConfigChange.getRepoKey()) &&
                RepoType.BuildInfo.equals(repoConfigChange.getNewRepoConfig().getPackageType()))) {
            throw new ConfigurationException("Can't save BuildInfo repository, all fallbacks names are taken.");
        }
    }
}
