/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.search.gavc;

import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.module.ModuleInfo;
import org.artifactory.api.module.ModuleInfoUtils;
import org.artifactory.api.search.ItemSearchResults;
import org.artifactory.api.search.gavc.GavcSearchControls;
import org.artifactory.api.search.gavc.GavcSearchResult;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.fs.FileInfo;
import org.artifactory.fs.ItemInfo;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.repo.HttpRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.Repo;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.remote.browse.RemoteItem;
import org.artifactory.repo.virtual.VirtualRepo;
import org.artifactory.sapi.search.VfsQuery;
import org.artifactory.sapi.search.VfsQueryResult;
import org.artifactory.sapi.search.VfsQueryResultType;
import org.artifactory.sapi.search.VfsQueryRow;
import org.artifactory.search.SearcherBase;
import org.artifactory.util.CollectionUtils;
import org.jfrog.client.util.PathUtils;

import javax.annotation.Nullable;
import java.io.IOException;
import java.util.Collection;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.maven.MavenModelUtils.escapeGroupPath;

/**
 * Holds the GAVC search logic
 *
 * @author Noam Tenne
 * @author Natalie Kirillov
 */

@Slf4j
public class GavcSearcher extends SearcherBase<GavcSearchControls, GavcSearchResult> {
    private static final String FILE_EXT_RE = ".*\\.\\w+$";
    private static final String EXCLUDED_FILES = ".*(\\.sha1|\\.sha256|\\.md5|\\.asc|maven-metadata.xml)$";

    @Override
    public ItemSearchResults<GavcSearchResult> doSearch(GavcSearchControls controls) {
        // regular gavc flow
        List<GavcSearchResult> localResults = doLocalSearch(controls, null, null);

        // verify we have any remote/virtual repos to search
        List<String> repos = controls.getSelectedRepoForSearch();
        repos = CollectionUtils.isNullOrEmpty(repos) ? List.of() : repos;

        // specific gavc flow for remote and virtual repositories
        Collection<GavcSearchResult> results;
        if (controls.isSpecific()) {
            results = new LinkedHashSet<>(localResults);
            results.addAll(doSpecificSearch(controls, repos));
        } else {
            results = Lists.newArrayList();
            results.addAll(localResults);
        }

        return new ItemSearchResults<>(Lists.newArrayList(results), results.size());
    }

    @SuppressWarnings("rawtypes")
    private List<GavcSearchResult> doLocalSearch(GavcSearchControls controls, List<String> localRepoKeys,
            @Nullable String virtualRepoKey) {
        //Validate and escape all input values
        String groupInput = escapeGroupPath(controls.getGroupId());

        if (StringUtils.isNotEmpty(virtualRepoKey)) {
            controls.setSelectedRepoForSearch(localRepoKeys);
        }

        //Build search path from inputted group
        VfsQuery query = createQuery(controls)
                .expectedResult(VfsQueryResultType.FILE)
                .addPathFilter(groupInput)
                .addPathFilters(controls.getArtifactId(), controls.getVersion());

        String classifier = controls.getClassifier();
        if (!StringUtils.isBlank(classifier)) {
            query.name("*-" + classifier + "*");
        }

        long limit = getLimit(controls);
        VfsQueryResult queryResult = query.execute(limit);
        List<GavcSearchResult> results = Lists.newArrayList();
        for (VfsQueryRow row : queryResult.getAllRows()) {
            if (results.size() >= limit) {
                break;
            }

            ItemInfo item = row.getItem();
            String repoKey = item.getRepoKey();
            LocalRepo localRepo = getRepoService().localFederatedOrCachedRepositoryByKey(repoKey);
            if (localRepo == null || !isResultAcceptable(item.getRepoPath(), localRepo)) {
                continue;
            }

            ModuleInfo moduleInfo = localRepo.getItemModuleInfo(item.getRelPath());
            if (StringUtils.isNotEmpty(virtualRepoKey)) {
                String filePath = String.join("/", virtualRepoKey, item.getRelPath());
                item = new FileInfoImpl(RepoPathFactory.create(filePath));
            }

            if (moduleInfo.isValid()) {
                results.add(new GavcSearchResult(item, moduleInfo));
            }
        }

        return results;
    }

    /**
     * Perform a GAVC search on a remote repository
     *
     * @param controls       - Search credentials
     * @param repoKey        - The remote repository that's being queried
     * @param virtualRepoKey - The virtual repository key the remote is aggregated to (if it is)
     * @return A list containing all the GavcSearchResult objects
     */
    private Set<GavcSearchResult> doRemoteSearch(GavcSearchControls controls, String repoKey,
            @Nullable String virtualRepoKey) {
        Set<GavcSearchResult> results = new LinkedHashSet<>();
        if (!canSearchRemote(controls)) {
            return results;
        }

        HttpRepo repo = (HttpRepo) getRepoService().remoteRepositoryByKey(repoKey);
        String groupId = escapeGroupPath(controls.getGroupId());
        String artifactId = controls.getArtifactId();
        String version = controls.getVersion();
        String url = getRemoteSearchUrl(repo, groupId, artifactId, version);

        try {
            List<RemoteItem> items = repo.getChildUrls(url);

            // search for specific versions for group id + artifact combination
            if (StringUtils.isEmpty(version)) {
                for (RemoteItem item : items) {
                    String itemUrl = item.getUrl();

                    // skip items that don't indicate a version directory for artifact
                    if (itemUrl.matches(FILE_EXT_RE)) {
                        continue;
                    }

                    List<RemoteItem> versionItems = repo.getChildUrls(itemUrl);
                    results.addAll(aggregateResults(versionItems, repo, virtualRepoKey));
                }
            } else {
                results.addAll(aggregateResults(items, repo, virtualRepoKey));
            }
        } catch (IOException e) {
            log.error("Failed to fetch items for Maven remote repository: {} - {}", repo.getUrl(), e.getMessage());
            log.debug("Failed to fetch items for Maven remote repository: {}", repo.getUrl(), e);
        }

        return results;
    }

    /**
     * Perform a GAVC search on a virtual repository
     *
     * @param controls - Search credentials
     * @param repo     - The virtual repository that's being queried
     * @return A list containing all the GavcSearchResult objects
     */
    private Set<GavcSearchResult> doVirtualSearch(GavcSearchControls controls, VirtualRepo repo) {
        List<String> remotes = repo.getResolvedRemoteRepos()
                .stream()
                .map(Repo::getKey)
                .collect(Collectors.toList());

        List<String> locals = repo.getResolvedLocalFederatedAndCachedRepos()
                .stream()
                .map(Repo::getKey)
                .collect(Collectors.toList());

        Set<GavcSearchResult> results = new LinkedHashSet<>(doLocalSearch(controls, locals, repo.getKey()));
        remotes.forEach(key -> results.addAll(doRemoteSearch(controls, key, repo.getKey())));

        return results;
    }

    /**
     * Perform an extended GAVC search on virtual and remote repositories
     *
     * @param controls - Search credentials
     * @param repos    - A list of all the repositories we want to search on
     * @return A list containing all the GavcSearchResult objects
     */
    private Set<GavcSearchResult> doSpecificSearch(GavcSearchControls controls, List<String> repos) {
        Set<GavcSearchResult> extendedResults = new LinkedHashSet<>();

        repos.forEach(repoKey -> {
            VirtualRepo virtualRepo = getRepoService().virtualRepositoryByKey(repoKey);
            if (virtualRepo != null) {
                extendedResults.addAll(doVirtualSearch(controls, virtualRepo));
                return;
            }

            if (getRepoService().remoteRepositoryByKey(repoKey) != null) {
                extendedResults.addAll(doRemoteSearch(controls, repoKey, null));
            }
        });

        return extendedResults;
    }

    /**
     * Aggregate the found results into a GavcSearchResult list
     *
     * @param items          - Current iteration found items
     * @param repo           - The remote repository being queried
     * @param virtualRepoKey - The virtual repository key, if a virtual repository is being queried
     * @return A list containing all the GavcSearchResult objects
     */
    private Set<GavcSearchResult> aggregateResults(List<RemoteItem> items, HttpRepo repo,
            @Nullable String virtualRepoKey) {
        Set<GavcSearchResult> results = new LinkedHashSet<>();
        String repoKey = StringUtils.isNotEmpty(virtualRepoKey) ? virtualRepoKey : repo.getKey();
        RepoLayout repoLayout = repo.getDescriptor().getRepoLayout();

        items.forEach(item -> {
            // filter out metadata and checksum files
            String url = item.getUrl();
            if (url.matches(EXCLUDED_FILES)) {
                return;
            }

            String relPath = url.replaceFirst(repo.getUrl(), "");
            String filePath = String.join("/", repoKey, relPath);
            FileInfo fileInfo = new FileInfoImpl(RepoPathFactory.create(filePath));
            ModuleInfo moduleInfo = ModuleInfoUtils.moduleInfoFromArtifactPath(relPath, repoLayout);
            results.add(new GavcSearchResult(fileInfo, moduleInfo));
        });

        return results;
    }

    /**
     * Verify there are remote repositories to run the query against
     * Verify group id parameter is present
     * Verify artifact id parameter is present
     *
     * @param controls - Contains the search parameters
     * @return A list of the remote repositories that are being queried if all validations have passed,
     * or an empty list otherwise
     */
    private boolean canSearchRemote(GavcSearchControls controls) {
        String groupId = controls.getGroupId();
        if (StringUtils.isEmpty(groupId)) {
            log.info("Group id parameter is mandatory for remote repository search");
            return false;
        }

        String artifactId = controls.getArtifactId();
        if (StringUtils.isEmpty(artifactId)) {
            log.info("Artifact id parameter is mandatory for remote repository search");
            return false;
        }

        return true;
    }

    /**
     * Constructs the search URL for the remote repository
     *
     * @param repo       - the remote repository
     * @param groupId    - group id for search
     * @param artifactId - artifact id for search
     * @param version    - version for search
     * @return The search URL for the remote repository
     */
    private String getRemoteSearchUrl(HttpRepo repo, String groupId, String artifactId, String version) {
        RepoLayout repoLayout = repo.getDescriptor().getRepoLayout();
        ModuleInfo moduleInfo = new ModuleInfo(groupId, artifactId, version, null, null, null, null, null, null);
        String baseItemPath = ModuleInfoUtils.constructBaseItemPath(moduleInfo, repoLayout);

        return String.join("/", PathUtils.trimTrailingSlashes(repo.getUrl()), baseItemPath);
    }
}