package org.artifactory.security.access;

import com.google.common.collect.Lists;
import lombok.Data;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.util.ListeningPortDetector;
import org.jfrog.access.common.ServiceId;
import org.jfrog.access.router.registration.EndpointProtocol;
import org.jfrog.access.router.registration.HealthCheckEndpoint;
import org.jfrog.access.router.registration.LocalEndpoint;
import org.jfrog.access.router.registration.RoutePath;
import org.jfrog.access.router.registration.RouterRegistrationData;

import java.util.List;

/**
 * @author Tamir Hadad
 */
@Data
public class ArtifactoryRouterRegistrationData {
    private static final String ARTIFACTORY_HEALTH_CHECK_ENDPOINT = "/api/v1/system/readiness";

    private final ServiceId serviceId;
    private final String nodeId;
    private final int artPort;
    private final int artMaintenancePort;
    private final String contextPath;

    ArtifactoryRouterRegistrationData(ServiceId serviceId, String nodeId, String contextPath) {
        this.serviceId = serviceId;
        this.nodeId = nodeId;
        this.artPort = ListeningPortDetector.detect();
        this.artMaintenancePort = ListeningPortDetector.getMaintenancePort();
        this.contextPath = contextPath;
    }

    public RouterRegistrationData build() {
        LocalEndpoint endpoint = new LocalEndpoint(artPort, EndpointProtocol.HTTP1, getRoutePaths());
        HealthCheckEndpoint healthCheckEndpoint =
                new HealthCheckEndpoint(artMaintenancePort, getArtifactoryHealthCheckEndpoint());
        return new RouterRegistrationData(serviceId, nodeId, Lists.newArrayList(endpoint), healthCheckEndpoint);
    }

    private String getArtifactoryHealthCheckEndpoint() {
        if (StringUtils.isNotBlank(contextPath)) {
            return "/" + contextPath + ARTIFACTORY_HEALTH_CHECK_ENDPOINT;
        }
        return ARTIFACTORY_HEALTH_CHECK_ENDPOINT;
    }

    /**
     * Note that the `/jfconfig/**` routes reside in Artifactory temporary,
     * until the new JFrog Config micro-service is implemented and released.
     * (it is going to replace these resources in Artifactory)
     */
    private List<RoutePath> getRoutePaths() {
        return Lists.newArrayList(
                new RoutePath("/artifactory"),
                new RoutePath("/artifactory/(.*)"),
                new RoutePath("/v2/(.*)"),
                new RoutePath("/.well-known/terraform.json"),
                new RoutePath("/jfconfig/api/v1/keys", "/artifactory/api/crypto/keys"),
                new RoutePath("/jfconfig/api/v1/keys/(.*)", "/artifactory/api/crypto/keys/$1")
        );
    }
}
