package org.artifactory.build;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.addon.plugin.PluginsAddon;
import org.artifactory.addon.retention.ArtifactRetentionAddon;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.build.BuildWrapper;
import org.artifactory.api.build.model.BuildGeneralInfo;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.exception.RepoRejectException;
import org.artifactory.api.request.UploadService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.checksum.ChecksumsInfo;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.FileInfo;
import org.artifactory.io.StringResourceStreamHandle;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.request.InternalCapturingResponse;
import org.artifactory.resource.ResourceStreamHandle;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.storage.db.build.service.InternalBuildStoreService;
import org.artifactory.storage.db.fs.dao.NodesDao;
import org.artifactory.storage.db.fs.entity.Node;
import org.jfrog.access.client.project.ProjectSlimEntity;
import org.jfrog.build.api.Build;
import org.jfrog.common.JsonUtils;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.sql.SQLException;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Set;

import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.reset;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertThrows;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

/**
 * @author Yevdo Abramov
 * Created on 03/12/2020
 */
public class BuildServiceImplTest {

    private static final String BUILD_REPO = "buildRepo";
    private static final String BUILD_STARTED = BuildInfoUtils.formatBuildTime(0);
    private static final String BUILD_NUMBER = "number";
    private static final String BUILD_NAME = "buildName";
    private static final String USER = "aUser";
    private static final String PROJECT_A = "proj";
    private static final Set<String> ALL_BUILD_REPOS =
            Set.of(BUILDINFO_REPO_KEY, PROJECT_A + BuildAddon.BUILD_INFO_REPO_EXTENSION);
    private final Build testBuild = new Build();

    @Mock
    private AuthorizationService authService;
    @Mock
    private InternalBuildStoreService buildStoreService;
    @Mock
    private InternalRepositoryService repoService;
    @Mock
    private CentralConfigService configService;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ProjectsService projectsService;
    @Mock
    private UploadService uploadService;
    @Mock
    private ArtifactRetentionAddon artifactRetentionAddon;

    private BuildServiceImpl buildService;
    private AutoCloseable autoCloseable;

    @BeforeMethod
    public void setUp() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        buildService = spy(new BuildServiceImpl());
        buildService.buildInfoReady = true;
        buildService.setAuthService(authService);
        buildService.setBuildStoreService(buildStoreService);
        buildService.setRepoService(repoService);
        buildService.setConfigService(configService);
        buildService.setAddonsManager(addonsManager);
        buildService.setProjectsService(projectsService);
        buildService.setUploadService(uploadService);
        buildService.init();

        testBuild.setName(BUILD_NAME);
        testBuild.setNumber(BUILD_NUMBER);
        testBuild.setStarted(BUILD_STARTED);
        testBuild.setVersion("1.0.0");
        testBuild.setDurationMillis(100);
        when(addonsManager.addonByType(ArtifactRetentionAddon.class)).thenReturn(artifactRetentionAddon);
        when(artifactRetentionAddon.isColdInstance()).thenReturn(false);

        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(context.beanForType(InternalBuildService.class)).thenReturn(buildService);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    public void tearDown() throws Exception {
        autoCloseable.close();
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testAddBuild() throws IOException, RepoRejectException {
        doReturn(Set.of(BUILDINFO_REPO_KEY)).when(buildService).getAllBuildInfoRepos();
        PluginsAddon pluginsAddon = mock(PluginsAddon.class);
        when(addonsManager.addonByType(PluginsAddon.class)).thenReturn(pluginsAddon);

        // Add build to an exist repo
        buildService.addBuild(testBuild, BUILDINFO_REPO_KEY);
        verify(pluginsAddon).execPluginActions(any(), any(), any());
        verify(uploadService).upload(any(), any());

        // Add build to an un-exist build repo with illegal name
        try {
            buildService.addBuild(testBuild, "repo");
            fail("Expected for an IllegalArgumentException");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), "Build info repository 'repo' does not exist");
        }

        // Add build to an un-exist build repo with illegal name
        try {
            buildService.addBuild(testBuild, "prjx" + BuildAddon.BUILD_INFO_REPO_EXTENSION);
            fail("Expected for an IllegalArgumentException");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), "Build info repository 'prjx-build-info' does not exist");
        }

        // Add build to an un-exist build repo
        reset(pluginsAddon);
        reset(uploadService);
        when(projectsService.getProjectSpec("prjx"))
                .thenReturn(ProjectSlimEntity.builder().projectKey("prjx").name("projectX").build());
        BuildAddon buildAddon = mock(BuildAddon.class);
        when(addonsManager.addonByType(BuildAddon.class)).thenReturn(buildAddon);
        when(buildAddon.createProjectRelatedBuildInfoRepo("projectX")).thenReturn("prjx-build-info");

        buildService.addBuild(testBuild, "prjx" + BuildAddon.BUILD_INFO_REPO_EXTENSION);
        verify(buildAddon).createProjectRelatedBuildInfoRepo("prjx");
        verify(uploadService).upload(any(), any());
        verify(pluginsAddon).execPluginActions(any(), any(), any());
    }

    @Test
    public void testAddBuildInternal() {
        BuildWrapper buildWrapper = new BuildWrapper(testBuild, BUILD_REPO);

        // Add build with no permission
        when(authService.canUploadBuild(any(), any(), any(), any())).thenReturn(false);
        try {
            buildService.addBuildInternal(buildWrapper);
            fail("Expected for a ForbiddenException");
        } catch (ForbiddenException e) {
            assertEquals(e.getMessage(), "User is not authorized to access build info");
        }

        // Add build with permission
        when(authService.canUploadBuild(any(), any(), any(), any())).thenReturn(true);
        XrayAddon xrayAddon = mock(XrayAddon.class);
        when(addonsManager.addonByType(XrayAddon.class)).thenReturn(xrayAddon);
        PluginsAddon pluginsAddon = mock(PluginsAddon.class);
        when(addonsManager.addonByType(PluginsAddon.class)).thenReturn(pluginsAddon);
        buildService.addBuildInternal(buildWrapper);

        verify(buildStoreService).addBuild(buildWrapper, true);
        verify(xrayAddon).callAddBuildInterceptor(buildWrapper);
        verify(pluginsAddon).execPluginActions(any(), any(), any());
    }

    @Test
    public void testGetBuild() {
        BuildRun buildRun = new BuildRunTestImpl();
        when(authService
                .canReadBuild(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(), buildRun.getBuildRepo()))
                .thenReturn(true);
        RepoPath repoPath = BuildServiceUtils
                .generateBuildJsonRepoPath(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                        buildRun.getBuildRepo());
        ResourceStreamHandle resourceStreamHandle = mock(ResourceStreamHandle.class);
        when(repoService.getResourceStreamHandle(repoPath)).thenReturn(resourceStreamHandle);
        when(resourceStreamHandle.getInputStream())
                .thenReturn(new ByteArrayInputStream(JsonUtils.getInstance().valueToByteArray(testBuild)));

        // act
        Build resultBuild = buildService.getBuild(buildRun).getBuild();
        assertEquals(resultBuild.toString(), testBuild.toString());
    }

    @Test
    public void testGetBuildInternally() {
    }

    @Test
    public void testGetBuildAsJson() {
        BuildRun buildRun = new BuildRunTestImpl();
        when(authService
                .canReadBuild(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(), buildRun.getBuildRepo()))
                .thenReturn(true);

        RepoPath repoPath = BuildServiceUtils
                .generateBuildJsonRepoPath(buildRun.getName(), buildRun.getNumber(), buildRun.getStarted(),
                        buildRun.getBuildRepo());
        ResourceStreamHandle resourceStreamHandle = mock(ResourceStreamHandle.class);
        when(repoService.getResourceStreamHandle(repoPath)).thenReturn(resourceStreamHandle);
        when(resourceStreamHandle.getInputStream())
                .thenReturn(new ByteArrayInputStream(JsonUtils.getInstance().valueToByteArray(testBuild)));
        String buildJson = buildService.getBuildAsJson(buildRun);
        assertTrue(buildJson.contains("\"version\" : \"1.0.0\""));
        assertTrue(buildJson.contains("\"name\" : \"buildName\""));
        assertTrue(buildJson.contains("\"number\" : \"number\""));
    }

    @Test
    public void testGetBuildInfoRepoForDefaultProject() {
        // when no relevant build-info repo, create
        doReturn(Set.of("build-repo-1", "artifactory-build-info")).when(buildService).getAllBuildInfoRepos();
        assertEquals(buildService.getBuildInfoRepoForDefaultProject(), BUILDINFO_REPO_KEY);

        // when no relevant build-info repo, create
        BuildAddon buildAddon = mock(BuildAddon.class);
        when(addonsManager.addonByType(BuildAddon.class)).thenReturn(buildAddon);
        when(buildAddon.createProjectRelatedBuildInfoRepo(ProjectsService.DEFAULT_PROJECT))
                .thenReturn(BUILDINFO_REPO_KEY);
        doReturn(Set.of("build-repo-1", "not-default-build-info")).when(buildService).getAllBuildInfoRepos();

        assertEquals(buildService.getBuildInfoRepoForDefaultProject(), "artifactory-build-info");
    }

    @Test
    public void testGetBuildInfoRepoKeyByProject() {
        doReturn(Set.of("prjx-build-info", "artifactory-build-info")).when(buildService).getAllBuildInfoRepos();

        // Get default build info repo
        assertEquals(buildService.getBuildInfoRepoKeyByProject(ProjectsService.DEFAULT_PROJECT),
                BUILDINFO_REPO_KEY);

        // Get a project related build info repo
        ProjectSlimEntity projectSlimEntity = ProjectSlimEntity.builder().projectKey("prjx").name("projectX").build();
        doReturn(projectSlimEntity).when(projectsService).getProjectSpec("projectX");
        assertEquals(buildService.getBuildInfoRepoKeyByProject("projectX"), "prjx-build-info");

        // Get a project related build info repo that was not yet created
        projectSlimEntity = ProjectSlimEntity.builder().projectKey("prjy").name("projectY").build();
        doReturn(projectSlimEntity).when(projectsService).getProjectSpec("projectY");
        BuildAddon buildAddon = mock(BuildAddon.class);
        when(addonsManager.addonByType(BuildAddon.class)).thenReturn(buildAddon);
        when(buildAddon.createProjectRelatedBuildInfoRepo("projectY")).thenReturn("prjy-build-info");
        assertEquals(buildService.getBuildInfoRepoKeyByProject("projectY"), "prjy-build-info");

        // Get a project related build info repo of an un-exist project
        doReturn(null).when(projectsService).getProjectSpec("projectZ");
        try {
            buildService.getBuildInfoRepoKeyByProject("projectZ");
            fail("An IllegalArgumentException was expected to be thrown");
        } catch (IllegalArgumentException e) {
            assertEquals(e.getMessage(), "Project projectZ was not found.");
        }
    }

    @Test
    public void testNormaliseBuildRepoName() {
        doReturn(Set.of(BUILDINFO_REPO_KEY)).when(buildService).getAllBuildInfoRepos();
        assertEquals(buildService.normaliseBuildRepoName(null), BUILDINFO_REPO_KEY);
        assertEquals(buildService.normaliseBuildRepoName(""), BUILDINFO_REPO_KEY);
        assertEquals(buildService.normaliseBuildRepoName(BUILDINFO_REPO_KEY),
                BUILDINFO_REPO_KEY);
        assertEquals(buildService.normaliseBuildRepoName("repo"), "repo");
    }

    @Test(dataProvider = "provideBuildRepoAndProject")
    public void testGetPreferredBuildRepo(String buildRepo, String projectKey, String expectedBuildRepo) {
        doReturn(ALL_BUILD_REPOS).when(buildService).getAllBuildInfoRepos();
        ProjectSlimEntity projectSlimEntity = ProjectSlimEntity.builder().projectKey(PROJECT_A).name(PROJECT_A).build();
        doReturn(projectSlimEntity).when(projectsService).getProjectSpec(PROJECT_A);

        String resultRepo = buildService.getPreferredBuildRepo(buildRepo, projectKey);
        assertEquals(expectedBuildRepo, resultRepo);
    }
    
    @Test
    public void testAssertBasicReadPermissions() {
        // given
        when(authService.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO))
                .thenReturn(false, true);
        when(authService.currentUsername()).thenReturn(USER);

        // when - there aren't read permissions
        assertThrows(ForbiddenException.class, () ->
                buildService.assertBasicReadPermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO));
        // then
        verify(authService).isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verify(authService).currentUsername();
        verifyNoMoreInteractions(authService);
        clearInvocations(authService);

        // when - there are read permissions
        buildService.assertBasicReadPermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);

        // then
        verify(authService).isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verifyNoMoreInteractions(authService);
    }

    @Test
    public void testAssertReadPermissions() {
        // given
        when(authService.canReadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO))
                .thenReturn(false, true);
        when(authService.currentUsername()).thenReturn(USER);

        // when - there aren't read permissions
        assertThrows(ForbiddenException.class, () ->
                buildService.assertReadPermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO));
        // then
        verify(authService).canReadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verify(authService).currentUsername();
        verifyNoMoreInteractions(authService);
        clearInvocations(authService);

        // when - there are read permissions
        buildService.assertReadPermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);

        // then
        verify(authService).canReadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verifyNoMoreInteractions(authService);
    }

    @Test
    public void testAssertDeletePermissions() {
        // given
        when(authService.canDeleteBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO))
                .thenReturn(false, true);
        when(authService.currentUsername()).thenReturn(USER);

        // when - there aren't delete permissions
        assertThrows(ForbiddenException.class, () ->
                buildService.assertDeletePermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO));
        // then
        verify(authService).canDeleteBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verify(authService).currentUsername();
        verifyNoMoreInteractions(authService);
        clearInvocations(authService);

        // when - there are delete permissions
        buildService.assertDeletePermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);

        // then
        verify(authService).canDeleteBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verifyNoMoreInteractions(authService);
    }

    @Test
    public void testAssertUploadPermissions() {
        // given
        when(authService.canUploadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO))
                .thenReturn(false, true);
        when(authService.currentUsername()).thenReturn(USER);

        // when - there aren't upload permissions
        assertThrows(ForbiddenException.class, () ->
                buildService.assertUploadPermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO));
        // then
        verify(authService).canUploadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verify(authService).currentUsername();
        verifyNoMoreInteractions(authService);
        clearInvocations(authService);

        // when - there are upload permissions
        buildService.assertUploadPermissions(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);

        // then
        verify(authService).canUploadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        verifyNoMoreInteractions(authService);
    }

    @Test
    public void testGetBuildGeneralInfo() throws IOException {
        // when - invalid build model
        when(buildStoreService.getBuildRun(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED))
                .thenReturn(mock(BuildRun.class));

        BuildGeneralInfo buildGeneralInfo =
                buildService.getBuildGeneralInfo(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);
        assertNull(buildGeneralInfo);

        verify(buildStoreService).getBuildRun(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED);
        verifyNoMoreInteractions(buildStoreService);
        when(authService.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO))
                .thenReturn(false);

        // when - no permissions
        Build build = new Build();
        build.setName(BUILD_NAME);
        build.setNumber(BUILD_NUMBER);
        build.setStarted(BUILD_STARTED);
        ResourceStreamHandle resourceStreamHandle =
                new StringResourceStreamHandle(JsonUtils.getInstance().valueToString(build));

        when(repoService.getResourceStreamHandle(any(RepoPath.class))).thenReturn(resourceStreamHandle);
        ChecksumsInfo checksumsInfoMock = mock(ChecksumsInfo.class);
        when(checksumsInfoMock.getSha256()).thenReturn("someChecksum");
        FileInfo fileInfoMock = mock(FileInfo.class);
        when(fileInfoMock.getChecksumsInfo()).thenReturn(checksumsInfoMock);
        when(repoService.getFileInfo(any(RepoPath.class))).thenReturn(fileInfoMock);
        when(buildStoreService.getBuildRun(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED))
                .thenReturn(new BuildRunTestImpl());

        assertThrows(ForbiddenException.class,
                () -> buildService.getBuildGeneralInfo(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO));

        // when - all good
        when(authService.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO)).thenReturn(true);

        buildGeneralInfo = buildService.getBuildGeneralInfo(BUILD_NAME, BUILD_NUMBER, BUILD_STARTED, BUILD_REPO);

        assertNotNull(buildGeneralInfo);
        assertEquals(buildGeneralInfo.getBuildName(), BUILD_NAME);
        assertEquals(buildGeneralInfo.getBuildNumber(), BUILD_NUMBER);
        assertEquals(buildGeneralInfo.getBuildRepo(), BUILD_REPO);
    }

    @Test
    public void testParseBuildSha256() {
        assertEquals(buildService.parseBuildSha256(createResponseMock(null)), "");
        assertEquals(buildService.parseBuildSha256(createResponseMock("")), "");
        assertEquals(buildService.parseBuildSha256(createResponseMock("test")), "");
        assertEquals(buildService.parseBuildSha256(createResponseMock("{\n" +
                "  \"originalChecksums\" : {\n" +
                "  }\n" +
                "}")), "");
        assertEquals(buildService.parseBuildSha256(createResponseMock("{\n" +
                "  \"originalChecksums\" : {\n" +
                "    \"sha256\" : \"706e99bde88ad92d5be14bc2eee0c56536ab617bb6edb782019df83a868f9fe4\"\n" +
                "  }\n" +
                "}")), "706e99bde88ad92d5be14bc2eee0c56536ab617bb6edb782019df83a868f9fe4");
    }

    @Test
    public void testGetAllBuildInfoRepos() {
        LocalRepoConfig localRepoConfig1 = LocalRepoConfig.create("local-repo", RepoType.BuildInfo);
        LocalRepoConfig localRepoConfig2 = LocalRepoConfig.create("federated-repo", RepoType.BuildInfo);

        when(repoService.getLocalFederatedRepoConfigsIncludingInfoRepos()).thenReturn(
                List.of(localRepoConfig1, localRepoConfig2));

        Set<String> myRepos = buildService.getAllBuildInfoRepos();

        assertTrue(myRepos.containsAll(List.of("local-repo", "federated-repo")));
    }

    @Test
    public void testGetFileNodesBySha1SameSha1On2RepoPaths() throws SQLException {
        NodesDao nodesDaoMock = mock(NodesDao.class);
        Date created = new Date(System.currentTimeMillis() - 30000);
        Date modified = new Date(System.currentTimeMillis() - 20000);
        Date updated = new Date(System.currentTimeMillis() - 10000);

        Node n1 = new Node(2, false, "repo", "path/to/there1", "name", (short) 3,
                created.getTime(), "dot", modified.getTime(), "modifier", updated.getTime(),
                0, "dddd89fc2a043c2479a6de676a2f7179e9eaddac", null, null, null, null);

        Node n2 = new Node(2, false, "repo", "path/to/there2", "name", (short) 3,
                created.getTime(), "dot", modified.getTime(), "modifier", updated.getTime(),
                0, "dddd89fc2a043c2479a6de676a2f7179e9eaddac", null, null, null, null);

        when(nodesDaoMock.getFileNodesBySha1(any())).thenReturn(List.of(n1, n2));
        buildService.setNodesDao(nodesDaoMock);
        Map<String, RepoPath> repoPathMap = buildService.getFileNodesBySha1(
                List.of("dddd89fc2a043c2479a6de676a2f7179e9eaddac"));
        assertEquals(repoPathMap.size(), 1);
        assertEquals(repoPathMap.get("dddd89fc2a043c2479a6de676a2f7179e9eaddac").getPath(), "path/to/there1");
    }

    private InternalCapturingResponse createResponseMock(String resultAsString) {
        InternalCapturingResponse responseMock = mock(InternalCapturingResponse.class);
        when(responseMock.getResultAsString()).thenReturn(resultAsString);
        return responseMock;
    }

    private static class BuildRunTestImpl implements BuildRun {
        @Override
        public String getName() {
            return BUILD_NAME;
        }

        @Override
        public String getNumber() {
            return BUILD_NUMBER;
        }

        @Override
        public String getStarted() {
            return BUILD_STARTED;
        }

        @Override
        public Date getStartedDate() {
            return new Date();
        }

        @Override
        public String getBuildRepo() {
            return BUILD_REPO;
        }

        @Override
        public String getCiUrl() {
            return "CI_URL";
        }

        @Override
        public String getReleaseStatus() {
            return "STATUS";
        }
    }

    @DataProvider
    public static Object[][] provideBuildRepoAndProject() {
        return new Object[][] {
                {null, null, BUILDINFO_REPO_KEY},
                {BUILDINFO_REPO_KEY, null, BUILDINFO_REPO_KEY},
                {PROJECT_A + BuildAddon.BUILD_INFO_REPO_EXTENSION, null,
                        PROJECT_A + BuildAddon.BUILD_INFO_REPO_EXTENSION},
                {null, PROJECT_A, PROJECT_A + BuildAddon.BUILD_INFO_REPO_EXTENSION},
                {null, "default", BUILDINFO_REPO_KEY},
                {BUILDINFO_REPO_KEY, PROJECT_A, BUILDINFO_REPO_KEY}
        };
    }

}