package org.artifactory.metrics.framework.providers;

import org.artifactory.common.ConstantValues;
import org.artifactory.repo.http.HttpConnectionPoolMonitorService;
import org.artifactory.repo.http.mbean.HTTPConnectionPool;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.metrics.model.Metric;
import org.jfrog.metrics.model.MetricProvider;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

public class HttpConnectionsMetricProviderTest extends ArtifactoryHomeBoundTest {

    private HttpConnectionsMetricProvider httpConnectionsMetricProvider;
    private HttpConnectionPoolMonitorService httpConnectionPoolMonitorService;

    @BeforeMethod
    public void setUp() {
        httpConnectionPoolMonitorService = mock(HttpConnectionPoolMonitorService.class);
        httpConnectionsMetricProvider = new HttpConnectionsMetricProvider(httpConnectionPoolMonitorService);
    }

    @Test
    public void testGetName() {
        assertEquals("jfrt_http_connections", httpConnectionsMetricProvider.getName());
    }

    @Test
    public void testGetDescription() {
        assertEquals("Artifactory HTTP Connection Metrics Provider", httpConnectionsMetricProvider.getDescription());
    }

    @Test
    public void testGetRefreshIntervalMillis() {
        assertEquals(0L, httpConnectionsMetricProvider.getRefreshIntervalMillis());
    }

    @Test
    public void testCollect() {
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", new WeakReference<>(pool));
        when(pool.getAvailable()).thenReturn(10);
        when(pool.getPending()).thenReturn(5);
        when(pool.getLeased()).thenReturn(2);
        when(pool.getMax()).thenReturn(100);
        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);


        List<String> allowedMetrics = new ArrayList<>();
        allowedMetrics.add("available_one");
        allowedMetrics.add("max_one");
        allowedMetrics.add("leased_one");
        allowedMetrics.add("pending_one");
        List<Metric> metricList = httpConnectionsMetricProvider.collect();

        assertEquals(4, metricList.size());
        assertTrue(metricList.stream().allMatch(metric -> allowedMetrics.contains(metric.getName())));
        Metric availableMetric = metricList.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("available_one"))
                .findFirst()
                .get();

        assertEquals(10, availableMetric.getValue().doubleValue());
        Metric pendingMetric = metricList.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("pending_one"))
                .findFirst()
                .get();

        assertEquals(5, pendingMetric.getValue().doubleValue());
        Metric leasedMetric = metricList.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("leased_one"))
                .findFirst()
                .get();
        assertEquals(2, leasedMetric.getValue().doubleValue());
    }

    @Test
    public void testCollectNullMetrics() {
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", null);
        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);

        List<Metric> metricList = httpConnectionsMetricProvider.collect();
        assertEquals(0, metricList.size());
    }

    @Test
    public void testCollectZeroAvailableMetrics() {
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", new WeakReference<>(pool));
        when(pool.getAvailable()).thenReturn(0);
        when(pool.getPending()).thenReturn(5);
        when(pool.getLeased()).thenReturn(2);
        when(pool.getMax()).thenReturn(100);
        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);

        List<Metric> metricList = httpConnectionsMetricProvider.collect();
        assertEquals(0, metricList.size());
    }

    @Test
    public void testCollectMultiPools() {
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool2 = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", new WeakReference<>(pool));
        poolsMap.put("two", new WeakReference<>(pool2));
        poolsMap.put("twos", null);
        when(pool.getAvailable()).thenReturn(40);
        when(pool.getPending()).thenReturn(15);
        when(pool.getLeased()).thenReturn(20);
        when(pool.getMax()).thenReturn(50);
        when(pool2.getAvailable()).thenReturn(10);
        when(pool2.getPending()).thenReturn(15);
        when(pool2.getLeased()).thenReturn(20);
        when(pool2.getMax()).thenReturn(100);
        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);

        List<String> allowedMetrics = new ArrayList<>();
        allowedMetrics.add("available_one");
        allowedMetrics.add("leased_one");
        allowedMetrics.add("pending_one");
        allowedMetrics.add("max_one");

        List<String> allowedMetrics2 = new ArrayList<>();
        allowedMetrics2.add("available_two");
        allowedMetrics2.add("leased_two");
        allowedMetrics2.add("pending_two");
        allowedMetrics2.add("max_two");

        List<Metric> metricList = httpConnectionsMetricProvider.collect();
        assertEquals(metricList.size(), 8);
        List<Metric> poolOneMetric =
                metricList.stream()
                        .filter(metric -> metric.getLabels().get("pool").toString().equalsIgnoreCase("one"))
                        .collect(Collectors.toList());
        assertTrue(poolOneMetric.stream().allMatch(metric -> allowedMetrics.contains(metric.getName())));
        assertEquals(poolOneMetric.size(), 4);

        Metric availableMetricOne = poolOneMetric.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("available_one"))
                .findFirst()
                .get();
        assertEquals(40, availableMetricOne.getValue().doubleValue());
        Metric pendingMetricOne = poolOneMetric.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("pending_one"))
                .findFirst()
                .get();
        assertEquals(15, pendingMetricOne.getValue().doubleValue());
        Metric leasedMetricOne = poolOneMetric.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("leased_one"))
                .findFirst()
                .get();
        assertEquals(20, leasedMetricOne.getValue().doubleValue());
        List<Metric> poolTwoMetric = metricList.stream()
                .filter(metric -> metric.getLabels().get("pool").toString().equalsIgnoreCase("two"))
                .collect(Collectors.toList());
        assertTrue(poolTwoMetric.stream().allMatch(metric -> allowedMetrics2.contains(metric.getName())));
        assertEquals(poolTwoMetric.size(), 4);

        Metric availableMetric = poolTwoMetric.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("available_two"))
                .findFirst()
                .get();
        assertEquals(10, availableMetric.getValue().doubleValue());
        Metric pendingMetric = poolTwoMetric.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("pending_two"))
                .findFirst()
                .get();
        assertEquals(15, pendingMetric.getValue().doubleValue());
        Metric leasedMetric = poolTwoMetric.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("leased_two"))
                .findFirst()
                .get();
        assertEquals(20, leasedMetric.getValue().doubleValue());
    }

    @Test
    public void testCollectMultiPoolsMorethan10() {
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool2 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool3 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool4 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool5 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool6 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool7 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool8 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool9 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool10 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool11 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool12 = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", new WeakReference<>(pool));
        poolsMap.put("two", new WeakReference<>(pool2));
        poolsMap.put("three", new WeakReference<>(pool3));
        poolsMap.put("four", new WeakReference<>(pool4));
        poolsMap.put("five", new WeakReference<>(pool5));
        poolsMap.put("six", new WeakReference<>(pool6));
        poolsMap.put("seven", new WeakReference<>(pool7));
        poolsMap.put("eight", new WeakReference<>(pool8));
        poolsMap.put("nine", new WeakReference<>(pool9));
        poolsMap.put("ten", new WeakReference<>(pool10));
        poolsMap.put("eleven", new WeakReference<>(pool11));
        poolsMap.put("twelve", new WeakReference<>(pool12));

        when(pool.getAvailable()).thenReturn(40);
        when(pool.getPending()).thenReturn(15);
        when(pool.getLeased()).thenReturn(20);
        when(pool.getMax()).thenReturn(50);

        when(pool2.getAvailable()).thenReturn(90);
        when(pool2.getPending()).thenReturn(15);
        when(pool2.getLeased()).thenReturn(20);
        when(pool2.getMax()).thenReturn(100);

        when(pool3.getAvailable()).thenReturn(70);
        when(pool3.getPending()).thenReturn(15);
        when(pool3.getLeased()).thenReturn(20);
        when(pool3.getMax()).thenReturn(100);

        when(pool4.getAvailable()).thenReturn(50);
        when(pool4.getPending()).thenReturn(15);
        when(pool4.getLeased()).thenReturn(20);
        when(pool4.getMax()).thenReturn(100);

        when(pool5.getAvailable()).thenReturn(20);
        when(pool5.getPending()).thenReturn(15);
        when(pool5.getLeased()).thenReturn(20);
        when(pool5.getMax()).thenReturn(100);


        when(pool6.getAvailable()).thenReturn(10);
        when(pool6.getPending()).thenReturn(15);
        when(pool6.getLeased()).thenReturn(20);
        when(pool6.getMax()).thenReturn(100);


        when(pool7.getAvailable()).thenReturn(80);
        when(pool7.getPending()).thenReturn(15);
        when(pool7.getLeased()).thenReturn(20);
        when(pool7.getMax()).thenReturn(100);


        when(pool8.getAvailable()).thenReturn(30);
        when(pool8.getPending()).thenReturn(15);
        when(pool8.getLeased()).thenReturn(20);
        when(pool8.getMax()).thenReturn(100);


        when(pool9.getAvailable()).thenReturn(100);
        when(pool9.getPending()).thenReturn(15);
        when(pool9.getLeased()).thenReturn(20);
        when(pool9.getMax()).thenReturn(100);


        when(pool10.getAvailable()).thenReturn(60);
        when(pool10.getPending()).thenReturn(15);
        when(pool10.getLeased()).thenReturn(20);
        when(pool10.getMax()).thenReturn(100);


        when(pool11.getAvailable()).thenReturn(90);
        when(pool11.getPending()).thenReturn(15);
        when(pool11.getLeased()).thenReturn(20);
        when(pool11.getMax()).thenReturn(100);


        when(pool12.getAvailable()).thenReturn(5);
        when(pool12.getPending()).thenReturn(15);
        when(pool12.getLeased()).thenReturn(20);
        when(pool12.getMax()).thenReturn(100);


        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);

        List<String> allowedMetrics = new ArrayList<>();
        allowedMetrics.add("available");
        allowedMetrics.add("leased");
        allowedMetrics.add("pending");
        List<Metric> metricList = httpConnectionsMetricProvider.collect();
        assertEquals(metricList.size(), 40);

        assertFalse(metricList.stream()
                .anyMatch(metric -> metric.getLabels().get("pool").toString().equalsIgnoreCase("six")));
        assertFalse(metricList.stream()
                .anyMatch(metric -> metric.getLabels().get("pool").toString().equalsIgnoreCase("twelve")));


    }

    @Test
    public void testCollectMultiPoolsMorethanConfiguredValue() {
        int configuredRepos = 5;
        homeStub.getArtifactoryProperties()
                .setProperty(ConstantValues.totalRepositoriesForHttpConnectionMetrics.getPropertyName(),
                        String.valueOf(configuredRepos));
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool2 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool3 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool4 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool5 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool6 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool7 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool8 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool9 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool10 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool11 = mock(HTTPConnectionPool.class);
        HTTPConnectionPool pool12 = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", new WeakReference<>(pool));
        poolsMap.put("two", new WeakReference<>(pool2));
        poolsMap.put("three", new WeakReference<>(pool3));
        poolsMap.put("four", new WeakReference<>(pool4));
        poolsMap.put("five", new WeakReference<>(pool5));
        poolsMap.put("six", new WeakReference<>(pool6));
        poolsMap.put("seven", new WeakReference<>(pool7));
        poolsMap.put("eight", new WeakReference<>(pool8));
        poolsMap.put("nine", new WeakReference<>(pool9));
        poolsMap.put("ten", new WeakReference<>(pool10));
        poolsMap.put("eleven", new WeakReference<>(pool11));
        poolsMap.put("twelve", new WeakReference<>(pool12));

        when(pool.getAvailable()).thenReturn(40);
        when(pool.getPending()).thenReturn(15);
        when(pool.getLeased()).thenReturn(20);
        when(pool.getMax()).thenReturn(50);

        when(pool2.getAvailable()).thenReturn(90);
        when(pool2.getPending()).thenReturn(15);
        when(pool2.getLeased()).thenReturn(20);
        when(pool2.getMax()).thenReturn(100);

        when(pool3.getAvailable()).thenReturn(70);
        when(pool3.getPending()).thenReturn(15);
        when(pool3.getLeased()).thenReturn(20);
        when(pool3.getMax()).thenReturn(100);

        when(pool4.getAvailable()).thenReturn(50);
        when(pool4.getPending()).thenReturn(15);
        when(pool4.getLeased()).thenReturn(20);
        when(pool4.getMax()).thenReturn(100);

        when(pool5.getAvailable()).thenReturn(20);
        when(pool5.getPending()).thenReturn(15);
        when(pool5.getLeased()).thenReturn(20);
        when(pool5.getMax()).thenReturn(100);


        when(pool6.getAvailable()).thenReturn(10);
        when(pool6.getPending()).thenReturn(15);
        when(pool6.getLeased()).thenReturn(20);
        when(pool6.getMax()).thenReturn(100);


        when(pool7.getAvailable()).thenReturn(80);
        when(pool7.getPending()).thenReturn(15);
        when(pool7.getLeased()).thenReturn(20);
        when(pool7.getMax()).thenReturn(100);


        when(pool8.getAvailable()).thenReturn(30);
        when(pool8.getPending()).thenReturn(15);
        when(pool8.getLeased()).thenReturn(20);
        when(pool8.getMax()).thenReturn(100);


        when(pool9.getAvailable()).thenReturn(100);
        when(pool9.getPending()).thenReturn(15);
        when(pool9.getLeased()).thenReturn(20);
        when(pool9.getMax()).thenReturn(100);


        when(pool10.getAvailable()).thenReturn(60);
        when(pool10.getPending()).thenReturn(15);
        when(pool10.getLeased()).thenReturn(20);
        when(pool10.getMax()).thenReturn(100);


        when(pool11.getAvailable()).thenReturn(90);
        when(pool11.getPending()).thenReturn(15);
        when(pool11.getLeased()).thenReturn(20);
        when(pool11.getMax()).thenReturn(100);


        when(pool12.getAvailable()).thenReturn(5);
        when(pool12.getPending()).thenReturn(15);
        when(pool12.getLeased()).thenReturn(20);
        when(pool12.getMax()).thenReturn(100);


        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);

        List<String> allowedMetrics = new ArrayList<>();
        allowedMetrics.add("available");
        allowedMetrics.add("leased");
        allowedMetrics.add("pending");
        List<Metric> metricList = httpConnectionsMetricProvider.collect();
        assertEquals(metricList.size(), configuredRepos * 4);
    }

    @Test
    public void testCollectNullPool() {
        HTTPConnectionPool pool = mock(HTTPConnectionPool.class);
        Map<String, WeakReference<HTTPConnectionPool>> poolsMap = new HashMap<>();
        poolsMap.put("one", new WeakReference<>(pool));
        poolsMap.put("two", null);
        when(pool.getAvailable()).thenReturn(10);
        when(pool.getPending()).thenReturn(5);
        when(pool.getLeased()).thenReturn(2);
        when(pool.getMax()).thenReturn(100);
        when(httpConnectionPoolMonitorService.getConnectionPools()).thenReturn(poolsMap);

        List<String> allowedMetrics = new ArrayList<>();
        allowedMetrics.add("available_one");
        allowedMetrics.add("max_one");
        allowedMetrics.add("leased_one");
        allowedMetrics.add("pending_one");
        List<Metric> metricList = httpConnectionsMetricProvider.collect();
        assertEquals(4, metricList.size());
        assertTrue(metricList.stream().allMatch(metric -> allowedMetrics.contains(metric.getName())));
        Metric availableMetric = metricList.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("available_one"))
                .findFirst()
                .get();
        assertEquals(10, availableMetric.getValue().doubleValue());
        Metric pendingMetric = metricList.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("pending_one"))
                .findFirst()
                .get();
        assertEquals(5, pendingMetric.getValue().doubleValue());
        Metric leasedMetric = metricList.stream()
                .filter(metric -> metric.getName().equalsIgnoreCase("leased_one"))
                .findFirst()
                .get();
        assertEquals(2, leasedMetric.getValue().doubleValue());
    }

    @Test
    public void testGetReportType() {
        assertEquals(MetricProvider.ReportType.CONTINUOUS, httpConnectionsMetricProvider.getReportType());
    }

    @Test
    public void testGetLastUpdatedMillis() {
        assertEquals(0L, httpConnectionsMetricProvider.getLastUpdatedMillis());
    }
}