package org.artifactory.ph;

import org.apache.http.HttpStatus;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.exception.PackageRepoDescriptorException;
import org.artifactory.core.api.ph.service.PackageArtifactoryConfigService;
import org.artifactory.core.api.ph.vcs.configuration.PhVcsGitConfiguration;
import org.artifactory.core.api.ph.vcs.configuration.PhVcsGitProvider;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.VcsConfiguration;
import org.artifactory.repo.Repo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.springframework.mock.web.MockHttpServletRequest;

import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

/**
 * @author Ori Yafe
 */
class PackageArtifactoryConfigServiceImplTest {

    @Mock
    private InternalRepositoryService repositoryService;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private ArtifactoryContext artifactoryContext;
    @Mock
    private CentralConfigDescriptor centralConfigDescriptor;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;

    private AutoCloseable mockitoAutoCloseable;

    private PackageArtifactoryConfigService packageArtifactoryConfigService;

    @BeforeEach
    void setUp() {
        mockitoAutoCloseable = MockitoAnnotations.openMocks(this);
        ArtifactoryHome home = Mockito.mock(ArtifactoryHome.class);
        when(home.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        ArtifactoryHome.bind(home);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);

        packageArtifactoryConfigService = new PackageArtifactoryConfigServiceImpl(repositoryService);
    }

    @AfterEach
    void tearDown() throws Exception {
        mockitoAutoCloseable.close();
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    void testGetStringFromConfig() throws PackageRepoDescriptorException {
        String repoKey = "local";
        String method = "getTerraformType";
        String value = "MODULE";
        prepareTest(repoKey);
        // no such field
        Assertions.assertThrows(
                PackageRepoDescriptorException.class, () ->
                        packageArtifactoryConfigService.getStringFromRepoConfig(repoKey, "nosuchmethod", "default"),
                "default");

        Assertions.assertEquals(packageArtifactoryConfigService.getStringFromRepoConfig(repoKey, method, "default"),
                value);
    }

    @Test
    void testGetIntegerFromConfig() throws PackageRepoDescriptorException {
        String repoKey = "local";
        String method = "getDockerTagRetention";
        int value = 1;
        prepareTest(repoKey);
        // no such method
        Assertions.assertThrows(PackageRepoDescriptorException.class, () ->
                packageArtifactoryConfigService.getIntFromRepoConfig(repoKey, "nosuchmethod", 33));
        // casting error
        Assertions.assertThrows(PackageRepoDescriptorException.class, () ->
                packageArtifactoryConfigService.getIntFromRepoConfig(repoKey, "isForceNugetAuthentication", 33));

        Assertions.assertEquals(
                value, packageArtifactoryConfigService.getIntFromRepoConfig(repoKey, method, 33));
    }

    @Test
    void testGetArtifactoryBaseUrlFromHttpUrl() {
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        String requestUrl = "https://lala.land.com";
        when(centralConfigDescriptor.getUrlBase()).thenReturn(null);
        MockHttpServletRequest mockRequest = new MockHttpServletRequest("GET", requestUrl);
        mockRequest.setServerName("lala.land.com");
        mockRequest.setScheme("https");
        mockRequest.setServerPort(443);
        String servletContextUrl = packageArtifactoryConfigService.getArtifactoryBaseUrl(mockRequest);
        Assertions.assertEquals(servletContextUrl, requestUrl);
    }

    @Test
    void testGetArtifactoryBaseUrlFromDescriptorWithoutContextPath() {
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        String contextPath = "/artifactory";
        String requestUrl = "https://lala.land.com";
        when(artifactorySystemProperties.getProperty(ConstantValues.contextPath)).thenReturn(contextPath);
        when(centralConfigDescriptor.getUrlBase()).thenReturn(requestUrl);
        MockHttpServletRequest mockRequest = new MockHttpServletRequest("GET", requestUrl);
        String servletContextUrl = packageArtifactoryConfigService.getArtifactoryBaseUrl(mockRequest);
        Assertions.assertEquals(servletContextUrl, requestUrl + contextPath);
    }

    @Test
    void testGetArtifactoryBaseUrlFromDescriptorWithContextPath() {
        when(artifactoryContext.getCentralConfig()).thenReturn(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(centralConfigDescriptor);
        String requestUrl = "https://lala.land.com/artifactory";
        when(centralConfigDescriptor.getUrlBase()).thenReturn(requestUrl);
        MockHttpServletRequest mockRequest = new MockHttpServletRequest("GET", requestUrl);
        String servletContextUrl = packageArtifactoryConfigService.getArtifactoryBaseUrl(mockRequest);
        Assertions.assertEquals(servletContextUrl, requestUrl);
    }

    @Test
    void testGetBooleanFromConfig() throws PackageRepoDescriptorException {
        String repoKey = "local";
        String method = "isDebianTrivialLayout";
        prepareTest(repoKey);
        // no such method
        Assertions.assertThrows(PackageRepoDescriptorException.class, () ->
                packageArtifactoryConfigService.getBooleanFromRepoConfig(repoKey, "nosuchmethod", true));

        // casting error
        Assertions.assertThrows(PackageRepoDescriptorException.class, () ->
                packageArtifactoryConfigService.getBooleanFromRepoConfig(repoKey, "getDockerTagRetention", true));

        Assertions.assertFalse(
                packageArtifactoryConfigService.getBooleanFromRepoConfig(repoKey, method, false));
    }

    @Test
    void testGetVcsConfigOfRemoteRepoWhenNotExist() {
        String repoKey = "local";
        when(repositoryService.remoteRepoDescriptorByKey(repoKey)).thenReturn(null);
        try {
            packageArtifactoryConfigService.getVcsConfigOfRemoteRepo(repoKey);
            Assertions.fail("Expected exception");
        } catch (PackageException e) {
            Assertions.assertEquals(HttpStatus.SC_NOT_FOUND, e.getStatus());
            Assertions.assertEquals("Remote repository 'local' was not found", e.getPublicReason());
        }
    }

    @Test
    void testGetVcsConfigOfRemoteRepoWhenNoVcs() {
        String repoKey = "local";
        RemoteRepoDescriptor remoteRepoDescriptor = mock(RemoteRepoDescriptor.class);
        when(repositoryService.remoteRepoDescriptorByKey(repoKey)).thenReturn(remoteRepoDescriptor);
        try {
            packageArtifactoryConfigService.getVcsConfigOfRemoteRepo(repoKey);
            Assertions.fail("Expected exception");
        } catch (PackageException e) {
            Assertions.assertEquals(HttpStatus.SC_BAD_REQUEST, e.getStatus());
            Assertions.assertEquals("Remote repository 'local' does not have VCS configuration", e.getPublicReason());
        }
    }

    @Test
    void testGetVcsConfigOfRemoteRepo() throws PackageRepoDescriptorException {
        String repoKey = "remote";
        String githubUrl = "https://github.com";
        VcsConfiguration vcs = new VcsConfiguration();
        vcs.getGit().setDownloadUrl(githubUrl);
        RemoteRepoDescriptor remoteRepoDescriptor = mock(RemoteRepoDescriptor.class);
        when(remoteRepoDescriptor.getVcs()).thenReturn(vcs);
        when(repositoryService.remoteRepoDescriptorByKey(repoKey)).thenReturn(remoteRepoDescriptor);

        PhVcsGitConfiguration vcsGitConfiguration = packageArtifactoryConfigService.getVcsConfigOfRemoteRepo(repoKey);
        Assertions.assertEquals(PhVcsGitProvider.GITHUB, vcsGitConfiguration.getProvider());
        Assertions.assertEquals(githubUrl, vcsGitConfiguration.getDownloadUrl());
    }

    private void prepareTest(String repoKey) {
        Repo repo = mock(Repo.class);
        when(repo.getKey()).thenReturn(repoKey);
        when(repositoryService.repositoryByKey(repoKey)).thenReturn(repo);
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        when(repo.getDescriptor()).thenReturn(localRepoDescriptor);
        when(repo.isReal()).thenReturn(true);
        when(repo.isLocal()).thenReturn(true);
    }
}
