package org.artifactory.ph;

import org.artifactory.common.repo.PackageType;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.service.PackageSpecificationAdapter;
import org.artifactory.descriptor.repo.RepoType;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.Mockito;

import java.util.stream.Stream;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

public class PackageSpecificationImplTest {

    private PackageSpecificationImpl packageIndexChecker;

    @Mock
    private PackageSpecificationAdapter packageSpecificationAdapter;


    @BeforeEach
    private void setup() {
        openMocks(this);
        packageIndexChecker = new PackageSpecificationImpl(packageSpecificationAdapter);
    }

    @Test
    public void testIsExpirable() throws PackageException {
        when(packageSpecificationAdapter.isExpirable(any(), any())).thenReturn(false);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);

        Assertions.assertFalse(packageIndexChecker.isExpirable(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isExpirable(eq(PackageType.CRAN), eq("a/b/c"));
        Mockito.reset(packageSpecificationAdapter);
        when(packageSpecificationAdapter.isExpirable(any(), any())).thenReturn(true);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.isExpirable(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isExpirable(eq(PackageType.CRAN), eq("a/b/c"));
    }

    @Test
    public void testIsLocalGenerated() throws PackageException {
        when(packageSpecificationAdapter.isLocalGenerated(any(), any())).thenReturn(false);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertFalse(packageIndexChecker.isLocalGenerated(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isLocalGenerated(eq(PackageType.CRAN), eq("a/b/c"));
        Mockito.reset(packageSpecificationAdapter);
        when(packageSpecificationAdapter.isLocalGenerated(any(), any())).thenReturn(true);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.isLocalGenerated(RepoType.CRAN, "repoKey", "a/b/c"));
        verify(packageSpecificationAdapter, times(1)).isLocalGenerated(eq(PackageType.CRAN), eq("a/b/c"));
    }

    @Test
    public void testShouldExcludePropertyFromMirroring() throws PackageException {
        when(packageSpecificationAdapter.isLocallyGeneratedProperty(PackageType.CRAN, "repoKey")).thenReturn(false);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertFalse(packageIndexChecker.shouldExcludePropertyFromMirroring(RepoType.CRAN, "repoKey"));
        verify(packageSpecificationAdapter, times(1)).isLocallyGeneratedProperty(eq(PackageType.CRAN), eq("repoKey"));
        Mockito.reset(packageSpecificationAdapter);
        when(packageSpecificationAdapter.isLocallyGeneratedProperty(PackageType.CRAN, "repoKey")).thenReturn(true);
        when(packageSpecificationAdapter.isSupportedPackage(any(PackageType.class))).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.shouldExcludePropertyFromMirroring(RepoType.CRAN, "repoKey"));
        verify(packageSpecificationAdapter, times(1)).isLocallyGeneratedProperty(eq(PackageType.CRAN), eq("repoKey"));
    }

    @Test
    public void testIsPackageArtifact() throws PackageException {
        when(packageSpecificationAdapter.isSupportedPackage(any(String.class))).thenReturn(true);
        Assertions.assertFalse(packageIndexChecker.isPackageArtifact("repoKey", "path"));
        verify(packageSpecificationAdapter, times(1)).isPackageArtifact(eq("repoKey"), eq("path"));
        Mockito.reset(packageSpecificationAdapter);

        when(packageSpecificationAdapter.isSupportedPackage(any(String.class))).thenReturn(true);
        when(packageSpecificationAdapter.isPackageArtifact("repoKey", "path")).thenReturn(true);
        Assertions.assertTrue(packageIndexChecker.isPackageArtifact("repoKey", "path"));
        verify(packageSpecificationAdapter, times(1)).isPackageArtifact(eq("repoKey"), eq("path"));
    }

    @ParameterizedTest
    @MethodSource("testIsPackageArtifactFallbackProvider")
    public void testIsPackageArtifactFallback(RepoType repoType, String path, boolean expectedResult)
            throws PackageException {
        when(packageSpecificationAdapter.isSupportedPackage(any(String.class))).thenReturn(false);
        Assertions.assertEquals(expectedResult, packageIndexChecker.isPackageArtifact("repoKey", path, repoType));
        verify(packageSpecificationAdapter, never()).isPackageArtifact(eq("repoKey"), eq(path));
    }

    private static Stream<Arguments> testIsPackageArtifactFallbackProvider() {
        return Stream.of(Arguments.of(RepoType.Npm, "path.tgz", true),
                Arguments.of(RepoType.Npm, "path.zip", false),
                Arguments.of(RepoType.NuGet, "path.tgz", false),
                Arguments.of(RepoType.NuGet, "path.nupkg", true)
        );
    }

}