package org.artifactory.projects;

import org.artifactory.aol.AolService;
import org.artifactory.api.common.BasicStatusHolder;
import org.artifactory.api.common.MoveMultiStatusHolder;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.common.StatusEntry;
import org.artifactory.model.xstream.security.UserGroupImpl;
import org.artifactory.projects.exception.AccessClientStateException;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.schedule.TaskService;
import org.artifactory.security.SimpleUser;
import org.artifactory.security.UserInfo;
import org.artifactory.security.access.AccessService;
import org.artifactory.storage.StorageService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.group.GroupsClient;
import org.jfrog.access.client.project.GrpcProjectClient;
import org.jfrog.access.client.project.ProjectEntity;
import org.jfrog.access.client.project.ProjectMembership;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.access.client.project.ProjectSlimEntity;
import org.jfrog.access.client.project.RoleActions;
import org.jfrog.access.client.project.StorageQuotaEntity;
import org.jfrog.access.client.user.UsersClient;
import org.jfrog.access.common.ResourceType;
import org.jfrog.access.common.RoleConstants;
import org.jfrog.access.rest.user.User;
import org.jfrog.access.rest.user.Users;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.context.SecurityContextImpl;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;

import static org.artifactory.projects.ProjectsService.DEFAULT_PROJECT;
import static org.artifactory.projects.ProjectsService.UNLIMITED_STORAGE;
import static org.artifactory.projects.ProjectsUtils.PROJ_LIMIT_MSG;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.ArgumentMatchers.argThat;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.clearInvocations;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author BarakH
 */
public class ProjectsServiceImplTest extends ArtifactoryHomeBoundTest {

    private static final String PROJECT_A = "proj1";
    private static final String PROJECT_B = "proj2";
    private static final String PROJECT_C = "proj3";
    private static final String USER = "user";
    private static final String ADMIN = "admin";
    private static final RepoPath REPO_A = RepoPathFactory.create("repo1");
    private static final RepoPath REPO_B = RepoPathFactory.create("repo2");
    private static final ProjectSlimEntity PROJECT_A_ENTITY =
            ProjectSlimEntity.builder().name(PROJECT_A).projectKey(PROJECT_A).build();
    private static final ProjectSlimEntity PROJECT_C_ENTITY =
            ProjectSlimEntity.builder().name(PROJECT_C).build();
    @Mock
    private GrpcProjectClient grpcProjectClient;
    @Mock
    private AccessService accessService;
    @Mock
    private SecurityContextImpl securityContext;
    @Mock
    private StorageService storageService;
    @Mock
    private AccessClient accessClient;
    @Mock
    private TaskService taskService;
    @Mock
    private AolService aolService;
    @Mock
    private ArtifactoryContext artifactoryContext;
    private ProjectsServiceImpl projectsService;

    private static SimpleUser createNonAdminUser(boolean isTransient, String... groups) {
        UserInfo userInfo;
        if (!isTransient) {
            userInfo = new UserInfoBuilder(ProjectsServiceImplTest.USER).updatableProfile(true)
                    .internalGroups(new HashSet<>(Arrays.asList(groups))).build();
        } else {
            userInfo = new UserInfoBuilder(ProjectsServiceImplTest.USER).updatableProfile(true)
                    .groups(Set.of(new UserGroupImpl(groups[0]))).transientUser().build();
        }
        return new SimpleUser(userInfo);
    }

    private static SimpleUser createAdminUser() {
        UserInfo userInfo = new UserInfoBuilder(ADMIN).admin(true).updatableProfile(true).build();
        return new SimpleUser(userInfo);
    }

    @DataProvider
    public static Object[][] provideUserTypeAndLicense() {
        return new Object[][] {
                {UserType.SIMPLE, false},
                {UserType.SIMPLE, false},
                {UserType.PROJECT_ADMIN, true},
                {UserType.ADMIN, true},
                {UserType.TRANSIENT_PROJECT_ADMIN, true},
        };
    }

    @DataProvider
    public static Object[][] provideUserTypeLicensesAndExpectedRepos() {
        return new Object[][] {
                {true, UserType.SIMPLE, Set.of()},
                {true, UserType.PROJECT_ADMIN, Set.of(REPO_A.getRepoKey(), REPO_B.getRepoKey())},
                {false, UserType.PROJECT_ADMIN, Set.of()},
        };
    }

    @DataProvider
    public static Object[][] provideLicenseRepoKeyAndExpectedProject() {
        return new Object[][] {
                {REPO_A.getRepoKey(), PROJECT_A},
                {"", DEFAULT_PROJECT},
                {REPO_B.getRepoKey(), null},
        };
    }

    @DataProvider
    public static Object[][] provideLicenseRepoKeyAndExpectedProjectInternal() {
        return new Object[][] {
                {true, REPO_A.getRepoKey(), PROJECT_A},
                {true, "", DEFAULT_PROJECT},
                {false, REPO_A.getRepoKey(), PROJECT_A},
                {true, REPO_B.getRepoKey(), null},
        };
    }

    @DataProvider
    public static Object[][] provideLicenseRepoConfigAndExpectedError() {
        return new Object[][] {
                {PROJECT_A, REPO_A.getRepoKey(), null, "Repository key: repo1, should start with the " +
                        "projectKey and a dash: 'proj1-'"},
                {DEFAULT_PROJECT, REPO_A.getRepoKey(), null, null},
                {PROJECT_B, REPO_A.getRepoKey(), Set.of(), "Project proj2 does not exist"},
                {PROJECT_A, REPO_A.getRepoKey(), Set.of(), "Repository key: repo1, should start with the " +
                        "projectKey and a dash: 'proj1-'"},
                {PROJECT_A, PROJECT_A + "-" +
                        REPO_A.getRepoKey(), Set.of(), "You must include at least one environment in the repository " +
                        "configuration"},
                {PROJECT_A, PROJECT_A + "-" + REPO_A.getRepoKey(), Set.of(
                        "LIE"), "Cannot set environments that are not part of the project. Project environments: " +
                        "[PROD, DEV] but got: [LIE]"},
        };
    }

    @BeforeMethod
    public void setUp() {
        MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        when(accessService.getAccessClient()).thenReturn(accessClient);
        when(accessClient.cachedGrpcProject()).thenReturn(grpcProjectClient);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);

        securityContext = new SecurityContextImpl();
        SecurityContextHolder.setContext(securityContext);
        projectsService = spy(new ProjectsServiceImpl());
        projectsService.setAccessService(accessService);
        projectsService.setStorageService(storageService);
        projectsService.setTaskService(taskService);
        projectsService.init();
    }

    @AfterMethod
    public void teardown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test(dataProvider = "provideUserTypeAndLicense")
    public void testIsProjectAdmin(UserType userType, boolean expectedResults) {
        SimpleUser simpleUser = null;
        switch (userType) {
            case ADMIN:
                simpleUser = setAdminAuthentication();
                when(grpcProjectClient.isProjectAdmin(ADMIN, PROJECT_A, null)).thenReturn(true);
                break;
            case PROJECT_ADMIN:
                simpleUser = setSimpleUserAuthentication(false);
                when(grpcProjectClient.isProjectAdmin(USER, PROJECT_A, null)).thenReturn(true);
                break;
            case TRANSIENT_PROJECT_ADMIN:
                simpleUser = setSimpleUserAuthentication(true, "externalGroup");
                when(grpcProjectClient.isProjectAdmin(USER, PROJECT_A, List.of("externalGroup"))).thenReturn(true);
                break;
            case SIMPLE:
                simpleUser = setSimpleUserAuthentication(false);
                when(grpcProjectClient.isProjectAdmin(USER, PROJECT_A, null)).thenReturn(false);
                break;
        }
        assertEquals(projectsService.isProjectAdmin(PROJECT_A, simpleUser), expectedResults);
    }

    @Test
    public void testIsResourceManager() {
        SimpleUser simpleUser = setSimpleUserAuthentication(false);
        when(grpcProjectClient.isProjectAdmin(USER, ProjectsService.DEFAULT_PROJECT, null)).thenReturn(true);
        assertTrue(projectsService.isResourceManager(simpleUser));
    }

    @Test(dataProvider = "provideUserTypeAndLicense")
    public void testCanMangeProjectResources(UserType userType, boolean expectedResults) {
        SimpleUser simpleUser = null;
        switch (userType) {
            case ADMIN:
                simpleUser = setAdminAuthentication();
                when(grpcProjectClient.canDoAction(ADMIN, PROJECT_A, null, RoleActions.MANAGE_RESOURCES))
                        .thenReturn(true);
                break;
            case PROJECT_ADMIN:
                simpleUser = setSimpleUserAuthentication(false);
                when(grpcProjectClient.canDoAction(USER, PROJECT_A, null, RoleActions.MANAGE_RESOURCES))
                        .thenReturn(true);
                break;
            case TRANSIENT_PROJECT_ADMIN:
                simpleUser = setSimpleUserAuthentication(true, "externalGroup");
                when(grpcProjectClient
                        .canDoAction(USER, PROJECT_A, List.of("externalGroup"), RoleActions.MANAGE_RESOURCES))
                        .thenReturn(true);
                break;
            case SIMPLE:
                simpleUser = setSimpleUserAuthentication(false);
                when(grpcProjectClient.canDoAction(USER, PROJECT_A, null, RoleActions.MANAGE_RESOURCES))
                        .thenReturn(false);
                break;
        }
        assertEquals(projectsService.canManageProjectResources(PROJECT_A, simpleUser), expectedResults);
    }

    @Test(dataProvider = "provideUserTypeAndLicense")
    public void testGetUserProjects(UserType userType, boolean projectShouldBeExist) {
        switch (userType) {
            case ADMIN:
                setAdminAuthentication();
                when(grpcProjectClient.getAllProjectKeys())
                        .thenReturn(List.of(ProjectsService.DEFAULT_PROJECT, PROJECT_A));
                break;
            case PROJECT_ADMIN:
                setSimpleUserAuthentication(false);
                when(grpcProjectClient.getAllProjectsManagedByUser(USER, null))
                        .thenReturn(List.of(PROJECT_A, ProjectsService.DEFAULT_PROJECT));
                break;
            case TRANSIENT_PROJECT_ADMIN:
                setSimpleUserAuthentication(true, "externalGroup");
                when(grpcProjectClient
                        .getAllProjectsManagedByUser(USER, List.of("externalGroup")))
                        .thenReturn(List.of(PROJECT_A, ProjectsService.DEFAULT_PROJECT));
                break;
            case SIMPLE:
                setSimpleUserAuthentication(false);
                when(grpcProjectClient.getAllProjectsManagedByUser(USER, null))
                        .thenReturn(List.of(ProjectsService.DEFAULT_PROJECT));
                break;
        }

        List<String> userProjects = projectsService.getAllProjectsManagedByUser();
        Assert.assertEquals(projectShouldBeExist, userProjects.contains(PROJECT_A));
        if (UserType.ADMIN.equals(userType) && projectShouldBeExist) {
            assertTrue(userProjects.contains(DEFAULT_PROJECT));
        } else {
            Assert.assertFalse(userProjects.contains(DEFAULT_PROJECT));
        }
    }

    @Test(dataProvider = "provideUserTypeAndLicense")
    public void testCanProjectAdminManageResources(UserType userType, boolean expectedResults) {
        SimpleUser simpleUser = null;
        switch (userType) {
            case ADMIN:
                simpleUser = setAdminAuthentication();
                when(grpcProjectClient.canProjectAdminDoOnResource(REPO_A.getRepoKey(), ResourceType.REPO,
                        List.of(RoleActions.MANAGE_RESOURCES), ADMIN, null))
                        .thenReturn(true);
                break;
            case PROJECT_ADMIN:
                simpleUser = setSimpleUserAuthentication(false);
                when(grpcProjectClient.canProjectAdminDoOnResource(REPO_A.getRepoKey(), ResourceType.REPO,
                        List.of(RoleActions.MANAGE_RESOURCES), USER, null))
                        .thenReturn(true);
                break;
            case TRANSIENT_PROJECT_ADMIN:
                simpleUser = setSimpleUserAuthentication(true, "externalGroup");
                when(grpcProjectClient
                        .canProjectAdminDoOnResource(REPO_A.getRepoKey(), ResourceType.REPO,
                                List.of(RoleActions.MANAGE_RESOURCES), USER, List.of("externalGroup")))
                        .thenReturn(true);
                break;
            case SIMPLE:
                simpleUser = setSimpleUserAuthentication(false);
                when(grpcProjectClient.canProjectAdminDoOnResource(REPO_A.getRepoKey(), ResourceType.REPO,
                        List.of(RoleActions.MANAGE_RESOURCES), USER, null))
                        .thenReturn(false);
                break;
        }
        assertEquals(projectsService.canProjectAdminManageResources(REPO_A.getRepoKey(), simpleUser), expectedResults);
    }

    @Test
    public void testGetProjectRepositories() {
        String repoKey = REPO_A.getRepoKey();
        when(grpcProjectClient.getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false))
                .thenReturn(Map.of(REPO_A.getRepoKey(), ProjectResourceEntity.builder().resourceName(repoKey).build()));

        Set<String> repositories = projectsService.getProjectRepositoriesKeys(PROJECT_A, false);
        assertTrue(repositories.contains(REPO_A.getRepoKey()));
    }

    @Test(dataProvider = "provideLicenseRepoKeyAndExpectedProject")
    public void testGetProjectOfRepository(String repoKey, String expectedProject) {
        when(grpcProjectClient.getProjectForResource(REPO_A.getRepoKey(), ResourceType.REPO))
                .thenReturn(PROJECT_A);
        when(grpcProjectClient.getProjectForResource(REPO_B.getRepoKey(), ResourceType.REPO))
                .thenReturn(null);

        String repoProject = projectsService.getProjectKeyOfRepository(repoKey);
        Assert.assertEquals(expectedProject, repoProject);
    }

    @Test(dataProvider = "provideLicenseRepoConfigAndExpectedError")
    public void testValidateRepoConfigIsMatchingToProjectSpec(String project, String repoKey,
            Set<String> environments, String expectedError) {
        when(grpcProjectClient.getProjectSlimEntity(PROJECT_A)).thenReturn(PROJECT_A_ENTITY);
        when(grpcProjectClient.getProjectSlimEntity(PROJECT_B)).thenReturn(null);
        when(grpcProjectClient.getProjectSlimEntity(PROJECT_C)).thenReturn(PROJECT_C_ENTITY);
        try {
            projectsService.validateRepoConfigIsMatchingToProjectSpec(project, repoKey, environments);
        } catch (IllegalArgumentException e) {
            Assert.assertEquals(e.getMessage(), expectedError);
        }
    }

    @Test
    public void testGetProjectRepositoryEntity() {
        when(grpcProjectClient.getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false))
                .thenReturn(new HashMap<>());

        ProjectResourceEntity projectEntity =
                projectsService.getProjectRepositoryEntity(PROJECT_A, REPO_B.getRepoKey());
        assertNull(projectEntity);
        verify(grpcProjectClient).getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false);
        verifyNoMoreInteractions(grpcProjectClient);
        clearInvocations(grpcProjectClient);
        when(grpcProjectClient.getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false))
                .thenReturn(Map.of(REPO_A.getRepoKey(),
                        ProjectResourceEntity.builder().resourceName(REPO_A.getRepoKey()).environment("dev")
                                .environment("test").build()));

        projectEntity = projectsService.getProjectRepositoryEntity(PROJECT_A, REPO_B.getRepoKey());
        assertNull(projectEntity);
        verify(grpcProjectClient).getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false);
        verifyNoMoreInteractions(grpcProjectClient);
        clearInvocations(grpcProjectClient);

        projectEntity = projectsService.getProjectRepositoryEntity(PROJECT_A, REPO_A.getRepoKey());
        assertNotNull(projectEntity);
        verify(grpcProjectClient).getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false);
        verifyNoMoreInteractions(grpcProjectClient);
    }

    @Test
    public void testGetAllProjectsRecourses() {
        when(grpcProjectClient.getAllProjectKeys()).thenReturn(List.of(PROJECT_A, PROJECT_B));
        when(grpcProjectClient.getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false))
                .thenReturn(Map.of(REPO_A.getRepoKey(),
                        ProjectResourceEntity.builder().resourceName(REPO_A.getRepoKey()).build()));
        when(grpcProjectClient.getProjectResourceEntitiesMap(PROJECT_B, ResourceType.REPO, false))
                .thenReturn(Map.of(REPO_B.getRepoKey(),
                        ProjectResourceEntity.builder().resourceName(REPO_B.getRepoKey()).build()));

        Map<String, Map<String, ProjectResourceEntity>> resourcesMap = projectsService.getAllProjectsRepositoriesMap(
                false);

        assertEquals(resourcesMap.keySet().size(), 2);
        assertEquals(resourcesMap.get(PROJECT_A).get(REPO_A.getRepoKey()).getResourceName(), REPO_A.getRepoKey());
        assertEquals(resourcesMap.get(PROJECT_B).get(REPO_B.getRepoKey()).getResourceName(), REPO_B.getRepoKey());

    }

    @Test
    public void testGetProjectRecourses() {
        when(grpcProjectClient.getProjectResourceEntitiesMap(PROJECT_A, ResourceType.REPO, false))
                .thenReturn(Map.of(REPO_A.getRepoKey(),
                        ProjectResourceEntity.builder().resourceName(REPO_A.getRepoKey()).build()));

        Map<String, ProjectResourceEntity> resourcesMap = projectsService.getProjectRepositoriesMap(PROJECT_A, false);
        assertEquals(resourcesMap.keySet().size(), 1);
        assertEquals(resourcesMap.get(REPO_A.getRepoKey()).getResourceName(), REPO_A.getRepoKey());
    }

    @Test
    public void testAssertProjectStorageQuotaForDefaultProject() {
        String repoKey = "generic";
        BasicStatusHolder statusHolder = new BasicStatusHolder();

        // when the storage quota feature is disabled
        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(false);
        assertFalse(statusHolder.hasErrors());
        verify(projectsService, never()).getProjectKeyOfRepository(anyString());
        verify(projectsService, never()).getProjectStorageQuotaBytes(anyString());

        // when the storage quota feature is enabled
        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        doReturn(DEFAULT_PROJECT).when(projectsService).getProjectKeyOfRepository(repoKey);
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, Long.MAX_VALUE);
        assertFalse(statusHolder.hasErrors());
        verify(projectsService).getProjectKeyOfRepository(repoKey);
        verify(projectsService, never()).getProjectStorageQuotaBytes(anyString());
    }

    @Test
    public void testAssertProjectStorageQuotaForUnlimitedQuota() {
        // given
        String repoKey = "proj-generic";
        String projectName = "the_project";
        BasicStatusHolder statusHolder = new BasicStatusHolder();

        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        doReturn(projectName).when(projectsService).getProjectKeyOfRepository(repoKey);
        doReturn(UNLIMITED_STORAGE).when(projectsService).getProjectStorageQuotaBytes(projectName);

        // when
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, Long.MAX_VALUE);

        // then
        assertFalse(statusHolder.hasErrors());
        verify(projectsService).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectName);
    }

    @Test
    public void testAssertProjectStorageQuotaForLimitedQuota() {
        // given
        String repoKey = "proj-generic";
        String projectKey = "the_project";
        ProjectSlimEntity projectEntity =
                ProjectSlimEntity.builder().storageQuota(StorageQuotaEntity.builder().isSoftLimit(false).build())
                        .build();
        BasicStatusHolder statusHolder = new BasicStatusHolder();

        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        doReturn(projectKey).when(projectsService).getProjectKeyOfRepository(repoKey);
        doReturn(100L).when(projectsService).getProjectStorageQuotaBytes(projectKey);
        doReturn(projectEntity).when(projectsService).getProjectSpec(projectKey);
        when(storageService.getProjectStorageUsage(projectKey)).thenReturn(50L);

        // when limit not yet reached
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, 49L);

        // then no errors
        assertFalse(statusHolder.hasErrors());
        verify(projectsService, times(2)).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectKey);
        verify(projectsService, times(2)).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectKey);
        clearInvocations(projectsService, storageService);

        // when limit reached and soft limit false
        statusHolder = new BasicStatusHolder();
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, 100L);

        // then an error message expected
        assertTrue(statusHolder.hasErrors());
        assertEquals(statusHolder.getLastError().getMessage(), PROJ_LIMIT_MSG);
        verify(projectsService, times(2)).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectKey);
        clearInvocations(projectsService);

        // when limit reached and soft limit true
        projectEntity =
                ProjectSlimEntity.builder().storageQuota(StorageQuotaEntity.builder().isSoftLimit(true).build())
                        .build();
        doReturn(projectEntity).when(projectsService).getProjectSpec(projectKey);
        statusHolder = new BasicStatusHolder();
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, 100L);
        StatusEntry statusEntry = statusHolder.getLastStatusEntry();

        // then no warnings are expected
        assertFalse(statusHolder.hasErrors());
        assertFalse(statusHolder.hasWarnings());
        verify(projectsService, times(2)).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectKey);
    }

    @Test
    public void testAssertProjectStorageQuotaWhenAOLAndHardLimit() {
        String repoKey = "proj-generic";
        String projectKey = "projx";
        ProjectSlimEntity
                projectEntity =
                ProjectSlimEntity.builder().storageQuota(StorageQuotaEntity.builder().isSoftLimit(false).build())
                        .build();
        BasicStatusHolder statusHolder = new BasicStatusHolder();

        when(aolService.isAol()).thenReturn(true);
        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        doReturn(projectKey).when(projectsService).getProjectKeyOfRepository(repoKey);
        doReturn(100L).when(projectsService).getProjectStorageQuotaBytes(projectKey);
        doReturn(projectEntity).when(projectsService).getProjectSpec(projectKey);
        when(storageService.getProjectStorageUsage(projectKey)).thenReturn(50L);

        // when limit reached and soft limit false, but is AOL
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, 100L);
        StatusEntry statusEntry = statusHolder.getLastStatusEntry();

        // then no warnings are expected
        assertFalse(statusHolder.hasErrors());
        assertFalse(statusHolder.hasWarnings());

        verify(projectsService, times(2)).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectKey);
        verify(aolService).isAol();
    }

    @Test(description = "checks a scenario of copy/move operations called by the scenario")
    public void testAssertProjectStorageQuotaWhenUsingMoveMultiStatusHolder() {
        String repoKey = "proj-generic";
        String projectName = "the_project";
        ProjectSlimEntity
                projectEntity =
                ProjectSlimEntity.builder().storageQuota(StorageQuotaEntity.builder().isSoftLimit(false).build())
                        .build();
        MoveMultiStatusHolder statusHolder = new MoveMultiStatusHolder();

        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        doReturn(projectName).when(projectsService).getProjectKeyOfRepository(repoKey);
        doReturn(100L).when(projectsService).getProjectStorageQuotaBytes(projectName);
        doReturn(projectEntity).when(projectsService).getProjectSpec(projectName);
        when(storageService.getProjectStorageUsage(projectName)).thenReturn(50L);

        // when limit reached and hard limit true
        projectsService.assertProjectStorageQuota(repoKey, statusHolder, 100L);

        // then a warning is expected
        assertTrue(statusHolder.hasErrors());
        assertNull(statusHolder.getLastWarning());
        verify(projectsService, times(2)).getProjectKeyOfRepository(repoKey);
        verify(projectsService).getProjectStorageQuotaBytes(projectName);
    }

    @Test
    public void testGetProjectAdmins() {
        // given
        UsersClient usersClient = mock(UsersClient.class);
        when(accessClient.users()).thenReturn(usersClient);
        GroupsClient groupsClient = mock(GroupsClient.class);
        when(accessClient.groups()).thenReturn(groupsClient);

        User adminUser = mock(User.class);
        when(usersClient.findUserByUsername("project_admin")).thenReturn(Optional.of(adminUser));
        Users groupUsersResponse = mock(Users.class);
        List<User> adminGroupUsers = List.of(mock(User.class));
        when(groupUsersResponse.getUsers()).thenReturn(adminGroupUsers);
        when(groupsClient.findGroupUsers("admin_group")).thenReturn(groupUsersResponse);

        ProjectEntity project = ProjectEntity.builder()
                .user("not_admin_user", ProjectMembership.builder().role(RoleConstants.CONTRIBUTOR).build())
                .user("project_admin", ProjectMembership.builder()
                        .role(RoleConstants.DEVELOPER)
                        .role(RoleConstants.PROJECT_ADMIN).build())
                .group("not_admin_group", ProjectMembership.builder().role(RoleConstants.CONTRIBUTOR).build())
                .group("admin_group", ProjectMembership.builder()
                        .role(RoleConstants.DEVELOPER)
                        .role(RoleConstants.PROJECT_ADMIN).build())
                .build();

        // when
        Set<User> foundProjectAdminUsers = projectsService.getProjectAdmins(project);
        assertEquals(foundProjectAdminUsers.size(), 2);
        assertTrue(foundProjectAdminUsers.containsAll(List.of(adminUser, adminGroupUsers.get(0))));

        // then
        verify(usersClient).findUserByUsername("project_admin");
        verify(groupsClient).findGroupUsers("admin_group");
        verifyNoMoreInteractions(usersClient);
        verifyNoMoreInteractions(groupsClient);
    }

    @Test
    public void testStorageQuotaPreconditionsAndGetProjQuotaSize() {
        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(false);
        assertEquals(projectsService.getProjectQuotaSizeByRepository(REPO_A.getRepoKey()),
                ProjectsService.UNLIMITED_STORAGE);

        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        when(projectsService.getProjectKeyOfRepository(REPO_A.getRepoKey()))
                .thenReturn(ProjectsService.DEFAULT_PROJECT);
        assertEquals(projectsService.getProjectQuotaSizeByRepository(REPO_A.getRepoKey()),
                ProjectsService.UNLIMITED_STORAGE);

        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        when(projectsService.getProjectKeyOfRepository(REPO_A.getRepoKey())).thenReturn(PROJECT_A);
        doReturn(UNLIMITED_STORAGE).when(projectsService).getProjectStorageQuotaBytes(PROJECT_A);
        assertEquals(projectsService.getProjectQuotaSizeByRepository(REPO_A.getRepoKey()),
                ProjectsService.UNLIMITED_STORAGE);

        when(storageService.isProjectsStorageQuotaEnabled()).thenReturn(true);
        when(projectsService.getProjectKeyOfRepository(REPO_A.getRepoKey())).thenReturn(PROJECT_A);
        when(projectsService.getProjectStorageQuotaBytes(PROJECT_A)).thenReturn(10L);
        assertEquals(projectsService.getProjectQuotaSizeByRepository(REPO_A.getRepoKey()), 10);
    }

    @Test
    public void testRemoveRepositoryFromProject() {
        when(grpcProjectClient.getProjectForResource(REPO_A.getRepoKey(), ResourceType.REPO))
                .thenReturn(PROJECT_A);
        projectsService.removeRepositoryFromCurrentProject(REPO_A.getRepoKey());
        verify(grpcProjectClient).getProjectForResource(REPO_A.getRepoKey(), ResourceType.REPO);
        verify(grpcProjectClient).removeProjectResource(PROJECT_A, ResourceType.REPO, REPO_A.getRepoKey());
        verifyNoMoreInteractions(grpcProjectClient);
    }

    @Test
    public void testRemoveRepositoryFromDefault() {
        when(grpcProjectClient.getProjectForResource(REPO_B.getRepoKey(), ResourceType.REPO))
                .thenReturn(DEFAULT_PROJECT);
        projectsService.removeRepositoryFromCurrentProject(REPO_B.getRepoKey());
        verify(grpcProjectClient).getProjectForResource(REPO_B.getRepoKey(), ResourceType.REPO);
        verify(grpcProjectClient).removeProjectResource(DEFAULT_PROJECT, ResourceType.REPO, REPO_B.getRepoKey());
        verifyNoMoreInteractions(grpcProjectClient);
    }


    @Test
    public void testgetRepoKeyToProjectMap() {
        // original data
        Set<String> repoKeys = Set.of("proj-remote-cache");
        Set<String> repoKeysExpected = Set.of("proj-remote");
        Map<String, String> repoToProjMap = new HashMap<>();
        // expected data
        repoToProjMap.put("proj-remote", "project");
        when(grpcProjectClient.getProjectForResourceMap(repoKeysExpected, ResourceType.REPO, false))
                .thenReturn(repoToProjMap);

        Map<String, String> repoKeyToProjectMap = projectsService.getRepoKeyToProjectMap(repoKeys, false);

        assertEquals(repoKeyToProjectMap.size(), 1);
        assertEquals(repoKeyToProjectMap.get("proj-remote"), "project");
        verify(grpcProjectClient).getProjectForResourceMap(repoKeysExpected, ResourceType.REPO, false);
        verifyNoMoreInteractions(grpcProjectClient);
    }

    @Test
    public void testGetAllProjects() {
        when(grpcProjectClient.getAllProjectKeys()).thenReturn(List.of(PROJECT_A));

        List<String> projectKeys = projectsService.getAllProjects();
        assertEquals(projectKeys.size(), 1);
        assertTrue(projectKeys.contains(PROJECT_A));
    }

    @Test
    public void testGetAllProjectsEntities() {
        ProjectEntity projectEntity = ProjectEntity.builder().projectKey(PROJECT_A).build();
        when(grpcProjectClient.getAllProjects()).thenReturn(List.of(projectEntity));

        List<ProjectEntity> projectEntities = projectsService.getAllProjectsEntities();
        assertEquals(projectEntities.size(), 1);
        assertTrue(projectEntities.contains(projectEntity));
    }

    @Test
    public void verifyProjectStorageQuotaNotificationJobStarted() {
        projectsService.onContextCreated();
        verify(taskService).startTask(
                argThat((task) -> task.getType().equals(ProjectStorageQuotaNotificationJob.class)), eq(false));
    }

    @Test(expectedExceptions = AccessClientStateException.class,
            expectedExceptionsMessageRegExp = "Failed to acquire Access project gRPC client. " +
                    "Access might not be initiated")
    public void testNpeWhenGettingGrpcClient() {
        when(accessService.getAccessClient()).thenReturn(null);
        projectsService.getGrpcProjectClient();
    }

    private SimpleUser setAdminAuthentication() {
        SimpleUser adminUser = createAdminUser();
        UsernamePasswordAuthenticationToken authenticationToken = new UsernamePasswordAuthenticationToken(
                adminUser, null, SimpleUser.ADMIN_GAS);
        securityContext.setAuthentication(authenticationToken);
        return adminUser;
    }

    private SimpleUser setSimpleUserAuthentication(boolean isTransient, String... groups) {
        SimpleUser simpleUser = createNonAdminUser(isTransient, groups);
        UsernamePasswordAuthenticationToken authenticationToken = new UsernamePasswordAuthenticationToken(
                simpleUser, "password", SimpleUser.USER_GAS);
        securityContext.setAuthentication(authenticationToken);
        return simpleUser;
    }

    private enum UserType {SIMPLE, PROJECT_ADMIN, TRANSIENT_PROJECT_ADMIN, ADMIN}
}