package org.artifactory.repo.cache;

import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;

import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Stream;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.Mockito.mock;

class RepoConfigCacheLoaderTest {

    @ParameterizedTest
    @MethodSource("getAllOptimizationRules")
    public void testGetAllOptimizationRules(int keysToLoad, int allRepoKeys, int batchSize, double optimizationRatio,
            boolean expectedResult) {
        RepoConfigCacheLoader cacheLoader = new RepoConfigCacheLoader(batchSize, optimizationRatio,
                () -> allRepoKeys, mock(Supplier.class), mock(Function.class), mock(Function.class));

        assertEquals(expectedResult, cacheLoader.shouldGetAllOptimizationBeApplied(keysToLoad));
    }

    private static Stream<Arguments> getAllOptimizationRules() {
        return Stream.of(
                /**
                 * keys to fetch = 1, total keys = 100, batch = 5
                 * keys to fetch <= batch
                 */
                Arguments.of(1, 100, 5, Double.MAX_VALUE, false),
                /**
                 * keys to fetch = 1, total keys = 100, batch = 5
                 * keys to fetch <= batch
                 */
                Arguments.of(1, 100, 5, Double.MAX_VALUE, false),
                /**
                 * keys to fetch = 1, total keys = 100, batch = 5
                 * allowed ration is 0 - means always apply optimization
                 */
                Arguments.of(1, 100, 5, 0, true),
                /**
                 * keys to fetch = 5, total keys = 100, batch = 5
                 * keys to fetch <= batch
                 */
                Arguments.of(5, 100, 5, Double.MAX_VALUE, false),
                /**
                 * keys to fetch = 10, total keys = 100, batch = 5
                 * 2 batches - useful, 18 batches - unuseful
                 * ratio = 2/18 = 0.11, optimization ratio = 1, getAll is not effective
                 */
                Arguments.of(10, 100, 5, 1.0, false),
                /**
                 * keys to fetch = 50, total keys = 100, batch = 5
                 * 10 batches - useful, 10 batches - unuseful
                 * ratio = 10/10 = 1, optimization ratio = 1, getAll is effective
                 */
                Arguments.of(50, 100, 5, 1.0, true),
                /**
                 * keys to fetch = 51, total keys = 100, batch = 5
                 * 11 batches - useful, 9 batches - unuseful
                 * ratio = 11/9 = 1.22, optimization ratio = 1, getAll is effective
                 */
                Arguments.of(51, 100, 5, 1.0, true),
                /**
                 * keys to fetch = 51, total keys = 100, batch = 5
                 * 11 batches - useful, 9 batches - unuseful
                 * ratio = 11/9 = 1.22, optimization ratio = 1.25, getAll is not effective
                 */
                Arguments.of(51, 100, 5, 1.25, false),
                /**
                 * keys to fetch = 60, total keys = 100, batch = 5
                 * 12 batches - useful, 8 batches - unuseful
                 * ratio = 12/8 = 1.5, optimization ratio = 1, getAll is effective
                 */
                Arguments.of(60, 100, 5, 1.0, true),
                /**
                 * keys to fetch = 60, total keys = 100, batch = 5
                 * do not apply optimization if allowed ratio < 0
                 */
                Arguments.of(60, 100, 5, -1, false),
                /**
                 * keys to fetch = 99, total keys = 100, batch = 5
                 * 20 batches - useful, 0 batches - unuseful
                 * if 0 unuseful batches - getAll is effective
                 */
                Arguments.of(99, 100, 5, Double.MAX_VALUE, true),
                /**
                 * keys to fetch = 99, total keys = 100, batch = 5
                 * 20 batches - useful, 0 batches - unuseful
                 * if 0 unuseful batches - getAll is effective
                 */
                Arguments.of(99, 100, 5, Double.MAX_VALUE, true),
                /**
                 * keys to fetch = 100, total keys = 100, batch = 5
                 * 20 batches - useful, 0 batches - unuseful
                 * if 0 unuseful batches - getAll is effective
                 */
                Arguments.of(100, 100, 5, Double.MAX_VALUE, true),
                /**
                 * keys to fetch = 110, total keys = 100, batch = 5
                 * number of keys > total keys (almost impossible) - getAll is effective
                 */
                Arguments.of(110, 100, 5, Double.MAX_VALUE, true),
                /**
                 * keys to fetch = 10_000, total keys = 50_000, batch = 500
                 * 20 batches - useful, 80 batches - unuseful
                 * ratio = 20/80 = 0.25, optimization ratio = 0.5, getAll is not effective
                 */
                Arguments.of(10000, 50000, 500, 0.5, false),
                /**
                 * keys to fetch = 20_000, total keys = 50_000, batch = 500
                 * 40 batches - useful, 60 batches - unuseful
                 * ratio = 40/60 = 0.25, optimization ratio = 0.66, getAll is effective
                 */
                Arguments.of(20000, 50000, 500, 0.5, true),
                /**
                 * keys to fetch = 24_500, total keys = 50_000, batch = 500
                 * 49 batches - useful, 51 batches - unuseful
                 * ratio = 49/51 = 0.25, optimization ratio = 0.96, getAll is not effective
                 */
                Arguments.of(24500, 50000, 500, 1.0, false),
                /**
                 * keys to fetch = 25_000, total keys = 50_000, batch = 500
                 * 50 batches - useful, 50 batches - unuseful
                 * ratio = 50/50 = 1, optimization ratio = 1, getAll is effective
                 */
                Arguments.of(25000, 50000, 500, 1.0, true),
                /**
                 * keys to fetch = 14, total keys = 121, batch = 5
                 * 3 batches - useful, 22 batches - unuseful
                 * ratio = 3/22 = 0.13, optimization ratio = 0.2, getAll is not effective
                 */
                Arguments.of(14, 121, 5, 0.2, false),
                /**
                 * keys to fetch = 68, total keys = 121, batch = 5
                 * 14 batches - useful, 11 batches - unuseful
                 * ratio = 14/11 = 1.27, optimization ratio = 1.2, getAll is not effective
                 */
                Arguments.of(68, 121, 5, 1.2, true)
        );
    }
}