package org.artifactory.repo.interceptor.config;

import org.artifactory.addon.AddonType;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.webstart.ArtifactWebstartAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.common.ConstantValues;
import org.artifactory.config.ConfigurationException;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.http.DenyListCacheHolderService;
import org.artifactory.request.UrlVerifier;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.test.ArtifactoryHomeStub;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Map;

import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_CREATE;
import static org.artifactory.repo.RepoConfigChange.ConfigAction.REPO_UPDATE;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.fail;
import static org.testng.AssertJUnit.assertTrue;

/**
 * @author Tamar Rosen
 */
public class RemoteReposConfigInterceptorTest extends ArtifactoryHomeBoundTest {

    private AddonsManager addonsManager;

    @BeforeMethod
    public void setup() {
        InternalArtifactoryContext contextMock = mock(InternalArtifactoryContext.class);
        AolService aolService = mock(AolService.class);
        when(aolService.isAol()).thenReturn(true);
        when(contextMock.beanForType(AolService.class)).thenReturn(aolService);
        ArtifactoryContextThreadBinder.bind(contextMock);
        addonsManager = mock(AddonsManager.class);
    }

    @Test
    public void testBlacklistedUrls() {
        // given - repo with whitelist url
        getBound().setProperty(ConstantValues.whitelistRemoteRepoUrls, "http://localhost");
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig remoteRepoConfig =
                RemoteRepoConfig.create("Remote-http://localhost:80/artifactory", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://localhost:80/artifactory");
        // when
        interceptor.onBeforeSave(Map.of(REMOTE,
                List.of(new RepoConfigChange("Remote-http://localhost:80/artifactory", REPO_CREATE, remoteRepoConfig,
                        null))));
        // then - save will be ok
        assertFalse(remoteRepoConfig.getRepoTypeConfig().isBlackedOut());

        // given
        getBound().setProperty(ConstantValues.whitelistRemoteRepoUrls, null);
        interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig googleRemoteRepoConfig = RemoteRepoConfig.create("Remote-http://google.com/artifactory", RepoType.Generic);
        googleRemoteRepoConfig.getRepoTypeConfig().setUrl("http://google.com/artifactory");
        // when - update repo with blacklisted url + create repo with whitelisted url
        interceptor.onBeforeSave(Map.of(REMOTE,
                List.of(new RepoConfigChange("Remote-http://localhost:80/artifactory", REPO_UPDATE, remoteRepoConfig,
                                remoteRepoConfig),
                        new RepoConfigChange("Remote-http://google.com/artifactory", REPO_CREATE, googleRemoteRepoConfig,
                                null))));
        // then - save will be ok - but repository should be blacked out
        assertTrue(remoteRepoConfig.getRepoTypeConfig().isBlackedOut());

        try {
            // when - create repo with blacklisted url
            interceptor.onBeforeSave(Map.of(REMOTE,
                    List.of(new RepoConfigChange("Remote-http://localhost:80/artifactory", REPO_CREATE, remoteRepoConfig,
                            null))));
        } catch (IllegalArgumentException expected) {
            // this is good!
            assertEquals(expected.getMessage(), "Found a remote repository(ies) containing blacklisted URLs");
            return;
        }
        fail("Should have found a remote repository(ies) containing blacklisted URLs");
    }

    @Test(expectedExceptions = ConfigurationException.class,
            expectedExceptionsMessageRegExp = "Found a remote repository with a missing URL")
    public void testAssertNoMissingUrlInRemoteReposEnabledException() {
        homeStub.setProperty(ConstantValues.centralConfigRemoteRepoUrlValidation, "true");
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig remoteRepoConfig1 = RemoteRepoConfig.create("cool-remote1", RepoType.Generic);
        remoteRepoConfig1.getRepoTypeConfig().setUrl("http://cool1.com");

        RemoteRepoConfig remoteRepoConfig2 = RemoteRepoConfig.create("cool-remote2", RepoType.Generic);
        remoteRepoConfig2.getRepoTypeConfig().setUrl(null);

        interceptor.onBeforeSave(
                Map.of(REMOTE, List.of(new RepoConfigChange("cool-remote1", REPO_CREATE, remoteRepoConfig1, null),
                        new RepoConfigChange("cool-remote2", REPO_CREATE, remoteRepoConfig2, null))));
    }

    @Test
    public void testCleanClientCertificate() {
        // given - webstartAddon is supported
        when(addonsManager.isAddonSupported(AddonType.WEBSTART)).thenReturn(true);
        ArtifactWebstartAddon webstartAddon = mock(ArtifactWebstartAddon.class);
        // given - addon doesn't know any ssl certificate names
        when(webstartAddon.getSslCertNames()).thenReturn(List.of());
        when(addonsManager.addonByType(ArtifactWebstartAddon.class)).thenReturn(webstartAddon);
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setClientTlsCertificate("doesn't exist");
        interceptor.onBeforeSave(
                Map.of(REMOTE, List.of(new RepoConfigChange("remote-repo", REPO_CREATE, remoteRepoConfig, null))));
        assertNull(remoteRepoConfig.getRepoTypeConfig().getClientTlsCertificate());
    }

    @Test
    public void testDoNotCleanClientCertificateIfWebstartIsDisabled() {
        // given - webstartAddon is supported
        when(addonsManager.isAddonSupported(AddonType.WEBSTART)).thenReturn(false);
        ArtifactWebstartAddon webstartAddon = mock(ArtifactWebstartAddon.class);
        // given - addon doesn't know any ssl certificate names
        when(webstartAddon.getSslCertNames()).thenReturn(List.of());
        when(addonsManager.addonByType(ArtifactWebstartAddon.class)).thenReturn(webstartAddon);
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setClientTlsCertificate("doesn't exist");
        interceptor.onBeforeSave(
                Map.of(REMOTE, List.of(new RepoConfigChange("remote-repo", REPO_CREATE, remoteRepoConfig, null))));
        assertEquals(remoteRepoConfig.getRepoTypeConfig().getClientTlsCertificate(), "doesn't exist");
    }

    @Test(expectedExceptions = IllegalArgumentException.class)
    public void testStrictPolicy() {
        // strict policy will fail by default, in cases of unresolvable URLs
        getBound().setProperty(ConstantValues.remoteRepoBlockUrlStrictPolicy, Boolean.toString(true));
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);

        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("Remote-" + "\\o//", RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setUrl("\\o//");

        interceptor.onBeforeSave(Map.of(REMOTE,
                List.of(new RepoConfigChange("Remote-" + "\\o//", REPO_CREATE, remoteRepoConfig, null))));
    }

    @Test
    public void testBlockedUrlsChangeUrl() {
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig oldRemoteRepo = createRemoteRepo(true);
        RemoteRepoConfig newRemoteRepo = createRemoteRepo(true);
        newRemoteRepo.getRepoTypeConfig().setUrl("http://google.com/artifactory");

        interceptor.onBeforeSave(Map.of(REMOTE,
                List.of(new RepoConfigChange("Repo-Valid-Cert-true", REPO_UPDATE, newRemoteRepo, oldRemoteRepo))));

        interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);

        newRemoteRepo.getRepoTypeConfig().setUrl("http://localhost:8080/artifactory");
        try {
            interceptor.onBeforeSave(Map.of(REMOTE,
                    List.of(new RepoConfigChange("Repo-Valid-Cert-true", REPO_UPDATE, newRemoteRepo, oldRemoteRepo))));
        } catch (IllegalArgumentException expected) {
            // this is good!
            assertEquals(expected.getMessage(), "Found a remote repository(ies) containing blacklisted URLs");
            return;
        }
        fail("Update URL should have been blocked!");
    }

    @Test(dataProvider = "testAssertNoMissingUrlInRemoteReposOtherDataProvider")
    public void testAssertNoMissingUrlInRemoteReposOther(String enabled, String url1) {
        homeStub.setProperty(ConstantValues.centralConfigRemoteRepoUrlValidation, enabled);
        RemoteReposConfigInterceptor interceptor =
                new RemoteReposConfigInterceptor(new UrlVerifier(new DenyListCacheHolderService()), addonsManager);
        RemoteRepoConfig remoteRepo1 = RemoteRepoConfig.create("cool-remote1", RepoType.Generic);
        remoteRepo1.getRepoTypeConfig().setUrl(url1);
        RemoteRepoConfig remoteRepo2 = RemoteRepoConfig.create("cool-remote2", RepoType.Generic);
        remoteRepo2.getRepoTypeConfig().setUrl("http://cool2.com");
        interceptor.onBeforeSave(Map.of(REMOTE,
                List.of(new RepoConfigChange("cool-remote1", REPO_CREATE, remoteRepo1, null),
                        new RepoConfigChange("cool-remote2", REPO_CREATE, remoteRepo2, null))));
    }

    @DataProvider
    private Object[][] testAssertNoMissingUrlInRemoteReposOtherDataProvider() {
        return new Object[][] {
                {"true", "http://cool1.com"},
                {"false", "http://cool1.com"},
                {"false", null},
        };
    }

    private RemoteRepoConfig createRemoteRepo(boolean withValidCert) {
        RemoteRepoConfig remoteRepoConfig =
                RemoteRepoConfig.create("Repo-Valid-Cert-" + withValidCert, RepoType.Generic);
        remoteRepoConfig.getRepoTypeConfig().setUrl("http://someurl.com");
        remoteRepoConfig.getRepoTypeConfig().setClientTlsCertificate(withValidCert ? "ValidTls" : "NotValidTLS");
        return remoteRepoConfig;
    }

    protected ArtifactoryHomeStub getBound() {
        return (ArtifactoryHomeStub) ArtifactoryHomeStub.get();
    }
}