package org.artifactory.repo.service;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.Logger;
import ch.qos.logback.classic.spi.ILoggingEvent;
import ch.qos.logback.core.read.ListAppender;
import com.google.common.base.Predicate;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import org.apache.commons.io.FileUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.home.TestArtifactoryHome;
import org.artifactory.config.InternalCentralConfigService;
import org.artifactory.descriptor.backup.BackupDescriptor;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.CentralConfigTestUtils;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.index.IndexerDescriptor;
import org.artifactory.descriptor.replication.LocalReplicationDescriptor;
import org.artifactory.descriptor.replication.RemoteReplicationDescriptor;
import org.artifactory.descriptor.repo.LocalCacheRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.ReverseProxyDescriptor;
import org.artifactory.metrics.MetricsInfo;
import org.artifactory.metrics.MetricsMetadata;
import org.artifactory.metrics.MetricsThreadLocal;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalCacheRepoConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RealRepoConfig;
import org.artifactory.model.ReleaseBundleRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RepoConfig;
import org.artifactory.model.RepoConfigPropagationModel;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.repo.ExtendedRepositoryConfig;
import org.artifactory.repo.PackageTypeConfigConverter;
import org.artifactory.repo.RepoConfigChange;
import org.artifactory.repo.RepoConfigReload;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.repo.RepositoryConfigBaseConverter;
import org.artifactory.repo.cache.RepoConfigCache;
import org.artifactory.repo.interceptor.RepoConfigChangeInterceptor;
import org.artifactory.repo.packagetype.configuration.MavenConfiguration;
import org.artifactory.repo.packagetype.configuration.NpmConfiguration;
import org.artifactory.repo.packagetype.PomCleanupPolicy;
import org.artifactory.repo.service.validation.ArtifactoryVersionValidator;
import org.artifactory.repo.service.validation.ProjectAdminRepoConfigValidator;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.service.ConfigValidationService;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.artifactory.storage.db.repo.service.RepositoryConfigStorageService;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigMapperImpl;
import org.artifactory.test.ArtifactoryHomeBoundJUnitTest;
import org.artifactory.version.ArtifactoryServiceVersionProvider;
import org.artifactory.version.CompoundVersionDetails;
import org.hibernate.validator.internal.util.Contracts;
import org.jfrog.common.config.validation.NotUnicode;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Disabled;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.slf4j.LoggerFactory;
import org.testng.Assert;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.stream.Stream;

import static java.util.stream.Collectors.toList;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.REMOTE;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL;
import static org.artifactory.repo.system.SystemRepos.SUPPORT_BUNDLE_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.TRASH_REPO_KEY;
import static org.artifactory.util.PackageTypeConfigFields.EXTERNAL_DEPENDENCIES_ENABLED;
import static org.artifactory.util.PackageTypeConfigFields.FETCH_JARS_EAGERLY;
import static org.artifactory.util.PackageTypeConfigFields.POM_REPOSITORY_REFERENCES_CLEANUP_POLICY;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyMap;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Tamar Rosen
 */
class RepositoryConfigServiceImplTest extends ArtifactoryHomeBoundJUnitTest {
    private RepositoryConfigMapperImpl repositoryConfigMapper;
    private ArtifactoryContext context;
    private ProjectAdminRepoConfigValidator projectAdminRepoConfigValidator;
    private ArtifactoryVersionValidator artifactoryVersionValidator;
    private InternalCentralConfigService configService;

    public static Stream<Arguments> providerAllRepoConfigTypesWithNonUnicodeKey() {
        return Stream.of(
                Arguments.of(LocalRepoConfig.create("\u0000", RepoType.Generic)),
                Arguments.of(RemoteRepoConfig.create("\u0000", RepoType.Generic)),
                Arguments.of(LocalCacheRepoConfig.create(RemoteRepoConfig.create("\u0000", RepoType.Generic))),
                Arguments.of(VirtualRepoConfig.create("\u0000", RepoType.Generic)),
                Arguments.of(ReleaseBundleRepoConfig.create("\u0000")),
                Arguments.of(FederatedRepoConfig.create("\u0000", RepoType.Generic))
        );
    }

    public static Stream<Arguments> provideRealRepoTypeConfigs() {
        return Stream.of(
                Arguments.of(LocalRepoConfig.create("\u0000", RepoType.Generic)),
                Arguments.of(RemoteRepoConfig.create("\u0000", RepoType.Generic)),
                Arguments.of(LocalCacheRepoConfig.create(RemoteRepoConfig.create("\u0000", RepoType.Generic))),
                Arguments.of(FederatedRepoConfig.create("\u0000", RepoType.Generic)),
                Arguments.of(ReleaseBundleRepoConfig.create("\u0000"))
        );
    }

    public static Stream<Arguments> provideCrudOperationAndExpectedLogMsg() {
        return Stream.of(
                Arguments.of("insert", "Cannot insert repository: repository config cannot be null"),
                Arguments.of("update", "Cannot update repository: repository config cannot be null"),
                Arguments.of("upsert", "Cannot upsert repository: repository config cannot be null"),
                Arguments.of("delete", "Cannot delete repository: repository key cannot be null")
        );
    }

    @BeforeEach
    public void setup() throws Exception {
        createAndBindToArtifactoryHome();
        repositoryConfigMapper = new RepositoryConfigMapperImpl();
        context = mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        configService = mock(InternalCentralConfigService.class);
        when(context.beanForType(InternalCentralConfigService.class)).thenReturn(configService);
        when(context.getCentralConfig()).thenReturn(configService);
        when(configService.getMutableDescriptor()).thenReturn(mock(MutableCentralConfigDescriptor.class));
        AddonsManager addonsManager = mock(AddonsManager.class);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(addonsManager.addonByType(HaAddon.class)).thenReturn(mock(HaAddon.class));
        projectAdminRepoConfigValidator = mock(ProjectAdminRepoConfigValidator.class);
        artifactoryVersionValidator = mock(ArtifactoryVersionValidator.class);
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        when(context.beanForType(InternalRepositoryService.class)).thenReturn(internalRepositoryService);
        when(context.beanForType(ProjectAdminRepoConfigValidator.class)).thenReturn(
                projectAdminRepoConfigValidator);
        Mockito.doNothing().when(internalRepositoryService).onRepoConfigReload(anyMap());
    }

    @AfterEach
    public void afterClass() throws Exception {
        unbindAndCleanHome();
    }

    @ParameterizedTest
    @MethodSource("provideCrudOperationAndExpectedLogMsg")
    public void crudNullRepoConfigShouldBeSwallowedAndLogged(String action, String expectedLogMsg) {
        Logger logger = (Logger) LoggerFactory.getLogger(RepositoryConfigServiceImpl.class);
        Level originalLevel = logger.getLevel();
        try {
            // given
            logger.setLevel(Level.DEBUG);
            ListAppender<ILoggingEvent> listAppender = new ListAppender<>();
            listAppender.start();
            logger.addAppender(listAppender);

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(mock(RepositoryConfigStorageService.class), Set.of());

            switch (action) {
                case "insert" -> repositoryConfigService.insertRepository(null);
                case "update" -> repositoryConfigService.updateRepository(null);
                case "upsert" -> repositoryConfigService.upsertRepository(null);
                case "delete" -> repositoryConfigService.deleteRepository(null);
            }

            // then
            List<ILoggingEvent> logsList = listAppender.list;
            boolean warnLogExists = logsList.stream().anyMatch(iLoggingEvent ->
                    iLoggingEvent.getMessage().equals(expectedLogMsg) &&
                            iLoggingEvent.getLevel().equals(Level.WARN));
            assertTrue(warnLogExists);

            boolean debugLogExists = logsList.stream().anyMatch(iLoggingEvent ->
                    iLoggingEvent.getMessage().equals(expectedLogMsg) &&
                            iLoggingEvent.getLevel().equals(Level.DEBUG));
            assertTrue(debugLogExists);
        } finally {
            logger.setLevel(originalLevel);
        }
    }

    @Test
    void testGetRepositoryConfigByKey() {
        // given
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
        DbRepositoryConfig dbRepositoryConfig = getDbRepositoryConfig(repositoryConfig);

        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepositoryConfig));

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig = repositoryConfigService.getRepoConfigByKey("repoKey");

        // then
        assertNotNull(retrievedRepositoryConfig);
        assertEquals(retrievedRepositoryConfig, repositoryConfig);
    }

    @ParameterizedTest
    @ValueSource(strings = {SUPPORT_BUNDLE_REPO_KEY, TRASH_REPO_KEY, "repo" + LocalCacheRepoDescriptor.PATH_SUFFIX})
    void testCannotGetSupportBundleOrTrashOrCacheRepoConfigs(String repoKey) {
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        assertThrows(IllegalArgumentException.class, () -> repositoryConfigService.getRepoConfigByKey(repoKey));
    }

    @Test
    void findRepositoryConfigsByKeysInSpecificOrder() {
        // given
        String repoKey1 = "repoKey";
        String repoKey2 = "Zopa";
        String repoKey3 = "Superman";
        LocalRepoConfig repositoryConfig1 = LocalRepoConfig.create(repoKey1, RepoType.Generic);
        LocalRepoConfig repositoryConfig2 = LocalRepoConfig.create(repoKey2, RepoType.Generic);
        LocalRepoConfig repositoryConfig3 = LocalRepoConfig.create(repoKey3, RepoType.Generic);
        Set<String> repoKeys = new LinkedHashSet<>();
        repoKeys.add(repoKey1);
        repoKeys.add(repoKey2);
        repoKeys.add(repoKey3);
        List<DbRepositoryConfig> dbRepositoryConfigs = new ArrayList<>();
        Set<LocalRepoConfig> expected = Sets.newLinkedHashSet();
        DbRepositoryConfig dbRepositoryConfig1 = getDbRepositoryConfig(repositoryConfig1);
        DbRepositoryConfig dbRepositoryConfig2 = getDbRepositoryConfig(repositoryConfig2);
        DbRepositoryConfig dbRepositoryConfig3 = getDbRepositoryConfig(repositoryConfig3);
        dbRepositoryConfigs.add(dbRepositoryConfig1);
        dbRepositoryConfigs.add(dbRepositoryConfig3);
        dbRepositoryConfigs.add(dbRepositoryConfig2);
        expected.add(repositoryConfig1);
        expected.add(repositoryConfig2);
        expected.add(repositoryConfig3);
        VirtualRepoConfig virtualRepoConfig = VirtualRepoConfig.create("repo-key", RepoType.Generic);
        virtualRepoConfig.getRepoTypeConfig().setRepositoryRefs(new ArrayList<>(repoKeys));
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(dbRepositoryConfigs);

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        Set<RepoConfig> actual =
                repositoryConfigService.findRepositoryConfigsByKeysInSpecificOrder(virtualRepoConfig);

        // then
        assertNotNull(actual);
        assertEquals(actual, expected);
    }

    @Test
    void findRepositoryConfigsByKeys() {
        // given
        String repoKey1 = "repoKey";
        String repoKey2 = "Zopa";
        String repoKey3 = "Superman";
        LocalRepoConfig repositoryConfig1 = LocalRepoConfig.create(repoKey1, RepoType.Generic);
        LocalRepoConfig repositoryConfig2 = LocalRepoConfig.create(repoKey2, RepoType.Generic);
        LocalRepoConfig repositoryConfig3 = LocalRepoConfig.create(repoKey3, RepoType.Generic);
        Set<String> repoKeys = new LinkedHashSet<>();
        repoKeys.add(repoKey1);
        repoKeys.add(repoKey2);
        repoKeys.add(repoKey3);
        List<DbRepositoryConfig> dbRepositoryConfigs = new ArrayList<>();
        Set<LocalRepoConfig> expected = Sets.newLinkedHashSet();
        DbRepositoryConfig dbRepositoryConfig1 = getDbRepositoryConfig(repositoryConfig1);
        DbRepositoryConfig dbRepositoryConfig2 = getDbRepositoryConfig(repositoryConfig2);
        DbRepositoryConfig dbRepositoryConfig3 = getDbRepositoryConfig(repositoryConfig3);
        dbRepositoryConfigs.add(dbRepositoryConfig1);
        dbRepositoryConfigs.add(dbRepositoryConfig3);
        dbRepositoryConfigs.add(dbRepositoryConfig2);
        expected.add(repositoryConfig1);
        expected.add(repositoryConfig3);
        expected.add(repositoryConfig2);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(dbRepositoryConfigs);

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        Set<RepoConfig> actual =
                repositoryConfigService.findRepositoryConfigsByKeys(repoKeys);

        // then
        assertNotNull(actual);
        assertEquals(actual, expected);
    }

    @Test
    void testLocalOrFederatedRepositoryByKeyForUnfoundRepo() {
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repo-key")).thenReturn(null);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNull(retrievedRepositoryConfig);
    }

    @Test
    void testLocalOrFederatedRepositoryByKeyForLocal() {
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepoConfig));

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNotNull(retrievedRepositoryConfig);
        assertEquals(retrievedRepositoryConfig, repositoryConfig);
    }

    @Test
    void testLocalOrFederatedRepositoryByKeyForFederated() {
        FederatedRepoConfig repositoryConfig = FederatedRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepoConfig));

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNotNull(retrievedRepositoryConfig);
        assertEquals(retrievedRepositoryConfig, repositoryConfig);
    }

    @Test
    void testLocalOrFederatedRepositoryByKeyForRemote() {
        RemoteRepoConfig repositoryConfig = RemoteRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repo-key")).thenReturn(dbRepoConfig);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNull(retrievedRepositoryConfig);
    }

    @Test
    void testLocalOrFederatedRepositoryByKeyForVirtual() {
        VirtualRepoConfig repositoryConfig = VirtualRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repo-key")).thenReturn(dbRepoConfig);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNull(retrievedRepositoryConfig);
    }

    @Test
    void testlocalOrFederatedRepoConfigByKeyForUnfoundRepo() {
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repo-key")).thenReturn(null);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNull(retrievedRepositoryConfig);
    }

    @Test
    void testlocalOrFederatedRepoConfigByKeyForLocal() {
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepoConfig));

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNotNull(retrievedRepositoryConfig);
        assertEquals(retrievedRepositoryConfig, repositoryConfig);
    }

    @Test
    void testlocalOrFederatedRepoConfigByKeyForFederated() {
        FederatedRepoConfig repositoryConfig = FederatedRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepoConfig));

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNotNull(retrievedRepositoryConfig);
        assertEquals(retrievedRepositoryConfig, repositoryConfig);
    }

    @Test
    void testlocalOrFederatedRepoConfigByKeyForRemote() {
        RemoteRepoConfig repositoryConfig = RemoteRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repo-key")).thenReturn(dbRepoConfig);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNull(retrievedRepositoryConfig);
    }

    @Test
    void testlocalOrFederatedRepoConfigByKeyForVirtual() {
        VirtualRepoConfig repositoryConfig = VirtualRepoConfig.create("repo-key", RepoType.Generic);
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        DbRepositoryConfig dbRepoConfig = repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repo-key")).thenReturn(dbRepoConfig);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        RepoConfig retrievedRepositoryConfig =
                repositoryConfigService.getLocalOrFederatedRepoConfigByKey("repo-key");
        assertNull(retrievedRepositoryConfig);
    }

    @Test
    void findRepositoriesByType() {
        LocalRepoConfig repositoryConfig1 = LocalRepoConfig.create("repo1", RepoType.Generic);
        VirtualRepoConfig repositoryConfig2 = VirtualRepoConfig.create("repo2", RepoType.Generic);

        List<RepoConfig> repoConfigs = List.of(repositoryConfig1, repositoryConfig2);
        List<DbRepositoryConfig> dbRepoConfigs = repoConfigs.stream()
                .map(repoConfig -> repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repoConfig))
                .collect(toList());
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(dbRepoConfigs);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        List<RepoConfig> returnedRepos =
                repositoryConfigService.findRepositoriesByType(LOCAL);

        assertEquals(1, returnedRepos.size());
        assertEquals(LOCAL, returnedRepos.get(0).getRepoType());
    }

    @Test
    void findRepositoriesByPackageType() {
        LocalRepoConfig repositoryConfig1 = LocalRepoConfig.create("repo1", RepoType.Docker);
        LocalRepoConfig repositoryConfig2 = LocalRepoConfig.create("repo2", RepoType.BuildInfo);

        List<LocalRepoConfig> repoConfigs = List.of(repositoryConfig1, repositoryConfig2);
        List<DbRepositoryConfig> dbRepoConfigs = repoConfigs.stream()
                .map(repoConfig -> repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repoConfig))
                .collect(toList());
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(dbRepoConfigs);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        List<RepoConfig> returnedRepos =
                repositoryConfigService.findRepositoriesByPackageType(RepoType.BuildInfo);

        assertEquals(1, returnedRepos.size());
        assertEquals(RepoType.BuildInfo, returnedRepos.get(0).getPackageType());
    }

    @Test
    void findRepositoriesByTypeAndPackageType() {
        VirtualRepoConfig repositoryConfig1 = VirtualRepoConfig.create("repoKey1", RepoType.Docker);
        RemoteRepoConfig repositoryConfig2 = RemoteRepoConfig.create("repoKey2", RepoType.Vagrant);
        LocalRepoConfig repositoryConfig3 = LocalRepoConfig.create("repoKey3", RepoType.Docker);

        List<RepoConfig> repoConfigs = List.of(repositoryConfig1, repositoryConfig2, repositoryConfig3);
        List<DbRepositoryConfig> dbRepoConfigs = repoConfigs.stream()
                .map(repoConfig -> repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repoConfig))
                .toList();
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);

        when(storageService.getAllRepositories()).thenReturn(dbRepoConfigs);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        List<RepoConfig> returnedRepos =
                repositoryConfigService.findRepositoriesByTypeAndPackageType(REMOTE, RepoType.Vagrant);

        assertEquals(1, returnedRepos.size());
        assertEquals(RepoType.Vagrant, returnedRepos.get(0).getPackageType());
        assertEquals(REMOTE, returnedRepos.get(0).getRepoType());
    }

    @Test
    void testFindRepositoriesByKeys() {

        List<DbRepositoryConfig> reposInStore = new ArrayList<>();
        Set<String> repoKeys = new HashSet<>();
        for (int i = 0; i < 3; i++) {
            RepoConfig repoConfig = LocalRepoConfig.create("repo-" + i, RepoType.Generic);
            repoKeys.add(repoConfig.getKey());
            reposInStore.add(repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repoConfig));
        }
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(reposInStore);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        Set<RepoConfig> repoConfigs = repositoryConfigService.findRepositoryConfigsByKeys(repoKeys);

        Contracts.assertNotNull(repoConfigs);
        assertTrue(repoConfigs.containsAll(reposInStore.stream()
                .map(dbRepoConfig -> repositoryConfigMapper
                        .dbInfoBlobToRepositoryConfig(dbRepoConfig))
                .toList()));
    }

    @Test
    void testGetExtendedRepositoryConfig() {
        // given
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Maven);
        repositoryConfig.getPackageTypeConfig()
                .put(POM_REPOSITORY_REFERENCES_CLEANUP_POLICY.getKey(), "discard_any_reference");
        repositoryConfig.getPackageTypeConfig().put(FETCH_JARS_EAGERLY.getKey(), Boolean.TRUE.toString());
        DbRepositoryConfig dbRepositoryConfig =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepositoryConfig));

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        ExtendedRepositoryConfig<MavenConfiguration> extendedConfig = repositoryConfigService
                .getExtendedRepositoryByKey("repoKey", MavenConfiguration.class);

        // then
        assertNotNull(extendedConfig);
        assertEquals(PomCleanupPolicy.discard_any_reference,
                extendedConfig.getExtendedRepoConfig().getPomRepositoryReferencesCleanupPolicy());
    }

    @Test
    void testInsertRepositoryConfigByKey() {
        try {
            // given
            LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
            repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());

            // when
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            repositoryConfigService.insertRepository(repositoryConfig);


            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> insertedDbRepositoryConfigs = insert.getValue();
            assertEquals(1, insertedDbRepositoryConfigs.size());
            RepoConfig insertedRepositoryConfig =
                    repositoryConfigMapper
                            .dbInfoBlobToRepositoryConfig(
                                    insertedDbRepositoryConfigs.stream().findFirst().orElse(new DbRepositoryConfig()));

            assertEquals(insertedRepositoryConfig, repositoryConfig);

            assertTrue(update.getValue().isEmpty());
            assertTrue(delete.getValue().isEmpty());
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testUpdateRepositoryConfigByKey() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            LocalRepoConfig oldRepoConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.getAllRepositoryKeysAndModificationDates()).thenReturn(Map.of("repoKey", 0L));
            when(storageService.getAllRepositories()).thenReturn(
                    List.of(repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(oldRepoConfig)));

            LocalRepoConfig newRepoConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            newRepoConfig.getBaseConfig().setDescription("description-y description");

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.updateRepository(newRepoConfig);


            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> updatedDbRepositoryConfigs = update.getValue();
            assertEquals(1, updatedDbRepositoryConfigs.size());
            RepoConfig updatedRepositoryConfig =
                    repositoryConfigMapper.dbInfoBlobToRepositoryConfig(updatedDbRepositoryConfigs.iterator().next());
            assertEquals(updatedRepositoryConfig.getBaseConfig().getDescription(),
                    newRepoConfig.getBaseConfig().getDescription());

            assertTrue(insert.getValue().isEmpty());
            assertTrue(delete.getValue().isEmpty());
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testCannotUpdateRepositoryThatDoesNotExist() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.findRepositoryByKey("repoKey")).thenReturn(null);

            LocalRepoConfig newRepoConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            newRepoConfig.getBaseConfig().setDescription("description-y description");

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            assertThrows(NotFoundException.class, () -> repositoryConfigService.updateRepository(newRepoConfig));
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testDeleteAndDereferenceRepository() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            String repoKey = "repoKey";
            LocalRepoConfig oldRepoConfig = LocalRepoConfig.create(repoKey, RepoType.Generic);
            DbRepositoryConfig oldRepoDbConfig =
                    repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(oldRepoConfig);

            String virtualRepoKey = "virtual-repo";
            VirtualRepoConfig virtualRepoConfig = VirtualRepoConfig.create(virtualRepoKey, RepoType.Generic);
            virtualRepoConfig.getRepoTypeConfig().getRepositoryRefs().add(repoKey);
            DbRepositoryConfig virtualDbRepositoryConfig =
                    repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(virtualRepoConfig);

            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.getAllRepositoryKeysAndModificationDates()).thenReturn(
                    Map.of(repoKey, 0L, virtualRepoKey, 0L));
            when(storageService.getAllRepositories()).thenReturn(List.of(oldRepoDbConfig, virtualDbRepositoryConfig));

            // config descriptor references
            CentralConfigDescriptorImpl configDescriptor = new CentralConfigDescriptorImpl();
            populateConfigDescriptorWithRefsToRepo(repoKey, configDescriptor);
            when(configService.getMutableDescriptor()).thenReturn(configDescriptor);

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.deleteRepository(repoKey);

            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService, times(1)).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            // repo was deleted
            Set<String> deletedRepositoryConfigs = delete.getValue();
            assertEquals(1, deletedRepositoryConfigs.size());
            assertEquals(deletedRepositoryConfigs.iterator().next(), repoKey);

            // referencing virtual repo was updated
            Set<DbRepositoryConfig> updatedRepoConfigs = update.getAllValues().get(0);

            assertEquals(1, updatedRepoConfigs.size());
            assertEquals(VIRTUAL.typeNameLowercase(), updatedRepoConfigs.iterator().next().getType());
            VirtualRepoConfig updatedVirtualRepo =
                    (VirtualRepoConfig) repositoryConfigMapper.dbInfoBlobToRepositoryConfig(
                            updatedRepoConfigs.iterator().next());
            assertFalse(updatedVirtualRepo.getRepoTypeConfig().getRepositoryRefs().contains(repoKey));

            // config descriptor references were deleted
            assertConfigDescriptorRefsWereRemoved(repoKey, configDescriptor);
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    public void testDereferenceRepositorySavesOnlyModifiedDescriptor() {
        CentralConfigDescriptorImpl configDescriptor = new CentralConfigDescriptorImpl();
        when(configService.getMutableDescriptor()).thenReturn(configDescriptor);

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(mock(RepositoryConfigStorageService.class), Set.of());
        repositoryConfigService.dereferenceRepositoryFromConfigDescriptor(
                List.of(LocalRepoConfig.create("local", RepoType.Generic)));

        verify(configService, times(0)).saveEditedDescriptorAndReload(any());
    }

    private void populateConfigDescriptorWithRefsToRepo(String repoKey, CentralConfigDescriptorImpl configDescriptor) {
        BackupDescriptor backupDescriptor = new BackupDescriptor();
        backupDescriptor.setKey("test");
        backupDescriptor.addExcludedRepository(repoKey);
        configDescriptor.addBackup(backupDescriptor);

        IndexerDescriptor indexer = new IndexerDescriptor();
        SortedSet<String> includedRepositories = new TreeSet<>();
        includedRepositories.add(repoKey);
        indexer.setIncludedRepositories(includedRepositories);
        configDescriptor.setIndexer(indexer);

        RemoteReplicationDescriptor remoteReplication = new RemoteReplicationDescriptor();
        remoteReplication.setReplicationKey("test");
        remoteReplication.setRepoKey(repoKey);
        configDescriptor.addRemoteReplication(remoteReplication);

        LocalReplicationDescriptor localReplication = new LocalReplicationDescriptor();
        localReplication.setReplicationKey("test");
        localReplication.setRepoKey(repoKey);
        configDescriptor.addLocalReplication(localReplication);

        ReverseProxyDescriptor reverseProxy = new ReverseProxyDescriptor();
        reverseProxy.setKey(repoKey);
        configDescriptor.addReverseProxy(reverseProxy);
    }

    private void assertConfigDescriptorRefsWereRemoved(String repoKey, CentralConfigDescriptorImpl configDescriptor) {
        assertFalse(configDescriptor.getBackup("test").getExcludedRepositories().contains(repoKey));
        assertFalse(configDescriptor.getIndexer().getIncludedRepositories().contains(repoKey));
        assertNull(configDescriptor.getRemoteReplication("test"));
        assertNull(configDescriptor.getLocalReplication("test"));
        assertNull(configDescriptor.getReverseProxy("test"));
    }

    @Test
    void testCannotDeleteRepositoryThatDoesNotExist() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.findRepositoryByKey("repoKey")).thenReturn(null);

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            assertNull(repositoryConfigService.deleteRepository("repoKey"));
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testUpsertNewRepositoryConfigByKey() {
        try {
            // given
            LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);

            // when
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.upsertRepository(repositoryConfig);

            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);

            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> insertedDbRepositoryConfigs = insert.getValue();
            assertEquals(1, insertedDbRepositoryConfigs.size());
            RepoConfig insertedRepositoryConfig =
                    repositoryConfigMapper.dbInfoBlobToRepositoryConfig(insertedDbRepositoryConfigs.iterator().next());

            assertTrue(update.getValue().isEmpty());
            assertTrue(delete.getValue().isEmpty());
            assertEquals(insertedRepositoryConfig.getKey(), repositoryConfig.getKey());
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testUpsertExistingRepositoryConfigByKey() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            LocalRepoConfig oldRepoConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.getAllRepositoryKeysAndModificationDates()).thenReturn(Map.of("repoKey", 0L));
            when(storageService.getAllRepositories()).thenReturn(
                    List.of(repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(oldRepoConfig)));

            LocalRepoConfig newRepoConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            newRepoConfig.getBaseConfig().setDescription("description-y description");

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.upsertRepository(newRepoConfig);

            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> updatedRepoConfigs = update.getValue();
            assertEquals(1, updatedRepoConfigs.size());
            RepoConfig updatedRepoConfig = repositoryConfigMapper.dbInfoBlobToRepositoryConfig(
                    updatedRepoConfigs.stream().findFirst().orElse(new DbRepositoryConfig())
            );

            assertEquals(newRepoConfig, updatedRepoConfig);

            assertTrue(insert.getValue().isEmpty());
            assertTrue(delete.getValue().isEmpty());
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @ParameterizedTest
    @MethodSource("providerAllRepoConfigTypesWithNonUnicodeKey")
    void testInsertRepoWithUnicodeKey(RepoConfig repositoryConfig) {
        // given
        AuthorizationService authService = mock(AuthorizationService.class);
        when(authService.isAnyProjectAdmin()).thenReturn(true);
        when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

        // when
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        repositoryConfigService.setConfigValidationService(new ConfigValidationService());
        when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
        assertThrows(IllegalStateException.class, () -> repositoryConfigService.insertRepository(repositoryConfig),
                ".* Unicode is not supported");
    }

    @ParameterizedTest
    @MethodSource("provideRealRepoTypeConfigs")
    void testUpsertRepoWithUnicodePropertySet(RealRepoConfig realRepoConfig) {
        // given
        realRepoConfig.getRepoTypeConfig().setPropertySetRefs(List.of("\u0000"));
        AuthorizationService authService = mock(AuthorizationService.class);
        when(authService.isAnyProjectAdmin()).thenReturn(true);
        when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

        // when
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        repositoryConfigService.setConfigValidationService(new ConfigValidationService());
        when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
        List<RealRepoConfig> repositoryConfigs = List.of(realRepoConfig);
        assertThrows(IllegalStateException.class,
                () -> repositoryConfigService.bulkInsertRepositoryConfigs(repositoryConfigs, false, true, false),
                ".* Unicode is not supported");
    }

    @Test
    void testGetPackageTypeConfigAsMap() {
        // given
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Maven);

        repositoryConfig.getPackageTypeConfig()
                .put(POM_REPOSITORY_REFERENCES_CLEANUP_POLICY.getKey(), "discard_any_reference");
        repositoryConfig.getPackageTypeConfig().put(FETCH_JARS_EAGERLY.getKey(), Boolean.TRUE.toString());
        DbRepositoryConfig dbRepositoryConfig =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepositoryConfig));

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        Map<String, String> retrievedMavenConfig = repositoryConfigService
                .getPackageConfigByRepoKey("repoKey");

        // then
        assertNotNull(retrievedMavenConfig);
        assertEquals(retrievedMavenConfig.get(POM_REPOSITORY_REFERENCES_CLEANUP_POLICY.getKey()),
                PomCleanupPolicy.discard_any_reference.getType());
    }

    @Disabled
    @Test
    void testGetPackageConfigWithNonTrivialPackageConfig() {
        // given
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("my-repo", RepoType.Maven);

        repositoryConfig.getPackageTypeConfig().put(EXTERNAL_DEPENDENCIES_ENABLED.getKey(), Boolean.TRUE.toString());
        DbRepositoryConfig dbRepositoryConfig =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.findRepositoryByKey("repoKey")).thenReturn(dbRepositoryConfig);

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        NpmConfiguration retrievedMavenConfig = repositoryConfigService
                .getPackageConfigByRepoKey("repoKey", NpmConfiguration.class);

        // then
        assertNotNull(retrievedMavenConfig);
    }

    @Test
    void testGetPackageConfig() {
        // given
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Maven);

        repositoryConfig.getPackageTypeConfig()
                .put(POM_REPOSITORY_REFERENCES_CLEANUP_POLICY.getKey(), "discard_any_reference");
        repositoryConfig.getPackageTypeConfig()
                .put(FETCH_JARS_EAGERLY.getKey(), Boolean.TRUE.toString());

        DbRepositoryConfig dbRepositoryConfig =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepositoryConfig));

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        MavenConfiguration retrievedMavenConfig = repositoryConfigService
                .getPackageConfigByRepoKey("repoKey", MavenConfiguration.class);

        // then
        assertNotNull(retrievedMavenConfig);
        assertEquals(PomCleanupPolicy.discard_any_reference,
                retrievedMavenConfig.getPomRepositoryReferencesCleanupPolicy());
    }

    @Test
    void testInsertPackageConfigAsMap() {
        try {
            // given
            LocalRepoConfig repositoryConfig = LocalRepoConfig.create("repoKey", RepoType.Generic);
            DbRepositoryConfig dbRepositoryConfig = getDbRepositoryConfig(repositoryConfig);
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.getAllRepositories()).thenReturn(List.of(dbRepositoryConfig));
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            repositoryConfigService
                    .insertPackageConfigByKey("repoKey", Map.of("foo", "bar"));

            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> updatedRepoConfigs = update.getValue();
            assertEquals(1, updatedRepoConfigs.size());
            LocalRepoConfig insertedRepositoryConfig =
                    (LocalRepoConfig) repositoryConfigMapper.dbInfoBlobToRepositoryConfig(
                            updatedRepoConfigs.iterator().next());
            assertEquals(insertedRepositoryConfig.getPackageTypeConfig(), Map.of("foo", "bar"));

            insertedRepositoryConfig.getPackageTypeConfig().
                    forEach((k, v) -> repositoryConfig.getPackageTypeConfig().put(k, v));
            assertEquals(insertedRepositoryConfig, repositoryConfig);
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testExists() {
        RepoConfig repoConfig = LocalRepoConfig.create("exists", RepoType.Generic);

        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(
                List.of(repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repoConfig)));

        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        boolean exists = repositoryConfigService.exists("exists");
        assertTrue(exists);
        exists = repositoryConfigService.exists("does-not-exist");
        assertFalse(exists);
    }

    @Test
    void testMutableChangesAreFlushed() {
        try {
            // given
            LocalRepoConfig localRepoConfigToCreate = LocalRepoConfig.create("create-me", RepoType.Generic);

            // when
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            RepoConfigChangeInterceptor mutatingInterceptor = mock(RepoConfigChangeInterceptor.class);
            when(context.beansForType(RepoConfigChangeInterceptor.class))
                    .thenReturn(Map.of("mutatingInterceptor", mutatingInterceptor));
            repositoryConfigService.init();

            Mockito.doAnswer(invocation -> {
                Object[] args = invocation.getArguments();
                Map<RepoDetailsType, List<RepoConfigChange>> mutableRepoConfigChanges =
                        ((Map<RepoDetailsType, List<RepoConfigChange>>) args[0]);
                RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create("delete-me", RepoType.Generic);
                mutableRepoConfigChanges.put(REMOTE,
                        List.of(new RepoConfigChange("delete-me", RepoConfigChange.ConfigAction.REPO_DELETE,
                                remoteRepoConfig, remoteRepoConfig)));
                VirtualRepoConfig virtualRepoConfigToUpdate = VirtualRepoConfig.create("update-me", RepoType.Generic);
                mutableRepoConfigChanges.put(VIRTUAL,
                        List.of(new RepoConfigChange("update-me", RepoConfigChange.ConfigAction.REPO_UPDATE,
                                virtualRepoConfigToUpdate, virtualRepoConfigToUpdate)));
                return null;
            }).when(mutatingInterceptor).onBeforeSave(anyMap());

            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.insertRepository(localRepoConfigToCreate);

            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            assertEquals(1, insert.getValue().size());
            assertEquals("create-me", insert.getValue().iterator().next().getKey());

            assertEquals(1, update.getValue().size());
            assertEquals("update-me", update.getValue().iterator().next().getKey());

            assertEquals(1, update.getValue().size());
            assertEquals("delete-me", delete.getValue().iterator().next());
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void orderOfRepositoryRefs() {
        // given
        List<String> expected = List.of("3", "C", "1", "4");
        VirtualRepoConfig repositoryConfig = VirtualRepoConfig.create("repoKey", RepoType.Generic);
        repositoryConfig.getRepoTypeConfig().setRepositoryRefs(expected);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        RepoConfigCache repoConfigCache = mock(RepoConfigCache.class);
        when(repoConfigCache.getByKey("repoKey")).thenReturn(repositoryConfig);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        repositoryConfigService.setCache(repoConfigCache);

        // when
        RepoConfig retrievedRepositoryConfig = repositoryConfigService.getRepoConfigByKey("repoKey");
        List<@NotUnicode String> actual =
                ((VirtualRepoConfig) retrievedRepositoryConfig).getRepoTypeConfig().getRepositoryRefs();

        // then
        assertEquals(actual, expected);
    }

    @Test
    void testGetVirtualReposContainingRepo() {
        // given
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        String referencedRepoKey = "referenced-repo";
        LocalRepoConfig referencedRepo = LocalRepoConfig.create(referencedRepoKey, RepoType.Generic);
        DbRepositoryConfig dbReferencedRepo =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(referencedRepo);
        String targetRepoKey = "target-repo";
        VirtualRepoConfig targetVirtualRepo = VirtualRepoConfig.create(targetRepoKey, RepoType.Generic);
        targetVirtualRepo.getRepoTypeConfig().setRepositoryRefs(List.of(referencedRepo.getKey()));
        DbRepositoryConfig dbTargetVirtualRepo =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(targetVirtualRepo);
        VirtualRepoConfig someVirtualRepo = VirtualRepoConfig.create("repo-key", RepoType.Generic);
        DbRepositoryConfig dbSomeVirtualRepo =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(someVirtualRepo);
        when(storageService.getAllRepositories()).thenReturn(
                List.of(dbSomeVirtualRepo, dbReferencedRepo, dbTargetVirtualRepo));
        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        List<VirtualRepoConfig> virtualReposContainingRepo =
                repositoryConfigService.getVirtualReposContainingRepo(referencedRepo, 1);

        // then
        assertEquals(1, virtualReposContainingRepo.size());
        assertEquals(virtualReposContainingRepo.get(0).getKey(), targetRepoKey);
    }

    @Test
    void testUpdateRemoteRepositoryConfigByKey() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            when(context.beanForType(AuthorizationService.class)).thenReturn(authService);

            MutableCentralConfigDescriptor configDescriptor = mock(MutableCentralConfigDescriptor.class);
            when(configService.getMutableDescriptor()).thenReturn(configDescriptor);

            RemoteRepoConfig oldRemoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.getAllRepositories()).thenReturn(
                    List.of(repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(oldRemoteRepoConfig)));

            RemoteRepoConfig newRemoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
            newRemoteRepoConfig.getBaseConfig().setDescription("description-y description");

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.upsertRepository(newRemoteRepoConfig);

            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);
            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> updatedDbRepositoryConfigs = update.getValue();
            assertEquals(1, updatedDbRepositoryConfigs.size());
            RepoConfig updatedRepositoryConfig =
                    repositoryConfigMapper.dbInfoBlobToRepositoryConfig(updatedDbRepositoryConfigs.iterator().next());
            assertEquals(updatedRepositoryConfig.getBaseConfig().getDescription(),
                    newRemoteRepoConfig.getBaseConfig().getDescription());
            verify(configDescriptor, times(0)).conditionallyAddToBackups(any());
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    void testInsertRemoteRepositoryConfigByKey() {
        try {
            // given
            AuthorizationService authService = mock(AuthorizationService.class);
            when(authService.isAnyProjectAdmin()).thenReturn(true);
            MutableCentralConfigDescriptor configDescriptor = mock(MutableCentralConfigDescriptor.class);
            when(configService.getMutableDescriptor()).thenReturn(configDescriptor);

            RemoteRepoConfig oldRemoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
            RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
            when(storageService.findRepositoryByKey("repoKey")).thenReturn(
                    repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(oldRemoteRepoConfig));
            when(storageService.exists("repoKey")).thenReturn(false);

            RemoteRepoConfig newRemoteRepoConfig = RemoteRepoConfig.create("repoKey", RepoType.Generic);
            newRemoteRepoConfig.getBaseConfig().setDescription("description-y description");

            // when
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(storageService, Set.of());
            repositoryConfigService.setConfigValidationService(new ConfigValidationService());
            when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);
            repositoryConfigService.upsertRepository(newRemoteRepoConfig);

            // then
            ArgumentCaptor<Set<DbRepositoryConfig>> insert = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<DbRepositoryConfig>> update = ArgumentCaptor.forClass(Set.class);
            ArgumentCaptor<Set<String>> delete = ArgumentCaptor.forClass(Set.class);

            verify(storageService).flushBulkChanges(insert.capture(), update.capture(), delete.capture());

            Set<DbRepositoryConfig> insertedDbRepositoryConfigs = insert.getValue();
            assertEquals(1, insertedDbRepositoryConfigs.size());
            RepoConfig insertedRepositoryConfig = repositoryConfigMapper.dbInfoBlobToRepositoryConfig(
                    insertedDbRepositoryConfigs.iterator().next());
            assertEquals(insertedRepositoryConfig.getBaseConfig().getDescription(),
                    newRemoteRepoConfig.getBaseConfig().getDescription());

            assertTrue(update.getValue().isEmpty());
            assertTrue(delete.getValue().isEmpty());
            verify(configDescriptor).conditionallyAddToBackups("repoKey");
        } finally {
            ArtifactoryContextThreadBinder.unbind();
        }
    }

    @Test
    public void testWriteRepoClassToMetricsThreadLocal() {
        try {
            MetricsThreadLocal.bind(new MetricsMetadata());
            RepositoryConfigServiceImpl repositoryConfigService =
                    new RepositoryConfigServiceImpl(mock(RepositoryConfigStorageService.class), Set.of());
            when(context.beanForType(ProjectAdminRepoConfigValidator.class)).thenReturn(
                    projectAdminRepoConfigValidator);

            repositoryConfigService.writeMetricsInfoToMetricsThreadLocal(Map.of());
            List<MetricsInfo> metricsInfos = MetricsThreadLocal.getMetricsInfoList();
            assertNotNull(metricsInfos);
            assertTrue(metricsInfos.isEmpty());

            RepoConfigChange createChange =
                    new RepoConfigChange("create-repo", RepoConfigChange.ConfigAction.REPO_CREATE,
                            LocalRepoConfig.create("create-repo", RepoType.Generic), null);
            RepoConfigChange updateChange =
                    new RepoConfigChange("update-repo", RepoConfigChange.ConfigAction.REPO_UPDATE,
                            LocalRepoConfig.create("update-repo", RepoType.Maven),
                            LocalRepoConfig.create("update-repo", RepoType.Maven));
            RepoConfigChange deleteChange =
                    new RepoConfigChange("delete-repo", RepoConfigChange.ConfigAction.REPO_DELETE,
                            LocalRepoConfig.create("delete-repo", RepoType.Docker),
                            LocalRepoConfig.create("delete-repo", RepoType.Docker));

            repositoryConfigService.writeMetricsInfoToMetricsThreadLocal(Map.of(
                    LOCAL, List.of(createChange),
                    REMOTE, List.of(createChange, updateChange),
                    VIRTUAL, List.of(createChange, updateChange, deleteChange)));

            metricsInfos = MetricsThreadLocal.getMetricsInfoList();
            assertNotNull(metricsInfos);
            assertEquals(6, metricsInfos.size());

            List<MetricsInfo> createMetrics =
                    metricsInfos.stream().filter(metricsInfo -> metricsInfo.getAction().equals("REPO_CREATED"))
                            .toList();
            assertEquals(3, createMetrics.size());
            List.of(LOCAL, REMOTE, VIRTUAL).forEach(repoDetailsType -> assertEquals(1, createMetrics.stream()
                    .filter(metricsInfo -> metricsInfo.getRepoClass().equals(repoDetailsType.typeNameLowercase()))
                    .count()));
            createMetrics.forEach(metricsInfo -> assertEquals(RepoType.Generic.getType(), metricsInfo.getRepoType()));

            List<MetricsInfo> updateMetrics =
                    metricsInfos.stream().filter(metricsInfo -> metricsInfo.getAction().equals("REPO_UPDATED"))
                            .toList();
            assertEquals(2, updateMetrics.size());
            List.of(REMOTE, VIRTUAL).forEach(repoDetailsType -> assertEquals(1, updateMetrics.stream()
                    .filter(metricsInfo -> metricsInfo.getRepoClass().equals(repoDetailsType.typeNameLowercase()))
                    .count()));
            updateMetrics.forEach(metricsInfo -> assertEquals(RepoType.Maven.getType(), metricsInfo.getRepoType()));

            List<MetricsInfo> deleteMetrics =
                    metricsInfos.stream().filter(metricsInfo -> metricsInfo.getAction().equals("REPO_DELETED"))
                            .toList();
            assertEquals(1, deleteMetrics.size());
            assertEquals(VIRTUAL.typeNameLowercase(), deleteMetrics.get(0).getRepoClass());
            assertEquals(RepoType.Docker.getType(), deleteMetrics.get(0).getRepoType());
        } finally {
            MetricsThreadLocal.unbind();
        }
    }

    @Disabled
    @Test
    void testConvert() {
        String convertMe = "{\n" +
                "  \"baseConfig\": {\n" +
                "    \"modelVersion\": 0,\n" +
                "    \"key\": \"my-repo\",\n" +
                "    \"type\": \"LOCAL\",\n" +
                "    \"packageType\": \"generic\"\n" +
                "  }" +
                "}";
        DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
        dbRepositoryConfig.setKey("my-repo");
        dbRepositoryConfig.setType("local");
        dbRepositoryConfig.setPackageType("generic");
        dbRepositoryConfig.setRepositoryInfoBlob(convertMe.getBytes());
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositories()).thenReturn(List.of(dbRepositoryConfig));

        // when
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService,
                        Set.of(new PackageTypeConfigConverter(), new RepositoryConfigBaseConverter()));

        repositoryConfigService
                .convert(new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v7140m001.get(), "build", 9),
                        new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v7140m001.get(), "build", 9));

        // then (dummy converters)
        ArgumentCaptor<Set<DbRepositoryConfig>> captor = ArgumentCaptor.forClass(Set.class);
        verify(storageService).bulkUpdateRepositoryConfigs(captor.capture());

        Set<DbRepositoryConfig> convertedRepos = captor.getValue();
        assertEquals(1, convertedRepos.size());

        RepoConfig repositoryConfig = convertedRepos.stream()
                .map(convertedRepo -> repositoryConfigMapper.dbInfoBlobToRepositoryConfig(convertedRepo))
                .collect(toList())
                .get(0);
        assertEquals("default notes", repositoryConfig.getBaseConfig().getNotes());
    }

    @Test
    public void reloadRepoConfigurationsDeleteTwoKeys() {
        RepoConfigPropagationModel repoConfigPropagationModel = new RepoConfigPropagationModel();
        Set<RepoConfigReload> repoConfigDeletes = new HashSet<>();
        RepoConfigReload repoConfigDelete1 =
                new RepoConfigReload("localKey1", LOCAL, RepoType.Docker, RepoConfigChange.ConfigAction.REPO_DELETE,
                        0L);
        RepoConfigReload repoConfigDelete2 =
                new RepoConfigReload("localKey2", LOCAL, RepoType.Docker, RepoConfigChange.ConfigAction.REPO_DELETE,
                        0L);
        repoConfigDeletes.add(repoConfigDelete1);
        repoConfigDeletes.add(repoConfigDelete2);
        repoConfigPropagationModel.setRepoConfigDelete(repoConfigDeletes);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        when(storageService.getAllRepositoryKeysAndModificationDates()).thenReturn(Map.of("repoKey", 0L));
        when(context.beanForType(RepositoryConfigService.class)).thenReturn(repositoryConfigService);
        repositoryConfigService.reloadRepoConfigurations(repoConfigPropagationModel);
        ArgumentCaptor<Map<RepoDetailsType, List<RepoConfigReload>>> repoConfigReload =
                ArgumentCaptor.forClass(Map.class);

        verify(ContextHelper.get().beanForType(InternalRepositoryService.class)).onRepoConfigReload(
                repoConfigReload.capture());
        assertEquals(2, repoConfigReload.getValue().get(LOCAL).size());
        assertTrue(repoConfigReload.getValue().get(LOCAL).contains(repoConfigDelete2));
        assertTrue(repoConfigReload.getValue().get(LOCAL).contains(repoConfigDelete1));

    }

    @Test
    public void reloadRepoConfigurationsDeleteOneKey() {
        RepoConfigPropagationModel repoConfigPropagationModel = new RepoConfigPropagationModel();
        Set<RepoConfigReload> repoConfigDeletes = new HashSet<>();
        RepoConfigReload repoConfigDelete1 = new RepoConfigReload("localKey1", LOCAL, RepoType.Docker,
                RepoConfigChange.ConfigAction.REPO_DELETE, 0L);
        RepoConfigReload repoConfigDelete2 = new RepoConfigReload("localKey2", LOCAL, RepoType.Docker,
                RepoConfigChange.ConfigAction.REPO_DELETE, 0L);
        repoConfigDeletes.add(repoConfigDelete1);
        repoConfigDeletes.add(repoConfigDelete2);
        repoConfigPropagationModel.setRepoConfigDelete(repoConfigDeletes);

        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        when(storageService.getAllRepositoryKeysAndModificationDates()).thenReturn(
                Map.of("localKey3", 0L, "localKey1", 0L));

        RepositoryConfigServiceImpl repositoryConfigService = new RepositoryConfigServiceImpl(storageService, Set.of());
        when(context.beanForType(RepositoryConfigService.class)).thenReturn(repositoryConfigService);
        repositoryConfigService.reloadRepoConfigurations(repoConfigPropagationModel);
        ArgumentCaptor<Map<RepoDetailsType, List<RepoConfigReload>>> repoConfigReload =
                ArgumentCaptor.forClass(Map.class);

        verify(ContextHelper.get().beanForType(InternalRepositoryService.class))
                .onRepoConfigReload(repoConfigReload.capture());

        assertEquals(1, repoConfigReload.getValue().get(LOCAL).size());
        assertTrue(repoConfigReload.getValue().get(LOCAL).contains(repoConfigDelete2));
    }

    @Test
    public void reloadRepoConfigurationsDeleteOneKeyCreateOneKey() {
        RepoConfigPropagationModel repoConfigPropagationModel = new RepoConfigPropagationModel();
        Set<RepoConfigReload> repoConfigDeletes = new HashSet<>();
        RepoConfigReload repoConfigDelete1 =
                new RepoConfigReload("localKey1", LOCAL, RepoType.Docker, RepoConfigChange.ConfigAction.REPO_DELETE,
                        0L);
        RepoConfigReload repoConfigDelete2 =
                new RepoConfigReload("localKey2", LOCAL, RepoType.Docker, RepoConfigChange.ConfigAction.REPO_DELETE,
                        0L);
        repoConfigDeletes.add(repoConfigDelete1);
        repoConfigDeletes.add(repoConfigDelete2);
        repoConfigPropagationModel.setRepoConfigDelete(repoConfigDeletes);
        Set<RepoConfigReload> repoConfigsCreateOrUpdate = new HashSet<>();
        RepoConfigReload repoConfigCreateOrUpdate1 =
                new RepoConfigReload("localKey4", LOCAL, RepoType.Docker, RepoConfigChange.ConfigAction.REPO_CREATE,
                        0L);
        repoConfigsCreateOrUpdate.add(repoConfigCreateOrUpdate1);
        repoConfigPropagationModel.setRepoConfigCreateOrUpdate(repoConfigsCreateOrUpdate);
        RepositoryConfigStorageService storageService = mock(RepositoryConfigStorageService.class);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(storageService, Set.of());
        when(storageService.getAllRepositoryKeysAndModificationDates()).thenReturn(
                Map.of("localKey3", 0L, "localKey1", 0L));
        when(context.beanForType(RepositoryConfigService.class)).thenReturn(repositoryConfigService);
        repositoryConfigService.reloadRepoConfigurations(repoConfigPropagationModel);
        ArgumentCaptor<Map<RepoDetailsType, List<RepoConfigReload>>> repoConfigReload =
                ArgumentCaptor.forClass(Map.class);

        verify(ContextHelper.get().beanForType(InternalRepositoryService.class)).onRepoConfigReload(
                repoConfigReload.capture());
        assertEquals(2, repoConfigReload.getValue().get(LOCAL).size());
        assertTrue(repoConfigReload.getValue().get(LOCAL).contains(repoConfigDelete2));
        assertTrue(repoConfigReload.getValue().get(LOCAL).contains(repoConfigCreateOrUpdate1));
    }

    @Test
    public void convertRepoConfigsToUpdateChangesCreateAndUpdate() {
        VirtualRepoConfig virtualRepoConfigOld1 = VirtualRepoConfig.create("updateVirtaulKey", RepoType.Maven);
        VirtualRepoConfig virtualRepoConfigNew1 = VirtualRepoConfig.create("updateVirtaulKey", RepoType.Maven);
        virtualRepoConfigOld1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1", "repokey2"));
        virtualRepoConfigNew1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1"));
        VirtualRepoConfig virtualRepoConfig2 = VirtualRepoConfig.create("createVirtaulKey", RepoType.Maven);
        List<VirtualRepoConfig> virtualRepoConfigs = List.of(virtualRepoConfigNew1, virtualRepoConfig2);
        Collection<RepoConfig> existingRepos = Set.of(virtualRepoConfigOld1);
        Map<RepoDetailsType, List<RepoConfigChange>> expected =
                Map.of(VIRTUAL, List.of(new RepoConfigChange("updateVirtaulKey",
                        RepoConfigChange.ConfigAction.REPO_UPDATE, virtualRepoConfigNew1, virtualRepoConfigOld1)));
        Map<RepoDetailsType, List<RepoConfigChange>> actual =
                RepositoryConfigServiceImpl.convertRepoConfigsToUpdateChanges(virtualRepoConfigs, existingRepos);
        assertEquals(expected, actual);
    }

    @Test
    public void convertRepoConfigsToUpdateChangesOnlyUpdate() {
        VirtualRepoConfig virtualRepoConfigOld1 = VirtualRepoConfig.create("updateVirtaulKey1", RepoType.Maven);
        VirtualRepoConfig virtualRepoConfigNew1 = VirtualRepoConfig.create("updateVirtaulKey1", RepoType.Maven);
        virtualRepoConfigOld1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1", "repokey2"));
        virtualRepoConfigNew1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1"));
        VirtualRepoConfig virtualRepoConfigOld2 = VirtualRepoConfig.create("updateVirtaulKey2", RepoType.Maven);
        VirtualRepoConfig virtualRepoConfigNew2 = VirtualRepoConfig.create("updateVirtaulKey2", RepoType.Maven);
        virtualRepoConfigNew1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey11"));
        List<VirtualRepoConfig> virtualRepoConfigs = List.of(virtualRepoConfigNew1, virtualRepoConfigNew2);
        Collection<RepoConfig> existingRepos = Set.of(virtualRepoConfigOld1, virtualRepoConfigOld2);
        Map<RepoDetailsType, List<RepoConfigChange>> expected =
                Map.of(VIRTUAL, List.of(new RepoConfigChange("updateVirtaulKey1",
                                RepoConfigChange.ConfigAction.REPO_UPDATE, virtualRepoConfigNew1,
                                virtualRepoConfigOld1),
                        new RepoConfigChange("updateVirtaulKey2",
                                RepoConfigChange.ConfigAction.REPO_UPDATE, virtualRepoConfigNew2,
                                virtualRepoConfigOld2)));
        Map<RepoDetailsType, List<RepoConfigChange>> actual =
                RepositoryConfigServiceImpl.convertRepoConfigsToUpdateChanges(virtualRepoConfigs, existingRepos);
        assertEquals(expected, actual);
    }

    @Test
    public void convertRepoConfigsToUpdateChangesNull() {
        VirtualRepoConfig virtualRepoConfigOld1 = VirtualRepoConfig.create("updateVirtaulKey1", RepoType.Maven);
        VirtualRepoConfig virtualRepoConfigNew1 = VirtualRepoConfig.create("updateVirtaulKey1", RepoType.Maven);
        virtualRepoConfigOld1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1", "repokey2"));
        virtualRepoConfigNew1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1"));
        VirtualRepoConfig virtualRepoConfigOld2 = VirtualRepoConfig.create("updateVirtaulKey2", RepoType.Maven);
        VirtualRepoConfig virtualRepoConfigNew2 = VirtualRepoConfig.create("updateVirtaulKey2", RepoType.Maven);
        virtualRepoConfigNew1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey11"));
        List<VirtualRepoConfig> virtualRepoConfigs = List.of();
        Collection<RepoConfig> existingRepos = Set.of(virtualRepoConfigOld1, virtualRepoConfigOld2);
        Map<RepoDetailsType, List<RepoConfigChange>> expected =
                Map.of();
        Map<RepoDetailsType, List<RepoConfigChange>> actual =
                RepositoryConfigServiceImpl.convertRepoConfigsToUpdateChanges(virtualRepoConfigs, existingRepos);
        assertEquals(expected, actual);
    }

    @Test
    public void convertRepoConfigsToUpdateChangesOnlyCreate() {
        VirtualRepoConfig virtualRepoConfig1 = VirtualRepoConfig.create("create1VirtaulKey", RepoType.Maven);
        virtualRepoConfig1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1", "repokey2"));
        VirtualRepoConfig virtualRepoConfig2 = VirtualRepoConfig.create("create2VirtaulKey", RepoType.Maven);
        List<VirtualRepoConfig> virtualRepoConfigs = List.of(virtualRepoConfig1, virtualRepoConfig2);
        Collection<RepoConfig> existingRepos = Set.of();
        Map<RepoDetailsType, List<RepoConfigChange>> expected =
                Map.of();
        Map<RepoDetailsType, List<RepoConfigChange>> actual =
                RepositoryConfigServiceImpl.convertRepoConfigsToUpdateChanges(virtualRepoConfigs, existingRepos);
        assertEquals(expected, actual);
    }

    @Test
    public void convertRepoConfigsToCreateUpdateChanges() {
        VirtualRepoConfig virtualRepoConfigOld1 = VirtualRepoConfig.create("updateVirtaulKey", RepoType.Maven);
        VirtualRepoConfig virtualRepoConfigNew1 = VirtualRepoConfig.create("updateVirtaulKey", RepoType.Maven);
        virtualRepoConfigOld1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1", "repokey2"));
        virtualRepoConfigNew1.getRepoTypeConfig().setRepositoryRefs(List.of("repokey1"));
        VirtualRepoConfig virtualRepoConfig2 = VirtualRepoConfig.create("createVirtaulKey", RepoType.Maven);
        List<VirtualRepoConfig> virtualRepoConfigs = List.of(virtualRepoConfigNew1, virtualRepoConfig2);
        Collection<RepoConfig> existingRepos = Set.of(virtualRepoConfigOld1);
        Map<RepoDetailsType, List<RepoConfigChange>> expected =
                Map.of(VIRTUAL, List.of(new RepoConfigChange("updateVirtaulKey",
                                RepoConfigChange.ConfigAction.REPO_UPDATE, virtualRepoConfigNew1,
                                virtualRepoConfigOld1),
                        new RepoConfigChange("createVirtaulKey", RepoConfigChange.ConfigAction.REPO_CREATE,
                                virtualRepoConfig2, null)));
        Map<RepoDetailsType, List<RepoConfigChange>> actual =
                RepositoryConfigServiceImpl.convertRepoConfigsToCreateUpdateChanges(virtualRepoConfigs, existingRepos);
        assertEquals(expected, actual);
    }


    @Test
    public void excludeNewRepoFromBackup() {
        // given
        CentralConfigDescriptorImpl cc = CentralConfigTestUtils.initCentralConfig();
        CentralConfigService centralConfigService = mock(CentralConfigService.class);
        when(centralConfigService.getMutableDescriptor()).thenReturn(cc);
        when(context.getCentralConfig()).thenReturn(centralConfigService);
        cc.conditionallyAddToBackups("newLocal");
        RemoteRepoConfig newRemote = RemoteRepoConfig.create("newRemote", RepoType.Generic);
        RepositoryConfigServiceImpl repositoryConfigService =
                new RepositoryConfigServiceImpl(mock(RepositoryConfigStorageService.class), Set.of());
        repositoryConfigService.setConfigValidationService(new ConfigValidationService());
        when(context.beanForType(ArtifactoryVersionValidator.class)).thenReturn(artifactoryVersionValidator);

        // when
        repositoryConfigService.upsertRepository(newRemote);

        Assert.assertTrue(cc.isBackupExists("backup1"));
        BackupDescriptor backup1 = Iterables.tryFind(cc.getBackups(), new Predicate<BackupDescriptor>() {
            @Override
            public boolean apply(BackupDescriptor input) {
                return "backup1".equals(input.getKey());
            }
        }).orNull();
        Assert.assertNotNull(backup1, "Should find backup1");
        Assert.assertEquals(backup1.getExcludedRepositories().size(), 0);

        Assert.assertTrue(cc.isBackupExists("backup3"));
        BackupDescriptor backup3 = Iterables.tryFind(cc.getBackups(), new Predicate<BackupDescriptor>() {
            @Override
            public boolean apply(BackupDescriptor input) {
                return "backup3".equals(input.getKey());
            }
        }).orNull();
        Assert.assertNotNull(backup3, "Should find backup3");

        List<String> excludedRepositories = backup3.getExcludedRepositories();
        Assert.assertEquals(excludedRepositories.size(), 2);
        Assert.assertTrue(excludedRepositories.contains("newLocal"));
        Assert.assertTrue(excludedRepositories.contains(newRemote.getKey()));
    }

    private void createAndBindToArtifactoryHome() throws IOException {
        File homeDir = Files.createTempDirectory(this.getClass().getSimpleName()
                + "_" + new Random().nextLong()).toFile();
        ArtifactoryHome artifactoryHome = new TestArtifactoryHome(homeDir);
        artifactoryHome.initArtifactorySystemProperties();
        ArtifactoryHome.bind(artifactoryHome);
    }

    private void unbindAndCleanHome() {
        if (ArtifactoryHome.isBound()) {
            FileUtils.deleteQuietly(ArtifactoryHome.get().getHomeDir());
            ArtifactoryHome.unbind();
        }
    }

    private DbRepositoryConfig getDbRepositoryConfig(LocalRepoConfig repositoryConfig) {
        repositoryConfig.getSecurityConfig().setXrayDataTtl(repositoryConfig.getSecurityConfig().getXrayDataTtl());
        repositoryConfig.getSecurityConfig().setXrayIndex(repositoryConfig.getSecurityConfig().isXrayIndex());
        return repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);
    }
}