package org.artifactory.repo.service;

import lombok.extern.slf4j.Slf4j;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.repo.LocalCacheRepo;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.service.versioning.RepositoriesCache;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.storage.db.security.service.VersioningCacheImpl;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.util.RepoPathUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.springframework.util.ReflectionUtils;

import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.openMocks;

/**
 * @author Omri Ziv
 */
@Slf4j
public class RepositoryServiceImplTest extends ArtifactoryHomeBoundTest {

    private RepositoryServiceImpl repositoryService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private RepositoriesCache repositoriesCache;
    @Mock
    private LocalCacheRepo repo1Cache;
    @Mock
    private LocalRepo repo2Cache;
    @Mock
    private LocalRepo trash;
    private AutoCloseable mocks;
    private RepoPath repoPath1;
    private RepoPath repoPath2;

    @BeforeEach
    void init() {
        mocks = openMocks(this);
        repositoryService = spy(new RepositoryServiceImpl());
        Field field = ReflectionUtils.findField(RepositoryServiceImpl.class, "authService");
        ReflectionUtils.makeAccessible(field);
        ReflectionUtils.setField(field, repositoryService, authorizationService);
        Map<String, LocalRepo> localRepoMap = new HashMap<>();
        when(repo1Cache.isCache()).thenReturn(true);
        localRepoMap.put("repo1-cache", repo1Cache);
        localRepoMap.put("repo2", repo2Cache);
        repositoriesCache = new RepositoriesCache(new HashMap<>(), new HashMap<>(), new HashMap<>(), new HashMap<>(),
                new HashMap<>(), localRepoMap, trash, new HashSet<>(),
                null, new HashMap<>(), new HashMap<>());
        repoPath1 = new RepoPathImpl("repo1-cache",
                "/mysql/connector-java/" + "5.1.7-SNAPSHOT/connector-java-5.1.7-20190219.131101-4.jar");
        repoPath2 = new RepoPathImpl("repo2",
                "/mysql/connector-java/" + "5.1.7-SNAPSHOT/connector-java-5.1.7-20190219.131101-4.jar");
        VersioningCacheImpl<RepositoriesCache> cache = spy(mock(VersioningCacheImpl.class));
        repositoryService.cache = cache;
        when(cache.get()).thenReturn(repositoriesCache);
    }

    @AfterEach
    void afterEach() throws Exception {
        mocks.close();
    }

    @Test
    public void testZap() {
        when(authorizationService.canDelete(repoPath1)).thenReturn(true);
        repositoryService.zap(repoPath1);
        verify(repo1Cache, times(1)).zap(repoPath1);
    }

    @Test
    public void testZapNotPermitted() {
        when(authorizationService.canDelete(repoPath1)).thenReturn(false);
        assertThrows(ForbiddenException.class, () -> repositoryService.zap(repoPath1));
    }

    @Test
    public void testZapNotCachedRepo() {
        when(authorizationService.canDelete(repoPath2)).thenReturn(true);
        assertThrows(BadRequestException.class, () -> repositoryService.zap(repoPath2));
    }

    @Test
    public void testZapTrash() {
        RepoPath repoPathTrash = RepoPathUtils.repoRootPath(TrashService.TRASH_KEY);
        when(authorizationService.canDelete(repoPathTrash)).thenReturn(true);
        assertThrows(BadRequestException.class, () -> repositoryService.zap(repoPathTrash));
    }
}