/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security;

import com.google.common.base.Charsets;
import com.google.common.collect.ImmutableList;
import com.google.common.collect.Lists;
import com.google.common.collect.Multimap;
import com.google.common.collect.Sets;
import org.apache.commons.codec.binary.Base64;
import org.apache.commons.collections.CollectionUtils;
import org.apache.commons.io.FileUtils;
import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.oauth.OAuthHandler;
import org.artifactory.addon.sso.CentralOAuthSSOAddon;
import org.artifactory.addon.sso.HttpSsoAddon;
import org.artifactory.addon.sso.crowd.CrowdAddon;
import org.artifactory.addon.sso.saml.SamlSsoAddon;
import org.artifactory.api.build.BuildWrapper;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.CentralOAuthSettings;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.ResetPasswordException;
import org.artifactory.api.security.SecurityListener;
import org.artifactory.api.security.SecurityService;
import org.artifactory.api.security.UserInfoBuilder;
import org.artifactory.api.security.ldap.LdapService;
import org.artifactory.build.InternalBuildService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.config.InternalCentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.security.PasswordExpirationPolicy;
import org.artifactory.descriptor.security.PasswordResetPolicy;
import org.artifactory.descriptor.security.PasswordSettings;
import org.artifactory.descriptor.security.SecurityDescriptor;
import org.artifactory.descriptor.security.ldap.LdapSetting;
import org.artifactory.descriptor.security.oauth.OAuthProviderSettings;
import org.artifactory.descriptor.security.oauth.OAuthSettings;
import org.artifactory.descriptor.security.sso.CrowdSettings;
import org.artifactory.descriptor.security.sso.HttpSsoSettings;
import org.artifactory.descriptor.security.sso.SamlSettings;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.exception.InvalidNameException;
import org.artifactory.factory.InfoFactory;
import org.artifactory.factory.InfoFactoryHolder;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.security.AceImpl;
import org.artifactory.model.xstream.security.BuildPermissionTargetImpl;
import org.artifactory.model.xstream.security.GroupImpl;
import org.artifactory.model.xstream.security.MutableBuildAclImpl;
import org.artifactory.model.xstream.security.PrincipalPermissionImpl;
import org.artifactory.model.xstream.security.RepoPermissionTargetImpl;
import org.artifactory.model.xstream.security.UserGroupImpl;
import org.artifactory.model.xstream.security.UserImpl;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.InternalRepoPathFactory;
import org.artifactory.repo.LocalRepo;
import org.artifactory.repo.RemoteRepo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.repo.trash.TrashService;
import org.artifactory.rest.exception.ForbiddenException;
import org.artifactory.security.access.AccessResponseAuthentication;
import org.artifactory.security.access.AccessService;
import org.artifactory.security.access.AccessTokenAuthentication;
import org.artifactory.security.access.emigrate.AccessConverter;
import org.artifactory.security.access.emigrate.AccessConverters;
import org.artifactory.security.access.emigrate.conveter.V5100ConvertResourceTypeToRepo;
import org.artifactory.security.access.emigrate.conveter.V600DecryptAllUsersCustomData;
import org.artifactory.security.access.emigrate.conveter.V6600CreateDefaultBuildAcl;
import org.artifactory.security.exceptions.PasswordChangeException;
import org.artifactory.security.interceptor.SecurityConfigurationChangesInterceptors;
import org.artifactory.security.props.auth.ApiKeyManager;
import org.artifactory.security.props.auth.PropsAuthenticationToken;
import org.artifactory.security.signature.SignedUrlAuthenticationToken;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.storage.db.converter.markers.MarkerConverterAdapter;
import org.artifactory.storage.db.converter.markers.version.v2.V5100AddAccessResourceTypeMarkerMigration;
import org.artifactory.storage.db.converter.markers.version.v3.V600AddAccessDecryptAllUsersMarkerMigration;
import org.artifactory.storage.db.converter.markers.version.v4.V658AnonAccessToBuildsMarkerMigration;
import org.artifactory.storage.db.init.CommonDataSourceFactory;
import org.artifactory.storage.db.init.migration.MarkerMigrationHelper;
import org.artifactory.storage.db.init.migration.MarkerMigrationInfoBlob;
import org.artifactory.storage.db.init.migration.dao.MigrationStatusDao;
import org.artifactory.storage.db.init.migration.entity.DbMigrationStatus;
import org.artifactory.storage.db.init.migration.model.MigrationStatus;
import org.artifactory.storage.db.init.migration.service.MigrationStatusStorageServiceImpl;
import org.artifactory.storage.db.init.migration.service.mapper.MigrationStatusMapperImpl;
import org.artifactory.storage.security.service.AclCache;
import org.artifactory.storage.security.service.AclStoreService;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.artifactory.storage.security.service.UserLockInMemoryService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.test.ArtifactoryHomeStub;
import org.artifactory.util.NameValidator;
import org.artifactory.version.ArtifactoryServiceVersion;
import org.artifactory.version.ArtifactoryServiceVersionProvider;
import org.artifactory.version.CompoundVersionDetails;
import org.easymock.EasyMock;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.AccessClientHttpException;
import org.jfrog.access.client.auth.AccessPrincipal;
import org.jfrog.access.client.auth.AuthClient;
import org.jfrog.access.client.auth.LoginResult;
import org.jfrog.access.client.user.UsersClient;
import org.jfrog.access.common.PermissionScope;
import org.jfrog.access.common.PermissionScopeAction;
import org.jfrog.access.common.PermissionScopeResourceType;
import org.jfrog.access.model.Realm;
import org.jfrog.access.model.UserStatus;
import org.jfrog.access.rest.group.Group;
import org.jfrog.access.rest.system.UserLockPolicyConfig;
import org.jfrog.access.rest.user.UserResponse;
import org.jfrog.access.rest.user.UsersResponse;
import org.jfrog.access.token.JwtAccessToken;
import org.jfrog.access.user.UserMFAStatus;
import org.jfrog.build.api.Build;
import org.jfrog.common.ClockUtils;
import org.jfrog.common.JsonUtils;
import org.jfrog.common.StreamSupportUtils;
import org.jfrog.converter.DbDetails;
import org.jfrog.security.crypto.CipherAlg;
import org.jfrog.security.file.SecurityFolderHelper;
import org.jfrog.storage.JdbcHelper;
import org.junit.jupiter.api.Assertions;
import org.mockito.Answers;
import org.mockito.Mockito;
import org.mockito.stubbing.Answer;
import org.springframework.security.authentication.UsernamePasswordAuthenticationToken;
import org.springframework.security.core.Authentication;
import org.springframework.security.core.context.SecurityContext;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.core.context.SecurityContextImpl;
import org.springframework.security.core.userdetails.UserDetails;
import org.springframework.test.util.ReflectionTestUtils;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;
import org.testng.collections.Maps;

import java.io.IOException;
import java.lang.reflect.Field;
import java.nio.charset.Charset;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.TreeSet;
import java.util.stream.Collectors;

import static org.artifactory.api.security.AuthorizationService.ROLE_ADMIN;
import static org.artifactory.api.security.SecurityService.INTERNAL_PERMISSION_PREFIX;
import static org.artifactory.api.security.SecurityService.USER_SYSTEM;
import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.LOG_SHIPPING_REPO_KEY;
import static org.artifactory.repo.system.SystemRepos.USAGE_REPO_KEY;
import static org.artifactory.security.PermissionTarget.ANY_DISTRIBUTION_REPO;
import static org.artifactory.security.PermissionTarget.ANY_LOCAL_REPO;
import static org.artifactory.security.PermissionTarget.ANY_REMOTE_REPO;
import static org.artifactory.security.PermissionTarget.ANY_REPO;
import static org.artifactory.security.props.auth.OauthManager.OAUTH_INFO_PROPERTY;
import static org.easymock.EasyMock.anyObject;
import static org.easymock.EasyMock.createMock;
import static org.easymock.EasyMock.createNiceMock;
import static org.easymock.EasyMock.eq;
import static org.easymock.EasyMock.expect;
import static org.easymock.EasyMock.replay;
import static org.easymock.EasyMock.reset;
import static org.easymock.EasyMock.verify;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertThrows;
import static org.testng.Assert.assertTrue;
import static org.testng.Assert.fail;

/**
 * SecurityServiceImpl unit tests.
 *
 * @author Yossi Shaul
 */
@Test
public class SecurityServiceImplTest extends ArtifactoryHomeBoundTest {

    public static final String SSH_KEY = "ssh.basictoken";
    private static final String PROJECT_A = "proj";
    private static final String PROJECT_B = "proj2";
    private static final RepoPath REPO = RepoPathFactory.create("my-repo");
    private static final RepoPath SHARED_REPO = RepoPathFactory.create("shared-repo");
    private static final String PERMISSION_ACTOR = "permission_actor";
    private static final String BUILD_NAME = "buildA";
    private static final String BUILD_NUMBER = "1";
    private static final String BUILD_DATE = "2015-05-15T11:49:04.804+02:00";
    public static final String IS_SUPER_ADMIN_ERROR = "The user in test is super admin";
    public static final String IS_PROJECT_ADMIN_ERROR = "The user in test is project admin";
    public static final String NOT_A_PROJECT_ADMIN_ERROR = "The user in test is not a project admin";
    private static final String ACCESS_TOKEN =
            "eyJ2ZXIiOiIyIiwidHlwIjoiSldUIiwiYWxnIjoiUlMyNTYiLCJraWQiOiJheHlVMFFzc3J3S0p3ZllqMUtO" +
                    "bzNMbzZ6TnhLcnc3NEJ5YkVGVUJrSjB3In0" +
                    ".eyJzdWIiOiJqZnJ0QDAxZGtqMTR5bXNiamczMDRtbWJ4MXExbnNnXC91c2Vyc1wvYWxl" +
                    "eCIsInNjcCI6Im1lbWJlci1vZi1ncm91cHM6cmVhZGVycyBhcGk6KiIsImF1ZCI6ImpmcnRAMDFka" +
                    "2oxNHltc2JqZzMwNG1tYngxcTFu" +
                    "c2ciLCJpc3MiOiJqZnJ0QDAxZGtqMTR5bXNiamczMDRtbWJ4MXExbnNnXC91c2Vyc1wvYWRtaW4iL" +
                    "CJleHAiOjE1NjczMjkyMDAsImlh" +
                    "dCI6MTU2NzMyNTYwMCwianRpIjoiYzBhY2QzYjMtZWI4OS00M2NhLWIxNTMtY2FlZjk5OTMxOWFhIn0" +
                    ".DDACJum60qihTypKCZ7irlU1" +
                    "c6AfEmxqFZW5seYwxLsiumbxF3eKiVTHcCaWXNP3fOjsnRKUqie0AerJB7KLlhSSlH2Iu2tFK0tQThT" +
                    "h_Wzt_D27BD1K7ZQAaNLy5p-Leons9jK6HKaDk70wxK3FT08gHaN-HiOslfifbYA_gfOl0EeH5" +
                    "-iW6zOj2sK7d1ixqsNbqbbW4VGK1lF1A9YASFWirHgH8Oph0zA02cowyAWGZ6N7NA-oinyldIGhExN" +
                    "kOzQ5HwrWZHyplK6UZwtZ2aXWHihOBwcSjGklzSGqA4Xs1n5IBmlLcm__78HxUR0ptOW6U7R6mME0IZ8zJ3-uuA";

    private SecurityContextImpl securityContext;
    private AccessService accessService;
    private SecurityServiceImpl service;
    private List<RepoAcl> testAcls;
    private AclStoreService aclStoreServiceMock;
    private InternalRepositoryService repositoryServiceMock;
    private InternalBuildService buildServiceMock;
    private LocalRepo localRepoMock;
    private LocalRepo cacheRepoMock;
    private InternalCentralConfigService centralConfigServiceMock;
    private UserGroupStoreService userGroupStoreService;
    private ApiKeyManager apiKeyManager;
    private SecurityListener securityListenerMock;
    private SecurityServiceImplTestHelper securityServiceImplTestHelper;
    private AclCache<RepoPermissionTarget> aclCache;
    private AclCache<RepoPermissionTarget> emptyAclCache;
    private UserPassAuthenticationProvider userPassAuthenticationProvider;
    private V600DecryptAllUsersCustomData decryptAllUsersCustomData;
    private V5100ConvertResourceTypeToRepo resourceTypeToRepoConverter;
    private V6600CreateDefaultBuildAcl createDefaultBuildAcl;
    private AccessConverters accessConverters;
    private MarkerMigrationHelper markerMigrationHelper;
    private SecurityConfigurationChangesInterceptors interceptors;
    private MigrationStatusDao migrationStatusDao = Mockito.mock(MigrationStatusDao.class);
    private ProjectsService projectsService;
    private UserLockInMemoryService userLockInMemoryService;
    private List<DbMigrationStatus> mockDbStatusMigrations = Lists.newArrayList();
    private Group groupMock = Mockito.mock(Group.class);

    @DataProvider
    public static Object[][] applyEnabled() {
        return new Object[][] {
                {true},
                {false}
        };
    }

    @BeforeClass
    public void initArtifactoryRoles() throws SQLException {
        securityServiceImplTestHelper = new SecurityServiceImplTestHelper();
        testAcls = securityServiceImplTestHelper.createTestAcls();
        aclCache = securityServiceImplTestHelper.createUserAndGroupResultMap();
        emptyAclCache = securityServiceImplTestHelper.createEmptyAclCache();
        aclStoreServiceMock = createMock(AclStoreService.class);
        repositoryServiceMock = createRepoServiceMock();
        projectsService = Mockito.mock(ProjectsService.class);
        buildServiceMock = createBuildServiceMock();
        centralConfigServiceMock = createMock(InternalCentralConfigService.class);
        userGroupStoreService = createMock(UserGroupStoreService.class);
        apiKeyManager = Mockito.mock(ApiKeyManager.class);
        localRepoMock = createLocalRepoMock();
        cacheRepoMock = createCacheRepoMock();
        securityListenerMock = createMock(SecurityListener.class);
        userPassAuthenticationProvider = createMock(UserPassAuthenticationProvider.class);
        accessService = createMock(AccessService.class);
        decryptAllUsersCustomData = Mockito.mock(V600DecryptAllUsersCustomData.class);
        resourceTypeToRepoConverter = Mockito.mock(V5100ConvertResourceTypeToRepo.class);
        createDefaultBuildAcl = Mockito.mock(V6600CreateDefaultBuildAcl.class);
        markerMigrationHelper = new MarkerMigrationHelper(new MigrationStatusStorageServiceImpl(migrationStatusDao),
                homeStub);
        projectsService = Mockito.mock(ProjectsService.class);
        interceptors = Mockito.mock(SecurityConfigurationChangesInterceptors.class);
        when(decryptAllUsersCustomData.getMarkerFile())
                .thenReturn(homeStub.getAccessUserCustomDataDecryptionMarkerFile());
        when(decryptAllUsersCustomData.getMarkerMigrationConverter())
                .thenReturn(new V600AddAccessDecryptAllUsersMarkerMigration());

        when(resourceTypeToRepoConverter.getMarkerFile())
                .thenReturn(homeStub.getAccessResourceTypeConverterMarkerFile());
        when(resourceTypeToRepoConverter.getMarkerMigrationConverter())
                .thenReturn(new V5100AddAccessResourceTypeMarkerMigration());

        when(createDefaultBuildAcl.getMarkerFile())
                .thenReturn(homeStub.getCreateDefaultBuildPermissionMarkerFile());
        when(createDefaultBuildAcl.getMarkerMigrationConverter())
                .thenReturn(new V658AnonAccessToBuildsMarkerMigration());
        MarkerMigrationHelper.bind(markerMigrationHelper);

        mockDaoRequests();
        accessConverters =
                new AccessConverters(resourceTypeToRepoConverter, decryptAllUsersCustomData,
                        createDefaultBuildAcl);
        userLockInMemoryService = createMock(UserLockInMemoryService.class);
    }

    @BeforeMethod
    public void setUp() {
        homeStub.setProperty(ConstantValues.allowAnyUpgrade, "7.0.0-m001");
        // create new security context
        securityContext = new SecurityContextImpl();
        SecurityContextHolder.setContext(securityContext);

        resetMockDbStatusMigrations();

        // new service instance
        service = spy(new SecurityServiceImpl());
        // set the aclManager mock on the security service
        ReflectionTestUtils.setField(service, "userGroupStoreService", userGroupStoreService);
        ReflectionTestUtils.setField(service, "apiKeyManager", apiKeyManager);
        ReflectionTestUtils.setField(service, "aclStoreService", aclStoreServiceMock);
        ReflectionTestUtils.setField(service, "repositoryService", repositoryServiceMock);
        ReflectionTestUtils.setField(service, "buildService", buildServiceMock);
        ReflectionTestUtils.setField(service, "centralConfig", centralConfigServiceMock);
        ReflectionTestUtils.setField(service, "userPassAuthenticationProvider", userPassAuthenticationProvider);
        ReflectionTestUtils.setField(service, "accessConverters", accessConverters);
        ReflectionTestUtils.setField(service, "userLockInMemoryService", userLockInMemoryService);
        ReflectionTestUtils.setField(service, "markerMigrationHelper", markerMigrationHelper);
        ReflectionTestUtils.setField(service, "projectsService", projectsService);
        ReflectionTestUtils.setField(service, "interceptors", interceptors);
        ReflectionTestUtils.setField(service, "accessService", accessService);

        // reset mocks
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock, centralConfigServiceMock, userLockInMemoryService,
                accessService, userGroupStoreService, accessService);
        Mockito.reset(apiKeyManager);
        FileUtils.deleteQuietly(homeStub.getAccessUserCustomDataDecryptionMarkerFile());
    }

    @AfterMethod
    public void afterMethod() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @DataProvider
    public static Object[][] scopedActionsAndExpected() {
        return new Object[][] {
                {PermissionScopeAction.WRITE, true, false, false, false, false},
                {PermissionScopeAction.READ, false, true, false, false, false},
                {PermissionScopeAction.ANNOTATE, false, false, true, false, false},
                {PermissionScopeAction.DELETE, false, false, false, true, false},
                {PermissionScopeAction.MANAGE, false, false, false, false, false}
        };
    }

    @Test(dataProvider = "scopedActionsAndExpected")
    public void testActionsByPermissionScopedToken(PermissionScopeAction permissionScopeAction,
            boolean canDeploy, boolean canRead, boolean canAnnotate, boolean canDelete, boolean canManage) {
        AccessService accessServiceMock = mock(AccessService.class, Answers.RETURNS_DEEP_STUBS);
        LoginResult loginResult = mock(LoginResult.class);
        AccessPrincipal accessPrincipal = mock(AccessPrincipal.class);
        when(loginResult.getPrincipal()).thenReturn(accessPrincipal);
        when(accessServiceMock.getAccessClient().auth().loginByAuthHeader(any(String.class))).thenReturn(loginResult);
        when(accessServiceMock.getAccessClient().auth().principalCanDo(any(), any())).thenAnswer(
                (Answer<Boolean>) invocationOnMock -> {
                    PermissionScope permissionScope = invocationOnMock.getArgument(1);
                    return permissionScope.getAction() == permissionScopeAction;
                });
        service.setAccessService(accessServiceMock);

        authenticateUserWithScopes(false, Lists.newArrayList("artifact:repoKey:" + permissionScopeAction.getCode()));

        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("repoKey")).andReturn(null).anyTimes();
        replay(repositoryServiceMock);
        expectAclScan();

        RepoPath repoPath = RepoPathFactory.create("repoKey", "path");
        assertEquals(service.canDeploy(repoPath), canDeploy);
        assertEquals(service.canRead(repoPath), canRead);
        assertEquals(service.canAnnotate(repoPath), canAnnotate);
        assertEquals(service.canDelete(repoPath), canDelete);
        assertEquals(service.canManage(repoPath), canManage);
    }

    private void authenticateUserWithScopes(boolean isAdmin, List<String> scopes) {
        SimpleUser simpleUser;
        if (isAdmin) {
            simpleUser = createAdminUser("admin");
        } else {
            simpleUser = createNonAdminUser("non-admin");
        }
        Authentication authentication = getAuthentication(scopes, simpleUser);
        SecurityContext securityContext = mock(SecurityContext.class);
        when(securityContext.getAuthentication()).thenReturn(authentication);
        SecurityContextHolder.setContext(securityContext);
    }

    private Authentication getAuthentication(List<String> scopes, SimpleUser simpleUser) {
        Authentication authentication;
        if (CollectionUtils.isEmpty(scopes)) { // Basic Auth
            authentication = mock(AccessResponseAuthentication.class);
        } else {
            JwtAccessToken jwtToken = mock(JwtAccessToken.class);
            when(jwtToken.getScope()).thenReturn(scopes);
            AccessTokenAuthentication tokenAuthentication = mock(AccessTokenAuthentication.class);
            when(tokenAuthentication.getAccessToken()).thenReturn(jwtToken);
            authentication = tokenAuthentication;
        }
        when(authentication.isAuthenticated()).thenReturn(true);
        when(authentication.getPrincipal()).thenReturn(simpleUser);
        return authentication;
    }

    public void currentUserEncryptedPasswordOrApiKey() {
        // Prepare regular user authentication
        String userName = "user";
        String userEmail = "user@gmail.com";
        UserInfo userInfoMock = mock(UserInfo.class);
        when(userInfoMock.getEmail()).thenReturn(userEmail);
        when(userInfoMock.isUpdatableProfile()).thenReturn(true);
        when(userInfoMock.isPasswordDisabled()).thenReturn(true);
        CentralConfigDescriptor centralConfigDescriptorMock = mock(CentralConfigDescriptor.class);
        SecurityDescriptor securityDescriptorMock = mock(SecurityDescriptor.class);
        PasswordSettings passwordSettingsMock = mock(PasswordSettings.class);
        when(centralConfigDescriptorMock.getSecurity()).thenReturn(securityDescriptorMock);
        when(securityDescriptorMock.getPasswordSettings()).thenReturn(passwordSettingsMock);
        expect(accessService.isEncryptionEnabled()).andReturn(false).anyTimes();
        expect(userGroupStoreService.findUser(eq(userName))).andReturn(userInfoMock).anyTimes();
        expect(centralConfigServiceMock.getDescriptor()).andReturn(centralConfigDescriptorMock).anyTimes();
        when(apiKeyManager.getToken(userName)).thenReturn(null);
        replay(userGroupStoreService, centralConfigServiceMock);

        setSimpleUserAuthentication();
        // And expect password returning
        String password = service.currentUserEncryptedPasswordOrApiKey();
        assertEquals(password, "password");

        // Now switch to external user with AccessTokenAuthentication e.g. LDAP
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(context);
        Authentication accessTokenAuthentication = mock(AccessTokenAuthentication.class);
        securityContext.setAuthentication(accessTokenAuthentication);
        when(userInfoMock.isExternal()).thenReturn(true);
        when(userInfoMock.getRealm()).thenReturn(Realm.LDAP.getName());

        // For LDAP expect null (instead of password and JWT-token)
        password = service.currentUserEncryptedPasswordOrApiKey();
        assertNull(password);
    }

    public void isAdminOnAdminUser() {
        Authentication authentication = setAdminAuthentication();

        boolean admin = service.isAdmin();
        assertTrue(admin, "The user in test is admin");
        // un-authenticate
        authentication.setAuthenticated(false);
        admin = service.isAdmin();
        assertFalse(admin, "Unauthenticated token");
    }

    public void isAdminOnSimpleUser() {
        setSimpleUserAuthentication();

        boolean admin = service.isAdmin();
        assertFalse(admin, "The user in test is not an admin");
    }

    public void isAdminOnProjectAdmin() {
        setProjectAdminAuthentication(true, false);
        boolean admin = service.isAdmin();
        assertFalse(admin, "The user in test is not an admin");
    }

    public void isProjectAdminOnSimpleUser() {
        setProjectAdminAuthentication(false, false);
        boolean projectAdmin = service.isProjectAdmin(PROJECT_A);
        assertFalse(projectAdmin, NOT_A_PROJECT_ADMIN_ERROR);
    }

    public void isResourceManagerOnSimpleUser() {
        setProjectAdminAuthentication(false, true);
        boolean resourceAdmin = service.isResourceManager();
        assertTrue(resourceAdmin);
    }

    public void isResourceManagerOnSimpleUser2() {
        setProjectAdminAuthentication(true, true);
        boolean resourceAdmin = service.isResourceManager();
        assertTrue(resourceAdmin);
    }

    public void isProjectAdminOnAdmin() {
        setAdminAuthentication();
        boolean projectAdmin = service.isProjectAdmin(PROJECT_A);
        assertTrue(projectAdmin, IS_SUPER_ADMIN_ERROR);
    }

    public void isProjectAdminOnProjectAdmin() {
        setProjectAdminAuthentication(true, false);
        boolean projectAdmin = service.isProjectAdmin(PROJECT_A);
        assertTrue(projectAdmin, IS_PROJECT_ADMIN_ERROR);
    }

    public void isProjectAdminOnResourceSimpleUser() {
        setProjectAdminAuthentication(false, false);
        boolean projectAdmin = service.isProjectAdmin(REPO);
        assertFalse(projectAdmin, NOT_A_PROJECT_ADMIN_ERROR);
    }

    public void isProjectAdminOnResourceAdminUser() {
        setAdminAuthentication();
        boolean projectAdmin = service.isProjectAdmin(REPO);
        assertTrue(projectAdmin, IS_SUPER_ADMIN_ERROR);
    }

    public void isProjectAdminOnResourceProjectAdminUser() {
        setProjectAdminAuthentication(true, false);
        boolean projectAdmin = service.isProjectAdmin(REPO);
        assertTrue(projectAdmin, IS_PROJECT_ADMIN_ERROR);
    }

    public void isAnyProjectAdminOnSimpleUser() {
        setProjectAdminAuthentication(false, false);
        boolean projectAdmin = service.isAnyProjectAdmin();
        assertFalse(projectAdmin, NOT_A_PROJECT_ADMIN_ERROR);
    }

    public void isAnyProjectAdminOnAdminUser() {
        setAdminAuthentication();
        boolean projectAdmin = service.isAnyProjectAdmin();
        assertTrue(projectAdmin, IS_PROJECT_ADMIN_ERROR);
    }

    public void isAnyProjectAdminOnProjectAdminUser() {
        setProjectAdminAuthentication(true, false);
        boolean projectAdmin = service.isAnyProjectAdmin();
        assertTrue(projectAdmin, IS_PROJECT_ADMIN_ERROR);
    }

    public void canManageProjectResourcesSuperAdmin() {
        setAdminAuthentication();
        boolean canManage = service.canManageProjectResources(PROJECT_A);
        assertTrue(canManage, IS_SUPER_ADMIN_ERROR);
        canManage = service.canManageResource(REPO);
        assertTrue(canManage, IS_SUPER_ADMIN_ERROR);
        canManage = service.canManageResource(SHARED_REPO);
        assertTrue(canManage, IS_SUPER_ADMIN_ERROR);
    }

    public void canManageProjectResourcesProjectAdmin() {
        setProjectAdminAuthentication(true, false);
        boolean canManage = service.canManageProjectResources(PROJECT_A);
        assertTrue(canManage, IS_PROJECT_ADMIN_ERROR);
        canManage = service.canManageResource(REPO);
        assertTrue(canManage, IS_PROJECT_ADMIN_ERROR);
        canManage = service.canManageResource(SHARED_REPO);
        assertTrue(canManage, IS_PROJECT_ADMIN_ERROR);
    }

    public void canManageProjectResourcesSimpleUser() {
        setProjectAdminAuthentication(false, false);
        boolean canManage = service.canManageProjectResources(PROJECT_A);
        assertFalse(canManage, NOT_A_PROJECT_ADMIN_ERROR);
        canManage = service.canManageResource(REPO);
        assertFalse(canManage, NOT_A_PROJECT_ADMIN_ERROR);
        canManage = service.canManageResource(SHARED_REPO);
        assertFalse(canManage, NOT_A_PROJECT_ADMIN_ERROR);
    }

    public void testIsValidAuthInCaseMfaEnabled() {
        setSimpleUserAuthentication("testUser", true);
        assertThrows(ForbiddenException.class, () -> service.assertValidAuthInCaseMfaEnabled());

        setSimpleUserAuthentication("testUser", false);
        Assertions.assertDoesNotThrow(() -> service.assertValidAuthInCaseMfaEnabled());

        setAccessTokenAuthentication("testUser", false);
        Assertions.assertDoesNotThrow(() -> service.assertValidAuthInCaseMfaEnabled());

        setAccessTokenAuthentication("testUser", true);
        Assertions.assertDoesNotThrow(() -> service.assertValidAuthInCaseMfaEnabled());
    }

    public void testAnotherProjectAdminHasNoPermissionsOnProjectRelatedPath() {
        setProjectAdminAuthentication(false, false);
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(REPO.getRepoKey())).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey(REPO.getRepoKey())).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        expectGetAllAclsCallWithAnyArray();
        expectAclScan();

        assertFalse(service.canManage(REPO), "User should not have manage permissions for this path");
        assertFalse(service.canDelete(REPO), "User should not have delete permissions for this path");
        assertFalse(service.canAnnotate(REPO), "User should not have annotate permissions for this path");
        assertFalse(service.canDeploy(REPO), "User should not have deploy permissions for this path");
        assertFalse(service.canRead(REPO), "User should not have read permissions for this path");

        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);
    }

    @Test(dependsOnMethods = "isAdminOnAdminUser")
    public void spidermanCanDoAnything() {
        setAdminAuthentication();
        assertFalse(service.isAnonymous());// sanity
        assertTrue(service.isAdmin());// sanity

        RepoPath path = InternalRepoPathFactory.create("someRepo", "blabla");
        boolean canRead = service.canRead(path);
        assertTrue(canRead);
        boolean canDeploy = service.canDeploy(path);
        assertTrue(canDeploy);
    }

    @Test
    public void getUserEffectiveSids() {
        SimpleUser userMock = Mockito.mock(SimpleUser.class);
        Mockito.when(userMock.getUsername()).thenReturn("tester");

        UserInfo userInfoMock = Mockito.mock(UserInfo.class);
        Mockito.when(userMock.getDescriptor()).thenReturn(userInfoMock);
        Set<UserGroupInfo> groups = new HashSet<>();
        for (int i = 1; i <= 1000; i++) {
            groups.add(new UserGroupImpl("grp" + i));
        }
        when(userInfoMock.getGroups()).thenReturn(groups);

        Set<ArtifactorySid> userEffectiveSids = SecurityServiceImpl.getUserEffectiveSids(userMock);
        // verify first sid is the actual user
        ArtifactorySid firstSid = userEffectiveSids.iterator().next();
        assertFalse(firstSid.isGroup());
        assertEquals(firstSid.getPrincipal(), "tester");

        // make sure all groups exists
        assertEquals(groups.size(), userEffectiveSids.stream().filter(ArtifactorySid::isGroup).count());

        // make sure all expected groups existed in the result
        Set<String> returnedGroupNames = userEffectiveSids.stream()
                .filter(sid -> !sid.getPrincipal().equals("tester"))
                .map(ArtifactorySid::getPrincipal)
                .collect(Collectors.toSet());

        Set<String> expectedGroupNames = groups.stream()
                .map(UserGroupInfo::getGroupName)
                .collect(Collectors.toSet());
        assertTrue(returnedGroupNames.containsAll(expectedGroupNames));
    }

    @Test(expectedExceptions = RuntimeException.class, expectedExceptionsMessageRegExp = ".*not permitted to reset " +
            "his password.")
    public void testResetPasswordWithExternalAuthFailedWhereUserPasswordDisabled() {
        String userName = "tester";
        String userEmail = "tester@gmail.com";
        UserInfo userInfoMock = mock(UserInfo.class);
        CentralConfigDescriptor centralConfigDescriptorMock = mock(CentralConfigDescriptor.class);
        SecurityDescriptor securityDescriptorMock = mock(SecurityDescriptor.class);
        PasswordSettings passwordSettingsMock = mock(PasswordSettings.class);
        PasswordResetPolicy passwordResetPolicyMock = mock(PasswordResetPolicy.class);
        when(centralConfigDescriptorMock.getSecurity()).thenReturn(securityDescriptorMock);
        when(securityDescriptorMock.getPasswordSettings()).thenReturn(passwordSettingsMock);
        when(passwordSettingsMock.getResetPolicy()).thenReturn(passwordResetPolicyMock);
        when(passwordResetPolicyMock.isEnabled()).thenReturn(false);
        when(userInfoMock.getEmail()).thenReturn(userEmail);
        when(userInfoMock.isUpdatableProfile()).thenReturn(true);
        when(userInfoMock.isPasswordDisabled()).thenReturn(true);
        expect(userGroupStoreService.findUser(userName)).andReturn(userInfoMock).anyTimes();
        expect(centralConfigServiceMock.getDescriptor()).andReturn(centralConfigDescriptorMock).anyTimes();
        replay(userGroupStoreService, centralConfigServiceMock);
        service.resetPassword(userName, null, null);
    }

    public void userReadAndDeployPermissions() {
        setSimpleUserAuthentication();

        RepoPath securedPath = InternalRepoPathFactory.create("securedRepo", "blabla");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey(securedPath.getRepoKey()))
                .andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(securedPath.getRepoKey())).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey(securedPath.getRepoKey())).andReturn(null)
                .anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        // cannot read the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean hasPermissions =
                service.canRead(securedPath) || service.canManage(securedPath) || service.canDeploy(securedPath) ||
                        service.canDelete(securedPath);
        assertFalse(hasPermissions, "User should not have permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        // cannot deploy to the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);

        boolean canDeploy = service.canDeploy(securedPath);
        assertFalse(canDeploy, "User should not have permissions for this path");
        EasyMock.verify(aclStoreServiceMock, repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        RepoPath allowedReadPath = InternalRepoPathFactory.create("testRepo1", "blabla");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey(allowedReadPath.getRepoKey()))
                .andReturn(localRepoMock).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(allowedReadPath.getRepoKey())).andReturn(null)
                .anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey(allowedReadPath.getRepoKey())).andReturn(null)
                .anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        // can read the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        hasPermissions = service.canRead(allowedReadPath);
        assertTrue(hasPermissions, "User should have read permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        // cannot deploy to the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        canDeploy = service.canDeploy(allowedReadPath);
        assertFalse(canDeploy, "User should not have permissions for this path");
        verify(aclStoreServiceMock, repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        // cannot admin the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean canAdmin = service.canManage(allowedReadPath);
        assertFalse(canAdmin, "User should not have permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
    }

    public void adminRolePermissions() {
        // user with admin role on permission target 'target1'
        setSimpleUserAuthentication("yossis");

        RepoPath allowedReadPath = InternalRepoPathFactory.create("testRepo1", "blabla");

        // can read the specified path
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectGetAllAclsCall();
        replay(aclStoreServiceMock, repositoryServiceMock);
        boolean canRead = service.canRead(allowedReadPath);
        assertTrue(canRead, "User should have permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        // can deploy to the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean canDeploy = service.canDeploy(allowedReadPath);
        assertTrue(canDeploy, "User should have permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        // can admin the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean canAdmin = service.canManage(allowedReadPath);
        assertTrue(canAdmin, "User should have permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);
    }

    public void groupPermissions() {
        RepoPath allowedReadPath = InternalRepoPathFactory.create("testRepo1", "**");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey(allowedReadPath.getRepoKey()))
                .andReturn(localRepoMock).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        replay(repositoryServiceMock);

        // cannot deploy to the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean canDeploy = service.canDeploy(allowedReadPath);
        assertFalse(canDeploy, "User should have permissions for this path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        // add the user to a group with permissions and expext permission garnted
        setSimpleUserAuthentication("userwithnopermissions", "deployGroup");
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canDeploy = service.canDeploy(allowedReadPath);
        assertTrue(canDeploy, "User in a group with permissions for this path");
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);
    }

    public void userWithPermissionsToAGroupWithTheSameName() {
        setSimpleUserAuthentication(securityServiceImplTestHelper.USER_AND_GROUP_SHARED_NAME,
                securityServiceImplTestHelper.USER_AND_GROUP_SHARED_NAME);

        RepoPath testRepo1Path = InternalRepoPathFactory.create("testRepo1", "**");
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canRead = service.canRead(testRepo1Path);
        assertTrue(canRead, "User should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        RepoPath testRepo2Path = InternalRepoPathFactory.create("testRepo2", "**");
        expectGetAllAclsCallWithAnyArray();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo2")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        replay(aclStoreServiceMock);
        canRead = service.canRead(testRepo2Path);
        assertTrue(canRead, "User belongs to a group with permissions to the path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);
    }

    public void userWithPermissionsToANonUniqueGroupName() {
        // here we test that a user that belongs to a group which has
        // the same name of a nother user will only get the group permissions
        // and not the user with the same name permissions
        setSimpleUserAuthentication(PERMISSION_ACTOR, securityServiceImplTestHelper.USER_AND_GROUP_SHARED_NAME);

        RepoPath testRepo1Path = InternalRepoPathFactory.create("testRepo1", "**");
        expectGetAllAclsCallWithAnyArray();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo1")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        replay(aclStoreServiceMock);
        boolean canRead = service.canRead(testRepo1Path);
        assertFalse(canRead, "User should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        RepoPath testRepo2Path = InternalRepoPathFactory.create("testRepo2", "**");
        expectGetAllAclsCallWithAnyArray();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo2")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        replay(aclStoreServiceMock, repositoryServiceMock);
        canRead = service.canRead(testRepo2Path);
        EasyMock.verify(repositoryServiceMock);
        assertTrue(canRead, "User belongs to a group with permissions to the path");
    }

    public void hasPermissionPassingUserInfo() {
        SimpleUser user = createNonAdminUser("yossis");
        UserInfo userInfo = user.getDescriptor();

        RepoPath testRepo1Path = InternalRepoPathFactory.create("testRepo1", "any/path");

        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock, repositoryServiceMock);
        boolean canRead = service.canRead(userInfo, testRepo1Path);
        assertTrue(canRead, "User should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canDeploy = service.canDeploy(userInfo, testRepo1Path);
        assertTrue(canDeploy, "User should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo1")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canDelete = service.canDelete(userInfo, testRepo1Path);
        assertFalse(canDelete, "User should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canAdmin = service.canManage(userInfo, testRepo1Path);
        assertTrue(canAdmin, "User should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        RepoPath testRepo2Path = InternalRepoPathFactory.create("testRepo2", "**");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo2")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canRead = service.canRead(userInfo, testRepo2Path);
        assertFalse(canRead, "User should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        SecurityDescriptor securityDescriptor = new SecurityDescriptor();

        CentralConfigDescriptor configDescriptor = createMock(CentralConfigDescriptor.class);
        expect(configDescriptor.getSecurity()).andReturn(securityDescriptor).anyTimes();
        replay(configDescriptor);
        expect(centralConfigServiceMock.getDescriptor()).andReturn(configDescriptor).anyTimes();
        replay(centralConfigServiceMock);

        SimpleUser anon = createNonAdminUser(UserInfo.ANONYMOUS);
        UserInfo anonUserInfo = anon.getDescriptor();

        RepoPath testMultiRepo = InternalRepoPathFactory.create("multi1", "**");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("multi1")).andReturn(cacheRepoMock)
                .anyTimes();
        expectGetAllAclsCallWithAnyArray();
        replay(repositoryServiceMock);

        canRead = service.canRead(anonUserInfo, testMultiRepo);
        EasyMock.verify(repositoryServiceMock);
        assertFalse(canRead, "Anonymous user should have permissions for this path");
    }

    public void hasPermissionWithSpecificTarget() {
        SimpleUser user = createNonAdminUser("shay");
        UserInfo userInfo = user.getDescriptor();
        RepoPath testRepo1Path = InternalRepoPathFactory.create("specific-repo", "com", true);

        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("specific-repo")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("specific-repo")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("specific-repo")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canRead = service.canRead(userInfo, testRepo1Path);
        assertTrue(canRead, "User should have read permissions for this path");
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canDeploy = service.canDeploy(userInfo, testRepo1Path);
        assertTrue(canDeploy, "User should have deploy permissions for this path");
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canDelete = service.canDelete(userInfo, testRepo1Path);
        assertFalse(canDelete, "User should not have delete permissions for this path");
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canAdmin = service.canManage(userInfo, testRepo1Path);
        assertFalse(canAdmin, "User should not have admin permissions for this path");
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);
    }

    @Test(dataProvider = "applyEnabled")
    public void hasPermissionsOnRemoteRepoWhenUsingLocalRepoPermissions(boolean applyEnabled) {
        getBound().setProperty(ConstantValues.applyLocalReposPermissionsOnRemoteRepos, String.valueOf(applyEnabled));
        setSimpleUserAuthentication();

        expect(repositoryServiceMock.repositoryByKey("testRepo1")).andReturn(createRemoteRepoMock()).anyTimes();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo1-cache")).andReturn(null)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1-cache"))
                .andReturn(null).anyTimes(); // Local remote repo
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1"))
                .andReturn(new HttpRepoDescriptor()).anyTimes(); // Local remote repo
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1-cache")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        // cannot read the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean hasPermissionOnRemoteRoot = service.userHasPermissionsOnRepositoryRoot("testRepo1");
        assertEquals(hasPermissionOnRemoteRoot, applyEnabled);
        EasyMock.verify(aclStoreServiceMock, repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);
    }

    @Test(dataProvider = "applyEnabled")
    public void hasPermissionsOnLocalRepoWhenUsingRemoteRepoPermissions(boolean applyEnabled) {
        getBound().setProperty(ConstantValues.applyRemoteReposPermissionsOnLocalRepos, String.valueOf(applyEnabled));
        setSimpleUserAuthentication();

        LocalRepo localRepo = createMock(LocalRepo.class);
        expect(localRepo.isLocal()).andReturn(true).anyTimes();
        expect(localRepo.isCache()).andReturn(false).anyTimes();
        expect(localRepo.isReal()).andReturn(true).anyTimes();
        replay(localRepo);

        expect(repositoryServiceMock.repositoryByKey("testRemote")).andReturn(localRepo).anyTimes();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRemote")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRemote")).andReturn(null)
                .anyTimes(); // Local remote repo
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRemote")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();

        // cannot read the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean hasPermissionOnRemoteRoot = service.userHasPermissionsOnRepositoryRoot("testRemote");
        assertEquals(hasPermissionOnRemoteRoot, applyEnabled);
        EasyMock.verify(aclStoreServiceMock, repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);
    }

    public void hasPermissionForGroupInfo() {
        GroupInfo groupInfo = InfoFactoryHolder.get().createGroup("deployGroup");

        RepoPath testRepo1Path = InternalRepoPathFactory.create("testRepo1", "any/path");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo1")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo1")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canRead = service.canRead(groupInfo, testRepo1Path);
        assertFalse(canRead, "Group should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canDeploy = service.canDeploy(groupInfo, testRepo1Path);
        assertTrue(canDeploy, "Group should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canDelete = service.canDelete(groupInfo, testRepo1Path);
        assertFalse(canDelete, "Group should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        boolean canAdmin = service.canManage(groupInfo, testRepo1Path);
        assertFalse(canAdmin, "Group should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        RepoPath testRepo2Path = InternalRepoPathFactory.create("testRepo2", "some/path");

        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRepo2")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRepo2")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canRead = service.canRead(groupInfo, testRepo2Path);
        assertFalse(canRead, "Group should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        GroupInfo anyRepoGroupRead = InfoFactoryHolder.get().createGroup("anyRepoReadersGroup");
        RepoPath somePath = InternalRepoPathFactory.create("blabla", "some/path");

        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("blabla")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("blabla")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("blabla")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canRead = service.canRead(anyRepoGroupRead, somePath);
        assertTrue(canRead, "Group should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canDeploy = service.canDeploy(anyRepoGroupRead, somePath);
        assertFalse(canDeploy, "Group should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);

        GroupInfo multiRepoGroupRead = InfoFactoryHolder.get().createGroup("multiRepoReadersGroup");
        RepoPath multiPath = InternalRepoPathFactory.create("multi1", "some/path");
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("multi1")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("multi1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("multi1")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("multi2")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("multi2")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("multi2")).andReturn(null).anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canRead = service.canRead(multiRepoGroupRead, multiPath);
        assertTrue(canRead, "Group should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        RepoPath multiPath2 = InternalRepoPathFactory.create("multi2", "some/path");
        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canRead = service.canRead(multiRepoGroupRead, multiPath2);
        assertTrue(canRead, "Group should have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock);

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock);
        canDeploy = service.canDeploy(multiRepoGroupRead, multiPath);
        assertFalse(canDeploy, "Group should not have permissions for this path");
        EasyMock.verify(repositoryServiceMock);
    }

    public void getAllPermissionTargetsForAdminUser() {
        setAdminAuthentication();
        Map<String, Set<PrincipalPermission<RepoPermissionTarget>>> adminRepoToAclMap =
                aclCache.getUserResultMap().get("yossis");
        for (String repoPath : adminRepoToAclMap.keySet()) {
            RepoPath path = RepoPathFactory.create(repoPath);
            assertTrue(service.canManage(path), "Admin should be manager of all repos in his cache map");
        }
        assertTrue(service.canManage(RepoPathFactory.create("multi1")),
                "Admin should be manager of repos not belonging to his cache map");
    }

    public void getAllPermissionTargetsForUserWithNoPermission() {
        setSimpleUserAuthentication("noadminpermissionsuser");

        expectAclScan();

        List<RepoPermissionTarget> permissionTargets = service.getRepoPermissionTargets(ArtifactoryPermission.MANAGE);
        assertEquals(permissionTargets.size(), 0);

        EasyMock.verify(aclStoreServiceMock);
    }

    @Test(enabled = false)
    public void getDeployPermissionTargetsForUserWithNoPermission() {
        setSimpleUserAuthentication("user");

        expectAclScan();

        List<RepoPermissionTarget> targets = service.getRepoPermissionTargets(ArtifactoryPermission.DEPLOY);
        assertEquals(targets.size(), 0);

        EasyMock.verify(aclStoreServiceMock);
    }

    public void getDeployPermissionTargetsForUserWithDeployPermission() {
        // yossis should have testRepo1 and testRemote-cache
        setSimpleUserAuthentication("yossis");

        expectAclScan();

        List<RepoPermissionTarget> targets = service.getRepoPermissionTargets(ArtifactoryPermission.DEPLOY);
        assertEquals(targets.size(), 2, "Expecting two deploy permission");

        EasyMock.verify(aclStoreServiceMock);
    }

    public void userPasswordMatches() {
        setSimpleUserAuthentication("user");

        assertTrue(service.userPasswordMatches("password"));
        assertFalse(service.userPasswordMatches(""));
        assertFalse(service.userPasswordMatches("Password"));
        assertFalse(service.userPasswordMatches("blabla"));
    }

    public void permissionOnRemoteRoot() {
        setSimpleUserAuthentication();

        expect(repositoryServiceMock.repositoryByKey("testRemote")).andReturn(createRemoteRepoMock()).anyTimes();
        expect(repositoryServiceMock.localFederatedOrCachedRepositoryByKey("testRemote-cache")).andReturn(null)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRemote-cache")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey("testRemote")).andReturn(new HttpRepoDescriptor())
                .anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey("testRemote-cache")).andReturn(null).anyTimes();
        replay(repositoryServiceMock);

        // cannot read the specified path
        expectGetAllAclsCall();
        replay(aclStoreServiceMock);
        boolean hasPermissionOnRemoteRoot = service.userHasPermissionsOnRepositoryRoot("testRemote");
        assertTrue(hasPermissionOnRemoteRoot, "User should have permissions for this path");
        EasyMock.verify(aclStoreServiceMock, repositoryServiceMock);
        EasyMock.reset(aclStoreServiceMock, repositoryServiceMock);
    }

    @Test(dependsOnMethods = {"permissionOnRemoteRoot"},
            expectedExceptions = {InvalidNameException.class})
    protected void findOrCreateExternalAuthUserWithIllegalChars() {
        EasyMock.reset(userGroupStoreService);
        String userName = new String(NameValidator.getForbiddenChars());
        expect(userGroupStoreService.findUser(userName)).andReturn(null).anyTimes();
        replay(userGroupStoreService);
        service.findOrCreateExternalAuthUser(userName, false);
    }

    @DataProvider
    public static Object[][] provideAccessResourceTypeData() {
        return new Object[][] {
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v540.get(), "",
                        System.currentTimeMillis()), new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(),
                        "", System.currentTimeMillis()), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v400.get(), "",
                        System.currentTimeMillis()), new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(),
                        "", System.currentTimeMillis()), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v590.get(), "",
                        System.currentTimeMillis()), new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(),
                        "", System.currentTimeMillis()), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v5100m009.get(), "",
                        System.currentTimeMillis()), new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(),
                        "", System.currentTimeMillis()), false, false},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v5100m009.get(), "",
                        System.currentTimeMillis()), new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(),
                        "", System.currentTimeMillis()), true, true}
        };
    }

    @DataProvider
    public static Object[][] provideAccessDecryptionTypeData() {
        return new Object[][] {
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v540.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v400.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v590.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v600.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, false},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v600.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), true, true}
        };
    }

    @DataProvider
    public static Object[][] provideCreateDefaultBuildAclTypeData() {
        return new Object[][] {
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v540.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v400.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v590.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v600.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v660.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, false},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v660.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), true, true}
        };
    }

    @DataProvider
    public static Object[][] provideAccessDecryptionAndResourceTypeDataMarker() {
        return new Object[][] {
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v540.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v400.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "", System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), false, false},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v540.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), true, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v400.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), true, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersionProvider.v590.get(), "",
                        System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), true, true},
                {new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "", System.currentTimeMillis()),
                        new CompoundVersionDetails(ArtifactoryServiceVersion.getCurrent(), "",
                                System.currentTimeMillis()
                        ), true, true},
        };
    }

    @Test(dataProvider = "provideAccessResourceTypeData")
    public void testConvertResourceType(CompoundVersionDetails source, CompoundVersionDetails target,
            boolean createMarker, boolean expected)
            throws IOException {
        testConvertAccessConverterAndItsPreInit(resourceTypeToRepoConverter, source, target, createMarker, expected);
    }

    @Test(dataProvider = "provideAccessDecryptionTypeData")
    public void testConvertDecryptPropsType(CompoundVersionDetails source, CompoundVersionDetails target,
            boolean createMarker, boolean expected) throws IOException {
        testConvertAccessConverterAndItsPreInit(decryptAllUsersCustomData, source, target, createMarker, expected);
    }

    @Test(dataProvider = "provideCreateDefaultBuildAclTypeData")
    public void testConvertCreateDefaultBuildAcl(CompoundVersionDetails source, CompoundVersionDetails target,
            boolean createMarker,
            boolean expected) throws IOException {
        testConvertAccessConverterAndItsPreInit(createDefaultBuildAcl, source, target, createMarker, expected);
    }

    @Test(dataProvider = "provideAccessDecryptionAndResourceTypeDataMarker")
    public void testIsInterestedDecryptProps(CompoundVersionDetails source, CompoundVersionDetails target,
            boolean createMarker, boolean expected) throws IOException, SQLException {
        if (createMarker) {
            FileUtils.write(homeStub.getAccessUserCustomDataDecryptionMarkerFile(), "", Charset.defaultCharset());
        }
        when(migrationStatusDao.findAllUnfinishedIn(any()))
                .thenReturn(Collections.emptyList());

        assertEquals(service.isInterested(source, target), expected);
    }

    @Test(dataProvider = "provideAccessDecryptionAndResourceTypeDataMarker")
    public void testIsInterestedResourceType(CompoundVersionDetails source, CompoundVersionDetails target,
            boolean createMarker, boolean expected) throws IOException, SQLException {
        if (createMarker) {
            FileUtils.write(homeStub.getAccessResourceTypeConverterMarkerFile(), "", Charset.defaultCharset());
        }
        when(migrationStatusDao.findAllUnfinishedIn(any()))
                .thenReturn(Collections.emptyList());

        assertEquals(service.isInterested(source, target), expected);
    }

    @Test(dataProvider = "provideGetBuildNameAcls")
    public void testGetBuildNameAcls(String buildName, List<BuildAcl> acls, boolean expectingValues) {
        expect(aclStoreServiceMock.getAllBuildAcls()).andReturn(acls).anyTimes();
        replay(aclStoreServiceMock);
        assertEquals(!service.getBuildPathAcls(buildName, "1", "2015-05-15T11:49:04.804+02:00", BUILDINFO_REPO_KEY)
                        .isEmpty(),
                expectingValues);
    }

    @DataProvider
    public static Object[][] provideGetBuildNameAcls() {
        return new Object[][] {
                // {Build name, permission target with pattern, should build be a match to pattern}
                {"test1", Collections.singletonList(createBuildAcl("test1/**")), true},
                {"test2", Collections.singletonList(createBuildAcl("no")), false},
                {"test3", Collections.singletonList(createBuildAcl("**/test3/**")), true},
                {"test4", Collections.singletonList(createBuildAcl("**/test4")), false},
                {"test5", Collections.singletonList(createBuildAcl("*/test5")), false},
                {"test6", Collections.singletonList(createBuildAcl("someDir/test6")), false},
                {"test7", Collections.singletonList(createBuildAcl("someDir/test7/**")), false},
                {"test8", Collections.singletonList(createBuildAcl("test8/1**")), true},
                {"test9", Collections.singletonList(createBuildAcl("test9/2**")), false},

        };
    }

    public void testIsBuildBasicRead() {
        SecurityDescriptor newSecurityDescriptor = new SecurityDescriptor();
        CentralConfigDescriptor newConfigDescriptor = createMock(CentralConfigDescriptor.class);
        expect(newConfigDescriptor.getSecurity()).andReturn(newSecurityDescriptor).anyTimes();
        replay(newConfigDescriptor);
        expect(accessService.isAnonymousAccessEnabled()).andReturn(false).anyTimes();
        replay(accessService);
        expect(centralConfigServiceMock.getDescriptor()).andReturn(newConfigDescriptor).anyTimes();
        expect(repositoryServiceMock.isPipeInfoRepository(EasyMock.anyString())).andReturn(false).anyTimes();

        replay(centralConfigServiceMock, repositoryServiceMock);

        // ANONYMOUS user with BuildGlobalBasicReadForAnonymous
        setSimpleUserAuthentication(UserInfo.ANONYMOUS);
        newSecurityDescriptor.setBuildGlobalBasicReadAllowed(true);
        newSecurityDescriptor.setBuildGlobalBasicReadForAnonymous(true);
        assertTrue(service.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertTrue(service.isBuildBasicRead(BUILD_NAME, BUILDINFO_REPO_KEY));

        // ANONYMOUS user without BuildGlobalBasicReadForAnonymous
        newSecurityDescriptor.setBuildGlobalBasicReadForAnonymous(false);
        assertFalse(service.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertFalse(service.isBuildBasicRead(BUILD_NAME, BUILDINFO_REPO_KEY));

        //Simple user with BuildGlobalBasicReadAllowed
        setSimpleUserAuthentication();
        newSecurityDescriptor.setBuildGlobalBasicReadAllowed(true);
        assertTrue(service.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertTrue(service.isBuildBasicRead(BUILD_NAME, BUILDINFO_REPO_KEY));

        //Simple user without BuildGlobalBasicReadAllowed and without read permissions
        newSecurityDescriptor.setBuildGlobalBasicReadAllowed(false);
        expectGetBuildsAclCacheCall();
        assertFalse(service.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        expectGetBuildsAclCacheCall();
        assertFalse(service.isBuildBasicRead(BUILD_NAME, BUILDINFO_REPO_KEY));

        // Admin user
        setAdminAuthentication();
        assertTrue(service.isBuildBasicRead(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertTrue(service.isBuildBasicRead(BUILD_NAME, BUILDINFO_REPO_KEY));
    }

    public void testCanReadBuildByNameAndNumberOnly() {
        setAdminAuthentication();
        assertTrue(service.canReadBuild(BUILD_NAME, BUILD_NUMBER, BUILDINFO_REPO_KEY));
        assertFalse(service.canReadBuild(BUILD_NAME, "2", BUILDINFO_REPO_KEY));
    }

    public void testCanUploadBuildByNameAndNumberOnly() {
        setAdminAuthentication();
        assertTrue(service.canUploadBuild(BUILD_NAME, BUILD_NUMBER, BUILDINFO_REPO_KEY));
        assertFalse(service.canUploadBuild(BUILD_NAME, "2", BUILDINFO_REPO_KEY));
    }

    public void testCanReadBuildPath() {
        // Set admin user
        setAdminAuthentication();
        assertTrue(service.canReadBuildPath(BUILD_NAME, BUILDINFO_REPO_KEY));

        //Set simple user with no permission target
        setSimpleUserAuthentication();
        expectGetBuildsAclCacheCall();
        assertFalse(service.canReadBuildPath(BUILD_NAME, BUILDINFO_REPO_KEY));
    }

    public void testCanDeleteBuildByName() {
        // Set admin user
        setAdminAuthentication();
        assertTrue(service.canDeleteBuildName(BUILD_NAME, BUILDINFO_REPO_KEY));

        //Set simple user with no permission target
        setSimpleUserAuthentication();
        expectGetBuildsAclCacheCall();
        assertFalse(service.canDeleteBuildName(BUILD_NAME, BUILDINFO_REPO_KEY));
    }

    public void testBuildPermissionsAsAdmin() {
        // Set admin user
        setAdminAuthentication();
        // Act
        assertTrue(service.canReadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertTrue(service.canUploadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertTrue(service.canDeleteBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        assertTrue(service.canManageBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
    }

    public void testBuildPermissionsAsSimpleUser() {
        //Set simple user with no permission target
        setSimpleUserAuthentication();
        // Act
        expectGetBuildsAclCacheCall();
        assertFalse(service.canReadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        expectGetBuildsAclCacheCall();
        assertFalse(service.canUploadBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        expectGetBuildsAclCacheCall();
        assertFalse(service.canDeleteBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
        expectGetBuildsAclCacheCall();
        assertFalse(service.canManageBuild(BUILD_NAME, BUILD_NUMBER, BUILD_DATE, BUILDINFO_REPO_KEY));
    }

    private static BuildAcl createBuildAcl(String pattern) {
        return createBuildAcl("1", pattern, 0);
    }

    private static BuildAcl createBuildAcl(String name, String pattern, int mask) {
        MutableBuildPermissionTarget buildPermissionTarget = InfoFactoryHolder.get()
                .createBuildPermissionTarget(name, Collections.singletonList("repoKey"));
        buildPermissionTarget.setIncludesPattern(pattern);
        return InfoFactoryHolder.get().createBuildAcl(
                buildPermissionTarget,
                Set.of(InfoFactoryHolder.get().createAce(PERMISSION_ACTOR, false, mask)),
                PERMISSION_ACTOR, 0, name);
    }

    private static RepoAcl createRepoAcl(String name, int mask) {
        InfoFactory factory = InfoFactoryHolder.get();
        return factory.createRepoAcl(
                factory.createRepoPermissionTarget(name, Collections.singletonList("repoKey")),
                Set.of(factory.createAce(PERMISSION_ACTOR, false, mask)),
                null, 0, name);
    }

    private static ReleaseBundleAcl createReleaseBundleAcl(String name, int mask) {
        InfoFactory factory = InfoFactoryHolder.get();
        return factory.createReleaseBundleAcl(
                factory.createReleaseBundlePermissionTarget(name, Collections.singletonList("repoKey")),
                Set.of(factory.createAce(PERMISSION_ACTOR, false, mask)),
                null, 0, name);
    }

    @Test
    public void testDecryptAllUserPropsWithOldEncryption() throws Exception {
        // unbind first, this test is already bound
        createAndSaveArtifactoryKey();
        AccessService accessService = Mockito.mock(AccessService.class);
        // dependency nightmare
        SecurityServiceImpl securityServiceImpl = new SecurityServiceImpl();
        injectField(securityServiceImpl, "accessService", accessService);
        // Prepare
        AccessClient accessClientMockWithData = Mockito.mock(AccessClient.class);
        UsersClient usersClient = Mockito.mock(UsersClient.class);
        when(accessClientMockWithData.users()).thenReturn(usersClient);
        UsersResponse usersResponse = new UsersResponse();
        usersResponse.users(Collections.singletonList(new UserResponse("johnny")
                .status(UserStatus.ENABLED)
                .mfaStatus(UserMFAStatus.NONE)
                .addCustomData(SSH_KEY,
                        ArtifactoryHome.get().getArtifactoryEncryptionWrapper().encryptIfNeeded("notsecure"))));
        when(usersClient.findUsers()).thenReturn(usersResponse);
        when(accessService.getAccessClient()).thenReturn(accessClientMockWithData);

        securityServiceImpl.decryptAllUserProps();
        // Assert - the user should have been updated
        Mockito.verify(usersClient, times(1)).updateUser(any());
    }

    @Test
    public void testDecryptAllUserPropsWithNoEncryptionAndSearchableProperty() throws Exception {
        // unbind first, this test is already bound
        createAndSaveArtifactoryKey();
        AccessService accessService = Mockito.mock(AccessService.class);
        // dependency nightmare
        SecurityServiceImpl securityServiceImpl = new SecurityServiceImpl();
        injectField(securityServiceImpl, "accessService", accessService);
        // Prepare
        AccessClient accessClientMockWithData = Mockito.mock(AccessClient.class);
        UsersClient usersClient = Mockito.mock(UsersClient.class);
        when(accessClientMockWithData.users()).thenReturn(usersClient);
        UsersResponse usersResponse = new UsersResponse();
        usersResponse.users(Collections.singletonList(new UserResponse("johnny")
                .status(UserStatus.ENABLED)
                .mfaStatus(UserMFAStatus.NONE)
                .addCustomData(SSH_KEY, "notsecure")));
        when(usersClient.findUsers()).thenReturn(usersResponse);
        when(accessService.getAccessClient()).thenReturn(accessClientMockWithData);

        securityServiceImpl.decryptAllUserProps();
        // Assert - the user should not be updated
        Mockito.verify(usersClient, times(1)).updateUser(any());
    }

    @Test
    public void testDecryptAllUserPropsWithNoEncryptionAndNonSearchableProperty() throws Exception {
        // unbind first, this test is already bound
        createAndSaveArtifactoryKey();
        AccessService accessService = Mockito.mock(AccessService.class);
        // dependency nightmare
        SecurityServiceImpl securityServiceImpl = new SecurityServiceImpl();
        injectField(securityServiceImpl, "accessService", accessService);
        // Prepare
        AccessClient accessClientMockWithData = Mockito.mock(AccessClient.class);
        UsersClient usersClient = Mockito.mock(UsersClient.class);
        when(accessClientMockWithData.users()).thenReturn(usersClient);
        UsersResponse usersResponse = new UsersResponse();
        usersResponse.users(Collections.singletonList(new UserResponse("johnny")
                .status(UserStatus.ENABLED)
                .mfaStatus(UserMFAStatus.NONE)
                .addCustomData(SSH_KEY, "notsecure")));
        usersResponse.users(Collections.singletonList(new UserResponse("johnny")
                .status(UserStatus.ENABLED)
                .mfaStatus(UserMFAStatus.NONE)
                .addCustomData(SSH_KEY + "_shash", "whatever")));
        when(usersClient.findUsers()).thenReturn(usersResponse);
        when(accessService.getAccessClient()).thenReturn(accessClientMockWithData);

        securityServiceImpl.decryptAllUserProps();
        // Assert - the user should still be updated due to the searchable props being added
        Mockito.verify(usersClient, times(1)).updateUser(any());
    }

    private void injectField(SecurityService securityService, String fieldName, Object object)
            throws Exception {
        Field field = securityService.getClass().getDeclaredField(fieldName);
        field.setAccessible(true);
        field.set(securityService, object);
    }

    private void createAndSaveArtifactoryKey() {
        ArtifactoryHome artifactoryHome = new ArtifactoryHomeStub(getClass().getName());
        artifactoryHome.getArtifactoryKey().delete();
        SecurityFolderHelper.createKeyFile(artifactoryHome.getArtifactoryKey(), CipherAlg.AES128);
    }

    public void testUserHasPermissions() {
        // user shouldn't have any permissions
        SimpleUser user = createNonAdminUser("noperm");
        EasyMock.reset(userGroupStoreService);
        expect(userGroupStoreService.findUser("noperm")).andReturn(user.getDescriptor()).once();

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock, userGroupStoreService);
        Map<String, Set<PrincipalPermission<RepoPermissionTarget>>> nopermPermMap =
                aclCache.getUserResultMap().get("noperm");
        boolean hasNoPermissions = nopermPermMap == null || nopermPermMap.size() == 0;
        assertTrue(hasNoPermissions, "User should not have permissions");
        EasyMock.reset(aclStoreServiceMock, userGroupStoreService);
    }

    public void testUserHasPermissionsFromGroup() {
        // user don't have permissions on his own, but should have permission from his group
        SimpleUser user = createNonAdminUser("noperm", securityServiceImplTestHelper.USER_AND_GROUP_SHARED_NAME);
        EasyMock.reset(userGroupStoreService);
        expect(userGroupStoreService.findUser("noperm")).andReturn(user.getDescriptor()).once();

        expectGetAllAclsCallWithAnyArray();
        replay(aclStoreServiceMock, userGroupStoreService);
        Map<String, Set<PrincipalPermission<RepoPermissionTarget>>> nopermPermMap =
                aclCache.getUserResultMap().get("noperm");
        boolean hasNoPermissions = nopermPermMap == null || nopermPermMap.size() == 0;
        assertTrue(hasNoPermissions, "User shouldn't have permissions on this path");
        hasNoPermissions = user.getDescriptor()
                .getGroups()
                .stream()
                .allMatch(group -> {
                    Map<String, Set<PrincipalPermission<RepoPermissionTarget>>> nopermGroupPermMap =
                            aclCache.getGroupResultMap()
                                    .get(group.getGroupName());
                    return nopermGroupPermMap == null || nopermGroupPermMap.size() == 0;
                });
        assertFalse(hasNoPermissions, "User should have permissions on this path from his groups");
        EasyMock.reset(aclStoreServiceMock, userGroupStoreService);
    }

    public void userReadPermissionsOnAnyRemote() {
        // user should have a read (only) permission on any remote
        setSimpleUserAuthentication("anyRemoteUser");
        //secureRepo is a cached remote repo
        RepoPath securedPath = InternalRepoPathFactory.create("securedRepo", "blabla");

        expectGetAllAclsCall();
        RemoteRepo remoteRepoMock = createRemoteRepoMock();
        expect(repositoryServiceMock.remoteRepositoryByKey("securedRepo")).andReturn(remoteRepoMock).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(securedPath.getRepoKey()))
                .andReturn(new HttpRepoDescriptor()).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey(securedPath.getRepoKey())).andReturn(null)
                .anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        replay(aclStoreServiceMock);
        assertTrue(service.canRead(securedPath), "User should have read permissions on any remote path");
        assertFalse(
                service.canDeploy(securedPath) || service.canManage(securedPath) || service.canDelete(securedPath) ||
                        service.canManage(securedPath), "User shouldn't have other permissions on any remote path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);
    }

    public void userDeployPermissionsOnAnyLocal() {
        // user should have deploy permission on any local
        setSimpleUserAuthentication("anyLocalUser");

        RepoPath securedPath = InternalRepoPathFactory.create("securedRepo", "blabla");

        expectGetAllAclsCall();
        expect(repositoryServiceMock.localOrFederatedRepositoryByKey("securedRepo")).andReturn(localRepoMock)
                .anyTimes();
        expect(repositoryServiceMock.remoteRepositoryByKey("securedRepo")).andReturn(null).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(securedPath.getRepoKey())).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey(securedPath.getRepoKey())).andReturn(null)
                .anyTimes();
        expectRemoteByKeyNullForAnyConsts();
        replay(aclStoreServiceMock);

        assertTrue(service.canDeploy(securedPath), "User should have deploy permissions on any local path");
        assertFalse(
                service.canRead(securedPath) || service.canManage(securedPath) || service.canDelete(securedPath) ||
                        service.canManage(securedPath), "User shouldn't have other permissions on any remote path");
        EasyMock.verify(aclStoreServiceMock);
        EasyMock.reset(aclStoreServiceMock);
    }

    public void updateLastLoginWithNotExistingUserTest() {
        // Make sure that if the user doesn't exists we stop the "updateLastLogin" process without  exception
        EasyMock.reset(userGroupStoreService);
        // Enable the update last login process "userLastAccessUpdatesResolutionSecs" must be greater or equals to "1"
        getBound().setProperty(ConstantValues.userLastAccessUpdatesResolutionSecs, "1");
        expect(userGroupStoreService.findUser("user")).andReturn(null).once();
        replay(userGroupStoreService);
        service.updateUserLastLogin("user", System.currentTimeMillis() + 1000, "momo");
        EasyMock.reset(userGroupStoreService);
    }

    public void testUserLastLoginTimeUpdateBuffer() {
        getBound().setProperty(ConstantValues.userLastAccessUpdatesResolutionSecs, "0");
        service.updateUserLastLogin("user", System.currentTimeMillis(), "127.0.0.1");
        getBound().setProperty(ConstantValues.userLastAccessUpdatesResolutionSecs,
                ConstantValues.userLastAccessUpdatesResolutionSecs.getDefValue());

        MutableUserInfo user = new UserInfoBuilder("user").build();
        user.setLastLoginTimeMillis(0);

        //Simulate No existing last login, expect an update
        expect(userGroupStoreService.findUser("user")).andReturn(user).times(1);
        long currentTimeMillis = System.currentTimeMillis();
        userGroupStoreService.updateUserLastLogin("user", currentTimeMillis, "127.0.0.1");
        EasyMock.expectLastCall();
        replay(userGroupStoreService);
        service.updateUserLastLogin("user", currentTimeMillis, "127.0.0.1");
        EasyMock.verify(userGroupStoreService);

        //Give a last login from the near past, expect no update
        long nearPastLogin = System.currentTimeMillis();
        EasyMock.reset(userGroupStoreService);
        user = new UserInfoBuilder("user").build();
        user.setLastLoginTimeMillis(nearPastLogin);
        expect(userGroupStoreService.findUser("user")).andReturn(user).once();
        replay(userGroupStoreService);
        service.updateUserLastLogin("user", nearPastLogin + 100L, "momo");
        EasyMock.verify(userGroupStoreService);

        //Give a last login from the future, expect an update
        EasyMock.reset(userGroupStoreService);
        expect(userGroupStoreService.findUser("user")).andReturn(user).times(1);
        userGroupStoreService.updateUserLastLogin("user", currentTimeMillis + 6000L, "127.0.0.1");
        EasyMock.expectLastCall();
        replay(userGroupStoreService);
        service.updateUserLastLogin("user", currentTimeMillis + 6000L, "127.0.0.1");
        EasyMock.verify(userGroupStoreService);

        // Give a last login from the future, expect an update (use an IPv6 Address)
        EasyMock.reset(userGroupStoreService);
        expect(userGroupStoreService.findUser("user")).andReturn(user).times(1);
        userGroupStoreService.updateUserLastLogin("user", currentTimeMillis + 10000L,
                "[2a05:d014:82c:c112:ecc8:404a:7917:dbc2]");
        EasyMock.expectLastCall();
        replay(userGroupStoreService);
        service.updateUserLastLogin("user", currentTimeMillis + 10000L,
                "[2a05:d014:82c:c112:ecc8:404a:7917:dbc2]");
        EasyMock.verify(userGroupStoreService);
    }

    public void testSelectiveReload() {
        TreeSet<SecurityListener> securityListeners = new TreeSet<>();
        securityListeners.add(securityListenerMock);
        ReflectionTestUtils.setField(service, "securityListeners", securityListeners);
        EasyMock.reset(securityListenerMock);
        securityListenerMock.onClearSecurity();
        expect(securityListenerMock.compareTo(securityListenerMock)).andReturn(0).anyTimes();
        replay(securityListenerMock);

        SecurityDescriptor newSecurityDescriptor = new SecurityDescriptor();
        SecurityDescriptor oldSecurityDescriptor = new SecurityDescriptor();
        oldSecurityDescriptor.addLdap(new LdapSetting());

        CentralConfigDescriptor newConfigDescriptor = createMock(CentralConfigDescriptor.class);
        expect(newConfigDescriptor.getSecurity()).andReturn(newSecurityDescriptor).anyTimes();
        replay(newConfigDescriptor);

        CentralConfigDescriptor oldConfigDescriptor = createMock(CentralConfigDescriptor.class);
        expect(oldConfigDescriptor.getSecurity()).andReturn(oldSecurityDescriptor).anyTimes();
        replay(oldConfigDescriptor);

        expect(centralConfigServiceMock.getDescriptor()).andReturn(newConfigDescriptor).anyTimes();
        replay(centralConfigServiceMock);

        service.reload(oldConfigDescriptor, ImmutableList.of());
        EasyMock.verify(securityListenerMock);

        // The security conf is the same, so onClearSecurity should NOT be called
        service.reload(newConfigDescriptor, ImmutableList.of());
        EasyMock.verify(securityListenerMock);
        ReflectionTestUtils.setField(service, "securityListeners", null);
    }

    public void testPermissionCheckOnAclFitPermission() {
        SimpleUser nonAdminUser = createNonAdminUser("user1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());

        Set<AceInfo> aces = Sets.newHashSet(aceU1);

        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList("repo-a"),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission = new PrincipalPermissionImpl<>(pt, aceU1);
        assertTrue(service.permissionCheckOnAcl(Collections.singletonList(principalPermission),
                new RepoPathImpl("repo-a", null),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, nonAdminUser.getUsername()));
    }

    public void testPermissionCheckOnAclDifferentPermission() {
        SimpleUser nonAdminUser = createNonAdminUser("user1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());

        Set<AceInfo> aces = Sets.newHashSet(aceU1);

        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList("repo-a"),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission = new PrincipalPermissionImpl<>(pt, aceU1);
        assertFalse(service.permissionCheckOnAcl(Collections.singletonList(principalPermission),
                new RepoPathImpl("repo-a", null),
                Collections.singletonList("repo-a"), ArtifactoryPermission.READ, nonAdminUser.getUsername()));
    }


    public void testPermissionCheckOnAclOnGroupFailed() {
        SimpleUser nonAdminUser = createNonAdminUser("user1", "g1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());

        Set<AceInfo> aces = Sets.newHashSet(aceU1);

        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList("repo-a"),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission = new PrincipalPermissionImpl<>(pt, aceU1);
        assertFalse(service.permissionCheckOnAcl(Collections.singletonList(principalPermission),
                new RepoPathImpl("repo-a", null),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, "g1"));
    }

    public void testPermissionCheckOnAclOnGroupSuccess() {
        SimpleUser nonAdminUser = createNonAdminUser("user1", "g1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl aceU2 = new AceImpl("g1", true, ArtifactoryPermission.DEPLOY.getMask());


        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList("repo-a"),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission1 = new PrincipalPermissionImpl<>(pt, aceU1);
        PrincipalPermission principalPermission2 = new PrincipalPermissionImpl<>(pt, aceU2);
        assertTrue(service.permissionCheckOnAcl(Arrays.asList(principalPermission1, principalPermission2),
                new RepoPathImpl("repo-a", null),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, "g1"));
    }

    public void testPermissionCheckOnAclOnGroupSuccess2() {
        SimpleUser nonAdminUser = createNonAdminUser("user1", "g1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl aceU2 = new AceImpl("g1", true, ArtifactoryPermission.DEPLOY.getMask());


        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList("repo-a"),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission1 = new PrincipalPermissionImpl<>(pt, aceU1);
        PrincipalPermission principalPermission2 = new PrincipalPermissionImpl<>(pt, aceU2);
        assertTrue(service.permissionCheckOnAcl(Arrays.asList(principalPermission1, principalPermission2),
                new RepoPathImpl("repo-a", null),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, "g1"));
    }

    public void testPermissionCheckOnAclAnySuccess() {
        SimpleUser nonAdminUser = createNonAdminUser("user1", "g1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl aceU2 = new AceImpl("g1", true, ArtifactoryPermission.DEPLOY.getMask());


        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList(ANY_REPO),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission1 = new PrincipalPermissionImpl<>(pt, aceU1);
        PrincipalPermission principalPermission2 = new PrincipalPermissionImpl<>(pt, aceU2);
        assertTrue(service.permissionCheckOnAcl(Arrays.asList(principalPermission1, principalPermission2),
                new RepoPathImpl("repo-a", null),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, "g1"));
    }

    public void testPermissionCheckOnAclPatternSuccess() {
        SimpleUser nonAdminUser = createNonAdminUser("user1", "g1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl aceU2 = new AceImpl("g1", true, ArtifactoryPermission.DEPLOY.getMask());


        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList(ANY_REPO),
                Lists.newArrayList("**dod**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission1 = new PrincipalPermissionImpl<>(pt, aceU1);
        PrincipalPermission principalPermission2 = new PrincipalPermissionImpl<>(pt, aceU2);
        assertTrue(service.permissionCheckOnAcl(Arrays.asList(principalPermission1, principalPermission2),
                new RepoPathImpl("repo-a", "dodge"),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, "g1"));
    }


    public void testPermissionCheckOnAclPatternFailed() {
        SimpleUser nonAdminUser = createNonAdminUser("user1", "g1");

        AceImpl aceU1 = new AceImpl(nonAdminUser.getUsername(), false, ArtifactoryPermission.DEPLOY.getMask());
        AceImpl aceU2 = new AceImpl("g1", true, ArtifactoryPermission.DEPLOY.getMask());


        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList(ANY_REPO),
                Lists.newArrayList("**do/e**"),
                Lists.newArrayList());

        PrincipalPermission principalPermission1 = new PrincipalPermissionImpl<>(pt, aceU1);
        PrincipalPermission principalPermission2 = new PrincipalPermissionImpl<>(pt, aceU2);
        assertFalse(service.permissionCheckOnAcl(Arrays.asList(principalPermission1, principalPermission2),
                new RepoPathImpl("repo-a", "dodge"),
                Collections.singletonList("repo-a"), ArtifactoryPermission.DEPLOY, "g1"));
    }

    public void createGroupWhichNotExistExpectCreatedAndGroupAdd() {
        final MutableGroupInfo groupInfo = new GroupImpl("testGroup");
        expect(userGroupStoreService.createGroup(groupInfo)).andReturn(groupMock).once();
        replay(userGroupStoreService);
        when(groupMock.getRealm()).thenReturn(Realm.INTERNAL);
        when(groupMock.getName()).thenReturn("testGroup");
        final Group expectedGroup = service.createGroup(groupInfo);
        assertEquals(groupMock, expectedGroup);
        Mockito.verify(interceptors).onGroupAdd("testGroup", Realm.INTERNAL.getName());
    }

    @Test(expectedExceptions = AccessClientHttpException.class)
    public void createGroupGetExceptionExpectInterceptorNotRun() {
        final MutableGroupInfo groupInfo = new GroupImpl("testGroup");
        expect(userGroupStoreService.createGroup(groupInfo))
                .andThrow(new AccessClientHttpException(409, "TestMessage")).once();
        replay(userGroupStoreService);
        service.createGroup(groupInfo);
        Mockito.verify(interceptors, never())
                .onGroupAdd("testGroup", Realm.INTERNAL.getName());
    }

    public void updateGroupWhichExistExpectCreatedAndGroupAddInterceptorRun() {
        final MutableGroupInfo groupInfo = new GroupImpl("testGroup");
        when(groupMock.getRealm()).thenReturn(Realm.INTERNAL);
        when(groupMock.getName()).thenReturn("testGroup");
        expect(userGroupStoreService.updateGroup(groupInfo)).andReturn(groupMock).once();
        replay(userGroupStoreService);
        final Group expectedGroup = service.updateGroup(groupInfo);
        assertEquals(groupMock, expectedGroup);
        Mockito.verify(interceptors).onGroupUpdate("testGroup", Realm.INTERNAL.getName());
    }

    @Test(expectedExceptions = AccessClientHttpException.class)
    public void updateGroupGetExceptionExpectInterceptorNotRun() {
        final MutableGroupInfo groupInfo = new GroupImpl("testGroup");
        expect(userGroupStoreService.updateGroup(groupInfo))
                .andThrow(new AccessClientHttpException(409, "TestMessage"));
        replay(userGroupStoreService);
        service.updateGroup(groupInfo);
        Mockito.verify(interceptors, never()).onGroupUpdate("testGroup", Realm.INTERNAL.getName());
    }

    @Test
    public void noPermissionsProjectAdminOnPipeInfoRepo() {
        setProjectAdminAuthentication(true, false);
        expect(aclStoreServiceMock.getAclCache()).andReturn(aclCache).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(REPO.getRepoKey())).andReturn(null).anyTimes();
        expect(repositoryServiceMock.releaseBundlesRepoDescriptorByKey(REPO.getRepoKey())).andReturn(null).anyTimes();
        expect(repositoryServiceMock.isPipeInfoRepository(REPO.getRepoKey())).andReturn(true).anyTimes();
        replay(aclStoreServiceMock, repositoryServiceMock);
        assertFalse(service.canManage(REPO));
        assertFalse(service.canDelete(REPO));
        assertFalse(service.canRead(REPO));
        assertFalse(service.canAnnotate(REPO));
        assertFalse(service.canDeploy(REPO));
    }

    @Test
    public void adminOnPipeInfoRepo() {
        reset(repositoryServiceMock);
        setAdminAuthentication();
        RepoPath repoPath = RepoPathFactory.create("pipe-info");
        assertTrue(service.canManage(repoPath));
        assertTrue(service.canDelete(repoPath));
        assertTrue(service.canDeploy(repoPath));
        assertTrue(service.canRead(repoPath));
    }

    @Test
    public void regularUserOnPipeInfoRepoNotAllowed() {
        reset(repositoryServiceMock);
        setSimpleUserAuthentication();
        RepoPath repoPath = RepoPathFactory.create("pipe-info");
        expect(repositoryServiceMock.isPipeInfoRepository("pipe-info")).andReturn(true).anyTimes();
        replay(repositoryServiceMock);

        assertFalse(service.canManage(repoPath));
        assertFalse(service.canDelete(repoPath));
        assertFalse(service.canRead(repoPath));
        assertFalse(service.canAnnotate(repoPath));
        assertFalse(service.canDeploy(repoPath));
    }

    @Test
    public void regularUserOnObservabilityRepoNotAllowed() {
        reset(repositoryServiceMock);
        InternalArtifactoryContext context = mock(InternalArtifactoryContext.class);
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(context.getAuthorizationService()).thenReturn(authorizationService);
        when(authorizationService.currentUsername()).thenReturn("user");
        service.setApplicationContext(context);
        setSimpleUserAuthentication();
        RepoPath repoPath = RepoPathFactory.create(LOG_SHIPPING_REPO_KEY);
        RepoPath repoPathConsumption = RepoPathFactory.create(USAGE_REPO_KEY);
        assertFalse(service.canManage(repoPath));
        assertFalse(service.canDelete(repoPath));
        assertFalse(service.canRead(repoPath));
        assertFalse(service.canAnnotate(repoPath));
        assertFalse(service.canDeploy(repoPath));
        assertFalse(service.canManage(repoPathConsumption));
        assertFalse(service.canDelete(repoPathConsumption));
        assertFalse(service.canRead(repoPathConsumption));
        assertFalse(service.canAnnotate(repoPathConsumption));
        assertFalse(service.canDeploy(repoPathConsumption));
    }

    @Test
    public void adminOnObservabilityRepo() {
        reset(repositoryServiceMock);
        InternalArtifactoryContext context = mock(InternalArtifactoryContext.class);
        AuthorizationService authorizationService = mock(AuthorizationService.class);
        when(context.getAuthorizationService()).thenReturn(authorizationService);
        when(authorizationService.currentUsername()).thenReturn(USER_SYSTEM);
        setAdminAuthentication();
        service.setApplicationContext(context);
        RepoPath repoPath = RepoPathFactory.create(LOG_SHIPPING_REPO_KEY);
        RepoPath repoPathConsumption = RepoPathFactory.create(USAGE_REPO_KEY);
        assertTrue(service.canManage(repoPath));
        assertTrue(service.canDelete(repoPath));
        assertTrue(service.canDeploy(repoPath));
        assertTrue(service.canRead(repoPath));
        assertTrue(service.canManage(repoPathConsumption));
        assertTrue(service.canDelete(repoPathConsumption));
        assertTrue(service.canDeploy(repoPathConsumption));
        assertTrue(service.canRead(repoPathConsumption));
    }

    private static SimpleUser createAdminUser(String name) {
        UserInfo userInfo = new UserInfoBuilder(name).admin(true).updatableProfile(true).build();
        return new SimpleUser(userInfo);
    }

    private static SimpleUser createNonAdminUser(String username, String... groups) {
        return createNonAdminUser(username, false, groups);
    }

    private static SimpleUser createNonAdminUser(String username, boolean mfaEnabled, String... groups) {
        UserInfo userInfo = new UserInfoBuilder(username).updatableProfile(true)
                .mfaStatus(mfaEnabled ? org.artifactory.security.UserMFAStatus.VERIFY :
                        org.artifactory.security.UserMFAStatus.NONE)
                .internalGroups(new HashSet<>(Arrays.asList(groups))).build();
        return new SimpleUser(userInfo);
    }


    private static SimpleUser createNonAdminExternalUser(String username, Realm realm, String... groups) {
        UserInfo userInfo = new UserInfoBuilder(username).updatableProfile(true).realm(realm.getName())
                .internalGroups(new HashSet<>(Arrays.asList(groups))).build();
        return new SimpleUser(userInfo);
    }

    @Test
    public void testHasPermission() {
        // Create mocks.
        RepoAcl generatedInternalRepoAcl =
                createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + "repo_perm_target", 1);
        BuildAcl generatedInternalBuildAcl =
                createBuildAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + "build_perm_target", "bu/**", 4);
        RepoAcl generatedInternalDefaultRepoAcl =
                createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + ProjectsService.DEFAULT_PROJECT +
                        "repo_perm_target", 16);
        BuildAcl generatedInternalDefaultBuildAcl =
                createBuildAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + ProjectsService.DEFAULT_PROJECT +
                        "build_perm_target", "bu/**", 32);
        RepoAcl manuallyCreatedRepoAcl = createRepoAcl("not_internal_repo_perm_target", 2);
        BuildAcl manuallyCreatedBuildAcl = createBuildAcl("not_internal_build_perm_target", "bu/**", 64);

        expect(aclStoreServiceMock.getAllRepoAcls())
                .andReturn(List.of(generatedInternalRepoAcl, manuallyCreatedRepoAcl, generatedInternalDefaultRepoAcl))
                .anyTimes();
        expect(aclStoreServiceMock.getAllBuildAcls()).andReturn(
                        List.of(generatedInternalBuildAcl, manuallyCreatedBuildAcl, generatedInternalDefaultBuildAcl))
                .anyTimes();
        doReturn(Collections.emptyList()).when(service).getAllReleaseBundleAclsByLicense();
        replay(aclStoreServiceMock);

        // Test simple user.
        setSimpleUserAuthentication(PERMISSION_ACTOR);
        assertHasPermission();

        // Test project-admin user.
        setProjectAdminAuthentication(PERMISSION_ACTOR, true, false);
        assertHasPermission();

        // Test admin.
        setAdminAuthentication(PERMISSION_ACTOR);
        // The result should always be true.
        Arrays.stream(ArtifactoryPermission.values()).forEach(permission -> {
            assertTrue(service.hasPermission(permission, true));
            assertTrue(service.hasPermission(permission, false));
        });
    }

    private void assertHasPermission() {
        // Test Repo permissions.
        // READ is internal-only.
        assertFalse(service.hasPermission(ArtifactoryPermission.READ, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.READ, false));
        // MANAGE is for internal for default project, so should not be skipped.
        assertTrue(service.hasPermission(ArtifactoryPermission.MANAGE, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.MANAGE, false));
        // DEPLOY is a user-created permission, so should not be skipped.
        assertTrue(service.hasPermission(ArtifactoryPermission.DEPLOY, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.DEPLOY, false));
        // Test Build permissions.
        // ANNOTATE is internal-only.
        assertFalse(service.hasPermission(ArtifactoryPermission.ANNOTATE, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.ANNOTATE, false));
        // MANAGED_XRAY_META is for internal for default project, so should not be skipped.
        assertTrue(service.hasPermission(ArtifactoryPermission.MANAGE, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.MANAGE, false));
        // DISTRIBUTE is a user-created permission, so should not be skipped.
        assertTrue(service.hasPermission(ArtifactoryPermission.DEPLOY, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.DEPLOY, false));
    }

    @Test
    public void testPermissionsForTrashcan() {
        // Prepare mocks.
        expectAclScan();
        expectIsPipeInfoRepositoryCheck();

        // User "shay" has read, deploy, annotate, delete permissions to testRepo1 as part of group permissions.
        setSimpleUserAuthentication("shay");

        // testRepo1 is one of the repos that a "shay" has a "deploy" permission for
        RepoPath trashPath = RepoPathFactory.create(TrashService.TRASH_KEY + "/testRepo1/path");
        assertTrue(service.canRead(trashPath));
        assertTrue(service.canAnnotate(trashPath));
        assertTrue(service.canDeploy(trashPath));
        assertTrue(service.canDelete(trashPath));
        assertFalse(service.canManage(trashPath));

        setSimpleUserAuthentication("yossis"); // has manage permission
        assertTrue(service.canManage(trashPath));

        // Test access to the root trashcan path. Only read allowed for non-admins
        setSimpleUserAuthentication("shay");
        RepoPath trashRootPath = RepoPathFactory.create(TrashService.TRASH_KEY);
        assertTrue(service.canRead(trashRootPath));
        assertFalse(service.canAnnotate(trashRootPath));
        assertFalse(service.canDeploy(trashRootPath));
        assertFalse(service.canDelete(trashRootPath));
        assertFalse(service.canManage(trashRootPath));

        // Test that admin has all the permissions
        setAdminAuthentication();
        assertTrue(service.canRead(trashPath));
        assertTrue(service.canAnnotate(trashPath));
        assertTrue(service.canDeploy(trashPath));
        assertTrue(service.canDelete(trashPath));
        assertTrue(service.canManage(trashPath));
        assertTrue(service.canRead(trashRootPath));
        assertTrue(service.canAnnotate(trashRootPath));
        assertTrue(service.canDeploy(trashRootPath));
        assertTrue(service.canDelete(trashRootPath));
        assertTrue(service.canManage(trashRootPath));
    }

    @Test
    public void testCanAccessTrashcan() {
        // Prepare mocks.
        expectAclScan();
        expectIsPipeInfoRepositoryCheck();

        // Simple user doesn't have deploy/manage permissions to any of the repos.
        setSimpleUserAuthentication(PERMISSION_ACTOR);
        assertFalse(service.canAccessTrashcan());

        // User "shay" has read, deploy, annotate, delete permissions to testRepo1 as part of group permissions.
        setSimpleUserAuthentication("shay");
        assertTrue(service.canAccessTrashcan());

        // User "manageOnlyUser" has manage permission for some repo. So also should be able to access trashcan.
        setSimpleUserAuthentication("manageOnlyUser");
        assertTrue(service.canAccessTrashcan());

        // Admin should be able to access trashcan.
        setAdminAuthentication();
        assertTrue(service.canAccessTrashcan());
    }

    @Test
    public void testCanRestoreFromTrashcan() {
        RepoPath trashPath = RepoPathFactory.create(TrashService.TRASH_KEY + "/testRepo1/path");
        RepoPath restrictedTrashPath =
                RepoPathFactory.create(TrashService.TRASH_KEY + "/someRepoThatNoOneHasPermissionsTo/path");

        // Prepare mocks.
        expectAclScan();
        expectIsPipeInfoRepositoryCheck();

        // Simple user doesn't have deploy/manage permissions to testRepo1.
        setSimpleUserAuthentication(PERMISSION_ACTOR);
        assertFalse(service.canRestoreFromTrashcan(trashPath));
        assertFalse(service.canRestoreFromTrashcan(restrictedTrashPath));

        // User "shay" has read, deploy, annotate, delete permissions to testRepo1 as part of group permissions.
        setSimpleUserAuthentication("shay");
        assertTrue(service.canRestoreFromTrashcan(trashPath));
        assertFalse(service.canRestoreFromTrashcan(restrictedTrashPath));

        // User "manageOnlyUser" has manage permission for testRepo1. So also should be able to restore.
        setSimpleUserAuthentication("manageOnlyUser");
        assertTrue(service.canRestoreFromTrashcan(trashPath));
        assertFalse(service.canRestoreFromTrashcan(restrictedTrashPath));

        // Admin should be able to restore anything from trashcan.
        setAdminAuthentication();
        assertTrue(service.canRestoreFromTrashcan(trashPath));
        assertTrue(service.canRestoreFromTrashcan(restrictedTrashPath)); // true
    }

    private void expectAclScan() {
        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(testAcls).anyTimes();
        expect(aclStoreServiceMock.getAclCache()).andReturn(aclCache).anyTimes();
        replay(aclStoreServiceMock);
    }

    private void expectIsPipeInfoRepositoryCheck() {
        expect(repositoryServiceMock.isPipeInfoRepository(TrashService.TRASH_KEY)).andReturn(false).anyTimes();
        replay(repositoryServiceMock);
    }

    private void expectGetAllAclsCall() {
        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(testAcls).anyTimes();
        expect(aclStoreServiceMock.getAclCache()).andReturn(aclCache).anyTimes();
    }

    private void expectGetBuildsAclCacheCall() {
        reset(repositoryServiceMock, aclStoreServiceMock);
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(BUILDINFO_REPO_KEY)).andReturn(null).anyTimes();
        expect(repositoryServiceMock.isPipeInfoRepository(BUILDINFO_REPO_KEY)).andReturn(false).anyTimes();
        expect(aclStoreServiceMock.getBuildsAclCache()).andReturn(emptyAclCache).anyTimes();
        replay(aclStoreServiceMock, repositoryServiceMock);
    }

    private void expectGetAllAclsCallWithAnyArray() {
        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(testAcls);
        expect(aclStoreServiceMock.getAclCache()).andReturn(aclCache);
    }

    @Test
    public void testHasPermissionForReleaseBundleLicenseAcl() {
        ReleaseBundleAcl generatedInternalRepoAcl =
                createReleaseBundleAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + "repo_perm_target", 1);
        ReleaseBundleAcl generatedInternalDefaultRepoAcl =
                createReleaseBundleAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + ProjectsService.DEFAULT_PROJECT +
                        "repo_perm_target", 16);
        ReleaseBundleAcl manuallyCreatedRepoAcl = createReleaseBundleAcl("not_internal_repo_perm_target", 2);

        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(Collections.emptyList()).anyTimes();
        expect(aclStoreServiceMock.getAllBuildAcls()).andReturn(Collections.emptyList()).anyTimes();
        doReturn(List.of(generatedInternalRepoAcl, generatedInternalDefaultRepoAcl, manuallyCreatedRepoAcl))
                .when(service).getAllReleaseBundleAclsByLicense();
        replay(aclStoreServiceMock);

        setSimpleUserAuthentication(PERMISSION_ACTOR);

        // READ is internal-only.
        assertFalse(service.hasPermission(ArtifactoryPermission.READ, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.READ, false));
        // MANAGE is for internal for default project, so should not be skipped.
        assertTrue(service.hasPermission(ArtifactoryPermission.MANAGE, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.MANAGE, false));
        // DEPLOY is a user-created permission, so should not be skipped.
        assertTrue(service.hasPermission(ArtifactoryPermission.DEPLOY, true));
        assertTrue(service.hasPermission(ArtifactoryPermission.DEPLOY, false));
    }

    @Test
    public void testIsNotInternalProjectSpecificAcl() {
        assertTrue(service.isNotInternalProjectSpecificAcl(createRepoAcl(null, 0)));
        assertTrue(service.isNotInternalProjectSpecificAcl(createRepoAcl("", 0)));
        assertTrue(service.isNotInternalProjectSpecificAcl(createRepoAcl("test", 0)));
        assertFalse(
                service.isNotInternalProjectSpecificAcl(createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX, 0)));
        assertFalse(service.isNotInternalProjectSpecificAcl(
                createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + "test", 0)));
        assertTrue(service.isNotInternalProjectSpecificAcl(
                createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + ProjectsService.DEFAULT_PROJECT, 0)));
        assertTrue(service.isNotInternalProjectSpecificAcl(
                createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + ProjectsService.DEFAULT_PROJECT + "test",
                        0)));
    }

    private Authentication setSimpleUserAuthentication() {
        return setSimpleUserAuthentication("user");
    }

    private Authentication setAdminAuthentication() {
        return setAdminAuthentication("spiderman");
    }

    private Authentication setAdminAuthentication(String name) {
        SimpleUser adminUser = createAdminUser(name);
        UsernamePasswordAuthenticationToken authenticationToken = new UsernamePasswordAuthenticationToken(
                adminUser, null, SimpleUser.ADMIN_GAS);
        securityContext.setAuthentication(authenticationToken);
        return authenticationToken;
    }

    private void setProjectAdminAuthentication(boolean enable, boolean enableResourceManager) {
        setProjectAdminAuthentication("user", enable, enableResourceManager);
    }

    private Authentication setSimpleUserAuthentication(String username, String... groups) {
        return setSimpleUserAuthentication(username, false, groups);
    }

    private Authentication setSimpleUserAuthentication(String username, boolean mfaEnabled, String... groups) {
        SimpleUser simpleUser = createNonAdminUser(username, mfaEnabled, groups);
        UsernamePasswordAuthenticationToken authenticationToken = new UsernamePasswordAuthenticationToken(
                simpleUser, "password", SimpleUser.USER_GAS);
        securityContext.setAuthentication(authenticationToken);
        return authenticationToken;
    }

    private Authentication setAccessTokenAuthentication(String username, boolean mfaEnabled, String... groups) {
        SimpleUser simpleUser = createNonAdminUser(username, mfaEnabled, groups);
        AccessTokenAuthentication authenticationToken = new AccessTokenAuthentication(ACCESS_TOKEN, simpleUser, null);
        securityContext.setAuthentication(authenticationToken);
        return authenticationToken;
    }

    private Authentication setSimpleExternalUserAuthentication(String username, Realm realm, String... groups) {
        SimpleUser simpleUser = createNonAdminExternalUser(username, realm, groups);
        UsernamePasswordAuthenticationToken authenticationToken = new UsernamePasswordAuthenticationToken(
                simpleUser, "password", SimpleUser.USER_GAS);
        securityContext.setAuthentication(authenticationToken);
        return authenticationToken;
    }

    private void setProjectAdminAuthentication(String name, boolean enable, boolean enableResourceManager) {
        Authentication authentication = setSimpleUserAuthentication(name);
        SimpleUser user = (SimpleUser) authentication.getPrincipal();
        if (enable) {
            when(projectsService.getAllProjectsManagedByUser()).thenReturn(List.of(PROJECT_A, PROJECT_B));
            when(projectsService.getProjectKeyOfRepository(REPO.getRepoKey())).thenReturn(PROJECT_A);
            when(projectsService.getProjectRepositoriesKeys(PROJECT_B, false))
                    .thenReturn(Set.of(SHARED_REPO.getRepoKey()));
            when(projectsService.getProjectRepositoriesKeys(PROJECT_A, false))
                    .thenReturn(Set.of(REPO.getRepoKey()));
            when(projectsService.canManageProjectResources(PROJECT_B, user)).thenReturn(true);
            when(projectsService.canProjectAdminManageResources(REPO.getRepoKey(), user)).thenReturn(true);
            when(projectsService.canProjectAdminManageResources(SHARED_REPO.getRepoKey(), user)).thenReturn(true);
        } else {
            when(projectsService.getAllProjectsManagedByUser()).thenReturn(Collections.emptyList());
            when(projectsService.getProjectKeyOfRepository(REPO.getRepoKey()))
                    .thenReturn(ProjectsService.DEFAULT_PROJECT);
            when(projectsService.canManageProjectResources(PROJECT_B, user)).thenReturn(false);
        }
        when(projectsService.isProjectAdmin(PROJECT_A, user)).thenReturn(enable);
        when(projectsService.isResourceManager(user)).thenReturn(enableResourceManager);
        when(projectsService.canManageProjectResources(PROJECT_A, user)).thenReturn(enable);
        when(projectsService.canProjectAdminManageResources(SHARED_REPO.getRepoKey(), user)).thenReturn(enable);
        when(projectsService.canProjectAdminManageResources(REPO.getRepoKey(), user)).thenReturn(enable);
    }

    private static LocalRepo createLocalRepoMock() {
        LocalRepo localRepo = createMock(LocalRepo.class);
        expect(localRepo.isLocal()).andReturn(true).anyTimes();
        expect(localRepo.isCache()).andReturn(false).anyTimes();
        replay(localRepo);
        return localRepo;
    }

    private static LocalRepo createCacheRepoMock() {
        LocalRepo localRepo = createMock(LocalRepo.class);
        expect(localRepo.isLocal()).andReturn(true).anyTimes();
        expect(localRepo.isCache()).andReturn(true).anyTimes();
        replay(localRepo);
        return localRepo;
    }

    private RemoteRepo createRemoteRepoMock() {
        RemoteRepo remoteRepo = createMock(RemoteRepo.class);
        expect(remoteRepo.isReal()).andReturn(true).anyTimes();
        replay(remoteRepo);
        return remoteRepo;
    }

    private InternalRepositoryService createRepoServiceMock() {
        InternalRepositoryService repositoryService = createNiceMock(InternalRepositoryService.class);
        replay(repositoryService);
        return repositoryService;
    }

    private InternalBuildService createBuildServiceMock() {
        InternalBuildService buildService = createMock(InternalBuildService.class);
        expect(buildService.getBuildInfoRepoForDefaultProject())
                .andReturn(BUILDINFO_REPO_KEY)
                .anyTimes();
        expect(buildService.getBuildInfoRepoKeyByProject(anyObject()))
                .andReturn(BUILDINFO_REPO_KEY).anyTimes();
        Build build = new Build();
        build.setName(BUILD_NAME);
        build.setNumber(BUILD_NUMBER);
        build.setStarted(BUILD_DATE);
        expect(buildService
                .getLatestBuildByNameAndNumberInternally(BUILD_NAME, BUILD_NUMBER, BUILDINFO_REPO_KEY))
                .andReturn(new BuildWrapper(build, BUILDINFO_REPO_KEY)).anyTimes();
        expect(buildService.getLatestBuildByNameAndNumberInternally(BUILD_NAME, "2", BUILDINFO_REPO_KEY))
                .andReturn(null).anyTimes();
        expect(buildService.normaliseBuildRepoName(BUILDINFO_REPO_KEY))
                .andReturn(BUILDINFO_REPO_KEY).anyTimes();
        expect(buildService.getAllBuildInfoRepos()).andReturn(Set.of(BUILDINFO_REPO_KEY)).anyTimes();
        replay(buildService);
        return buildService;
    }

    @Test(
            expectedExceptions = {PasswordChangeException.class},
            expectedExceptionsMessageRegExp = "Old password is incorrect"
    )
    public void changePasswordUsingIncorrectOldPassword() {
        SaltedPassword sp = new SaltedPassword("foo", "salt");

        UserInfo user = new UserImpl() {
            {
                setUsername("test");
                setUpdatableProfile(true);
                setPassword(sp);
            }
        };

        expect(userGroupStoreService.findUser("test")).andReturn(user).anyTimes();
        expect(userPassAuthenticationProvider.canUserLogin(EasyMock.anyObject(), EasyMock.anyObject()))
                .andReturn(false);

        MutableCentralConfigDescriptor mutableCentralConfigDescriptor = createMock(
                MutableCentralConfigDescriptor.class);
        PasswordExpirationPolicy expirationPolicy = new PasswordExpirationPolicy();
        expirationPolicy.setEnabled(Boolean.TRUE);

        SecurityDescriptor securityDescriptor = new SecurityDescriptor();
        securityDescriptor.setPasswordSettings(new PasswordSettings() {
            {
                setExpirationPolicy(expirationPolicy);
            }
        });

        expect(centralConfigServiceMock.getDescriptor()).andReturn(mutableCentralConfigDescriptor).anyTimes();
        expect(centralConfigServiceMock.getMutableDescriptor()).andReturn(mutableCentralConfigDescriptor).anyTimes();
        expect(mutableCentralConfigDescriptor.getSecurity()).andReturn(securityDescriptor).anyTimes();
        UserLockPolicyConfig userLockPolicyConfig = UserLockPolicyConfig.builder().attempts(0).build();
        expect(accessService.getUserLockPolicyConfig()).andReturn(userLockPolicyConfig).anyTimes();

        expect(userLockInMemoryService.getNextLoginTime("test")).andReturn(-1L).anyTimes();
        userLockInMemoryService.registerIncorrectLoginAttempt("test");
        EasyMock.expectLastCall();
        replay(userGroupStoreService, mutableCentralConfigDescriptor, centralConfigServiceMock, accessService);

        service.changePassword("test", "bar", "pass", "pass");
    }

    @Test(
            dependsOnMethods = {"changePasswordUsingIncorrectOldPassword"},
            expectedExceptions = {PasswordChangeException.class},
            expectedExceptionsMessageRegExp = "Login has failed. " +
                    "Due to Incorrect username/password or locked user."
    )

    public void changePasswordForLockedOutUserTest() {
        SaltedPassword sp = new SaltedPassword("foo", "salt");

        UserInfo user = new UserImpl() {
            {
                setUsername("test");
                setUpdatableProfile(true);
                setPassword(sp);
            }
        };
        EasyMock.reset(userGroupStoreService);
        expect(userGroupStoreService.findUser("test")).andReturn(user).anyTimes();

        MutableCentralConfigDescriptor mutableCentralConfigDescriptor = createMock(
                MutableCentralConfigDescriptor.class);
        PasswordExpirationPolicy expirationPolicy = new PasswordExpirationPolicy();
        expirationPolicy.setEnabled(Boolean.TRUE);

        SecurityDescriptor securityDescriptor = new SecurityDescriptor();
        securityDescriptor.setPasswordSettings(new PasswordSettings() {
            {
                setExpirationPolicy(expirationPolicy);
            }
        });
        UserLockPolicyConfig userLockPolicyConfig = UserLockPolicyConfig.builder().attempts(2).build();
        expect(accessService.getUserLockPolicyConfig()).andReturn(userLockPolicyConfig).anyTimes();

        expect(centralConfigServiceMock.getDescriptor()).andReturn(mutableCentralConfigDescriptor).anyTimes();
        expect(centralConfigServiceMock.getMutableDescriptor()).andReturn(mutableCentralConfigDescriptor).anyTimes();
        expect(mutableCentralConfigDescriptor.getSecurity()).andReturn(securityDescriptor).anyTimes();
        expect(userGroupStoreService.isUserLocked("test")).andReturn(true);
        replay(userGroupStoreService, mutableCentralConfigDescriptor, centralConfigServiceMock, accessService);

        service.changePassword("test", "bar", "pass", "pass");
    }

    public void testValidateResetPasswordAttempt() {
        PasswordResetPolicy policy = new PasswordResetPolicy() {
            {
                setEnabled(true);
                setMaxAttemptsPerAddress(3);
                setTimeToBlockInMinutes(60);
            }
        };
        PasswordSettings passwordSettings = new PasswordSettings() {
            {
                setResetPolicy(policy);
            }
        };
        SecurityDescriptor securityDescriptor = new SecurityDescriptor() {
            {
                setPasswordSettings(passwordSettings);
            }
        };

        PasswordResetPolicy oldPolicy = new PasswordResetPolicy();
        PasswordSettings oldPasswordSettings = new PasswordSettings() {
            {
                setResetPolicy(oldPolicy);
            }
        };
        SecurityDescriptor oldSecurityDescriptor = new SecurityDescriptor() {
            {
                setPasswordSettings(oldPasswordSettings);
            }
        };

        CentralConfigDescriptor oldConfigDescriptorMock = createMock(CentralConfigDescriptor.class);
        expect(oldConfigDescriptorMock.getSecurity()).andReturn(oldSecurityDescriptor).anyTimes();

        CentralConfigDescriptor configDescriptorMock = createMock(CentralConfigDescriptor.class);
        expect(centralConfigServiceMock.getDescriptor()).andReturn(configDescriptorMock).anyTimes();
        expect(configDescriptorMock.getSecurity()).andReturn(securityDescriptor).anyTimes();
        replay(centralConfigServiceMock, configDescriptorMock, oldConfigDescriptorMock);

        service.reload(oldConfigDescriptorMock, ImmutableList.of());

        int timeBetweenAttempts = 600;
        String remoteAddress1 = "remote.address.1." + System.currentTimeMillis();

        service.validateResetPasswordAttempt(remoteAddress1);
        try {
            service.validateResetPasswordAttempt(remoteAddress1);
            fail("Validation should fail due to too frequent requests");
        } catch (ResetPasswordException e) {
            assertTrue(e.getMessage().toLowerCase().contains("too frequent"));
        }
        sleep(timeBetweenAttempts);
        service.validateResetPasswordAttempt(remoteAddress1);
        sleep(timeBetweenAttempts);
        service.validateResetPasswordAttempt(remoteAddress1);
        try {
            service.validateResetPasswordAttempt(remoteAddress1);
            fail("Validation should fail due to too many requests");
        } catch (ResetPasswordException e) {
            assertTrue(e.getMessage().toLowerCase().contains("too many"));
        }

        String remoteAddress2 = "remote.address.2." + System.currentTimeMillis();
        service.validateResetPasswordAttempt(remoteAddress2);
    }

    public void testAuthenticateAsSystem() {
        service.authenticateAsSystem();
        assertAuthenticatedAsSystem();
    }

    public void testDoAsSystem() {
        Authentication mockAuthentication = createMock(Authentication.class);
        SecurityContextHolder.getContext().setAuthentication(mockAuthentication);
        service.doAsSystem(this::assertAuthenticatedAsSystem);
        assertEquals(SecurityContextHolder.getContext().getAuthentication(), mockAuthentication,
                "original authentication was not restored");

        try {
            service.doAsSystem(() -> {
                throw new RuntimeException("expected exception");
            });
            fail("Exception was expected");
        } catch (Exception e) {
            //ignore expected exception
        }
        assertEquals(SecurityContextHolder.getContext().getAuthentication(), mockAuthentication,
                "original authentication was not restored");
    }

    /**
     * Test if user require password to unlock profile edit
     *
     * @param authClass             type of Authentication
     * @param allowSsoAuth          if Sso users are allowed to edit their profiles
     * @param allowSamlAuth         if saml users are allowed to edit their profiles
     * @param allowPropsAuth        if oauth users are allowed to edit their profiles
     * @param realm                 current user realm
     * @param authProviderName      current user auth provider name
     * @param requireUnlockExpected is password unlock is required
     */
    @Test(dataProvider = "authProvider")
    public void requireProfileUnlockTest(Class<Authentication> authClass, boolean allowSsoAuth, boolean allowSamlAuth,
            boolean allowPropsAuth, String realm, String authProviderName, boolean requireUnlockExpected) {
        ArtifactoryHome artifactoryHome = new ArtifactoryHomeStub(getClass().getName());
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryHome.bind(artifactoryHome);
        CentralConfigService configService = mock(CentralConfigService.class);
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        SecurityDescriptor security = new SecurityDescriptor();
        descriptor.setSecurity(security);
        when(configService.getDescriptor()).thenReturn(descriptor);

        OAuthProviderSettings googleProvider = new OAuthProviderSettings();
        googleProvider.setEnabled(true);
        googleProvider.setName("Google Central");
        CentralOAuthSSOAddon centralOAuthSSOAddon = mock(CentralOAuthSSOAddon.class);
        when(centralOAuthSSOAddon.getConfiguration()).thenReturn(
                new CentralOAuthSettings(true, List.of(googleProvider)));

        AddonsManager addonsManager = mock(AddonsManager.class);
        when(addonsManager.addonByType(CentralOAuthSSOAddon.class)).thenReturn(centralOAuthSSOAddon);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        service.setAddons(addonsManager);
        when(context.getCentralConfig()).thenReturn(configService);

        ArtifactoryContextThreadBinder.bind(context);
        Authentication accessTokenAuthentication = mock(authClass);
        securityContext.setAuthentication(accessTokenAuthentication);

        SamlSsoAddon samlSsoAddon = mock(SamlSsoAddon.class);
        if (authClass.equals(SamlTestAuthentication.class)) {
            when(samlSsoAddon.isSamlAuthentication()).thenReturn(true);
        }
        when(addonsManager.addonByType(SamlSsoAddon.class)).thenReturn(samlSsoAddon);
        HttpSsoAddon httpSsoAddon = mock(HttpSsoAddon.class);
        if (authClass.equals(HttpSsoTestAuthentication.class)) {
            when(httpSsoAddon.isHttpSsoAuthentication()).thenReturn(true);
        }
        when(addonsManager.addonByType(HttpSsoAddon.class)).thenReturn(httpSsoAddon);

        CrowdAddon crowdAddon = mock(CrowdAddon.class);
        if (authClass.equals(CrowdTestAuthentication.class)) {
            when(crowdAddon.isCrowdAuthenticationSupported(AuthenticationHelper.getAuthentication().getClass()))
                    .thenReturn(true);
        }
        when(addonsManager.addonByType(CrowdAddon.class)).thenReturn(crowdAddon);

        HttpSsoSettings httpSsoSettings = new HttpSsoSettings();
        security.setHttpSsoSettings(httpSsoSettings);
        httpSsoSettings.setAllowUserToAccessProfile(allowSsoAuth);
        SamlSettings samlSettings = new SamlSettings();
        security.setSamlSettings(samlSettings);
        samlSettings.setAllowUserToAccessProfile(allowSamlAuth);
        OAuthSettings oAuthSettings = new OAuthSettings();
        security.setOauthSettings(oAuthSettings);
        oAuthSettings.setAllowUserToAccessProfile(allowPropsAuth);

        boolean requireUnlock = service.requireProfileUnlock(realm, authProviderName);
        assertEquals(requireUnlock, requireUnlockExpected);
    }

    @Test
    public void requireProfileUnlockWithCentralOAuthLoginTest() {
        ArtifactoryHome artifactoryHome = new ArtifactoryHomeStub(getClass().getName());
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryHome.bind(artifactoryHome);
        CentralConfigService configService = mock(CentralConfigService.class);
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        SecurityDescriptor security = new SecurityDescriptor();
        descriptor.setSecurity(security);
        when(configService.getDescriptor()).thenReturn(descriptor);

        OAuthProviderSettings googleProvider = new OAuthProviderSettings();
        googleProvider.setEnabled(true);
        googleProvider.setName("Google Central");
        CentralOAuthSSOAddon centralOAuthSSOAddon = mock(CentralOAuthSSOAddon.class);
        when(centralOAuthSSOAddon.getConfiguration()).thenReturn(
                new CentralOAuthSettings(true, List.of(googleProvider)));
        AddonsManager addonsManager = mock(AddonsManager.class);
        when(addonsManager.addonByType(CentralOAuthSSOAddon.class)).thenReturn(centralOAuthSSOAddon);
        when(context.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        service.setAddons(addonsManager);
        when(context.getCentralConfig()).thenReturn(configService);
        ArtifactoryContextThreadBinder.bind(context);

        PropsAuthenticationToken accessTokenAuthentication = mock(PropsAuthenticationToken.class);
        when(accessTokenAuthentication.getPropsKey()).thenReturn(OAUTH_INFO_PROPERTY + "Google Central");
        securityContext.setAuthentication(accessTokenAuthentication);

        SamlSsoAddon samlSsoAddon = mock(SamlSsoAddon.class);
        when(addonsManager.addonByType(SamlSsoAddon.class)).thenReturn(samlSsoAddon);
        HttpSsoAddon httpSsoAddon = mock(HttpSsoAddon.class);
        when(addonsManager.addonByType(HttpSsoAddon.class)).thenReturn(httpSsoAddon);
        CrowdAddon crowdAddon = mock(CrowdAddon.class);
        when(addonsManager.addonByType(CrowdAddon.class)).thenReturn(crowdAddon);

        HttpSsoSettings httpSsoSettings = new HttpSsoSettings();
        security.setHttpSsoSettings(httpSsoSettings);
        httpSsoSettings.setAllowUserToAccessProfile(false);
        SamlSettings samlSettings = new SamlSettings();
        security.setSamlSettings(samlSettings);
        samlSettings.setAllowUserToAccessProfile(false);
        OAuthSettings oAuthSettings = new OAuthSettings();
        security.setOauthSettings(oAuthSettings);
        oAuthSettings.setAllowUserToAccessProfile(false);

        boolean requireUnlock = service.requireProfileUnlock(null, null);
        assertFalse(requireUnlock);
    }

    @Test(dataProvider = "profileUpdateProvider")
    public void isProfileUpdatableTest(String requestedRealm, boolean userProfileUpdatable, String ldapSetting,
            boolean allowSsoAuth, boolean allowSamlAuth, boolean allowPropsAuth, boolean allowCrowdAuth,
            boolean allowLdapAuth, boolean profileUpdatableExpected) {
        ArtifactoryHome artifactoryHome = new ArtifactoryHomeStub(getClass().getName());
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryHome.bind(artifactoryHome);
        CentralConfigService configService = mock(CentralConfigService.class);
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        LdapService ldapService = mock(LdapService.class);
        service.setLdapService(ldapService);
        UserInfo userInfo = mock(UserInfo.class);
        when(userInfo.getRealm()).thenReturn(requestedRealm);
        when(userInfo.getUsername()).thenReturn("test");
        when(userInfo.isUpdatableProfile()).thenReturn(userProfileUpdatable);
        when(ldapService.getUserLdapSetting("test")).thenReturn(ldapSetting);
        SecurityDescriptor security = new SecurityDescriptor();
        descriptor.setSecurity(security);
        when(configService.getDescriptor()).thenReturn(descriptor);
        when(context.getCentralConfig()).thenReturn(configService);
        ArtifactoryContextThreadBinder.bind(context);

        HttpSsoSettings httpSsoSettings = new HttpSsoSettings();
        security.setHttpSsoSettings(httpSsoSettings);
        httpSsoSettings.setAllowUserToAccessProfile(allowSsoAuth);
        SamlSettings samlSettings = new SamlSettings();
        security.setSamlSettings(samlSettings);
        samlSettings.setAllowUserToAccessProfile(allowSamlAuth);
        OAuthSettings oAuthSettings = new OAuthSettings();
        security.setOauthSettings(oAuthSettings);
        oAuthSettings.setAllowUserToAccessProfile(allowPropsAuth);
        CrowdSettings crowdSettings = new CrowdSettings();
        security.setCrowdSettings(crowdSettings);
        crowdSettings.setAllowUserToAccessProfile(allowCrowdAuth);
        if (ldapSetting != null) {
            LdapSetting ldap = new LdapSetting();
            ldap.setKey(ldapSetting);
            ldap.setAllowUserToAccessProfile(allowLdapAuth);
            security.setLdapSettings(List.of(ldap));
        }

        boolean isProfileUpdatable = service.isProfileUpdatable(userInfo);
        assertEquals(isProfileUpdatable, profileUpdatableExpected);
    }

    public void isProfileUpdatableWithCentralOAuthTest() {
        ArtifactoryHome artifactoryHome = new ArtifactoryHomeStub(getClass().getName());
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        ArtifactoryHome.bind(artifactoryHome);
        CentralConfigService configService = mock(CentralConfigService.class);
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        LdapService ldapService = mock(LdapService.class);
        when(ldapService.getUserLdapSetting("test")).thenReturn(null);
        service.setLdapService(ldapService);

        OAuthProviderSettings googleProvider = new OAuthProviderSettings();
        googleProvider.setEnabled(true);
        googleProvider.setName("Google Central");
        CentralOAuthSSOAddon centralOAuthSSOAddon = mock(CentralOAuthSSOAddon.class);
        when(centralOAuthSSOAddon.getConfiguration()).thenReturn(
                new CentralOAuthSettings(true, List.of(googleProvider)));
        AddonsManager addonsManager = mock(AddonsManager.class);
        when(addonsManager.addonByType(CentralOAuthSSOAddon.class)).thenReturn(centralOAuthSSOAddon);
        service.setAddons(addonsManager);

        PropsAuthenticationToken accessTokenAuthentication = mock(PropsAuthenticationToken.class);
        when(accessTokenAuthentication.getPropsKey()).thenReturn(OAUTH_INFO_PROPERTY + "Google Central");
        securityContext.setAuthentication(accessTokenAuthentication);

        SecurityDescriptor security = new SecurityDescriptor();
        descriptor.setSecurity(security);
        when(configService.getDescriptor()).thenReturn(descriptor);
        when(context.getCentralConfig()).thenReturn(configService);
        ArtifactoryContextThreadBinder.bind(context);

        HttpSsoSettings httpSsoSettings = new HttpSsoSettings();
        security.setHttpSsoSettings(httpSsoSettings);
        httpSsoSettings.setAllowUserToAccessProfile(false);
        SamlSettings samlSettings = new SamlSettings();
        security.setSamlSettings(samlSettings);
        samlSettings.setAllowUserToAccessProfile(false);
        OAuthSettings oAuthSettings = new OAuthSettings();
        security.setOauthSettings(oAuthSettings);
        oAuthSettings.setAllowUserToAccessProfile(false);
        CrowdSettings crowdSettings = new CrowdSettings();
        security.setCrowdSettings(crowdSettings);
        crowdSettings.setAllowUserToAccessProfile(false);

        UserInfo userInfo = mock(UserInfo.class);
        when(userInfo.getRealm()).thenReturn(OAuthHandler.REALM);
        when(userInfo.getUsername()).thenReturn("test");
        when(userInfo.isUpdatableProfile()).thenReturn(false);

        boolean isProfileUpdatable = service.isProfileUpdatable(userInfo);
        assertTrue(isProfileUpdatable);
    }

    @Test
    public void testIsInternalAcl() {
        Acl<? extends PermissionTarget> acl = mock(Acl.class);
        when(acl.getAccessIdentifier()).thenReturn("INTERNAL_permission");
        boolean isInternal = service.isInternalAcl(acl);
        assertTrue(isInternal);

        when(acl.getAccessIdentifier()).thenReturn("permission");
        isInternal = service.isInternalAcl(acl);
        assertFalse(isInternal);
    }

    @Test
    public void testGetAllAclsExcludingGenerated() {
        RepoAcl generatedRepoAcl =
                createRepoAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + "repo_perm_target", 0);
        BuildAcl generatedBuildAcl =
                createBuildAcl(SecurityService.INTERNAL_PERMISSION_PREFIX + "build_perm_target", "bu/**", 0);
        RepoAcl manuallyCreatedRepoAcl = createRepoAcl("not_internal_repo_perm_target", 0);
        BuildAcl manuallyCreatedBuildAcl = createBuildAcl("not_internal_build_perm_target", "bu/**", 0);

        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(List.of(generatedRepoAcl, manuallyCreatedRepoAcl));
        expect(aclStoreServiceMock.getAllBuildAcls()).andReturn(List.of(generatedBuildAcl, manuallyCreatedBuildAcl));
        doReturn(Collections.emptyList()).when(service).getAllReleaseBundleAclsByLicense();
        replay(aclStoreServiceMock);

        List<Acl<? extends PermissionTarget>> aclList = service.getAllAclsExcludingInternal();
        assertEquals(aclList.size(), 2);
        assertTrue(aclList.contains(manuallyCreatedRepoAcl));
        assertTrue(aclList.contains(manuallyCreatedBuildAcl));
    }

    @Test
    public void testGetUserPermissionsExcludingInternalOnce() {
        InfoFactory factory = InfoFactoryHolder.get();

        RepoAcl generatedRepoAcl = factory.createRepoAcl(factory.createRepoPermissionTarget(),
                Set.of(factory.createAce(PERMISSION_ACTOR, false, 0)),
                null, 0, SecurityService.INTERNAL_PERMISSION_PREFIX + "perm_target");

        RepoAcl manuallyCreatedRepoAcl = factory.createRepoAcl(factory.createRepoPermissionTarget(),
                Set.of(factory.createAce(PERMISSION_ACTOR, false, 0)),
                null, 0, "not_internal_repo_perm_target");

        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(List.of(generatedRepoAcl, manuallyCreatedRepoAcl));
        replay(aclStoreServiceMock);

        Map<PermissionTarget, AceInfo> permissionTargetAceInfoMap =
                service.getUserPermissions(PERMISSION_ACTOR, ArtifactoryResourceType.REPO);
        assertEquals(permissionTargetAceInfoMap.size(), 1);
        assertNotNull(permissionTargetAceInfoMap.get(manuallyCreatedRepoAcl.getPermissionTarget()));
    }

    @Test
    public void testGetUserPermissionAndItsGroupsExcludingInternalOnce() {
        InfoFactory factory = InfoFactoryHolder.get();

        RepoAcl generatedRepoAcl = factory.createRepoAcl(factory.createRepoPermissionTarget(),
                Set.of(factory.createAce(PERMISSION_ACTOR, false, 0)),
                null, 0, SecurityService.INTERNAL_PERMISSION_PREFIX + "perm_target");

        RepoAcl manuallyCreatedRepoAcl = factory.createRepoAcl(factory.createRepoPermissionTarget(),
                Set.of(factory.createAce(PERMISSION_ACTOR, false, 0)),
                null, 0, "not_internal_repo_perm_target");

        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(List.of(generatedRepoAcl, manuallyCreatedRepoAcl));
        expect(userGroupStoreService.findUser(PERMISSION_ACTOR)).andReturn(new UserImpl(PERMISSION_ACTOR));
        replay(aclStoreServiceMock, userGroupStoreService);

        Multimap<PermissionTarget, AceInfo> permissionTargetAceInfoMap =
                service.getUserPermissionAndItsGroups(PERMISSION_ACTOR, ArtifactoryResourceType.REPO);
        assertEquals(permissionTargetAceInfoMap.size(), 1);
        assertNotNull(permissionTargetAceInfoMap.get(manuallyCreatedRepoAcl.getPermissionTarget()));
    }

    @Test
    public void testGetGroupsPermissionsExcludingInternalOnce() {
        InfoFactory factory = InfoFactoryHolder.get();

        RepoAcl generatedRepoAcl = factory.createRepoAcl(factory.createRepoPermissionTarget(),
                Set.of(factory.createAce(PERMISSION_ACTOR, true, 0)),
                null, 0, SecurityService.INTERNAL_PERMISSION_PREFIX + "perm_target");

        RepoAcl manuallyCreatedRepoAcl = factory.createRepoAcl(factory.createRepoPermissionTarget(),
                Set.of(factory.createAce(PERMISSION_ACTOR, true, 0)),
                null, 0, "not_internal_repo_perm_target");

        expect(aclStoreServiceMock.getAllRepoAcls()).andReturn(List.of(generatedRepoAcl, manuallyCreatedRepoAcl));
        replay(aclStoreServiceMock);

        Multimap<PermissionTarget, AceInfo> permissionTargetAceInfoMap =
                service.getGroupsPermissions(List.of(PERMISSION_ACTOR), ArtifactoryResourceType.REPO);
        assertEquals(permissionTargetAceInfoMap.size(), 1);
        assertNotNull(permissionTargetAceInfoMap.get(manuallyCreatedRepoAcl.getPermissionTarget()));
    }

    @Test
    public void testGetAllBuildAcls() {
        BuildPermissionTargetImpl buildPermissionTarget = new BuildPermissionTargetImpl();
        MutableBuildAclImpl buildAcl = new MutableBuildAclImpl(buildPermissionTarget,
                Set.of(), "access", 12345L, INTERNAL_PERMISSION_PREFIX);
        expect(aclStoreServiceMock.getAllBuildAcls()).andReturn(List.of(buildAcl)).times(2);
        replay(aclStoreServiceMock);

        List<BuildAcl> results = service.getAllBuildAcls(false);
        assertEquals(results.size(), 1);

        results = service.getAllBuildAcls(true);
        assertEquals(results.size(), 0);
    }

    @Test
    public void testGetBuildPathAcls() {
        BuildPermissionTargetImpl buildPermissionTarget = new BuildPermissionTargetImpl();
        buildPermissionTarget.setIncludes(List.of("proj-build-info"));
        MutableBuildAclImpl buildAcl = new MutableBuildAclImpl(buildPermissionTarget,
                Set.of(), "access", 12345L, INTERNAL_PERMISSION_PREFIX);
        expect(aclStoreServiceMock.getAllBuildAcls()).andReturn(List.of(buildAcl)).times(2);
        replay(aclStoreServiceMock);

        RepoPath repoPath = RepoPathFactory.create("proj-build-info");
        List<BuildAcl> results = service.getBuildPathAcls(repoPath, false);
        assertEquals(results.size(), 1);

        results = service.getBuildPathAcls(repoPath, true);
        assertEquals(results.size(), 0);
    }

    @Test(dataProvider = "isValidGenPasswordProvider", description = "RTDEV-24476")
    public void isValidGenPassword(boolean expectedResult, String expirationValue) {
        homeStub.setProperty(ConstantValues.generatedPasswordExpirationPeriodMinutes, expirationValue);
        String genKey = ":" + System.currentTimeMillis() + ":";
        byte[] genKeyBytes = Base64.encodeBase64URLSafe(genKey.getBytes(Charsets.UTF_8));
        assertEquals(expectedResult, service.isValidGenPassword(new String(genKeyBytes, Charsets.UTF_8)));
    }

    @Test
    public void testUpdateGroupUsers() {
        String groupName = "test-group";
        List<String> currentUsers = new ArrayList<>(List.of("user1", "user2"));
        List<String> givenUsers = new ArrayList<>(List.of("user1", "user3"));
        MutableGroupInfo mutableGroupInfo = mock(MutableGroupInfo.class);
        when(mutableGroupInfo.getGroupName()).thenReturn(groupName);
        expect(userGroupStoreService.findUsersInGroup(groupName)).andReturn(currentUsers);
        replay(userGroupStoreService);
        Mockito.doAnswer((Answer<?>) invocationOnMock -> {
            assert invocationOnMock.getArgument(0).toString().equals(groupName) &&
                    invocationOnMock.getArgument(1).equals(Set.of("user2"));
            return null;
        }).when(service).removeUsersFromGroup(any(), any());

        Mockito.doAnswer((Answer<?>) invocationOnMock -> {
            assert invocationOnMock.getArgument(0).toString().equals(groupName) &&
                    invocationOnMock.getArgument(1).equals(Set.of("user3"));
            return null;
        }).when(service).addUsersToGroup(any(), any());

        service.updateGroupUsers(mutableGroupInfo, givenUsers);
    }

    public void testReturnTrueIfGranted() throws SQLException {
        Authentication authentication = mock(Authentication.class);

        RepoPath repoPath = RepoPathFactory.create("test-repo");
        UserImpl userInfo = new UserImpl();
        userInfo.setUsername("Genos-Sama");
        when(authentication.getPrincipal()).thenReturn(new SimpleUser(userInfo));

        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        when(internalRepositoryService.releaseBundlesRepoDescriptorByKey(anyString())).thenReturn(null);
        InternalBuildService buildService = mock(InternalBuildService.class);
        when(buildService.getAllBuildInfoRepos()).thenReturn(new HashSet<>());
        AclStoreService aclStoreService = mock(AclStoreService.class);
        Map<String, Map<String, Set<PrincipalPermission<RepoPermissionTarget>>>> userToRepoToAclMap = Maps.newHashMap();

        AceImpl aceU1 = new AceImpl(userInfo.getUsername(), false, ArtifactoryPermission.READ.getMask());
        RepoPermissionTargetImpl pt = new RepoPermissionTargetImpl("pt1",
                Lists.newArrayList(repoPath.getRepoKey()),
                Lists.newArrayList("**"),
                Lists.newArrayList());

        PrincipalPermission<RepoPermissionTarget> principalPermission = new PrincipalPermissionImpl<>(pt, aceU1);
        userToRepoToAclMap.put(userInfo.getUsername(), Map.of(repoPath.getRepoKey(), Set.of(principalPermission)));
        AclCache<RepoPermissionTarget> aclCache = new AclCache<>(Maps.newHashMap(), userToRepoToAclMap);
        when(aclStoreService.getAclCache()).thenReturn(aclCache);

        SecurityServiceImpl realSecurityService = new SecurityServiceImpl();
        realSecurityService.setRepositoryService(internalRepositoryService);
        realSecurityService.setAclStoreService(aclStoreService);
        realSecurityService.setBuildService(buildService);

        SecurityServiceImpl securityService = spy(realSecurityService);
        // granted by permission target
        boolean granted = securityService.isGranted(PermissionScopeResourceType.ARTIFACT,
                repoPath, ArtifactoryPermission.READ,
                authentication);

        assertTrue(granted);
        Mockito.verify(securityService).isGranted(repoPath, ArtifactoryPermission.READ, userInfo);
        Mockito.verify(securityService, times(0)).isGrantedByAccessToken(repoPath, authentication);
        Mockito.verify(securityService, times(0))
                .isGrantedBySignedUrl(repoPath, authentication, ArtifactoryPermission.READ);
        Mockito.verify(securityService, times(0))
                .isGrantedByResourcePermissionScopedToken(PermissionScopeResourceType.ARTIFACT, repoPath,
                        ArtifactoryPermission.READ, authentication);

        // granted by access token
        securityService = spy(realSecurityService);

        when(aclStoreService.getAclCache()).thenReturn(new AclCache<>(Maps.newHashMap(), Maps.newHashMap()));

        AccessTokenAuthentication accessTokenAuthentication = mock(AccessTokenAuthentication.class);
        when(accessTokenAuthentication.getPrincipal()).thenReturn(new SimpleUser(userInfo));
        JwtAccessToken jwtAccessToken = mock(JwtAccessToken.class);
        when(jwtAccessToken.getScope())
                .thenReturn(List.of("path.checksum:" + "c6c78f7c0994364f0057e63f756c018343f7278a" + ":read"));
        when(accessTokenAuthentication.getAccessToken()).thenReturn(jwtAccessToken);

        granted = securityService.isGranted(PermissionScopeResourceType.ARTIFACT,
                repoPath, ArtifactoryPermission.READ,
                accessTokenAuthentication);
        assertTrue(granted);
        Mockito.verify(securityService).isGranted(repoPath, ArtifactoryPermission.READ, userInfo);
        Mockito.verify(securityService).isGrantedByAccessToken(repoPath, accessTokenAuthentication);
        Mockito.verify(securityService, times(0))
                .isGrantedBySignedUrl(repoPath, accessTokenAuthentication, ArtifactoryPermission.READ);
        Mockito.verify(securityService, times(0))
                .isGrantedByResourcePermissionScopedToken(PermissionScopeResourceType.ARTIFACT, repoPath,
                        ArtifactoryPermission.READ, accessTokenAuthentication);

        // granted by signed url
        securityService = spy(realSecurityService);

        when(aclStoreService.getAclCache()).thenReturn(new AclCache<>(Maps.newHashMap(), Maps.newHashMap()));

        SignedUrlAuthenticationToken signedUrlAuthentication = mock(SignedUrlAuthenticationToken.class);
        when(signedUrlAuthentication.getPrincipal()).thenReturn(new SimpleUser(userInfo));

        granted = securityService.isGranted(PermissionScopeResourceType.ARTIFACT,
                repoPath, ArtifactoryPermission.READ,
                signedUrlAuthentication);

        assertTrue(granted);
        Mockito.verify(securityService).isGranted(repoPath, ArtifactoryPermission.READ, userInfo);
        Mockito.verify(securityService).isGrantedByAccessToken(repoPath, signedUrlAuthentication);
        Mockito.verify(securityService)
                .isGrantedBySignedUrl(repoPath, signedUrlAuthentication, ArtifactoryPermission.READ);
        Mockito.verify(securityService, times(0))
                .isGrantedByResourcePermissionScopedToken(PermissionScopeResourceType.ARTIFACT, repoPath,
                        ArtifactoryPermission.READ, signedUrlAuthentication);

        // granted by scoped token
        securityService = spy(realSecurityService);

        when(aclStoreService.getAclCache()).thenReturn(new AclCache<>(Maps.newHashMap(), Maps.newHashMap()));
        when(jwtAccessToken.getScope()).thenReturn(List.of("artifact:test-repo/**:r"));
        when(accessTokenAuthentication.getAccessToken()).thenReturn(jwtAccessToken);

        try {
            SecurityContextImpl securityContext = new SecurityContextImpl();
            securityContext.setAuthentication(accessTokenAuthentication);
            SecurityContextHolder.setContext(securityContext);

            // accessService.getAccessClient().auth().principalCanDo(principal, permissionScope);
            AccessService accessService = mock(AccessService.class);
            AccessClient accessClient = mock(AccessClient.class);
            AuthClient authClient = mock(AuthClient.class);
            when(authClient.principalCanDo(any(), any())).thenReturn(true);
            LoginResult loginResult = mock(LoginResult.class);
            when(loginResult.getPrincipal()).thenReturn(mock(AccessPrincipal.class));
            when(authClient.loginByAuthHeader(anyString())).thenReturn(loginResult);
            when(accessClient.auth()).thenReturn(authClient);
            when(accessService.getAccessClient()).thenReturn(accessClient);
            securityService.setAccessService(accessService);

            granted = securityService.isGranted(PermissionScopeResourceType.ARTIFACT,
                    repoPath, ArtifactoryPermission.READ,
                    accessTokenAuthentication);

            assertTrue(granted);
            Mockito.verify(securityService).isGranted(repoPath, ArtifactoryPermission.READ, userInfo);
            Mockito.verify(securityService).isGrantedByAccessToken(repoPath, accessTokenAuthentication);
            Mockito.verify(securityService)
                    .isGrantedBySignedUrl(repoPath, accessTokenAuthentication, ArtifactoryPermission.READ);
            Mockito.verify(securityService)
                    .isGrantedByResourcePermissionScopedToken(PermissionScopeResourceType.ARTIFACT, repoPath,
                            ArtifactoryPermission.READ, accessTokenAuthentication);
        } finally {
            SecurityContextHolder.setContext(new SecurityContextImpl());
        }
    }

    @DataProvider
    public static Object[][] isValidGenPasswordProvider() {
        return new Object[][] {
                //expectedResult,expirationValue
                {true, "10"},
                {false, "-1"}};
    }

    @DataProvider
    public static Object[][] authProvider() {
        return new Object[][] {
                //Class<Authentication> authClass, boolean allowSsoAuth, boolean allowSamlAuth, boolean
                //allowPropsAuth, String realm, String authProviderName, boolean requireUnlockExpected
                {AccessTokenAuthentication.class, true, false, true, HttpSsoAddon.REALM, null, false},
                {AccessTokenAuthentication.class, false, false, true, HttpSsoAddon.REALM, null, true},
                {AccessTokenAuthentication.class, true, false, true, OAuthHandler.REALM, null, false},
                {AccessTokenAuthentication.class, false, false, false, OAuthHandler.REALM, "Google Central", false},
                {AccessTokenAuthentication.class, false, false, false, null, "other provider", true},
                {AccessTokenAuthentication.class, true, false, true, SamlSsoAddon.REALM, null, true},
                {AccessTokenAuthentication.class, false, true, false, SamlSsoAddon.REALM, null, false},
                {PropsAuthenticationToken.class, false, false, true, null, null, false},
                {UsernamePasswordAuthenticationToken.class, true, false, false, HttpSsoAddon.REALM, null, true},
                {UsernamePasswordAuthenticationToken.class, false, true, false, SamlSsoAddon.REALM, null, true},
                {UsernamePasswordAuthenticationToken.class, false, false, true, OAuthHandler.REALM, null, true},
                {HttpSsoTestAuthentication.class, true, false, false, OAuthHandler.REALM, null, false},
                {HttpSsoTestAuthentication.class, false, false, true, OAuthHandler.REALM, null, true},
                {SamlTestAuthentication.class, false, true, false, OAuthHandler.REALM, null, false},
                {SamlTestAuthentication.class, false, false, true, OAuthHandler.REALM, null, true},
                {SamlTestAuthentication.class, false, true, false, CrowdAddon.REALM, null, false},
                {CrowdTestAuthentication.class, false, false, true, CrowdAddon.REALM, null, true},
                {CrowdTestAuthentication.class, false, false, false, LdapService.REALM, null, true},

        };
    }

    @DataProvider
    public static Object[][] profileUpdateProvider() {
        return new Object[][] {
                //String requestedRealm, boolean userProfileUpdatable, String ldapSetting,
                //boolean allowSsoAuth, boolean allowSamlAuth, boolean allowPropsAuth, boolean allowCrowdAuth,
                //boolean allowLdapAuth, boolean profileUpdatableExpected
                {"bla", true, null, false, false, false, false, false, true},
                {CrowdAddon.REALM, false, null, false, false, false, false, false, false},
                {CrowdAddon.REALM, false, null, false, false, false, true, false, true},
                {SamlSsoAddon.REALM, false, null, false, false, false, false, false, false},
                {SamlSsoAddon.REALM, false, null, false, true, false, false, false, true},
                {HttpSsoAddon.REALM, false, null, false, false, false, false, false, false},
                {HttpSsoAddon.REALM, false, null, true, false, false, false, false, true},
                {OAuthHandler.REALM, false, null, false, false, false, false, false, false},
                {OAuthHandler.REALM, false, null, false, false, true, false, false, true},
                {LdapService.REALM, false, null, false, false, false, false, false, false},
                {LdapService.REALM, false, "ldap", false, false, false, false, false, false},
                {LdapService.REALM, false, "ldap", false, false, false, false, true, true}

        };
    }

    private interface HttpSsoTestAuthentication extends Authentication {
    }

    private interface SamlTestAuthentication extends Authentication {
    }

    private interface CrowdTestAuthentication extends Authentication {
    }

    private void expectRemoteByKeyNullForAnyConsts() {
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(ANY_LOCAL_REPO)).andReturn(null).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(ANY_REMOTE_REPO)).andReturn(null).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(ANY_DISTRIBUTION_REPO)).andReturn(null).anyTimes();
        expect(repositoryServiceMock.remoteRepoDescriptorByKey(ANY_REPO)).andReturn(null).anyTimes();
        replay(repositoryServiceMock);
    }

    private void assertAuthenticatedAsSystem() {
        Authentication authentication = SecurityContextHolder.getContext().getAuthentication();
        assertTrue(authentication.getAuthorities().stream()
                        .anyMatch(authority -> ROLE_ADMIN.equals(authority.getAuthority())),
                "Authentication does not apply the admin role");
        assertEquals(((UserDetails) authentication.getPrincipal()).getUsername(), USER_SYSTEM);
    }

    private static void sleep(long ms) {
        try {
            Thread.sleep(ms);
        } catch (InterruptedException e) {
            e.printStackTrace();
        }
    }

    private void mockDaoRequests() throws SQLException {

        doAnswer(i -> {
            DbMigrationStatus dbMigrationStatus = i.getArgument(0);
            mockDbStatusMigrations.add(dbMigrationStatus);
            return null;
        })
                .when(migrationStatusDao)
                .insertMigration(any(DbMigrationStatus.class));

        doAnswer(i -> {
            String identifier = i.getArgument(0);
            mockDbStatusMigrations.removeIf(dbMigrationStatus ->
                    StringUtils.equals(dbMigrationStatus.getIdentifier(), identifier));
            return null;
        })
                .when(migrationStatusDao)
                .delete(Mockito.anyString());

        doAnswer(i -> {
            String identifier = i.getArgument(0);
            return StreamSupportUtils.stream(mockDbStatusMigrations)
                    .filter(dbMigrationStatus -> dbMigrationStatus.getIdentifier().equals(identifier))
                    .findFirst();
        })
                .when(migrationStatusDao)
                .findMigrationById(Mockito.anyString());

        when(migrationStatusDao.findAllUnfinishedIn(Mockito.anyList()))
                .thenAnswer(list ->
                        StreamSupportUtils.stream(mockDbStatusMigrations)
                                .filter(dbMigrationStatus ->
                                        ((List<String>) list).contains(dbMigrationStatus.getIdentifier())
                                                && dbMigrationStatus.getFinished() == 0)
                                .collect(Collectors.toList()));

        doAnswer(i -> {
            String identifier = i.getArgument(0);
            long finishTime = i.getArgument(1);
            Optional<DbMigrationStatus> optionalDbMigrationStatus = mockDbStatusMigrations.stream()
                    .filter(dbMigrationStatus -> dbMigrationStatus.getIdentifier().equals(identifier))
                    .findFirst();
            optionalDbMigrationStatus.ifPresent(dbMigrationStatus -> dbMigrationStatus.setFinished(finishTime));
            return null;
        }).when(migrationStatusDao).updateFinishTimeById(Mockito.anyString(), Mockito.anyLong());
    }

    private MigrationStatus createMigrationStatus(String identifier, String serviceId) {
        final long started = ClockUtils.epochMillis();

        MigrationStatus migrationStatus = new MigrationStatus();
        migrationStatus.setIdentifier(identifier);
        migrationStatus.setStarted(started);
        migrationStatus.setFinished(0);
        if (StringUtils.isNotBlank(serviceId)) {
            migrationStatus.setMigrationInfoBlob(new MarkerMigrationInfoBlob(serviceId));
        }
        return migrationStatus;
    }

    private MigrationStatus convert(DbMigrationStatus dbMigrationStatus) {
        MigrationStatus migrationStatus = new MigrationStatusMapperImpl()
                .dbMigrationStatusToMigrationStatus(dbMigrationStatus);
        migrationStatus.setMigrationInfoBlob(
                JsonUtils
                        .getInstance()
                        .readValue(dbMigrationStatus.getMigrationInfoBlob(), MarkerMigrationInfoBlob.class));
        return migrationStatus;
    }

    private void resetMockDbStatusMigrations() {
        mockDbStatusMigrations.clear();
    }

    private void testConvertAccessConverterAndItsPreInit(AccessConverter accessConverter, CompoundVersionDetails source,
            CompoundVersionDetails target,
            boolean createMarker, boolean expected) throws IOException {
        ArtifactoryServiceVersion versionToUpgrade = accessConverter.getMarkerMigrationConverter().
                getVersionToUpgrade();
        boolean shouldPreInitCreateARow = source.getVersion().before(versionToUpgrade) &&
                target.getVersion().afterOrEqual(versionToUpgrade);

        ArtifactoryDbProperties dbProperties = ArtifactoryHome.get().getDBProperties();
        JdbcHelper jdbcHelper = new JdbcHelper(CommonDataSourceFactory.createDataSource(dbProperties, false));
        DbDetails dbDetails = new DbDetails(dbProperties.getDbType(), jdbcHelper);
        new MarkerConverterAdapter(homeStub, dbDetails).convert(source, target);
        if (createMarker) {
            FileUtils.write(accessConverter.getMarkerFile(), "");
        }
        assertEquals(mockDbStatusMigrationsContains(accessConverter.getMarkerMigrationConverter().uniqueName()),
                shouldPreInitCreateARow);
        service.convert(source, target);
        assertEquals(mockDbStatusMigrationsContains(accessConverter.getMarkerMigrationConverter().uniqueName()),
                expected);
        assertFalse(accessConverter.getMarkerFile().exists());
    }

    private boolean mockDbStatusMigrationsContains(String uniqueName) {
        return StreamSupportUtils.stream(mockDbStatusMigrations)
                .anyMatch(dbMigrationStatus -> StringUtils.equals(dbMigrationStatus.getIdentifier(), uniqueName));
    }

    @DataProvider
    public static Object[][] supportBundlePermissionScopesAndExpected() {
        return new Object[][] {
                //tokenScope, expectedCanDeploy, expectedCanRead, expectedCanAnnotate, expectedCanDelete,
                //                                                                   expectedCanManage, isAdmin, isEdge
                {"internal:supportbundle:w", true, false, false, false, false, false, false}, // not admin, not Edge
                {"internal:supportbundle:r", false, true, false, false, false, false, false},
                {"internal:supportbundle:a", false, false, false, false, false, false, false},
                {"internal:supportbundle:d", false, false, false, true, false, false, false},
                {"internal:supportbundle:m", false, false, false, false, false, false, false},
                {"internal:supportbundle:r,w,d", true, true, false, true, false, false, false},
                {"artifact:jfrog-support-bundle:w", false, false, false, false, false, false, false},
                {"artifact:jfrog-support-bundle:r", false, false, false, false, false, false, false},
                {"artifact:jfrog-support-bundle:a", false, false, false, false, false, false, false},
                {"artifact:jfrog-support-bundle:d", false, false, false, false, false, false, false},
                {"artifact:jfrog-support-bundle:m", false, false, false, false, false, false, false},
                {"artifact:jfrog-support-bundle:r,w,d", false, false, false, false, false, false, false},
                {"", false, false, false, false, false, false, false},
                {"internal:supportbundle:w", true, true, true, true, true, true, false},      // admin, not Edge
                {"internal:supportbundle:r", true, true, true, true, true, true, false},
                {"internal:supportbundle:r,d", true, true, true, true, true, true, false},
                {"artifact:jfrog-support-bundle:w", true, true, true, true, true, true, false},
                {"artifact:jfrog-support-bundle:r,w", true, true, true, true, true, true, false},
                {"", true, true, true, true, true, true, false},

                {"internal:supportbundle:w", true, false, false, false, false, false, true},  // not admin, Edge
                {"internal:supportbundle:r", false, true, false, false, false, false, true},
                {"internal:supportbundle:a", false, false, false, false, false, false, true},
                {"internal:supportbundle:d", false, false, false, true, false, false, true},
                {"internal:supportbundle:m", false, false, false, false, false, false, true},
                {"internal:supportbundle:r,w,d", true, true, false, true, false, false, true},
                {"artifact:jfrog-support-bundle:w", false, false, false, false, false, false, true},
                {"artifact:jfrog-support-bundle:r", false, false, false, false, false, false, true},
                {"artifact:jfrog-support-bundle:a", false, false, false, false, false, false, true},
                {"artifact:jfrog-support-bundle:d", false, false, false, false, false, false, true},
                {"artifact:jfrog-support-bundle:m", false, false, false, false, false, false, true},
                {"artifact:jfrog-support-bundle:r,w,d", false, false, false, false, false, false, true},
                {"", false, false, false, false, false, false, false},
                {"internal:supportbundle:w", true, true, true, true, true, true, true},       // admin, Edge
                {"internal:supportbundle:r", false, true, true, true, true, true, true},
                {"internal:supportbundle:r,d,m", false, true, true, true, true, true, true},
                {"internal:supportbundle:r,w,d", true, true, true, true, true, true, true},
                {"artifact:jfrog-support-bundle:w", false, true, true, true, true, true, true},
                {"artifact:jfrog-support-bundle:r,w", false, true, true, true, true, true, true},
                {"", false, true, true, true, true, true, true},

                // note: of scope "applied-permissions/admin" isn't needed to be covered here, as in real flow it
                // translates to role admin, so SecurityServiceImpl.isAdmin() returns true, and this is covered
        };
    }

    @Test(dataProvider = "supportBundlePermissionScopesAndExpected")
    public void testSupportBundleActionsByPermissionScopedToken(String userTokenPermissionScope,
            boolean expectedCanDeploy, boolean expectedCanRead, boolean expectedCanAnnotate, boolean expectedCanDelete,
            boolean expectedCanManage, boolean isAdmin, boolean isEdge) {
        AccessService accessServiceMock = mock(AccessService.class, Answers.RETURNS_DEEP_STUBS);
        LoginResult loginResult = mock(LoginResult.class);
        AccessPrincipal accessPrincipal = mock(AccessPrincipal.class);
        when(loginResult.getPrincipal()).thenReturn(accessPrincipal);
        when(accessServiceMock.getAccessClient().auth().loginByAuthHeader(any(String.class))).thenReturn(loginResult);
        when(accessServiceMock.getAccessClient().auth().principalCanDo(any(), any())).thenAnswer(
                (Answer<Boolean>) invocationOnMock -> {
                    PermissionScope requiredPermissionScope = invocationOnMock.getArgument(1);
                    String[] userScopeArr = userTokenPermissionScope.split(":");
                    return userScopeArr[0].equals(requiredPermissionScope.getResourceType().getName()) &&
                            userScopeArr[1].equals(requiredPermissionScope.getTarget()) &&
                            userScopeArr[2].contains(requiredPermissionScope.getAction().getCode());
                });
        service.setAccessService(accessServiceMock);
        AddonsManager addonsManager = mock(AddonsManager.class);
        when(addonsManager.isEntitled(EntitlementType.EDGE_MIXED)).thenReturn(isEdge);
        service.setAddons(addonsManager);
        List<String> scopes = // null scopes will use Basic Auth instead of token
                StringUtils.isEmpty(userTokenPermissionScope) ? null : Lists.newArrayList(userTokenPermissionScope);
        authenticateUserWithScopes(isAdmin, scopes);
        RepoPath repoPath = RepoPathFactory.create("jfrog-support-bundle", "p/a/t/h");
        assertEquals(service.canDeploy(repoPath), expectedCanDeploy);
        assertEquals(service.canRead(repoPath), expectedCanRead);
        assertEquals(service.canAnnotate(repoPath), expectedCanAnnotate);
        assertEquals(service.canDelete(repoPath), expectedCanDelete);
        assertEquals(service.canManage(repoPath), expectedCanManage);
    }
}
