/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.security.access;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.JWSAlgorithm;
import org.apache.commons.io.FileUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.api.config.ImportSettingsImpl;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationException;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.api.security.UserGroupService;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.entitlement.EntitlementType;
import org.artifactory.sapi.common.ImportKind;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.storage.fs.service.ConfigsService;
import org.artifactory.storage.security.service.UserGroupStoreService;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jfrog.access.client.AccessClient;
import org.jfrog.access.client.AccessClientException;
import org.jfrog.access.client.AccessClientHttpException;
import org.jfrog.access.client.auth.AuthClientImpl;
import org.jfrog.access.client.auth.AuthenticationResult;
import org.jfrog.access.client.auth.MTlsClientImpl;
import org.jfrog.access.client.cert.CertClient;
import org.jfrog.access.client.config.ConfigClient;
import org.jfrog.access.client.platformconfig.GrpcPlatformConfigClient;
import org.jfrog.access.client.registry.RegistryClient;
import org.jfrog.access.client.system.SystemClient;
import org.jfrog.access.client.token.TokenClient;
import org.jfrog.access.client.token.TokenRequest;
import org.jfrog.access.client.token.TokenResponseImpl;
import org.jfrog.access.router.RouterGrpcClient;
import org.jfrog.access.token.JwtAccessTokenImpl;
import org.jfrog.access.util.JoinKeyUtils;
import org.jfrog.client.http.RestResponse;
import org.jfrog.security.crypto.JFrogCryptoHelper;
import org.jfrog.security.ssl.CertificateGenerationException;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.springframework.context.ApplicationEventPublisher;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.crypto.SecretKey;
import javax.crypto.spec.SecretKeySpec;
import javax.servlet.http.HttpServletRequest;
import java.io.File;
import java.io.IOException;
import java.security.cert.Certificate;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.UUID;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

import static org.artifactory.test.TestUtils.createCertificate;
import static org.easymock.EasyMock.createNiceMock;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.argThat;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;
import static org.mockito.MockitoAnnotations.initMocks;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertThrows;

/**
 * @author Yinon Avraham
 */
public class AccessServiceImplTest extends ArtifactoryHomeBoundTest {

    private static final CharSequence TEST_JFROG_JOIN_KEY =
            "cc949ef041b726994a225dc20e018f2310101010101010101010101010101010";
    //default value of ConstantValues.accessTokenNonAdminMaxExpiresIn
    private static long constValueDefaultMaxExpiresIn = TimeUnit.HOURS.toSeconds(1);
    private static long USER_TOKEN_MAX_EXPIRES_IN_MINUTES_DEFAULT =
            TimeUnit.SECONDS.toMinutes(constValueDefaultMaxExpiresIn);
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private UserGroupStoreService userGroupStore;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private SecurityService securityService;
    @Mock
    private UserGroupService userGroupService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private ApplicationEventPublisher publisher;
    @Mock
    private ConfigsService configsService;
    @Mock
    private AccessClient accessClient;
    @Mock
    private ArtifactoryApplicationContext applicationContext;
    @Mock
    private SystemClient systemClient;
    @Mock
    private GrpcPlatformConfigClient platformConfigClient;
    private AccessServiceImpl accessServiceSpy;

    @BeforeMethod
    public void resetAll() throws CertificateGenerationException, IOException {
        initMocks(this);

        when(applicationContext.getContextPath()).thenReturn("artifactory");
        AccessServiceImpl accessService =
                new AccessServiceImpl(publisher, configsService, applicationContext);
        accessService.setAddonsManager(addonsManager);
        accessService.setSecurityService(securityService);
        accessService.setAuthorizationService(authorizationService);
        accessService.setUserGroupService(userGroupService);
        accessService.setUserGroupStore(userGroupStore);
        accessService.setRepositoryService(repositoryService);
        accessService.setAuthFailureMaxRetry(2);


        accessServiceSpy = Mockito.spy(accessService);
        when(accessClient.useAuth(argThat(i -> true))).thenReturn(accessClient);
        when(accessClient.useAuth(argThat(i -> true), argThat(i -> true))).thenReturn(accessClient);
        TokenClient tokenClient = mock(TokenClient.class);
        //noinspection ConstantConditions
        when(tokenClient.create(argThat(i -> true)))
                .thenReturn(new TokenResponseImpl("blah", "blah", "blah", Long.MAX_VALUE, null));
        when(accessClient.token()).thenReturn(tokenClient);
        systemClient = mock(SystemClient.class);
        when(systemClient.getAccessServerName()).thenReturn("blah");
        when(systemClient.getStartupPingRetries()).thenReturn(5);
        when(accessClient.system()).thenReturn(systemClient);
        RestResponse restResponse = mock(RestResponse.class);
        when(restResponse.isSuccessful()).thenReturn(true);
        when(restResponse.getBodyAsString()).thenReturn("{\"token\":\"aToken\"}");
        when(accessClient.restCall(argThat(i -> true))).thenReturn(restResponse);
        CertClient certClient = mock(CertClient.class);

        Certificate certificate = createCertificate(JFrogCryptoHelper.generateKeyPair());
        when(certClient.getRootCertificate()).thenReturn(certificate);
        when(accessClient.cert()).thenReturn(certClient);
        when(accessClient.router()).thenReturn(mock(RouterGrpcClient.class));

        //newClientBuilder()->getAccessServerSettings()->detectBundledAccessServerUrl() is hard to mock, so we stub
        doReturn(accessClient).when(accessServiceSpy).buildAccessClient();

        HaCommonAddon haCommonAddon = mock(HaCommonAddon.class);
        when(addonsManager.addonByType(HaCommonAddon.class)).thenReturn(haCommonAddon);
        ConfigClient configClient = mock(ConfigClient.class);
        when(accessClient.config()).thenReturn(configClient);

        MTlsClientImpl mTlsClient = mock(MTlsClientImpl.class);
        AuthClientImpl authClient = mock(AuthClientImpl.class);

        when(accessClient.auth()).thenReturn(authClient);
        when(accessClient.auth().mtls()).thenReturn(mTlsClient);

        RegistryClient registryClient = mock(RegistryClient.class);
        SecretKey secretKey =
                new SecretKeySpec(JoinKeyUtils.getSigningKey(""), JWSAlgorithm.HS256.getName());
        String secretKeyId = JoinKeyUtils.getKeyId("");
        String tokenValue =
                JwtAccessTokenImpl.builder()
                        .buildAndSign(secretKey, secretKeyId).getTokenValue();
        when(registryClient.join(anyString())).thenReturn(new ObjectMapper().readTree
                ("{\"token\":\"" + tokenValue + "\"}"));
        when(accessClient.registry()).thenReturn(registryClient);
        File joinKeyFile = new File(ArtifactoryHome.get().getSysConfig().helper().getJoinKeyFile());
        FileUtils.write(joinKeyFile, TEST_JFROG_JOIN_KEY);
        when(accessClient.cachedGrpcPlatformConfigs()).thenReturn(platformConfigClient);
        accessServiceSpy.init();
    }

    @Test(dataProvider = "provideNormalizeInstanceId")
    public void testNormalizeInstanceId(String original, String expected) {
        assertEquals(AccessServiceImpl.normalizeInstanceId(original), expected);
    }

    @Test(expectedExceptions = {IllegalArgumentException.class})
    public void testNormalizeInstanceIdTooShort() {
        AccessServiceImpl.normalizeInstanceId("foo");
    }

    @Test(dataProvider = "provideValidExpiresInForNonAdmin")
    public void testAssertExpiresInForNonAdminUsers(boolean nonnull, Long maxExpiresIn, Long expiresIn,
            Long maxExpiresInConstValue) {
        Long finalExpire = maxExpiresInConstValue;
        if (nonnull) {
            finalExpire = Optional.ofNullable(maxExpiresIn)
                    .map(TimeUnit.MINUTES::toSeconds)
                    .orElse(maxExpiresInConstValue);
        }
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.accessTokenNonAdminMaxExpiresIn.getPropertyName(),
                        finalExpire.toString());

        TokenRequest tokenRequest = TokenRequest.scopes("api:*").nonRefreshable().expiresIn(expiresIn).build();
        accessServiceSpy.assertValidExpiresInForNonAdmin(tokenRequest, "testuser", false);
    }

    @Test(dataProvider = "provideInvalidExpiresInForNonAdmin", expectedExceptions = AuthorizationException.class)
    public void failAssertExpiresInForNonAdminUsers(boolean nonnull, Long maxExpiresIn, Long expiresIn,
            Long maxExpiresInConstValue) {
        testAssertExpiresInForNonAdminUsers(nonnull, maxExpiresIn, expiresIn, maxExpiresInConstValue);
    }


    @Test
    public void testAccessTimeOutDuringBootstrap() throws Exception {
        ArtifactoryHome.get().getArtifactoryProperties()
                .setProperty(ConstantValues.accessClientWaitForServer.getPropertyName(), "2");
        doThrow(new AccessClientException("message")).when(this.accessClient).ping();
        try {
            accessServiceSpy.initAccessService("service-id");
        } catch (Exception e) {
        }
        verify(this.accessClient, times(0)).close();
    }

    @Test(expectedExceptions = AccessClientException.class,
            expectedExceptionsMessageRegExp = "HTTP response status 401:Test")
    public void testEnsureAuthMaxRetriesReached() {
        accessServiceSpy.ensureAuth(() -> {
            throw new AccessClientHttpException(HttpStatus.SC_UNAUTHORIZED, "Test");
        });
    }

    @Test
    public void testEnsureAuth2ndCallSuccess() {
        AtomicBoolean failedOnce = new AtomicBoolean(false);
        String result = accessServiceSpy.ensureAuth(() -> {
            if (!failedOnce.get()) {
                failedOnce.set(true);
                throw new AccessClientHttpException(HttpStatus.SC_UNAUTHORIZED, "Test");
            } else {
                return "Test";
            }
        });
        Assert.assertEquals(result, "Test");
    }

    @Test
    public void testExtractMtlsUsername() {
        String expectedUsername = "vova";

        AuthenticationResult<String> extractedUserName = new AuthenticationResult<>() {
            @Nullable
            @Override
            public String getPrincipal() {
                return expectedUsername;
            }

            @Override
            public boolean isSuccessful() {
                return true;
            }

            @NotNull
            @Override
            public String getReason() {
                return "none";
            }
        };
        when(addonsManager.isEntitled(EntitlementType.MTLS)).thenReturn(true);
        when(accessClient.auth().mtls().extractUsername(any())).thenReturn(extractedUserName);
        HttpServletRequest request = createNiceMock(HttpServletRequest.class);
        String actualUsername = accessServiceSpy.extractMtlsUsername(request);
        verify(accessClient.auth().mtls()).extractUsername(any());

        assertEquals(actualUsername, expectedUsername);
    }

    @Test
    public void moveTwoValuesConfigResources() {
        Map<String, Object> data = new HashMap<>();
        Map<String, Object> security = new HashMap<>();
        Map<String, Object> userLockPolicy = new HashMap<>();
        userLockPolicy.put("enabled", true);
        userLockPolicy.put("loginAttempts", 3);
        security.put("anonAccessEnabled", true);
        security.put("userLockPolicy",userLockPolicy);
        data.put("security",security);
        int expected = 2;
        int actual = accessServiceSpy.moveConfigResourcesToAccess(data);
        assertEquals(actual, expected);
        assertEquals(0,((Map)data.get("security")).size());
    }


    @Test
    public void moveOneValuesConfigResourcesAndKeepOne() {
        Map<String, Object> data = new HashMap<>();
        Map<String, Object> security = new HashMap<>();
        Map<String, Object> userLockPolicy = new HashMap<>();
        userLockPolicy.put("enabled", true);
        userLockPolicy.put("loginAttempts", 3);
        data.put("urlBase", "someStaff");
        security.put("userLockPolicy",userLockPolicy);
        data.put("security",security);
        int expected = 1;
        int actual = accessServiceSpy.moveConfigResourcesToAccess(data);
        assertEquals(actual, expected);
        assertEquals(2,data.size());
        assertEquals("someStaff",data.get("urlBase"));
    }
    @Test
    public void moveOneValueConfigResources() {
        Map<String, Object> data = new HashMap<>(Map.of("security", new HashMap<>(Map.of("anonAccessEnabled", true))));
        int expected = 1;
        int actual = accessServiceSpy.moveConfigResourcesToAccess(data);
        assertEquals(actual, expected);
    }

    @Test
    public void testExtractMtlsUsernameNotEntitled() {
        String expectedUsername = "vova";

        AuthenticationResult<String> extractedUserName = new AuthenticationResult<>() {
            @Nullable
            @Override
            public String getPrincipal() {
                return expectedUsername;
            }

            @Override
            public boolean isSuccessful() {
                return true;
            }

            @NotNull
            @Override
            public String getReason() {
                return "none";
            }
        };
        when(addonsManager.isEntitled(EntitlementType.MTLS)).thenReturn(false);
        when(accessClient.auth().mtls().extractUsername(any())).thenReturn(extractedUserName);
        HttpServletRequest request = createNiceMock(HttpServletRequest.class);
        String actualUsername = accessServiceSpy.extractMtlsUsername(request);
        verify(accessClient.auth().mtls(), never()).extractUsername(any());

        assertNull(actualUsername);
    }

    @Test
    public void testExtractMtlsUsernameNull() {

        AuthenticationResult<String> extractedUserName = new AuthenticationResult<>() {
            @Nullable
            @Override
            public String getPrincipal() {
                return null;
            }

            @Override
            public boolean isSuccessful() {
                return false;
            }

            @NotNull
            @Override
            public String getReason() {
                return "some fake reason";
            }
        };
        when(addonsManager.isEntitled(EntitlementType.MTLS)).thenReturn(true);
        when(accessClient.auth().mtls().extractUsername(any())).thenReturn(extractedUserName);
        HttpServletRequest request = createNiceMock(HttpServletRequest.class);
        String actualUsername = accessServiceSpy.extractMtlsUsername(request);
        assertNull(actualUsername);
    }

    @Test(expectedExceptions = AccessClientException.class,
            expectedExceptionsMessageRegExp = "java.lang.Exception: Test")
    public void testEnsureAuthFailsByCodeException() {
        accessServiceSpy.ensureAuth(() -> {
            throw new Exception("Test");
        });
    }

    @Test
    public void testAssertAcceptedScopePermissionScopes() {
        List<String> scopes =
                List.of("system:health:r", "system:livelogs:r", "system:health:r", "system:supportbundle:x");
        accessServiceSpy.assertAcceptedScope(scopes);
        assertThrows(IllegalArgumentException.class,
                () -> accessServiceSpy.assertAcceptedScope(Collections.singletonList("invalid:health:r")));
        assertThrows(IllegalArgumentException.class,
                () -> accessServiceSpy
                        .assertAcceptedScope(Collections.singletonList("system:health:somethinginvalid")));
    }

    @Test
    public void whenHybridMigrationThenSaasImportPerformed() throws IOException {
        ImportSettingsImpl importSettings = new ImportSettingsImpl(null);
        importSettings.setImportKind(ImportKind.HYBRID_MIGRATION);

        accessServiceSpy.performImport(importSettings);

        verify(systemClient, times(0)).importAccessServer();
        verify(systemClient, times(1)).importAccessServerSaasMigration();
    }

    @Test
    public void whenStandartImportThenNotSaasImportPerformed() throws IOException {
        ImportSettingsImpl importSettings = new ImportSettingsImpl(null);
        importSettings.setImportKind(ImportKind.STANDARD);

        accessServiceSpy.performImport(importSettings);

        verify(systemClient, times(1)).importAccessServer();
        verify(systemClient, times(0)).importAccessServerSaasMigration();
    }

    @DataProvider
    private static Object[][] provideNormalizeInstanceId() {
        String uuid = UUID.randomUUID().toString();
        return new Object[][] {
                {uuid, uuid},
                {"b0902ad17dfa28f5:7c169c4a:15849ae7faa:-8000", "b0902ad17dfa28f5_7c169c4a_15849ae7faa_-8000"},
                {" asdf 1234-erty_dfg5+sfg:sdf65=ghtr54 ", "_asdf_1234-erty_dfg5_sfg_sdf65_ghtr54"}
        };
    }

    @DataProvider
    private static Object[][] provideValidExpiresInForNonAdmin() {
        long validDefaultExpiresIn = USER_TOKEN_MAX_EXPIRES_IN_MINUTES_DEFAULT * 60 / 2;
        return new Object[][] {
                //{nonnull, constValueDefaultMaxExpiresIn, expiresIn, maxExpiresInConstValue}
                {false, null, validDefaultExpiresIn, constValueDefaultMaxExpiresIn}, //expects default max expires in
                {true, null, validDefaultExpiresIn, constValueDefaultMaxExpiresIn}, //expects default max expires in
                {true, null, USER_TOKEN_MAX_EXPIRES_IN_MINUTES_DEFAULT, constValueDefaultMaxExpiresIn}, //expects
                // default max expires in
                {true, 10L, 5L * 60, constValueDefaultMaxExpiresIn},
                {true, 10L, 10L * 60, constValueDefaultMaxExpiresIn},
                {true, 0L, 10L, constValueDefaultMaxExpiresIn},
                {true, 0L, null, constValueDefaultMaxExpiresIn},
                {true, 0L, Long.MAX_VALUE, constValueDefaultMaxExpiresIn},
                {true, null, constValueDefaultMaxExpiresIn - 60, constValueDefaultMaxExpiresIn - 60},
                {true, null, constValueDefaultMaxExpiresIn, constValueDefaultMaxExpiresIn},
                {true, null, constValueDefaultMaxExpiresIn + 60, constValueDefaultMaxExpiresIn + 60},
        };
    }

    @DataProvider
    private static Object[][] provideInvalidExpiresInForNonAdmin() {
        long invalidDefaultExpiresIn = USER_TOKEN_MAX_EXPIRES_IN_MINUTES_DEFAULT * 60 + 1;
        return new Object[][] {
                //{nonnull, constValueDefaultMaxExpiresIn, expiresIn, maxExpiresInConstValue}
                {false, null, invalidDefaultExpiresIn, constValueDefaultMaxExpiresIn}, //expects default max expires in
                {true, null, invalidDefaultExpiresIn, constValueDefaultMaxExpiresIn}, //expects default max expires in
                {true, null, 0L, constValueDefaultMaxExpiresIn},
                {true, 10L, 10L * 60 + 1, constValueDefaultMaxExpiresIn},
                {true, 10L, null, constValueDefaultMaxExpiresIn},
                {true, null, Long.MAX_VALUE, constValueDefaultMaxExpiresIn},
                {true, null, constValueDefaultMaxExpiresIn, constValueDefaultMaxExpiresIn - 60},
                {true, null, constValueDefaultMaxExpiresIn + 60, constValueDefaultMaxExpiresIn},
                {true, null, constValueDefaultMaxExpiresIn + 60 + 60, constValueDefaultMaxExpiresIn + 60},
        };
    }
}