/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.util;

import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpRequestInterceptor;
import org.apache.http.auth.AuthSchemeProvider;
import org.apache.http.auth.AuthScope;
import org.apache.http.auth.Credentials;
import org.apache.http.auth.UsernamePasswordCredentials;
import org.apache.http.client.config.AuthSchemes;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.impl.auth.BasicSchemeFactory;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.crypto.CryptoHelper;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.mbean.MBeanRegistrationServiceImpl;
import org.artifactory.repo.http.DenyListCacheHolderService;
import org.artifactory.repo.http.HttpConnectionPoolMonitorService;
import org.artifactory.repo.http.IdleConnectionMonitorService;
import org.artifactory.repo.http.mbean.HTTPConnectionPool;
import org.artifactory.util.bearer.RepoSpecificBearerSchemeFactory;
import org.jfrog.client.http.CloseableHttpClientDecorator;
import org.jfrog.client.http.CloseableObserver;
import org.jfrog.client.http.DenylistHttpBuilder;
import org.jfrog.client.http.auth.PreemptiveAuthInterceptor;
import org.jfrog.client.http.model.ProxyConfig;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nullable;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import static org.apache.commons.lang.StringUtils.isNotBlank;

/**
 * Artifactory builder extension for the common HTTP client builder.
 *
 * @author Yossi Shaul
 */
public class HttpClientConfigurator extends DenylistHttpBuilder<HttpClientConfigurator> {
    private static final Logger log = LoggerFactory.getLogger(HttpClientConfigurator.class);

    private String repoKey;
    private String jmxName;

    public HttpClientConfigurator() {
        super(ArtifactoryHome.get().getSysConfig(),
                ContextHelper.get().beanForType(DenyListCacheHolderService.class).getDenyListCache());
        userAgent(HttpUtils.getArtifactoryUserAgent());
        handleGzipResponse(ConstantValues.httpAcceptEncodingGzip.getBoolean());
        connectionPoolTTL(ConstantValues.httpConnectionPoolTimeToLive.getInt());
        config.setMaxRedirects(20);
        config.setCircularRedirectsAllowed(true);
        maxConnectionsTotal = ConstantValues.httpClientMaxTotalConnections.getInt();
        maxConnectionsPerRoute = ConstantValues.httpClientMaxConnectionsPerRoute.getInt();
    }

    @Override
    public CloseableHttpClient build(boolean withParams) {
        // TODO: [by fsi] This all system looks messy. Too many fields and connection between elements.
        CloseableHttpClientDecorator client = (CloseableHttpClientDecorator) super.build(withParams);
        IdleConnectionMonitorService idleConnectionMonitorService =
                ContextHelper.get().beanForType(IdleConnectionMonitorService.class);
        idleConnectionMonitorService.add(client.getId(), client.getClientConnectionManager());
        client.registerCloseableObserver((CloseableObserver) idleConnectionMonitorService);
        if (StringUtils.isNotBlank(jmxName)) {
            HTTPConnectionPool pool = new HTTPConnectionPool(client.getClientConnectionManager());
            ContextHelper.get().beanForType(HttpConnectionPoolMonitorService.class).monitorHttpPool(jmxName, pool);
            try {
                ContextHelper.get().beanForType(MBeanRegistrationServiceImpl.class)
                        .register(pool, "HTTPConnectionPool",
                                jmxName);
            } catch (Exception e) {
                String err = "Failed register HTTPConnectionPool mbean for " + jmxName + ": " + e.getMessage();
                log.error(err);
                log.debug(err, e);
            }
        }
        return client;
    }

    public HttpClientConfigurator registerMBean(String jmxName) {
        this.jmxName = jmxName;
        return this;
    }

    @Override
    protected PoolingHttpClientConnectionManager configConnectionManager() {
        additionalConfigByAuthScheme();
        return super.configConnectionManager();
    }

    @Override
    public boolean isCookieSupportEnabled() {
        return cookieSupportEnabled || ConstantValues.enableCookieManagement.getBoolean();
    }

    /**
     * Disable the automatic gzip compression on read.
     * Once disabled cannot be activated.
     */
    public HttpClientConfigurator handleGzipResponse(boolean handleGzipResponse) {
        if (!handleGzipResponse) {
            disableGzipResponse();
        }
        return this;
    }

    public HttpClientConfigurator enableTokenAuthentication(boolean enableTokenAuthentication, String repoKey,
            @Nullable HttpRequestInterceptor requestInterceptor) {
        if (enableTokenAuthentication) {
            if (defaultHost == null || StringUtils.isBlank(defaultHost.getHostName())) {
                throw new IllegalStateException("Cannot configure authentication when host is not set.");
            }
            this.repoKey = repoKey;
            config.setTargetPreferredAuthSchemes(Collections.singletonList("Bearer"));
            // The repository key is passed to the Bearer to reuse it's http client
            Registry<AuthSchemeProvider> bearerRegistry = RegistryBuilder.<AuthSchemeProvider>create()
                    .register("Bearer", new RepoSpecificBearerSchemeFactory(repoKey))
                    .build();
            builder.setDefaultAuthSchemeRegistry(bearerRegistry);
            if (requestInterceptor != null) {
                addRequestInterceptor(requestInterceptor);
            }
            chosenAuthScheme = JFrogAuthScheme.BEARER;
        }
        return this;
    }

    public HttpClientConfigurator proxy(@Nullable ProxyDescriptor proxyDescriptor) {
        if (proxyDescriptor != null) {
            ProxyConfig proxyConfig = proxyDescriptor.toProxyConfig();
            if (isNotBlank(proxyConfig.getPassword())) {
                proxyConfig.setPassword(CryptoHelper.decryptIfNeeded(ArtifactoryHome.get(), proxyConfig.getPassword()));
            }
            proxy(proxyConfig);
        }
        return this;
    }

    public RequestConfig.Builder getConfig() {
        return this.config;
    }

    /**
     * Sets required configuration based on the final chosen config for this client so that configurations
     * don't interfere with each other based on when they were created in the chain (i.e. auth before token etc.)
     */
    protected void additionalConfigByAuthScheme() {
        switch (chosenAuthScheme) {
            case BASIC:
                builder.addInterceptorFirst(new PreemptiveAuthInterceptor());
                break;
            case BEARER:
                if (shouldConfigureBearerDummyCredentials()) {
                    // We need dummy credentials hack to enforce httpClient behavior, otherwise we won't respond to a
                    //challenge properly... Dummy:dummy is the specification for forcing token authentication
                    credsProvider.setCredentials(
                            new AuthScope(defaultHost.getHostName(), AuthScope.ANY_PORT, AuthScope.ANY_REALM),
                            new UsernamePasswordCredentials("dummy", "dummy"));
                } else {
                    //Valid credentials exist for target host - set basic auth preference and register additional scheme
                    //so we can respond to basic challenges from target as required
                    List<String> authPrefs = Arrays.asList("Bearer", AuthSchemes.BASIC);
                    config.setTargetPreferredAuthSchemes(authPrefs);
                    Registry<AuthSchemeProvider> bearerRegistry = RegistryBuilder.<AuthSchemeProvider>create()
                            .register("Bearer", new RepoSpecificBearerSchemeFactory(repoKey))
                            .register(AuthSchemes.BASIC, new BasicSchemeFactory())
                            .build();
                    builder.setDefaultAuthSchemeRegistry(bearerRegistry);
                }
                break;
        }
    }

    /**
     * @return false if credentials were configured for this client's host (or any host if lenient) -> also
     * verifies the credential set is not the one configured for the proxy (proxy credentials are not considered
     * host credentials), true if dummy credentials should be configured for Bearer auth
     */
    private boolean shouldConfigureBearerDummyCredentials() {
        boolean shouldSetDummy = false;
        Credentials hostCreds = credsProvider
                .getCredentials(new AuthScope(defaultHost.getHostName(), AuthScope.ANY_PORT, AuthScope.ANY_REALM));
        Credentials anyHostCreds = credsProvider
                .getCredentials(new AuthScope(AuthScope.ANY_HOST, AuthScope.ANY_PORT, AuthScope.ANY_REALM));
        Credentials proxyCreds = null;
        if (proxyHost != null && StringUtils.isNotBlank(proxyHost.getHostName())) {
            proxyCreds =
                    credsProvider.getCredentials(
                            new AuthScope(proxyHost.getHostName(), AuthScope.ANY_PORT, AuthScope.ANY_REALM));
        }

        //Any host allowed - make sure credentials were set and that the credsProvider didn't just return
        //the proxy's credentials for the ANY_HOST scope.
        if (allowAnyHostAuth && anyHostCreds != null
                && (proxyCreds == null || (!proxyCreds.getUserPrincipal().equals(anyHostCreds.getUserPrincipal())))) {
            shouldSetDummy = true;
        } else if (hostCreds == null) {
            shouldSetDummy = true;
        }
        return shouldSetDummy;
    }

}
