package org.artifactory.lifecycle.converter.postinit.v105;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.npm.NpmAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.SecurityService;
import org.artifactory.aql.AqlService;
import org.artifactory.aql.api.domain.sensitive.AqlApiItem;
import org.artifactory.aql.api.internal.AqlBase;
import org.artifactory.aql.model.AqlItemTypeEnum;
import org.artifactory.aql.result.AqlEagerResult;
import org.artifactory.aql.result.rows.AqlItem;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.fs.FileInfo;
import org.artifactory.repo.RepoPath;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.version.CompoundVersionDetails;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import java.util.List;
import java.util.stream.Stream;

import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.name;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.path;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.property;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.repo;
import static org.artifactory.aql.api.domain.sensitive.AqlApiItem.type;
import static org.artifactory.aql.api.internal.AqlBase.and;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v560m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v650m001;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7482;
import static org.artifactory.version.ArtifactoryServiceVersionProvider.v7490m001;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;

class NpmDeprecateConverterTest {

    @Mock
    ArtifactoryContext artifactoryContext;
    @Mock
    RepositoryService repositoryService;
    @Mock
    AddonsManager addonsManager;
    @Mock
    NpmAddon npmAddon;
    @Mock
    AqlService aqlService;
    AutoCloseable closeable;
    NpmDeprecateConverter npmDeprecateConverter;

    @BeforeEach
    void setUp() {
        closeable = MockitoAnnotations.openMocks(this);
        when(artifactoryContext.beanForType(RepositoryService.class)).thenReturn(repositoryService);
        when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(addonsManager.addonByType(NpmAddon.class)).thenReturn(npmAddon);
        SecurityService securityService = mock(SecurityService.class);
        when(artifactoryContext.beanForType(SecurityService.class)).thenReturn(securityService);
        when(artifactoryContext.beanForType(AqlService.class)).thenReturn(aqlService);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        npmDeprecateConverter = new NpmDeprecateConverter(null);
    }

    @AfterEach
    void tearDown() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        closeable.close();
    }

    @ParameterizedTest
    @MethodSource("isInterestedProvider")
    void isInterested(CompoundVersionDetails source, CompoundVersionDetails target, boolean expectedResult) {
        assertEquals(expectedResult, npmDeprecateConverter.isInterested(source, target));
    }

    @Test
    void convert() {
        String repoKey = "repoKey";
        String path = "package/-";
        String name = "package-100.100.100.tgz";
        LocalRepoDescriptor localRepoDescriptor = mock(LocalRepoDescriptor.class);
        LocalRepoDescriptor localRepoDescriptor2 = mock(LocalRepoDescriptor.class);
        when(localRepoDescriptor.getType()).thenReturn(RepoType.Npm);
        when(localRepoDescriptor.getKey()).thenReturn(repoKey);
        when(localRepoDescriptor2.getType()).thenReturn(RepoType.Debian);
        when(repositoryService.getLocalAndFederatedRepoDescriptorsExcludingInfoRepos()).thenReturn(
                List.of(localRepoDescriptor2, localRepoDescriptor));
        doAnswer(invocationOnMock -> {
            AqlApiItem query = invocationOnMock.getArgument(0);
            AqlApiItem expectedQuery = AqlApiItem.create()
                    .filter(
                            and(
                                    repo().equal(repoKey),
                                    type().equal(AqlItemTypeEnum.file.signature),
                                    property().key().equal("npm.deprecated")
                            ))
                    .include(name(), repo(), path());
            assertEquals(expectedQuery.toNative(0), query.toNative(0));
            AqlEagerResult aqlEagerResult = mock(AqlEagerResult.class);
            AqlItem aqlBaseItem = mock(AqlItem.class);
            when(aqlBaseItem.getName()).thenReturn(name);
            when(aqlBaseItem.getRepo()).thenReturn(repoKey);
            when(aqlBaseItem.getPath()).thenReturn(path);
            when(aqlEagerResult.getResults()).thenReturn(List.of(aqlBaseItem));
            return aqlEagerResult;
        }).when(aqlService).executeQueryEager(any(AqlBase.class));
        RepoPath repoPath = RepoPathFactory.create(repoKey, path + "/" + name);
        FileInfo fileInfo =mock(FileInfo.class);
        when(repositoryService.getFileInfo(repoPath)).thenReturn(fileInfo);
        npmDeprecateConverter.convertAsync();

        verify(repositoryService, times(1)).getFileInfo(repoPath);
        verify(npmAddon,times(1)).handleDeprecate(fileInfo);
        verifyNoMoreInteractions(npmAddon);
    }

    public static Stream<Arguments> isInterestedProvider() {
        CompoundVersionDetails version1 = new CompoundVersionDetails(v560m001.get(), null, 0L);
        CompoundVersionDetails version2 = new CompoundVersionDetails(v650m001.get(), null, 0L);
        CompoundVersionDetails version3 = new CompoundVersionDetails(v7482.get(), null, 0L);
        CompoundVersionDetails version4 = new CompoundVersionDetails(v7490m001.get(), null, 0L);
        return Stream.of(
                Arguments.of(version1, version4, true),
                Arguments.of(version2, version4, true),
                Arguments.of(version3, version4, true),
                Arguments.of(version4, version4, false)
        );
    }

}