/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.descriptor.repo;

import com.google.common.annotations.VisibleForTesting;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang.StringUtils;
import org.artifactory.descriptor.Descriptor;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.property.PropertySet;
import org.artifactory.descriptor.repo.vcs.VcsGitProvider;
import org.artifactory.descriptor.repo.vcs.VcsType;
import org.artifactory.model.BaseConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.RemoteRepositoryRepoTypeConfig;
import org.artifactory.model.SecurityConfig;
import org.artifactory.repo.remote.ContentSynchronisation;
import org.artifactory.repo.remote.PropertiesContent;
import org.artifactory.repo.remote.SourceContent;
import org.artifactory.repo.remote.StatisticsContent;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.config.diff.DiffReference;
import org.jfrog.common.config.diff.GenerateDiffFunction;
import org.jfrog.common.config.validation.NotUnicode;

import javax.validation.Valid;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlIDREF;
import javax.xml.bind.annotation.XmlType;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import static org.artifactory.util.VcsPackageTypeConfigMapper.VCS_GIT_DOWNLOAD_URL;
import static org.artifactory.util.VcsPackageTypeConfigMapper.VCS_GIT_PROVIDER;
import static org.artifactory.util.VcsPackageTypeConfigMapper.VCS_TYPE;

/**
 * @author Fred Simon
 */
@XmlType(name = "RemoteRepoType",
        propOrder = {"username", "password", "allowAnyHostAuth", "socketTimeoutMillis", "enableCookieManagement",
                "enableTokenAuthentication", "localAddress", "proxy", "disableProxy", "queryParams",
                "propagateQueryParams", "clientTlsCertificate", "disableUrlNormalization"},
        namespace = Descriptor.NS)
@GenerateDiffFunction
@Slf4j
public class HttpRepoDescriptor extends RemoteRepoDescriptor implements ProxyRepoDescriptor {

    @NotUnicode
    private String username;

    @NotUnicode
    private String password;

    private boolean allowAnyHostAuth;

    @XmlElement(defaultValue = "15000", required = false)
    private int socketTimeoutMillis = 15000;//Default socket timeout

    private boolean enableCookieManagement;

    private boolean enableTokenAuthentication;

    private String localAddress;

    @XmlIDREF
    @XmlElement(name = "proxyRef")
    @DiffReference
    @Valid
    private ProxyDescriptor proxy;

    private boolean disableProxy;

    @NotUnicode
    private String queryParams;

    private boolean propagateQueryParams;

    @XmlElement(name = "clientTlsCertificate", required = false)
    @NotUnicode
    private String clientTlsCertificate;

    @XmlElement(defaultValue = "false")
    private boolean disableUrlNormalization;

    public HttpRepoDescriptor() {
    }

    public HttpRepoDescriptor(RemoteRepoConfig repositoryConfig, List<RepoLayout> repoLayouts,
            List<ProxyDescriptor> proxies, List<PropertySet> propertySets, List<KeyPairDescriptor> keyPairs) {
        setKey(repositoryConfig.getKey());
        setType(repositoryConfig.getPackageType());
        setBaseConfig(repositoryConfig.getBaseConfig(), repoLayouts);
        setRepoTypeConfig(repositoryConfig.getRepoTypeConfig(), repoLayouts, proxies, propertySets);
        setSecurityConfig(repositoryConfig.getSecurityConfig(), keyPairs);
        setPackageTypeConfig(repositoryConfig.getPackageTypeConfig(), repositoryConfig.getPackageType());
    }

    @Override
    public RemoteRepoConfig toRepositoryConfig() {
        RemoteRepoConfig remoteRepoConfig = RemoteRepoConfig.create(getKey(), getType());
        buildBaseConfig(remoteRepoConfig.getBaseConfig());
        buildRepoTypeConfig(remoteRepoConfig.getRepoTypeConfig());
        buildSecurityConfig(remoteRepoConfig.getSecurityConfig());
        buildPackageTypeConfig(getType(), remoteRepoConfig.getPackageTypeConfig());
        return remoteRepoConfig;
    }

    public static HttpRepoDescriptor fromRepositoryConfig(CentralConfigDescriptor centralConfigDescriptor,
            RemoteRepoConfig sourceRepoConfig) {
        return new HttpRepoDescriptor(sourceRepoConfig, centralConfigDescriptor.getRepoLayouts(),
                centralConfigDescriptor.getProxies(), centralConfigDescriptor.getPropertySets(),
                centralConfigDescriptor.getKeyPairs());
    }

    public String getUsername() {
        return username;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }


    public boolean isAllowAnyHostAuth() {
        return allowAnyHostAuth;
    }

    public void setAllowAnyHostAuth(boolean allowAnyHostAuth) {
        this.allowAnyHostAuth = allowAnyHostAuth;
    }

    public int getSocketTimeoutMillis() {
        return socketTimeoutMillis;
    }

    public void setSocketTimeoutMillis(int socketTimeoutMillis) {
        this.socketTimeoutMillis = socketTimeoutMillis;
    }

    public String getLocalAddress() {
        return localAddress;
    }

    public void setLocalAddress(String localAddress) {
        this.localAddress = localAddress;
    }

    public ProxyDescriptor getProxy() {
        return proxy;
    }

    @Override
    public void setProxy(ProxyDescriptor proxy) {
        this.proxy = proxy;
    }

    public boolean isDisableProxy() {
        return disableProxy;
    }

    @Override
    public void setDisableProxy(boolean disableProxy) {
        this.disableProxy = disableProxy;
    }

    public String getQueryParams() {
        return StringUtils.removeStart(queryParams, "?");
    }

    public void setQueryParams(String queryParams) {
        this.queryParams = queryParams;
    }

    public boolean isEnableCookieManagement() {
        return enableCookieManagement;
    }

    public void setEnableCookieManagement(boolean enableCookieManagement) {
        this.enableCookieManagement = enableCookieManagement;
    }

    public boolean isEnableTokenAuthentication() {
        return enableTokenAuthentication;
    }

    public void setEnableTokenAuthentication(boolean enableTokenAuthentication) {
        this.enableTokenAuthentication = enableTokenAuthentication;
    }

    public boolean isPropagateQueryParams() {
        return propagateQueryParams;
    }

    public void setPropagateQueryParams(boolean propagateQueryParams) {
        this.propagateQueryParams = propagateQueryParams;
    }

    public String getClientTlsCertificate() {
        return clientTlsCertificate;
    }

    public void setClientTlsCertificate(String clientTlsCertificate) {
        this.clientTlsCertificate = clientTlsCertificate;
    }

    public boolean isDisableUrlNormalization() {
        return disableUrlNormalization;
    }

    public void setDisableUrlNormalization(boolean disableUrlNormalization) {
        this.disableUrlNormalization = disableUrlNormalization;
    }

    @Override
    public void processUrl() {
        final String[] credentials = extractCredentialsFromUrl(getUrl());
        if (credentials != null) {
            setUsername(credentials[0]);
            setPassword(credentials[1]);
            setUrl(getUrl().replace(getUsername() + ":" + getPassword() + "@", ""));
        }
    }

    String[] extractCredentialsFromUrl(String url) {
        return Optional.ofNullable(url)
                .map(l -> {
                    try {
                        return new URL(l);
                    } catch (MalformedURLException e) {
                        log.warn("Can not parse URL: {}", url);
                        return null;
                    }
                })
                .map(URL::getUserInfo)
                .map(l -> l.split(":"))
                .filter(l -> l.length > 1)
                .orElse(null);
    }

    private void setBaseConfig(BaseConfig baseConfig, List<RepoLayout> repoLayouts) {
        setDescription(baseConfig.getDescription());
        setNotes(baseConfig.getNotes());
        for (RepoLayout repoLayout : repoLayouts) {
            if (repoLayout.getName().equals(baseConfig.getRepoLayoutRef())) {
                setRepoLayout(repoLayout);
            }
        }
        setIncludesPattern(baseConfig.getIncludesPattern());
        setExcludesPattern(baseConfig.getExcludesPattern());
    }

    private void buildBaseConfig(BaseConfig baseConfig) {
        baseConfig.setDescription(getDescription());
        baseConfig.setNotes(getNotes());
        if (getRepoLayout() != null) {
            baseConfig.setRepoLayoutRef(getRepoLayout().getName());
        }
        baseConfig.setIncludesPattern(getIncludesPattern());
        baseConfig.setExcludesPattern(getExcludesPattern());
    }

    private void setRepoTypeConfig(RemoteRepositoryRepoTypeConfig remoteConfig,
            List<RepoLayout> repoLayouts, List<ProxyDescriptor> proxies,
            List<PropertySet> propertySets) {
        setHandleReleases(remoteConfig.isHandleReleases());
        setHandleSnapshots(remoteConfig.isHandleSnapshots());
        setMaxUniqueSnapshots(remoteConfig.getMaxUniqueSnapshots());
        org.artifactory.repo.CdnRedirectRepoConfig cdnRedirectRepoConfig = remoteConfig.getCdnRedirectRepoConfig();
        if (cdnRedirectRepoConfig != null) {
            org.artifactory.descriptor.repo.CdnRedirectRepoConfig cdnRedirectRepoConfigDescriptor =
                    new org.artifactory.descriptor.repo.CdnRedirectRepoConfig();
            cdnRedirectRepoConfigDescriptor.setEnabled(cdnRedirectRepoConfig.isEnabled());
            setCdnRedirectConfig(cdnRedirectRepoConfigDescriptor);
        }
        org.artifactory.repo.DownloadRedirectRepoConfig downloadRedirectConfig =
                remoteConfig.getDownloadRedirectConfig();
        if (downloadRedirectConfig != null) {
            DownloadRedirectRepoConfig downloadRedirectDescriptor = new DownloadRedirectRepoConfig();
            downloadRedirectDescriptor.setEnabled(downloadRedirectConfig.isEnabled());
            setDownloadRedirectConfig(downloadRedirectDescriptor);
        }
        setUsername(remoteConfig.getUsername());
        setPassword(remoteConfig.getPassword());
        setLocalAddress(remoteConfig.getLocalAddress());
        for (RepoLayout repoLayout : repoLayouts) {
            if (repoLayout.getName().equals(remoteConfig.getRemoteRepoLayoutRef())) {
                setRemoteRepoLayout(repoLayout);
            }
        }
        for (ProxyDescriptor proxyDescriptor : proxies) {
            if (proxyDescriptor.getKey().equals(remoteConfig.getProxyRef())) {
                setProxy(proxyDescriptor);
            }
        }
        setDisableProxy(remoteConfig.isDisableProxy());
        ArrayList<PropertySet> newPropertySets = new ArrayList<>();
        for (String propertySetRef : remoteConfig.getPropertySetRefs()) {
            for (PropertySet propertySet : propertySets) {
                if (propertySet.getName().equals(propertySetRef)) {
                    newPropertySets.add(propertySet);
                }
            }
        }
        setPropertySets(newPropertySets);
        setChecksumPolicyType(ChecksumPolicyType.valueOf(remoteConfig.getChecksumPolicyType().name()));
        setAssumedOfflinePeriodSecs(remoteConfig.getAssumedOfflinePeriodSecs());
        setRetrievalCachePeriodSecs(remoteConfig.getRetrievalCachePeriodSecs());
        setMissedRetrievalCachePeriodSecs(remoteConfig.getMissedRetrievalCachePeriodSecs());
        setMetadataRetrievalTimeoutSecs(remoteConfig.getMetadataRetrievalTimeoutSecs());
        setSocketTimeoutMillis(remoteConfig.getSocketTimeoutMillis());
        setUnusedArtifactsCleanupPeriodHours(remoteConfig.getUnusedArtifactsCleanupPeriodHours());
        setClientTlsCertificate(remoteConfig.getClientTlsCertificate());
        ContentSynchronisation contentSynchronisation = remoteConfig.getContentSynchronisation();
        if (contentSynchronisation != null) {
            org.artifactory.descriptor.delegation.ContentSynchronisation contentSynchronisationDescriptor =
                    new org.artifactory.descriptor.delegation.ContentSynchronisation();
            contentSynchronisationDescriptor.setEnabled(contentSynchronisation.isEnabled());
            contentSynchronisationDescriptor.getStatistics()
                    .setEnabled(remoteConfig.getContentSynchronisation().getStatistics().isEnabled());
            contentSynchronisationDescriptor.getProperties()
                    .setEnabled(remoteConfig.getContentSynchronisation().getProperties().isEnabled());
            contentSynchronisationDescriptor.getSource()
                    .setOriginAbsenceDetection(
                            remoteConfig.getContentSynchronisation().getSource().isOriginAbsenceDetection());
            setContentSynchronisation(contentSynchronisationDescriptor);
        }
        setQueryParams(remoteConfig.getQueryParams());
        setUrl(remoteConfig.getUrl());
        setLocalAddress(remoteConfig.getLocalAddress());
        setMismatchingMimeTypesOverrideList(remoteConfig.getMismatchingMimeTypesOverrideList());
        setSynchronizeProperties(remoteConfig.isSynchronizeProperties());
        setShareConfiguration(remoteConfig.isShareConfiguration());
        setStoreArtifactsLocally(remoteConfig.isStoreArtifactsLocally());
        setHardFail(remoteConfig.isHardFail());
        setEnableCookieManagement(remoteConfig.isEnableCookieManagement());
        setBlackedOut(remoteConfig.isBlackedOut());
        setAllowAnyHostAuth(remoteConfig.isAllowAnyHostAuth());
        setArchiveBrowsingEnabled(remoteConfig.isArchiveBrowsingEnabled());
        setBlockMismatchingMimeTypes(remoteConfig.isBlockMismatchingMimeTypes());
        setBypassHeadRequests(remoteConfig.isBypassHeadRequests());
        setDisableUrlNormalization(remoteConfig.isDisableUrlNormalization());
        setListRemoteFolderItems(remoteConfig.isListRemoteFolderItems());
        setOffline(remoteConfig.isOffline());
        setPropagateQueryParams(remoteConfig.isPropagateQueryParams());
        setEnableTokenAuthentication(remoteConfig.isEnableTokenAuthentication());
        setPriorityResolution(remoteConfig.isPriorityResolution());
    }

    private void buildRepoTypeConfig(@NotNull RemoteRepositoryRepoTypeConfig remoteConfig) {
        remoteConfig.setHandleReleases(isHandleReleases());
        remoteConfig.setHandleSnapshots(isHandleSnapshots());
        remoteConfig.setMaxUniqueSnapshots(getMaxUniqueSnapshots());
        if (getCdnRedirectRepoConfig() != null) {
            org.artifactory.repo.CdnRedirectRepoConfig cdnRedirectRepoConfig =
                    new org.artifactory.repo.CdnRedirectRepoConfig();
            cdnRedirectRepoConfig.setEnabled(getCdnRedirectRepoConfig().isEnabled());
            remoteConfig.setCdnRedirectRepoConfig(cdnRedirectRepoConfig);
        }
        if (getDownloadRedirectConfig() != null) {
            org.artifactory.repo.DownloadRedirectRepoConfig downloadRedirectConfig =
                    new org.artifactory.repo.DownloadRedirectRepoConfig();
            downloadRedirectConfig.setEnabled(getDownloadRedirectConfig().isEnabled());
            remoteConfig.setDownloadRedirectConfig(downloadRedirectConfig);
        }
        remoteConfig.setUsername(getUsername());
        remoteConfig.setPassword(getPassword());
        remoteConfig.setLocalAddress(getLocalAddress());
        if (getRemoteRepoLayout() != null) {
            remoteConfig.setRemoteRepoLayoutRef(getRemoteRepoLayout().getName());
        }
        if (getProxy() != null) {
            remoteConfig.setProxyRef(getProxy().getKey());
        }
        remoteConfig.setDisableProxy(isDisableProxy());
        List<String> propertySetRefs = new ArrayList<>();
        getPropertySets().forEach(p -> propertySetRefs.add(p.getName()));
        remoteConfig.setPropertySetRefs(propertySetRefs);
        remoteConfig
                .setChecksumPolicyType(org.artifactory.repo.ChecksumPolicyType.valueOf(getChecksumPolicyType().name()));
        remoteConfig.setAssumedOfflinePeriodSecs(getAssumedOfflinePeriodSecs());
        remoteConfig.setRetrievalCachePeriodSecs(getRetrievalCachePeriodSecs());
        remoteConfig.setMissedRetrievalCachePeriodSecs(getMissedRetrievalCachePeriodSecs());
        remoteConfig.setMetadataRetrievalTimeoutSecs(getMetadataRetrievalTimeoutSecs());
        remoteConfig.setSocketTimeoutMillis(getSocketTimeoutMillis());
        remoteConfig.setUnusedArtifactsCleanupPeriodHours(getUnusedArtifactsCleanupPeriodHours());
        remoteConfig.setClientTlsCertificate(getClientTlsCertificate());

        if (getContentSynchronisation() != null) {
            ContentSynchronisation contentSynchronisation = new ContentSynchronisation();
            contentSynchronisation.setEnabled(getContentSynchronisation().isEnabled());
            if (getContentSynchronisation().getProperties() != null) {
                PropertiesContent properties = new PropertiesContent();
                properties.setEnabled(getContentSynchronisation().getProperties().isEnabled());
                contentSynchronisation.setProperties(properties);
            }
            if (getContentSynchronisation().getSource() != null) {
                SourceContent source = new SourceContent();
                source.setOriginAbsenceDetection(getContentSynchronisation().getSource().isOriginAbsenceDetection());
                contentSynchronisation.setSource(source);
            }
            if (getContentSynchronisation().getStatistics() != null) {
                StatisticsContent statistics = new StatisticsContent();
                statistics.setEnabled(getContentSynchronisation().getStatistics().isEnabled());
                contentSynchronisation.setStatistics(statistics);
            }
            remoteConfig.setContentSynchronisation(contentSynchronisation);
        }
        remoteConfig.setQueryParams(getQueryParams());
        remoteConfig.setUrl(getUrl());
        remoteConfig.setLocalAddress(getLocalAddress());
        remoteConfig.setMismatchingMimeTypesOverrideList(getMismatchingMimeTypesOverrideList());
        remoteConfig.setSynchronizeProperties(isSynchronizeProperties());
        remoteConfig.setShareConfiguration(isShareConfiguration());
        remoteConfig.setStoreArtifactsLocally(isStoreArtifactsLocally());
        remoteConfig.setHardFail(isHardFail());
        remoteConfig.setEnableCookieManagement(isEnableCookieManagement());
        remoteConfig.setBlackedOut(isBlackedOut());
        remoteConfig.setAllowAnyHostAuth(isAllowAnyHostAuth());
        remoteConfig.setArchiveBrowsingEnabled(isArchiveBrowsingEnabled());
        remoteConfig.setBlockMismatchingMimeTypes(isBlockMismatchingMimeTypes());
        remoteConfig.setBypassHeadRequests(isBypassHeadRequests());
        remoteConfig.setDisableUrlNormalization(isDisableUrlNormalization());
        remoteConfig.setListRemoteFolderItems(isListRemoteFolderItems());
        remoteConfig.setOffline(isOffline());
        remoteConfig.setPropagateQueryParams(isPropagateQueryParams());
        remoteConfig.setEnableTokenAuthentication(isEnableTokenAuthentication());
        remoteConfig.setPriorityResolution(isPriorityResolution());
    }

    @VisibleForTesting
    void setPackageTypeConfig(Map<String, String> packageTypeConfig,
            RepoType packageType) {
        if (packageType.isMavenGroup()) {
            setSuppressPomConsistencyChecks(
                    Boolean.parseBoolean(packageTypeConfig.get("suppressPomConsistencyChecks")));
            setFetchJarsEagerly(
                    Boolean.parseBoolean(packageTypeConfig.get("fetchJarsEagerly")));
            setFetchSourcesEagerly(
                    Boolean.parseBoolean(packageTypeConfig.get("fetchSourcesEagerly")));
            setRejectInvalidJars(
                    Boolean.parseBoolean(packageTypeConfig.get("rejectInvalidJars")));
            if (RepoType.P2.equals(packageType)) {
                setP2OriginalUrl(packageTypeConfig.get("p2OriginalUrl"));
            }
        } else if (RepoType.Debian.equals(packageType)) {
            setDdebSupported(Boolean.parseBoolean(packageTypeConfig.get("ddebSupported")));
        } else if (RepoType.Cargo.equals(packageType)) {
            CargoConfiguration cargoConfiguration = new CargoConfiguration();
            cargoConfiguration
                    .setCargoAnonymousAccess(Boolean.parseBoolean(packageTypeConfig.get("cargoAnonymousAccess")));
            cargoConfiguration
                    .setCargoInternalIndex(Boolean.parseBoolean(packageTypeConfig.get("cargoInternalIndex")));
            setCargoConfig(cargoConfiguration);
            if (packageTypeConfig.get("gitRegistryUrl") != null) {
                GitConfiguration git = new GitConfiguration();
                git.setRegistryUrl(packageTypeConfig.get("gitRegistryUrl"));
                setGit(git);
            }
        } else if (RepoType.Bower.equals(packageType)) {
            BowerConfiguration bowerConfiguration = new BowerConfiguration();
            bowerConfiguration.setBowerRegistryUrl(packageTypeConfig.get("bowerRegistryUrl"));
            if (packageTypeConfig.containsKey(VCS_TYPE)) {
                VcsConfiguration vcs = new VcsConfiguration();
                vcs.setGit(vcsGitConfigurationToDescriptor(packageTypeConfig));
                vcs.setType(VcsType.valueOf(packageTypeConfig.get(VCS_TYPE)));
                setVcs(vcs);
            }
            setBower(bowerConfiguration);
        } else if (RepoType.CocoaPods.equals(packageType)) {
            CocoaPodsConfiguration cocoaPodsConfiguration = new CocoaPodsConfiguration();
            if (packageTypeConfig.get("specsRepoDownloadProvider") != null) {
                VcsGitConfiguration specRepoProvider = new VcsGitConfiguration();
                if (packageTypeConfig.get("specsRepoDownloadUrl") != null) {
                    specRepoProvider.setDownloadUrl(packageTypeConfig.get("specsRepoDownloadUrl"));
                }
                specRepoProvider.setProvider(
                        VcsGitProvider.valueOf(packageTypeConfig.get("specsRepoDownloadProvider")));
                cocoaPodsConfiguration.setSpecRepoProvider(specRepoProvider);
            }
            cocoaPodsConfiguration.setCocoaPodsSpecsRepoUrl(packageTypeConfig.get("cocoaPodsSpecsRepoUrl"));
            setCocoaPods(cocoaPodsConfiguration);
            if (packageTypeConfig.containsKey(VCS_TYPE)) {
                VcsConfiguration vcs = new VcsConfiguration();
                vcs.setGit(vcsGitConfigurationToDescriptor(packageTypeConfig));
                vcs.setType(VcsType.valueOf(packageTypeConfig.get(VCS_TYPE)));
                setVcs(vcs);
            }
            setExternalDependencies(externalDependenciesToDescriptor(packageTypeConfig));
        } else if (RepoType.Composer.equals(packageType)) {
            ComposerConfiguration composerConfiguration = new ComposerConfiguration();
            composerConfiguration.setComposerRegistryUrl(packageTypeConfig.get("composerRegistryUrl"));
            if (packageTypeConfig.containsKey(VCS_TYPE)) {
                VcsConfiguration vcs = new VcsConfiguration();
                vcs.setGit(vcsGitConfigurationToDescriptor(packageTypeConfig));
                vcs.setType(VcsType.valueOf(packageTypeConfig.get(VCS_TYPE)));
                setVcs(vcs);
            }
            setComposer(composerConfiguration);
        } else if (RepoType.Docker.equals(packageType)) {
            Optional.ofNullable(packageTypeConfig.get("dockerApiVersion"))
                    .ifPresent(this::setDockerApiVersion);
            setBlockPushingSchema1(Boolean.parseBoolean(packageTypeConfig.get("blockPushingSchema1")));
            Optional.ofNullable(packageTypeConfig.get("maxUniqueTags"))
                    .map(Integer::parseInt)
                    .ifPresent(this::setMaxUniqueTags);
            setExternalDependencies(externalDependenciesToDescriptor(packageTypeConfig));
        } else if (RepoType.Helm.equals(packageType)) {
            HelmConfiguration configuration = new HelmConfiguration();
            configuration
                    .setChartsBaseUrl(packageTypeConfig.get("chartsBaseUrl"));
            setHelm(configuration);
            setExternalDependencies(externalDependenciesToDescriptor(packageTypeConfig));
        } else if (RepoType.Swift.equals(packageType)) {
            setExternalDependencies(externalDependenciesToDescriptor(packageTypeConfig));
        } else if (RepoType.NuGet.equals(packageType)) {
            NuGetConfiguration configuration = new NuGetConfiguration();
            Optional.ofNullable(packageTypeConfig.get("v3FeedUrl"))
                    .ifPresent(configuration::setV3FeedUrl);
            Optional.ofNullable(packageTypeConfig.get("feedContextPath"))
                    .ifPresent(configuration::setFeedContextPath);
            Optional.ofNullable(packageTypeConfig.get("downloadContextPath"))
                    .ifPresent(configuration::setDownloadContextPath);
            Optional.ofNullable(packageTypeConfig.get("symbolServerUrl"))
                    .ifPresent(configuration::setSymbolServerUrl);
            setForceNugetAuthentication(
                    Boolean.parseBoolean(packageTypeConfig.get("forceNugetAuthentication")));
            setNuget(configuration);
        } else if (RepoType.Pypi.equals(packageType)) {
            PypiConfiguration pypiConfiguration = new PypiConfiguration();
            // TODO [u.l] what an ugly field name
            Optional.ofNullable(packageTypeConfig.get("pyPIRegistryUrl"))
                    .ifPresent(pypiConfiguration::setPyPIRegistryUrl);
            Optional.ofNullable(packageTypeConfig.get("packagesContextPath"))
                    .ifPresent(pypiConfiguration::setPackagesContextPath);
            Optional.ofNullable(packageTypeConfig.get("repositorySuffix"))
                    .ifPresent(pypiConfiguration::setRepositorySuffix);
            Optional.ofNullable(packageTypeConfig.get("indexContextPath"))
                    .ifPresent(pypiConfiguration::setIndexContextPath);
            setPypi(pypiConfiguration);
        } else if (RepoType.VCS.equals(packageType)) {
            VcsConfiguration vcsConfiguration = new VcsConfiguration();
            vcsConfiguration.setGit(vcsGitConfigurationToDescriptor(packageTypeConfig));
            Optional.ofNullable(packageTypeConfig.get(VCS_TYPE))
                    .ifPresent(type -> vcsConfiguration.setType(VcsType.valueOf(type)));
            setVcs(vcsConfiguration);
        } else if (RepoType.Terraform.equals(packageType)) {
            TerraformConfiguration terraformConfiguration = new TerraformConfiguration();
            Optional.ofNullable(packageTypeConfig.get("terraformProvidersUrl"))
                    .ifPresent(terraformConfiguration::setTerraformProvidersUrl);
            Optional.ofNullable(packageTypeConfig.get("terraformRegistryUrl"))
                    .ifPresent(terraformConfiguration::setTerraformRegistryUrl);
            setTerraform(terraformConfiguration);
            if (packageTypeConfig.containsKey(VCS_TYPE)) {
                VcsConfiguration vcs = new VcsConfiguration();
                vcs.setGit(vcsGitConfigurationToDescriptor(packageTypeConfig));
                vcs.setType(VcsType.valueOf(packageTypeConfig.get(VCS_TYPE)));
                setVcs(vcs);
            }
        } else if (RepoType.Go.equals(packageType)) {
            if (packageTypeConfig.containsKey(VCS_TYPE)) {
                VcsConfiguration vcs = new VcsConfiguration();
                vcs.setGit(vcsGitConfigurationToDescriptor(packageTypeConfig));
                vcs.setType(VcsType.valueOf(packageTypeConfig.get(VCS_TYPE)));
                setVcs(vcs);
            }
            setExternalDependencies(externalDependenciesToDescriptor(packageTypeConfig));
        } else if (RepoType.Conan.equals(packageType)) {
            setForceConanAuthentication(Boolean.parseBoolean(packageTypeConfig.get("forceConanAuthentication")));
        }
    }

    void buildPackageTypeConfig(RepoType packageType,
            Map<String, String> packageTypeConfig) {
        if (packageType.isMavenGroup()) {
            packageTypeConfig.put("suppressPomConsistencyChecks", String.valueOf(isSuppressPomConsistencyChecks()));
            packageTypeConfig.put("fetchJarsEagerly", String.valueOf(isFetchJarsEagerly()));
            packageTypeConfig.put("fetchSourcesEagerly", String.valueOf(isFetchSourcesEagerly()));
            packageTypeConfig.put("rejectInvalidJars", String.valueOf(isRejectInvalidJars()));
            if (RepoType.P2.equals(packageType) && getP2OriginalUrl() != null) {
                packageTypeConfig.put("p2OriginalUrl", String.valueOf(getP2OriginalUrl()));
            }
        } else if (RepoType.Debian.equals(packageType)) {
            packageTypeConfig.put("ddebSupported", String.valueOf(isDdebSupported()));
        } else if (RepoType.Cargo.equals(packageType)) {
            packageTypeConfig.put("cargoAnonymousAccess", String.valueOf(isCargoAnonymousAccess()));
            packageTypeConfig.put("cargoInternalIndex", String.valueOf(isCargoInternalIndex()));
            if (getGit() != null && getGit().getRegistryUrl() != null) {
                packageTypeConfig.put("gitRegistryUrl", getGit().getRegistryUrl());
            }
        } else if (RepoType.Bower.equals(packageType)) {
            if (getBower() != null) {
                packageTypeConfig.put("bowerRegistryUrl", getBower().getBowerRegistryUrl());
            }
            if (getVcs() != null) {
                packageTypeConfig.put(VCS_TYPE, getVcs().getType().name());
                packageTypeConfig.put(VCS_GIT_PROVIDER, getVcs().getGit().getProvider().name());
                if (getVcs().getGit().getDownloadUrl() != null) {
                    packageTypeConfig.put(VCS_GIT_DOWNLOAD_URL, getVcs().getGit().getDownloadUrl());
                }
            }
        } else if (RepoType.CocoaPods.equals(packageType)) {
            CocoaPodsConfiguration cocoaPodsConfiguration = getCocoaPods();
            if (cocoaPodsConfiguration != null) {
                packageTypeConfig.put("cocoaPodsSpecsRepoUrl", cocoaPodsConfiguration.getCocoaPodsSpecsRepoUrl());
                VcsGitConfiguration specRepoProvider = cocoaPodsConfiguration.getSpecRepoProvider();
                if (specRepoProvider != null) {
                    if (specRepoProvider.getDownloadUrl() != null) {
                        packageTypeConfig.put("specsRepoDownloadUrl", specRepoProvider.getDownloadUrl());
                    }
                    packageTypeConfig.put("specsRepoDownloadProvider", specRepoProvider.getProvider().name());
                }
            }
            if (getVcs() != null) {
                packageTypeConfig.put(VCS_TYPE, getVcs().getType().name());
                packageTypeConfig.put(VCS_GIT_PROVIDER, getVcs().getGit().getProvider().name());
                if (getVcs().getGit().getDownloadUrl() != null) {
                    packageTypeConfig.put(VCS_GIT_DOWNLOAD_URL, getVcs().getGit().getDownloadUrl());
                }
            }
            externalDependenciesToConfig(packageTypeConfig, getExternalDependencies());
        } else if (RepoType.Composer.equals(packageType)) {
            if (getComposer() != null) {
                packageTypeConfig.put("composerRegistryUrl", getComposer().getComposerRegistryUrl());
            }
            if (getVcs() != null) {
                packageTypeConfig.put(VCS_TYPE, getVcs().getType().name());
                packageTypeConfig.put(VCS_GIT_PROVIDER, getVcs().getGit().getProvider().name());
                if (getVcs().getGit().getDownloadUrl() != null) {
                    packageTypeConfig.put(VCS_GIT_DOWNLOAD_URL, getVcs().getGit().getDownloadUrl());
                }
            }
        } else if (RepoType.Docker.equals(packageType)) {
            packageTypeConfig.put("dockerApiVersion", getDockerApiVersion().name());
            packageTypeConfig.put("blockPushingSchema1", String.valueOf(isBlockPushingSchema1()));
            packageTypeConfig.put("maxUniqueTags", String.valueOf(getMaxUniqueTags()));
            externalDependenciesToConfig(packageTypeConfig, getExternalDependencies());
        } else if (RepoType.Swift.equals(packageType)) {
            externalDependenciesToConfig(packageTypeConfig, getExternalDependencies());
        } else if (RepoType.Helm.equals(packageType)) {
            HelmConfiguration helmConfig = getHelm();
            if (helmConfig != null && helmConfig.getChartsBaseUrl() != null) {
                packageTypeConfig.put("chartsBaseUrl", String.valueOf(getHelm().getChartsBaseUrl()));
            }
            externalDependenciesToConfig(packageTypeConfig, getExternalDependencies());
        } else if (RepoType.NuGet.equals(packageType)) {
            if (getNuget() != null) {
                Optional.ofNullable(getNuget().getV3FeedUrl())
                        .ifPresent(value -> packageTypeConfig.put("v3FeedUrl", value));
                Optional.ofNullable(getNuget().getFeedContextPath())
                        .ifPresent(value -> packageTypeConfig.put("feedContextPath", value));
                Optional.ofNullable(getNuget().getDownloadContextPath())
                        .ifPresent(value -> packageTypeConfig.put("downloadContextPath", value));
                Optional.ofNullable(getNuget().getSymbolServerUrl())
                        .ifPresent(value -> packageTypeConfig.put("symbolServerUrl", value));
            }
            packageTypeConfig.put("forceNugetAuthentication", String.valueOf(isForceNugetAuthentication()));
        } else if (RepoType.Pypi.equals(packageType)) {
            // TODO [u.l] what an ugly field name
            if (getPypi() != null) {
                Optional.ofNullable(getPypi().getPyPIRegistryUrl())
                        .ifPresent(value -> packageTypeConfig.put("pyPIRegistryUrl", value));
                Optional.ofNullable(getPypi().getPackagesContextPath())
                        .ifPresent(value -> packageTypeConfig.put("packagesContextPath", value));
                Optional.ofNullable(getPypi().getRepositorySuffix())
                        .ifPresent(value -> packageTypeConfig.put("repositorySuffix", value));
                Optional.ofNullable(getPypi().getIndexContextPath())
                        .ifPresent(value -> packageTypeConfig.put("indexContextPath", value));
            }
        } else if (RepoType.VCS.equals(packageType)) {
            if (getVcs() != null) {
                if (getVcs().getGit() != null) {
                    packageTypeConfig.put(VCS_GIT_PROVIDER, getVcs().getGit().getProvider().name());
                    Optional.ofNullable(getVcs().getGit().getDownloadUrl())
                            .ifPresent(value -> packageTypeConfig.put(VCS_GIT_DOWNLOAD_URL, value));
                }
                packageTypeConfig.put(VCS_TYPE, getVcs().getType().name());
            }
        } else if (RepoType.Terraform.equals(packageType)) {
            if (getTerraform() != null) {
                Optional.ofNullable(getTerraform().getTerraformProvidersUrl())
                        .ifPresent(value -> packageTypeConfig.put("terraformProvidersUrl", value));
                Optional.ofNullable(getTerraform().getTerraformRegistryUrl())
                        .ifPresent(value -> packageTypeConfig.put("terraformRegistryUrl", value));
                if (getVcs() != null) {
                    packageTypeConfig.put(VCS_TYPE, getVcs().getType().name());
                    packageTypeConfig.put(VCS_GIT_PROVIDER, getVcs().getGit().getProvider().name());
                    Optional.ofNullable(getVcs().getGit().getDownloadUrl())
                            .ifPresent(value -> packageTypeConfig.put(VCS_GIT_DOWNLOAD_URL, value));
                }
            }
        } else if (RepoType.Go.equals(packageType)) {
            if (getVcs() != null) {
                packageTypeConfig.put(VCS_TYPE, getVcs().getType().name());
                packageTypeConfig.put(VCS_GIT_PROVIDER, getVcs().getGit().getProvider().name());
                Optional.ofNullable(getVcs().getGit().getDownloadUrl())
                        .ifPresent(value -> packageTypeConfig.put(VCS_GIT_DOWNLOAD_URL, value));
            }
            externalDependenciesToConfig(packageTypeConfig, getExternalDependencies());
        } else if (RepoType.Conan.equals(packageType)) {
            packageTypeConfig.put("forceConanAuthentication", String.valueOf(isForceConanAuthentication()));
        }
    }

    private void setSecurityConfig(SecurityConfig securityConfig,
            List<KeyPairDescriptor> keyPairs) {
        if (securityConfig.getXrayConfig() != null) {
            XrayRepoConfig xrayRepoConfig = new XrayRepoConfig();
            xrayRepoConfig.setDataTtl(securityConfig.getXrayConfig().getDataTtl());
            xrayRepoConfig.setEnabled(securityConfig.getXrayConfig().isEnabled());
            setXrayConfig(xrayRepoConfig);
        }
        setSignedUrlTtl(securityConfig.getSignedUrlTtl());

        for (KeyPairDescriptor keyPair : keyPairs) {
            if (keyPair.getKey().equals(securityConfig.getPrimaryKeyPairRef())) {
                setPrimaryKeyPairDescriptor(keyPair);
            }
            if (keyPair.getKey().equals(securityConfig.getSecondaryKeyPairRef())) {
                setSecondaryKeyPairDescriptor(keyPair);
            }
        }
    }

    private void buildSecurityConfig(@NotNull SecurityConfig securityConfig) {
        if (getXrayConfig() != null) {
            org.artifactory.model.XrayRepoConfig xrayConfig =
                    new org.artifactory.model.XrayRepoConfig();
            XrayRepoConfig xrayConfigDescriptor = getXrayConfig();
            xrayConfig.setDataTtl(xrayConfigDescriptor.getDataTtl());
            xrayConfig.setEnabled(xrayConfigDescriptor.isEnabled());
            securityConfig.setXrayConfig(xrayConfig);
        }
        if (getPrimaryKeyPairDescriptor() != null) {
            securityConfig.setPrimaryKeyPairRef(getPrimaryKeyPairDescriptor().getKey());
        }
        if (getSecondaryKeyPairDescriptor() != null) {
            securityConfig.setSecondaryKeyPairRef(getSecondaryKeyPairDescriptor().getKey());
        }
        securityConfig.setSignedUrlTtl(getSignedUrlTtl());
    }

    @NotNull
    private VcsGitConfiguration vcsGitConfigurationToDescriptor(Map<String, String> packageTypeConfig) {
        VcsGitConfiguration vcsGitConfiguration = new VcsGitConfiguration();
        Optional.ofNullable(packageTypeConfig.get(VCS_GIT_PROVIDER))
                .ifPresent(provider -> vcsGitConfiguration.setProvider(VcsGitProvider.valueOf(provider)));
        Optional.ofNullable(packageTypeConfig.get(VCS_GIT_DOWNLOAD_URL))
                .ifPresent(vcsGitConfiguration::setDownloadUrl);
        return vcsGitConfiguration;
    }

    private ExternalDependenciesConfig externalDependenciesToDescriptor(Map<String, String> packageTypeConfig) {
        ExternalDependenciesConfig externalDependenciesConfig = null;
        if (packageTypeConfig.containsKey("externalDependenciesEnabled")) {
            externalDependenciesConfig = new ExternalDependenciesConfig();
            externalDependenciesConfig.setEnabled(
                    Boolean.parseBoolean(packageTypeConfig.get("externalDependenciesEnabled")));
            List<String> patterns =
                    RepositoryConfigSource.jsonStringToList(packageTypeConfig, "externalDependenciesPatterns");
            externalDependenciesConfig.setPatterns(patterns);
        }
        return externalDependenciesConfig;
    }

    private void externalDependenciesToConfig(Map<String, String> packageTypeConfig,
            ExternalDependenciesConfig externalDependencies) {
        if (externalDependencies != null) {
            packageTypeConfig.put("externalDependenciesEnabled", String.valueOf(externalDependencies.isEnabled()));
            RepositoryConfigSource.entityToFlatMapKeys(externalDependencies.getPatterns(),
                    "externalDependenciesPatterns", packageTypeConfig);
        } else {
            packageTypeConfig.remove("externalDependenciesEnabled");
            packageTypeConfig.remove("externalDependenciesPatterns");
            packageTypeConfig.remove("externalDependenciesRemoteRepoRef");
        }
    }
}