package org.artifactory.core.api.ph.service;

import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpRequestBase;
import org.artifactory.common.resource.BaseResource;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.PackageArtifact;
import org.artifactory.core.api.ph.model.PackageDownloadContext;

import javax.ws.rs.core.Response;
import java.io.InputStream;

/**
 * This interface provides artifact metadata and or content to the package handler layer
 */
public interface PackageArtifactService {

    /**
     * The method downloads artifact from all repositories (local, remote and virtual), to the
     * package handler layer
     *
     * @param downloadContext information related to the download such as(repoKey, path, alternate-url...)
     * @return jersey response
     * @throws PackageException in case of not found, un authorized or internal error
     */
    Response download(PackageDownloadContext downloadContext) throws PackageException;

    /**
     * The method downloads artifact only from remote-cache repository, if exists.
     * The repoKey should be the remote key without the '-cache' part.
     *
     * @param downloadContext information related to the download such as(repoKey, path, alternate-url...)
     * @return jersey response
     * @throws PackageException in case of not found, un authorized or internal error
     */
    Response downloadFromRemoteCache(PackageDownloadContext downloadContext) throws PackageException;

    /**
     * The method returns artifact metadata from all local and cached (local and remote/virtual cache) repositories,
     * to the package handler layer
     *
     * @param downloadContext information related to the download such as(repoKey, path, alternate-url...)
     * @return returns Package Artifact
     * @throws PackageException in case of not found, un authorized or internal error
     */
    PackageArtifact getArtifact(PackageDownloadContext downloadContext) throws PackageException;

    /**
     * The method returns artifact metadata from all local and cached (local and remote/virtual cache) repositories,
     * to the package handler layer
     *
     * @param repoKey The repository to download from
     * @param path    The path to download from
     * @return returns Package Artifact
     * @throws PackageException in case of not found, un authorized or internal error
     */
    PackageArtifact getArtifact(String repoKey, String path) throws PackageException;

    /**
     * The method returns artifact stream (content) from all local and cached (local and remote/virtual cache)
     * repositories, to the package handler layer
     *
     * @param repoKey The repository to download from
     * @param path    The path to download from
     * @return returns Package Artifact
     * @throws PackageException in case of not found, un authorized or internal error
     */
    InputStream getStream(String repoKey, String path) throws PackageException;

    /**
     * The method returns artifact stream (content) from all local and cached (local and remote/virtual cache)
     * repositories, to the package handler layer
     *
     * @param downloadContext information related to the download such as(repoKey, path,alternate-url...)
     * @return returns Package Artifact
     * @throws PackageException in case of not found, un authorized or internal error
     */
    InputStream getStream(PackageDownloadContext downloadContext) throws PackageException;

    /**
     * Provides artifact resource, the method is being used internally by the package handler (Virtual repo traversal)
     * framework, and it is not exposed outside to the packages implementation
     *
     * @param packageDownloadContext information related to the download such as(repoKey, path,alternate-url...)
     * @return returns Package Artifact
     * @throws PackageException in case of not found, un authorized or internal error
     */
    BaseResource getResource(PackageDownloadContext packageDownloadContext) throws PackageException;

    /**
     * Provides artifact resource in virtual cache, the method is being used internally by the package handler
     * (Virtual repo traversal)
     * framework, and it is not exposed outside to the packages implementation
     *
     * @param repoKey The repository to download from
     * @param path    The path to download from
     * @return returns Package Artifact
     * @throws PackageException in case of not found, un authorized or internal error
     */
    BaseResource getResourceFromVirtualCache(String repoKey, String path) throws PackageException;

    /**
     * Execute http request with the remote repo context.
     * The response will not get cached.
     *
     * @return {@link CloseableHttpResponse} The http response.
     */
    CloseableHttpResponse executeRemoteHttpRequest(String remoteRepoKey, HttpRequestBase httpRequest)
            throws PackageException;

    /**
     * Return Artifact input stream from virtual cache
     *
     * @param repoKey Repository key
     * @param path    Repository path
     * @return InputStream contains the real resource of the resolved file
     * @throws PackageException In case of expired resource or not exists repo/path
     */
    InputStream getStreamFromVirtualCache(String repoKey, String path) throws PackageException;
}
