package org.artifactory.core.api.ph.service;

import lombok.NonNull;
import org.artifactory.core.api.ph.exception.PackageException;
import org.artifactory.core.api.ph.model.SimpleUser;

import java.util.concurrent.Callable;

public interface PackageSecurityService {

    /**
     * Check read permission for artifact
     *
     * @param repoKey Repository key
     * @param path    Path to file
     * @return true if user can read file
     */
    boolean canRead(String repoKey, String path);

    /**
     * Check write permission for artifact
     *
     * @param repoKey Repository key
     * @param path    Path to file
     * @return true if user can write file
     */
    boolean canWrite(String repoKey, String path);

    /**
     * Check annotate permission for artifact
     *
     * @param repoKey Repository key
     * @param path    Path to file
     * @return true if user can annotate file
     */
    boolean canAnnotate(String repoKey, String path);

    /**
     * Check delete permission for artifact
     *
     * @param repoKey Repository key
     * @param path    Path to file
     * @return true if user can read file
     */
    boolean canDelete(String repoKey, String path);

    /**
     * Check manage permission for artifact
     *
     * @param repoKey Repository key
     * @param path    Path to file
     * @return true if user can manage file
     */
    boolean canManage(String repoKey, String path);

    /**
     * Do an action (given as {@link Callable}) as the system user. This method ensures to authenticate as the system
     * user before the action and restore the original authentication after.
     *
     * @param callable the action to perform
     */
    <V> V callAsSystem(Callable<V> callable) throws PackageException;

    /**
     * @return True if the current user is a anonymous.
     */
    boolean isAnonymous();

    /**
     * @return True if anonymous access is allowed.
     */
    boolean isAnonAccessEnabled();

    /**
     * @return Current authenticated username
     */
    @NonNull
    String getCurrentUserName();

    /**
     * @return True if the current user is authenticated
     */
    boolean isAuthenticated();

    /**
     * Generating Access token for the user with following specs: no Expiry, non Refreshable, no groups limitation.
     * We set the scope to be "member-of-groups:*" since it will allow the user to obtain groups access according
     * to his group memberships AT THE TIME OF EACH LOGIN.
     */
    String loginAndGenerateAccessTokenWithBasicAuth(String userName, String password)
            throws PackageException;

    /**
     * Revoke the access token, if exist.
     */
    void revokeToken(String token) throws PackageException;

    /**
     * @return A {@link SimpleUser} with the current authenticated user details.
     * @throws PackageException if the user is unauthenticated or was not found.
     */
    SimpleUser getCurrentUserDetails() throws PackageException;
}
