package org.artifactory.storage.db.hqc.dao;

import com.google.common.collect.Lists;
import lombok.extern.slf4j.Slf4j;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.storage.db.hqc.entity.HeavyQueryCacheEntry;
import org.artifactory.storage.db.util.BaseDao;
import org.artifactory.storage.db.util.JdbcHelper;
import org.jfrog.storage.util.DbUtils;
import org.jfrog.storage.util.querybuilder.QueryWriter;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;

import static org.artifactory.common.ConstantValues.dbOperationsBulkSize;
import static org.artifactory.storage.db.hqc.enums.HeavyQueryCacheFields.CACHE_KEY;
import static org.artifactory.storage.db.hqc.enums.HeavyQueryCacheFields.CACHE_TIME;
import static org.artifactory.storage.db.hqc.enums.HeavyQueryCacheFields.DATA;
import static org.artifactory.storage.db.hqc.enums.HeavyQueryCacheFields.EXPIRY_TIME;
import static org.artifactory.storage.db.hqc.enums.HeavyQueryCacheFields.ID;

/**
 * @author Tomer Slutsky
 * @author Andrey Tashlitsky
 */
@Slf4j
@Repository
public class HeavyQueryCacheDao extends BaseDao {

    private static final String TABLE_NAME = "heavy_query_cache";
    private static final String INSERT_CACHE_SQL =
            "INSERT INTO " + TABLE_NAME + " (id, cache_key, cache_time, expiry_time, data) VALUES (?, ?, ?, ?, ?)";
    private static final String SELECT_CACHE_SQL =
            "SELECT * FROM " + TABLE_NAME + " WHERE cache_key = ? AND expiry_time > ?";
    private static final String DELETE_CACHE_BY_CACHE_KEY_SQL =
            "DELETE FROM " + TABLE_NAME + " WHERE cache_key = ?";
    private static final String BULK_DELETE_CACHE_BY_ID_SQL =
            "DELETE FROM " + TABLE_NAME + " WHERE id IN (#)";
    private final ArtifactoryDbProperties dbProperties;

    @Autowired
    public HeavyQueryCacheDao(JdbcHelper jdbcHelper, ArtifactoryDbProperties dbProperties) {
        super(jdbcHelper);
        this.dbProperties = dbProperties;
    }

    /**
     * Gets the requested entry by key.
     *
     * @return Entry from the heavy_query_cache table if found.
     */
    public HeavyQueryCacheEntry getCachedResponse(String cacheKey, long time) throws SQLException {
        ResultSet resultSet = null;
        try {
            log.debug("Executing: '{}', cache_key = '{}', time = '{}'", SELECT_CACHE_SQL, cacheKey, time);
            resultSet = jdbcHelper.executeSelect(SELECT_CACHE_SQL, cacheKey, time);
            if (resultSet.next()) {
                return fromResultSet(resultSet);
            }
            return null;
        } finally {
            DbUtils.close(resultSet);
        }
    }

    /**
     * Inserts an entry into the heavy_query_cache table.
     *
     * @param cacheEntry HeavyQueryCacheEntry object.
     */
    public int create(HeavyQueryCacheEntry cacheEntry) throws SQLException {
        log.debug("Executing: '{}', cacheEntry = '{}'", INSERT_CACHE_SQL, cacheEntry);
        return jdbcHelper.executeUpdate(INSERT_CACHE_SQL, cacheEntry.getId(), cacheEntry.getCacheKey(),
                cacheEntry.getCacheTime(), cacheEntry.getExpiryTime(), cacheEntry.getData());
    }

    /**
     * Get entries where expiry_time < System.currentTimeMillis().
     *
     * @param time The time of expiry. Entries with expiry_time less than this value will be considered as expired.
     * @return List of all expired entries (bulk of {@link ConstantValues#dbOperationsBulkSize}). <b>Note:</b> The
     * entries contain only id, cache_key, expiry_time.
     */
    public List<HeavyQueryCacheEntry> getExpiredEntries(long time) throws SQLException {
        String query = new QueryWriter(dbProperties.getDbType())
                .select("id, cache_key, expiry_time")
                .from(TABLE_NAME)
                .where("expiry_time < ?")
                .limit(dbOperationsBulkSize.getLong())
                .build();
        List<HeavyQueryCacheEntry> expiredEntries = Lists.newArrayList();
        ResultSet resultSet = null;
        try {
            log.debug("Executing: '{}', time = '{}'", query, time);
            resultSet = jdbcHelper.executeSelect(query, time);
            while (resultSet.next()) {
                HeavyQueryCacheEntry cacheEntry = HeavyQueryCacheEntry.builder()
                        .id(resultSet.getLong(ID.toString()))
                        .cacheKey(resultSet.getString(CACHE_KEY.toString()))
                        .expiryTime(resultSet.getLong(EXPIRY_TIME.toString()))
                        .build();
                expiredEntries.add(cacheEntry);
            }
            return expiredEntries;
        } finally {
            DbUtils.close(resultSet);
        }
    }

    /**
     * Deletes all the given entries from the heavy_query_cache table (by 'id').
     *
     * @param cacheEntries Entries to delete.
     */
    public void delete(List<HeavyQueryCacheEntry> cacheEntries) throws SQLException {
        List<Long> ids = cacheEntries.stream().map(HeavyQueryCacheEntry::getId).toList();
        log.debug("Executing: '{}', ids = '{}'", BULK_DELETE_CACHE_BY_ID_SQL, ids);
        jdbcHelper.executeUpdate(BULK_DELETE_CACHE_BY_ID_SQL, ids);
    }

    /**
     * Deletes a given entry from the heavy_query_cache table (by 'cache_key').
     *
     * @param cacheEntry Entry to delete.
     */
    public void delete(HeavyQueryCacheEntry cacheEntry) throws SQLException {
        log.debug("Executing: '{}', cache_key = '{}'", DELETE_CACHE_BY_CACHE_KEY_SQL, cacheEntry.getCacheKey());
        jdbcHelper.executeUpdate(DELETE_CACHE_BY_CACHE_KEY_SQL, cacheEntry.getCacheKey());
    }

    /**
     * Converts a result set to an object.
     *
     * @param resultSet Given result set.
     * @return HeavyQueryCacheEntry object.
     */
    private HeavyQueryCacheEntry fromResultSet(ResultSet resultSet) throws SQLException {
        return HeavyQueryCacheEntry.builder()
                .id(resultSet.getLong(ID.toString()))
                .cacheKey(resultSet.getString(CACHE_KEY.toString()))
                .cacheTime(resultSet.getLong(CACHE_TIME.toString()))
                .expiryTime(resultSet.getLong(EXPIRY_TIME.toString()))
                .data(resultSet.getBytes(DATA.toString()))
                .build();
    }
}
