package org.artifactory.storage.db.repo.itest;

import lombok.AllArgsConstructor;
import lombok.Data;
import org.apache.commons.lang3.builder.EqualsBuilder;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.storage.db.itest.DbBaseTest;
import org.artifactory.storage.db.repo.dao.RepositoryConfigDao;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.artifactory.storage.db.repo.service.mapper.RepositoryConfigMapperImpl;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.JsonUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.Set;

import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertTrue;

/**
 * @author Tamar Rosen
 */
@Test
public class RepositoryConfigDaoTest extends DbBaseTest {
    @Autowired
    private RepositoryConfigDao dao;

    RepositoryConfigMapperImpl repositoryConfigMapper;

    @BeforeMethod
    public void setUp() throws SQLException {
        repositoryConfigMapper = new RepositoryConfigMapperImpl();
    }

    @AfterMethod
    public void tearDown() throws SQLException {
        dao.deleteAll();
    }

    public void testBulkDelete() throws SQLException {
        testBulkInsertRepositoryConfig();
        List<String> repoKeysToDelete = new ArrayList<>();
        for (int i = 0; i < 5; i++) {
            repoKeysToDelete.add("key" + i);
        }

        dao.bulkDelete(repoKeysToDelete);
        List<DbRepositoryConfig> retrievedRepoConfigs = dao.getAll();
        assertEquals(retrievedRepoConfigs.size(), 5);
    }

    public void testBulkInsertRepositoryConfig() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = getDbRepositoryConfigs();
        dao.bulkInsert(repoConfigs);
        List<DbRepositoryConfig> retrievedRepoConfigs = dao.getAll();
        assertEquals(repoConfigs.size(), retrievedRepoConfigs.size());
        for (int i = 0; i < 10; i++) {
            EqualsBuilder.reflectionEquals(repoConfigs.get(i), retrievedRepoConfigs.get(i));
        }
    }

    @NotNull
    private List<DbRepositoryConfig> getDbRepositoryConfigs() {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i, null));
        }
        return repoConfigs;
    }

    public void findRepositoryConfigByKey() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        DbRepositoryConfig repoToRetrieve = new DbRepositoryConfig("key-to-retrieve", "local",
                RepoType.Gradle.getType(), 1, 2, 3, null);
        repoConfigs.add(repoToRetrieve);
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i, null));
        }
        dao.bulkInsert(repoConfigs);
        Optional<DbRepositoryConfig> retrievedRepoConfig = dao.findRepositoryByKey("key-to-retrieve");
        assertTrue(retrievedRepoConfig.isPresent());
        EqualsBuilder.reflectionEquals(retrievedRepoConfig, retrievedRepoConfig);
    }

    public void testInsertRepositoryConfig() throws SQLException {
        LocalRepoConfig repositoryConfig = LocalRepoConfig.create("my-repo", RepoType.Generic);
        repositoryConfig.getBaseConfig().setModelVersion(0);
        repositoryConfig.getPackageTypeConfig().put("modelVersion", "0");
        DbRepositoryConfig insertMe =
                repositoryConfigMapper.repositoryConfigToDbRepositoryConfig(repositoryConfig);

        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1, 2, 3,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        dao.insert(insertMe);
        Optional<DbRepositoryConfig> retrievedRepoConfig = dao.findRepositoryByKey("my-repo");
        assertTrue(retrievedRepoConfig.isPresent());
        EqualsBuilder.reflectionEquals(retrievedRepoConfig, repoConfig);
    }

    public void testUpdateRepositoryConfigBlob() throws SQLException {
        // insert repo
        DbRepositoryConfig repoConfig =
                new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(), 1, 2,
                        3, JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        dao.insert(repoConfig);

        // get repo
        Optional<DbRepositoryConfig> optionalRetrievedRepoConfig = dao.findRepositoryByKey("my-repo");
        assertTrue(optionalRetrievedRepoConfig.isPresent());
        DbRepositoryConfig retrievedRepoConfig = optionalRetrievedRepoConfig.get();
        RepositoryConfigDaoTest.TestBlob returnedTestBlob =
                JsonUtils.getInstance().readValue(retrievedRepoConfig.getRepositoryInfoBlob(),
                        RepositoryConfigDaoTest.TestBlob.class);
        long num = returnedTestBlob.getNum();
        assertEquals(num, 1L);

        // update config blob
        returnedTestBlob.setNum(2L);
        retrievedRepoConfig.setRepositoryInfoBlob(JsonUtils.getInstance().valueToByteArray(returnedTestBlob));
        dao.update(retrievedRepoConfig);

        // get repo and validate config blob was updated
        Optional<DbRepositoryConfig> optionalUpdatedRepoConfig = dao.findRepositoryByKey("my-repo");
        assertTrue(optionalUpdatedRepoConfig.isPresent());
        DbRepositoryConfig updatedRepoConfig = optionalUpdatedRepoConfig.get();
        RepositoryConfigDaoTest.TestBlob updatedTestBlob =
                JsonUtils.getInstance().readValue(updatedRepoConfig.getRepositoryInfoBlob(),
                        RepositoryConfigDaoTest.TestBlob.class);
        assertEquals(updatedTestBlob.getNum(), 2L);
    }

    public void testDeleteByKey() throws SQLException {
        dao.insert(new DbRepositoryConfig("delete-me", "local", RepoType.YUM.getType(), 1,
                2, 3, null));
        int deleted = dao.delete("does-not-exist");
        assertEquals(deleted, 0);

        deleted = dao.delete("delete-me");
        assertEquals(deleted, 1);
    }

    public void testExists() throws SQLException {
        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1, 2, 3,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        dao.insert(repoConfig);
        assertTrue(dao.exists("my-repo"));
        assertFalse(dao.exists("does-not-exist"));
    }

    public void testBulkUpdate() throws SQLException {
        // insert repos
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i,
                    JsonUtils.getInstance().valueToByteArray(new TestBlob(i))));
        }
        dao.bulkInsert(repoConfigs);

        // get repos
        List<DbRepositoryConfig> reposToUpdate = dao.getAll();
        for (int i = 0; i < reposToUpdate.size(); i++) {
            RepositoryConfigDaoTest.TestBlob returnedTestBlob =
                    JsonUtils.getInstance().readValue(reposToUpdate.get(i).getRepositoryInfoBlob(),
                            RepositoryConfigDaoTest.TestBlob.class);
            assertEquals(returnedTestBlob.getNum(), i);

            // update repos
            returnedTestBlob.setNum(i + 1);
            reposToUpdate.get(i)
                    .setRepositoryInfoBlob(JsonUtils.getInstance().valueToByteArray(returnedTestBlob));
        }

        dao.bulkUpdate(reposToUpdate);

        // get repos and validate config blob was updated
        List<DbRepositoryConfig> retrievedRepoConfigs = dao.getAll();
        for (int i = 0; i < retrievedRepoConfigs.size(); i++) {
            RepositoryConfigDaoTest.TestBlob returnedTestBlob =
                    JsonUtils.getInstance().readValue(retrievedRepoConfigs.get(i).getRepositoryInfoBlob(),
                            RepositoryConfigDaoTest.TestBlob.class);
            assertEquals(returnedTestBlob.getNum(), i + 1);
        }
    }

    public void testGetConfigsByKeys() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 10; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i,
                    JsonUtils.getInstance().valueToByteArray(new TestBlob(i))));
        }
        dao.bulkInsert(repoConfigs);

        Set<DbRepositoryConfig> configsByKeys = dao.findByKeysWithPagination(
                Set.of("key0", "key1", "key2", "key3", "non-existing-repo"));
        Set<DbRepositoryConfig> dbRepositoryConfigs = new HashSet<>(configsByKeys);

        Set<DbRepositoryConfig> expectedRepos = new HashSet<>(repoConfigs.subList(0, 4));
        assertEquals(dbRepositoryConfigs, expectedRepos);
    }

    public void testFindConfigsByDetailsType() throws SQLException {
        List<DbRepositoryConfig> repoConfigs = new ArrayList<>();
        for (int i = 0; i < 3; i++) {
            repoConfigs.add(new DbRepositoryConfig("key" + i, "type" + i,
                    RepoType.values()[i].getType(), i, i, i,
                    JsonUtils.getInstance().valueToByteArray(new TestBlob(i))));
        }
        dao.bulkInsert(repoConfigs);

        Collection<DbRepositoryConfig> configsByKeys = dao.findRepositoriesByDetailsType("type1");
        assertEquals(configsByKeys.size(), 1);
        assertEquals(configsByKeys.stream().findFirst().get().getType(), "type1");
    }

    public void testCountAll() throws SQLException {
        assertEquals(dao.countAll(), 0);

        DbRepositoryConfig repoConfig = new DbRepositoryConfig("my-repo", "local", RepoType.Maven.getType(),
                1L, 2L, 3L,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(1)));
        DbRepositoryConfig repoConfig1 = new DbRepositoryConfig("my-repo1", "remote", RepoType.Conan.getType(),
                10, 20, 30,
                JsonUtils.getInstance().valueToByteArray(new TestBlob(2)));

        dao.insert(repoConfig);
        dao.insert(repoConfig1);
        assertEquals(dao.countAll(), 2);
    }

    @Data
    @AllArgsConstructor
    public static class TestBlob {
        private long num;
    }
}