package org.artifactory.storage.db.repo.service;

import org.apache.commons.lang3.builder.EqualsBuilder;
import org.artifactory.storage.db.fs.service.RepositoryPersistingService;
import org.artifactory.storage.db.repo.dao.RepositoryConfigDao;
import org.artifactory.storage.db.repo.entity.DbRepositoryConfig;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoSession;
import org.mockito.quality.Strictness;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.function.Function;
import java.util.stream.Collectors;

import static org.hibernate.validator.internal.util.Contracts.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertArrayEquals;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

/**
 * @author Tamar Rosen
 */
class RepositoryConfigStorageServiceImplTest {
    private MockitoSession mockitoSession;
    private RepositoryConfigStorageService repositoryConfigStorageService;

    @Mock
    private RepositoryConfigDao repositoryConfigDao;
    @Mock
    private RepositoryPersistingService repositoryPersistingService;

    @BeforeEach
    public void setUp() {
        mockitoSession = Mockito.mockitoSession()
                .strictness(Strictness.STRICT_STUBS)
                .initMocks(this)
                .startMocking();

        repositoryConfigStorageService =
                spy(new RepositoryConfigStorageServiceImpl(repositoryConfigDao, repositoryPersistingService));
    }

    @AfterEach
    public void tearDown() {
        mockitoSession.finishMocking();
    }

    @Test
    void testFindRepositoryByKey() throws Exception {
        String repoKey = "find-me-repo";
        DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
        when(repositoryConfigDao.findRepositoryByKey(repoKey)).thenReturn(Optional.of(dbRepositoryConfig));

        DbRepositoryConfig retrievedRepoConfig = repositoryConfigStorageService.findRepositoryByKey(repoKey);
        assertNotNull(retrievedRepoConfig);
        EqualsBuilder.reflectionEquals(retrievedRepoConfig, dbRepositoryConfig);
    }

    @Test
    void testGetAllRepositories() throws Exception {
        List<DbRepositoryConfig> reposInStore = new ArrayList<>();
        for (int i = 0; i < 3; i++) {
            DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
            dbRepositoryConfig.setKey("repo-" + i);
            reposInStore.add(dbRepositoryConfig);
        }
        when(repositoryConfigDao.getAll()).thenReturn(reposInStore);

        List<DbRepositoryConfig> fetchedRepos = repositoryConfigStorageService.getAllRepositories();
        assertNotNull(fetchedRepos);
        assertEquals(fetchedRepos.size(), 3);
        for (int i = 0; i < 3; i++) {
            assertEquals(fetchedRepos.get(i).getKey(), "repo-" + i);
        }
    }

    @Test
    void testFindRepositoriesByKeys() throws Exception {
        Set<DbRepositoryConfig> reposInStore = new HashSet<>();
        Set<String> repoKeys = new HashSet<>();
        for (int i = 0; i < 3; i++) {
            DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
            dbRepositoryConfig.setKey("repo-" + i);
            reposInStore.add(dbRepositoryConfig);
            repoKeys.add(dbRepositoryConfig.getKey());
        }
        repoKeys.add("non-existing-repo");
        when(repositoryConfigDao.findByKeysWithPagination(repoKeys)).thenReturn(reposInStore);

        Set<DbRepositoryConfig> repoConfigs =
                repositoryConfigStorageService.findRepositoryConfigsByKeys(repoKeys);

        assertNotNull(repoConfigs);
        assertTrue(repoConfigs.containsAll(reposInStore));
    }

    @Test
    void testInsert() throws SQLException {
        DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
        dbRepositoryConfig.setKey("insert-me");
        List<DbRepositoryConfig> dbNaiveSimulator = new ArrayList<>();
        doAnswer(repo -> dbNaiveSimulator.add(repo.getArgument(0))).when(repositoryConfigDao)
                .insert(dbRepositoryConfig);

        repositoryConfigStorageService.insertRepositoryConfig(dbRepositoryConfig);

        assertTrue(dbNaiveSimulator.contains(dbRepositoryConfig));
    }

    @Test
    void testBulkInsert() throws SQLException {
        List<DbRepositoryConfig> reposToInsert = new ArrayList<>();
        for (int i = 3; i > 0; i--) {
            DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
            dbRepositoryConfig.setKey("repo-" + i);
            reposToInsert.add(dbRepositoryConfig);
        }
        List<DbRepositoryConfig> dbNaiveSimulator = new ArrayList<>();
        doAnswer(repo -> dbNaiveSimulator.addAll(repo.getArgument(0))).when(repositoryConfigDao)
                .bulkInsert(reposToInsert.stream()
                        .sorted(Comparator.comparing(DbRepositoryConfig::getKey))
                        .collect(Collectors.toList()));

        repositoryConfigStorageService.bulkInsertRepositoryConfigs(new HashSet<>(reposToInsert));

        assertTrue(dbNaiveSimulator.containsAll(reposToInsert));
    }

    @Test
    void testUpdate() throws SQLException {
        DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
        dbRepositoryConfig.setKey("update-me");
        dbRepositoryConfig.setRepositoryInfoBlob(new byte[] {1, 2, 3});
        Map<String, DbRepositoryConfig> dbNaiveSimulator = new HashMap<>();
        dbNaiveSimulator.put("update-me", dbRepositoryConfig);

        doAnswer(answer -> {
            DbRepositoryConfig updatedRepoConfig = answer.getArgument(0);
            dbNaiveSimulator.put("update-me", updatedRepoConfig);
            return null;
        }).when(repositoryConfigDao).update(any());

        DbRepositoryConfig updateAccordingToMe = new DbRepositoryConfig();
        byte[] newInfoBlob = {4, 5, 6};
        updateAccordingToMe.setRepositoryInfoBlob(newInfoBlob);
        repositoryConfigStorageService.updateRepositoryConfig(updateAccordingToMe);

        DbRepositoryConfig updatedRepoConfig = dbNaiveSimulator.get("update-me");
        assertNotNull(updatedRepoConfig);
        assertNotNull(updatedRepoConfig.getRepositoryInfoBlob());
        assertEquals(updatedRepoConfig.getRepositoryInfoBlob(), newInfoBlob);
    }

    @Test
    void testBulkUpdate() throws SQLException {
        Map<String, DbRepositoryConfig> dbNaiveSimulator = new HashMap<>();
        for (int i = 0; i < 3; i++) {
            DbRepositoryConfig dbRepositoryConfig = new DbRepositoryConfig();
            dbRepositoryConfig.setKey("update-me-" + i);
            dbRepositoryConfig.setRepositoryInfoBlob(new byte[] {(byte) i});
            dbNaiveSimulator.put("update-me-" + i, dbRepositoryConfig);
        }

        doAnswer(answer -> {
            List<DbRepositoryConfig> updatedReposConfig = answer.getArgument(0);
            updatedReposConfig.forEach(repo -> dbNaiveSimulator.put(repo.getKey(), repo));
            return null;
        }).when(repositoryConfigDao).bulkUpdate(any());

        Set<DbRepositoryConfig> dbRepositoryConfigs = new HashSet<>();
        for (int i = 0; i < 3; i++) {
            DbRepositoryConfig updateAccordingToMe = new DbRepositoryConfig();
            updateAccordingToMe.setKey("update-me-" + i);
            byte[] newInfoBlob = {(byte) (i + 1)};
            updateAccordingToMe.setRepositoryInfoBlob(newInfoBlob);
            dbRepositoryConfigs.add(updateAccordingToMe);
        }

        repositoryConfigStorageService.bulkUpdateRepositoryConfigs(dbRepositoryConfigs);

        assertEquals(dbNaiveSimulator.size(), 3);
        for (int i = 0; i < 3; i++) {
            DbRepositoryConfig updatedRepoConfig = dbNaiveSimulator.get("update-me-" + i);
            assertNotNull(updatedRepoConfig);
            assertNotNull(updatedRepoConfig.getRepositoryInfoBlob());
            assertArrayEquals(updatedRepoConfig.getRepositoryInfoBlob(), new byte[] {(byte) (i + 1)});
        }
    }

    @Test
    void testDelete() throws SQLException {
        List<String> dbNaiveSimulator = new ArrayList<>();
        dbNaiveSimulator.add("delete-me");
        when(repositoryConfigDao.delete("delete-me"))
                .thenAnswer(repo -> {
                    dbNaiveSimulator.remove(repo.getArgument(0));
                    return 0;
                });

        repositoryConfigStorageService.delete("delete-me");

        assertFalse(dbNaiveSimulator.contains("delete-me"));
    }

    @Test
    void testBulkDelete() throws SQLException {
        List<String> dbNaiveSimulator = new ArrayList<>();
        List<String> reposToDelete = new ArrayList<>();
        for (int i = 3; i > 0; i--) {
            dbNaiveSimulator.add("delete-me-" + i);
            reposToDelete.add("delete-me-" + i);
        }
        doAnswer(repo -> {
            dbNaiveSimulator.removeAll(repo.getArgument(0));
            return null;
        }).when(repositoryConfigDao).bulkDelete(reposToDelete.stream().sorted(Comparator.comparing(Function.identity()))
                .collect(Collectors.toList()));

        repositoryConfigStorageService.bulkDeleteRepositoryConfigs(new HashSet<>(reposToDelete));
        reposToDelete.forEach(repoToDelete -> assertFalse(dbNaiveSimulator.contains(repoToDelete)));
    }

    @Test
    void testExists() throws SQLException {
        List<String> dbNaiveSimulator = new ArrayList<>();
        dbNaiveSimulator.add("exists");
        when(repositoryConfigDao.exists("exists"))
                .thenAnswer(repo -> dbNaiveSimulator.contains(repo.getArgument(0)));

        boolean exists = repositoryConfigStorageService.exists("exists");
        assertTrue(exists);
        exists = repositoryConfigStorageService.exists("does-not-exist");
        assertFalse(exists);
    }

    @Test
    void testSortBeforePersist() throws SQLException {
        RepositoryConfigStorageServiceImpl storageService =
                new RepositoryConfigStorageServiceImpl(repositoryConfigDao, repositoryPersistingService);

        DbRepositoryConfig repoA = new DbRepositoryConfig();
        repoA.setKey("a");
        DbRepositoryConfig repoB = new DbRepositoryConfig();
        repoB.setKey("b");

        Set<DbRepositoryConfig> dbRepositoryConfigs = Set.of(repoB, repoA);

        storageService.bulkUpdateRepositoryConfigs(dbRepositoryConfigs);

        // Repos were sorted
        verify(repositoryConfigDao).bulkUpdate(List.of(repoA, repoB));

        storageService.bulkInsertRepositoryConfigs(dbRepositoryConfigs);
        verify(repositoryConfigDao).bulkInsert(List.of(repoA, repoB));

        storageService.bulkDeleteRepositoryConfigs(dbRepositoryConfigs.stream().map(DbRepositoryConfig::getKey).collect(
                Collectors.toSet()));

        verify(repositoryConfigDao).bulkDelete(List.of(repoA.getKey(), repoB.getKey()));
    }
}