/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.configuration.repositories;

import com.google.common.collect.Lists;
import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.model.RepoConfig;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.admin.configuration.repository.RepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.releasebundles.ReleaseBundlesRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.remote.RemoteRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.virtual.VirtualRepositoryConfigModel;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder.RepoConfigModelBuilder;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.jfrog.common.BiOptional;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Optional;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.repo.RepoDetailsType.FEDERATED_REPO;
import static org.artifactory.repo.RepoDetailsType.LOCAL_REPO;
import static org.artifactory.repo.RepoDetailsType.RELEASE_BUNDLE_REPO;
import static org.artifactory.repo.RepoDetailsType.REMOTE_REPO;
import static org.artifactory.repo.RepoDetailsType.VIRTUAL_REPO;

/**
 * @author Aviad Shikloshi
 * @author Dan Feldman
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetRepositoryConfigService implements RestService<Void> {

    private ProjectsService projectsService;
    private RepositoryService repositoryService;
    private RepoConfigModelBuilder configBuilder;

    @Override
    public void execute(ArtifactoryRestRequest<Void> request, RestResponse response) {
        String repoType = request.getPathParamByKey("repoType");
        String repoKey = request.getPathParamByKey("repoKey");
        List<? extends RepoConfig> allRepositoriesConfigs = getAllRepositoriesConfigs(repoType);
        if (StringUtils.isBlank(repoKey)) {
            List<RepositoryConfigModel<?, ?, ?>> repositoryConfigModels =
                    allRepositoriesConfigs.stream()
                            .map(repoConfig -> createModelFromRepoConfig(repoType, repoConfig))
                            .collect(Collectors.toList());

            response.iModelList(repositoryConfigModels)
                    .responseCode(HttpStatus.SC_OK);
        } else {
            BiOptional.of(allRepositoriesConfigs.stream()
                            .filter(repoConfig -> repoConfig.getKey().equals(repoKey))
                            .findFirst())
                    .ifPresent(repoConfig -> response.iModel(createModelFromRepoConfig(repoType, repoConfig)))
                    .ifNotPresent(() -> response.error("Repository '" + repoKey + "' does not exists.")
                            .responseCode(HttpStatus.SC_NOT_FOUND));
        }
    }

    private RepositoryConfigModel<?, ? ,?> createModelFromRepoConfig(String repoType, RepoConfig repoConfig) {
        RepositoryConfigModel<?, ? ,?> repositoryConfigModel =
                getRepoDescriptorByType(repoType).fromRepoConfig(configBuilder, repoConfig);

        String projectKey = projectsService.getProjectKeyOfRepository(repoConfig.getKey());
        ProjectResourceEntity projectRepoEntity =
                Optional.ofNullable(
                                projectsService.getProjectRepositoryEntity(projectKey,
                                        repoConfig.getKey()))
                        .orElse(ProjectResourceEntity.builder().resourceName(repoConfig.getKey())
                                .environment(ProjectsService.DEFAULT_ENVIRONMENT).build());

        repositoryConfigModel.getGeneral().setProjectKey(projectKey);
        repositoryConfigModel.getGeneral().setEnvironments(Set.copyOf(projectRepoEntity.getEnvironments()));
        return repositoryConfigModel;
    }

    private List<? extends RepoConfig> getAllRepositoriesConfigs(String repoType) {
        if (LOCAL_REPO.equals(repoType)) {
            return repositoryService.getLocalRepoConfigsExcludingInfoRepos();
        } else if (REMOTE_REPO.equals(repoType)) {
            return repositoryService.getRemoteRepoConfigs();
        } else if (FEDERATED_REPO.equals(repoType)) {
            return repositoryService.getAllFederatedRepoConfigsWithSystemRepo();
        } else if (VIRTUAL_REPO.equals(repoType)) {
            return repositoryService.getVirtualRepoConfigs();
        } else if (RELEASE_BUNDLE_REPO.equals(repoType)) {
            return repositoryService.getReleaseBundleRepoConfigs();
        }
        return Lists.newArrayList();
    }

    private RepositoryConfigModel<?, ?, ?> getRepoDescriptorByType(String repoType) {
        if (REMOTE_REPO.equals(repoType)) {
            return new RemoteRepositoryConfigModel();
        } else if (VIRTUAL_REPO.equals(repoType)) {
            return new VirtualRepositoryConfigModel();
        } else if (RELEASE_BUNDLE_REPO.equals(repoType)) {
            return new ReleaseBundlesRepositoryConfigModel();
        } else if (FEDERATED_REPO.equals(repoType)) {
            return new FederatedRepositoryConfigModel();
        }
        return new LocalRepositoryConfigModel();
    }

    @Autowired
    public void setProjectsService(@NonNull ProjectsService projectsService) {
        this.projectsService = projectsService;
    }

    @Autowired
    public void setRepositoryService(@NonNull RepositoryService repositoryService) {
        this.repositoryService = repositoryService;
    }

    @Autowired
    public void setConfigBuilder(@NonNull RepoConfigModelBuilder configBuilder) {
        this.configBuilder = configBuilder;
    }
}
