package org.artifactory.ui.rest.service.utils.repositories;

import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.rest.common.model.continues.ContinueIndexPage;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.api.rest.common.model.continues.FetchFunction;
import org.artifactory.api.rest.common.model.continues.util.PagingUtils;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalCacheRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RealRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.repo.system.SystemRepos;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.utils.repositories.RepoKeyType;
import org.artifactory.ui.utils.RequestUtils;
import org.jfrog.common.StreamSupportUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Function;

import static org.artifactory.api.rest.common.model.continues.util.ContinueResultIndexUtils.getLastIndex;
import static org.artifactory.descriptor.repo.RepoType.ReleaseBundles;
import static org.artifactory.repo.RepoDetailsType.FEDERATED;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoDetailsType.REMOTE;

/**
 * @author Dvir Twina
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetAllRepositoriesPagedService implements RestService {


    private RepositoryService repoService;

    private String projectKey;
    private String repoSearchStr;

    @Autowired
    public void setRepoService(RepositoryService repoService) {
        this.repoService = repoService;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        projectKey = RequestUtils.getProjectKeyFromRequest(request);
        repoSearchStr = request.getQueryParamByKey("repoSearchStr");
        boolean isAnyLocalRepo = Boolean.parseBoolean(request.getQueryParamByKey("isAnyLocalRepo"));
        boolean isAnyRemoteRepo = Boolean.parseBoolean(request.getQueryParamByKey("isAnyRemoteRepo"));
        boolean isAnyFederatedRepo = Boolean.parseBoolean(request.getQueryParamByKey("isAnyFederatedRepo"));
        String continueState = request.getQueryParamByKey("continueState");
        if (StringUtils.isBlank(continueState)) {
            continueState = "0";
        }
        Long limit = StringUtils.isNotBlank(request.getQueryParamByKey("limit")) ?
                Long.parseLong(request.getQueryParamByKey("limit")) : null;

        ContinueIndexPage cip = new ContinueIndexPage(Long.parseLong(continueState), limit);
        List<FetchFunction<RepoKeyType>> functions = generateFetchFunctions(
                isAnyLocalRepo, isAnyRemoteRepo, isAnyFederatedRepo);

        Long totalRequiredRepos = null;
        if (Long.parseLong(continueState) == 0) {
            //add to response a total amount of requested repos
            totalRequiredRepos = getTotalRequiredRepos(isAnyLocalRepo, isAnyRemoteRepo, isAnyFederatedRepo);
        }
        ContinueResult<RepoKeyType> cr = PagingUtils.getPagingFromMultipleFunctions(cip, functions);

        cr.setTotalRequiredDataAmount(totalRequiredRepos);
        response.iModel(cr);
    }

    private List<FetchFunction<RepoKeyType>> generateFetchFunctions(
            boolean isAnyLocalRepo, boolean isAnyRemoteRepo, boolean isAnyFederatedRepo) {

        if (!isAnyLocalRepo && !isAnyRemoteRepo && !isAnyFederatedRepo) {
            return List.of(new FetchFunction<>(this::getLocalRemoteAndFederatedRepos,
                    (long) repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(projectKey).size()));
        } else {
            List<FetchFunction<RepoKeyType>> functions = new ArrayList<>();
            if (isAnyLocalRepo) {
                functions.add(new FetchFunction<>(this::getLocalRepos,
                        (long) repoService.getLocalRepoConfigsByProject(projectKey).size()));
            }
            if (isAnyRemoteRepo) {
                functions.add(new FetchFunction<>(this::getRemoteRepos,
                        (long) repoService.getRemoteRepoDescriptorsByProject(projectKey).size()));
            }
            if (isAnyFederatedRepo) {
                functions.add(new FetchFunction<>(this::getFederatedRepos,
                        (long) repoService.getFederatedRepoDescriptorsByProject(projectKey).size()));
            }

            return functions;
        }
    }

    private ContinueResult<RepoKeyType> getLocalRemoteAndFederatedRepos(Integer skip, Integer limit) {
        List<RealRepoDescriptor> allRequiredLocalRemoteAndFederatedRepoDescriptors =
                repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(projectKey);
        List<RealRepoDescriptor> offsetRepoDescriptors =
                StreamSupportUtils.stream(allRequiredLocalRemoteAndFederatedRepoDescriptors)
                        .skip(skip)
                        .filter(repo -> requiredReposFilter().apply(repo))
                        .limit(limit)
                        .toList();

        int lastIndex = getLastIndex(allRequiredLocalRemoteAndFederatedRepoDescriptors, offsetRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getRepoKeyType(offsetRepoDescriptors));
    }

    private ContinueResult<RepoKeyType> getLocalRepos(Integer skip, Integer limit) {
        List<LocalConfig> allRequiredLocalRepoDescriptors =
                repoService.getLocalRepoConfigsByProject(projectKey);
        List<LocalConfig> offsetRepoDescriptors = StreamSupportUtils.stream(allRequiredLocalRepoDescriptors)
                .skip(skip)
                .filter(repo -> requiredRepoConfigsFilter().apply(repo))
                .limit(limit)
                .toList();

        int lastIndex = getLastIndex(allRequiredLocalRepoDescriptors, offsetRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getRepoConfigKeyType(offsetRepoDescriptors));
    }

    private ContinueResult<RepoKeyType> getRemoteRepos(Integer skip, Integer limit) {
        List<RemoteRepoDescriptor> allRequiredRemoteRepoDescriptors =
                repoService.getRemoteRepoDescriptorsByProject(projectKey);
        List<RemoteRepoDescriptor> offsetRepoDescriptors = StreamSupportUtils.stream(allRequiredRemoteRepoDescriptors)
                .skip(skip)
                .filter(repo -> requiredReposFilter().apply(repo))
                .limit(limit)
                .toList();

        int lastIndex = getLastIndex(allRequiredRemoteRepoDescriptors, offsetRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getRepoKeyType(offsetRepoDescriptors));
    }

    private ContinueResult<RepoKeyType> getFederatedRepos(Integer skip, Integer limit) {
        List<FederatedRepoDescriptor> allRequiredFederatedRepoDescriptors =
                repoService.getFederatedRepoDescriptorsByProject(projectKey);
        List<FederatedRepoDescriptor> offsetRepoDescriptors =
                StreamSupportUtils.stream(allRequiredFederatedRepoDescriptors)
                        .skip(skip)
                        .filter(repo -> requiredReposFilter().apply(repo))
                        .limit(limit)
                        .toList();

        int lastIndex = getLastIndex(allRequiredFederatedRepoDescriptors, offsetRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getRepoKeyType(offsetRepoDescriptors));
    }

    private Function<RealRepoConfig, Boolean> specialRepoConfigFilter() {
        return repoDesc ->
                repoService.isInfoRepo(repoDesc.getPackageType())
                || ReleaseBundles.equals(repoDesc.getPackageType())
                || SystemRepos.isInTransitRepo(repoDesc.getKey())
                || SystemRepos.isObservabilityRepo(repoDesc.getKey());
    }

    private Function<RealRepoDescriptor, Boolean> specialRepoFilter() {
        return repoDesc -> repoService.isInfoRepo(repoDesc.getType())
                || ReleaseBundles.equals(repoDesc.getType())
                || SystemRepos.isInTransitRepo(repoDesc.getKey())
                || SystemRepos.isObservabilityRepo(repoDesc.getKey());
    }

    private Function<RealRepoDescriptor, Boolean> requiredReposFilter() {
        return repo -> !specialRepoFilter().apply(repo)
                && repo.getKey().contains(repoSearchStr);
    }

    private Function<RealRepoConfig, Boolean> requiredRepoConfigsFilter() {
        return repo -> !specialRepoConfigFilter().apply(repo)
                && repo.getKey().contains(repoSearchStr);
    }

    private List<RepoKeyType> getRepoKeyType(@NonNull List<? extends RealRepoDescriptor> filteredRepoDescriptors) {
        List<RepoKeyType> repos = new ArrayList<>();
        filteredRepoDescriptors
                .forEach(repoDesc -> {
                    if (repoDesc instanceof FederatedRepoDescriptor) {
                        repos.add(new RepoKeyType(FEDERATED.typeNameLowercase(), repoDesc.getKey()));
                    } else if (repoDesc instanceof LocalRepoDescriptor) {
                        repos.add(new RepoKeyType(LOCAL.typeNameLowercase(), repoDesc.getKey()));
                    } else if (repoDesc instanceof RemoteRepoDescriptor) {
                        repos.add(new RepoKeyType(REMOTE.typeNameLowercase(), repoDesc.getKey()));
                    }
                });
        return repos;
    }

    private List<RepoKeyType> getRepoConfigKeyType(@NonNull List<? extends LocalConfig> localConfigs) {
        List<RepoKeyType> repos = new ArrayList<>();
        localConfigs
                .forEach(localConfig -> {
                    if (localConfig instanceof FederatedRepoConfig) {
                        repos.add(new RepoKeyType(FEDERATED.typeNameLowercase(), localConfig.getKey()));
                    } else if (localConfig instanceof LocalRepoConfig) {
                        repos.add(new RepoKeyType(LOCAL.typeNameLowercase(), localConfig.getKey()));
                    } else if (localConfig instanceof LocalCacheRepoConfig || localConfig instanceof RemoteRepoConfig) {
                        repos.add(new RepoKeyType(REMOTE.typeNameLowercase(), localConfig.getKey()));
                    }
                });
        return repos;
    }

    private long getTotalRequiredRepos(boolean isAnyLocalRepo, boolean isAnyRemoteRepo, boolean isAnyFederatedRepo) {
        if (!isAnyLocalRepo && !isAnyRemoteRepo && !isAnyFederatedRepo) {
            return repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(projectKey).stream()
                    .filter(repo -> requiredReposFilter().apply(repo)).count();
        } else {
            long sum = 0;
            if (isAnyLocalRepo) {
                sum += repoService.getLocalRepoConfigsByProject(projectKey).stream()
                        .filter(repo -> requiredRepoConfigsFilter().apply(repo)).count();
            }
            if (isAnyRemoteRepo) {
                sum += repoService.getRemoteRepoDescriptorsByProject(projectKey).stream()
                        .filter(repo -> requiredReposFilter().apply(repo)).count();
            }
            if (isAnyFederatedRepo) {
                sum += repoService.getFederatedRepoDescriptorsByProject(projectKey).stream()
                        .filter(repo -> requiredReposFilter().apply(repo)).count();
            }
            return sum;
        }
    }
}
