package org.artifactory.ui.rest.model.admin.configuration.repository.info;

import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.webstart.ArtifactWebstartAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.repo.RepoPath;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.artifactory.storage.binstore.service.BinaryService;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertTrue;

/**
 * Author: Ivan Perepelytsya
 * Date: 6/3/20.
 */
public class RepositoryFieldsValuesModelTest {

    private AutoCloseable closeable;

    @Mock
    private ArtifactoryApplicationContext artifactoryContext;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private ArtifactWebstartAddon artifactWebstartAddon;
    @Mock
    private CentralConfigDescriptor descriptor;
    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;
    @Mock
    private AolService aolService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private BinaryService binaryService;

    @BeforeEach
    void setup() {
        closeable = MockitoAnnotations.openMocks(this);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);

        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);

        when(addonsManager.addonByType(ArtifactWebstartAddon.class)).thenReturn(artifactWebstartAddon);
        when(artifactWebstartAddon.getKeyPairNames()).thenReturn(List.of());
        when(ContextHelper.get().beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(ContextHelper.get().beanForType(CentralConfigService.class)).thenReturn(centralConfigService);
        when(centralConfigService.getDescriptor()).thenReturn(descriptor);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
        when(aolService.skipDockerDns()).thenReturn(false);
    }

    @AfterEach
    void cleanup() throws Exception {
        ArtifactoryContextThreadBinder.unbind();
        ArtifactoryHome.unbind();
        if (closeable != null) {
            closeable.close();
        }
    }

    @Test
    void testCdn() {
        when(binaryService.isCdnAllowed()).thenReturn(true);
        when(descriptor.getProxies()).thenReturn(List.of());
        final RepositoryFieldsValuesModel model =
                new RepositoryFieldsValuesModel(descriptor, repositoryService, binaryService, authorizationService,
                        List.of("mvn", "npm", "nuget"), List.of("Docker", "mvn"));
        final List<String> allowedRedirects = model.getCdnRedirectAllowedRepos();

        assertEquals(allowedRedirects.size(), 3);
        assertEquals(allowedRedirects.get(0), "mvn");
        assertEquals(allowedRedirects.get(1), "npm");
        assertEquals(allowedRedirects.get(2), "nuget");

        final List<String> allowedCdnRedirects = model.getForcedCdnRedirectRepos();
        assertEquals(allowedCdnRedirects.size(), 2);
        assertEquals(allowedCdnRedirects.get(0), "Docker");
        assertEquals(allowedCdnRedirects.get(1), "mvn");

        assertTrue(model.isCdnAllowedPerRepo());
    }

    @ParameterizedTest
    @ValueSource(booleans = {true, false})
    void testWithFederationRepos(boolean returnNonProjectRepos) {
        final String federatedRepoKey = "key";
        final BinaryService binaryService = mock(BinaryService.class);
        when(binaryService.isCdnAllowed()).thenReturn(true);
        when(authorizationService.userHasPermissionsOnRepositoryRoot(federatedRepoKey))
                .thenReturn(returnNonProjectRepos);
        FederatedRepoDescriptor federatedRepoDescriptor = new FederatedRepoDescriptor();
        federatedRepoDescriptor.setKey(federatedRepoKey);
        when(repositoryService.getFederatedRepoDescriptors()).thenReturn(List.of(federatedRepoDescriptor));
        final RepositoryFieldsValuesModel model =
                new RepositoryFieldsValuesModel(descriptor, repositoryService, binaryService, authorizationService,
                        List.of("mvn", "npm", "nuget"), List.of("Docker"));
        List<String> availableFederatedRepos = model.getAvailableFederatedRepos();
        if (returnNonProjectRepos) {
            assertEquals(availableFederatedRepos.size(), 1);
            assertEquals(availableFederatedRepos.get(0), federatedRepoKey);
        } else {
            assertEquals(availableFederatedRepos.size(), 0);
        }
    }

    @ParameterizedTest
    @ValueSource(booleans = {true, false})
    void testWithLocalRepos(boolean returnNonProjectRepos) {
        final String localRepoKey = "key";
        when(binaryService.isCdnAllowed()).thenReturn(true);
        when(authorizationService.userHasPermissionsOnRepositoryRoot(localRepoKey))
                .thenReturn(returnNonProjectRepos);

        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setKey(localRepoKey);
        when(repositoryService.getLocalRepoDescriptorsExcludingInfoRepos()).thenReturn(List.of(localRepoDescriptor));
        final RepositoryFieldsValuesModel model =
                new RepositoryFieldsValuesModel(descriptor, repositoryService, binaryService, authorizationService,
                        List.of("mvn", "npm", "nuget"), List.of("Docker"));
        List<String> availableLocalRepos = model.getAvailableLocalRepos();
        if (returnNonProjectRepos) {
            assertEquals(availableLocalRepos.size(), 1);
            assertEquals(availableLocalRepos.get(0), localRepoKey);
        } else {
            assertEquals(availableLocalRepos.size(), 0);
        }
    }

    @ParameterizedTest
    @ValueSource(booleans = {true, false})
    void testWithRemoteRepos(boolean returnNonProjectRepos) {
        final String remoteRepoKey = "key";
        when(binaryService.isCdnAllowed()).thenReturn(true);
        when(authorizationService.userHasPermissionsOnRepositoryRoot(remoteRepoKey))
                .thenReturn(returnNonProjectRepos);

        RemoteRepoDescriptor remoteRepoDescriptor = new HttpRepoDescriptor();
        remoteRepoDescriptor.setKey(remoteRepoKey);
        when(repositoryService.getRemoteRepoDescriptors()).thenReturn(List.of(remoteRepoDescriptor));
        final RepositoryFieldsValuesModel model =
                new RepositoryFieldsValuesModel(descriptor, repositoryService, binaryService, authorizationService,
                        List.of("mvn", "npm", "nuget"), List.of("Docker"));
        List<String> availableRepos = model.getAvailableRemoteRepos();
        if (returnNonProjectRepos) {
            assertEquals(availableRepos.size(), 1);
            assertEquals(availableRepos.get(0), remoteRepoKey);
        } else {
            assertEquals(availableRepos.size(), 0);
        }
    }

    @ParameterizedTest
    @ValueSource(booleans = {true, false})
    void testWithVirtualRepos(boolean returnNonProjectRepos) {
        final String virtualRepoKey = "key";

        VirtualRepoDescriptor repoDescriptor = new VirtualRepoDescriptor();
        repoDescriptor.setKey(virtualRepoKey);
        when(authorizationService.isProjectAdmin(any(RepoPath.class))).thenReturn(returnNonProjectRepos);
        when(repositoryService.getVirtualRepoDescriptors()).thenReturn(List.of(repoDescriptor));
        final RepositoryFieldsValuesModel model =
                new RepositoryFieldsValuesModel(descriptor, repositoryService, binaryService, authorizationService,
                        List.of("mvn", "npm", "nuget"), List.of("Docker"));
        List<String> availableRepos = model.getAvailableVirtualRepos();
        if (returnNonProjectRepos) {
            assertEquals(availableRepos.size(), 1);
            assertEquals(availableRepos.get(0), virtualRepoKey);
        } else {
            assertEquals(availableRepos.size(), 0);
        }
    }

    @ParameterizedTest
    @ValueSource(booleans = {true, false})
    void testDockerV1Enabled(boolean skipDns) {
        when(descriptor.getProxies()).thenReturn(List.of());
        when(aolService.skipDockerDns()).thenReturn(skipDns);

        final RepositoryFieldsValuesModel model =
                new RepositoryFieldsValuesModel(descriptor, repositoryService, binaryService, authorizationService,
                        List.of(), List.of());
        assertEquals(!skipDns, model.isDockerV1Enabled());
    }
}