package org.artifactory.ui.rest.service.admin.configuration.repositories;

import org.apache.http.HttpStatus;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.admin.configuration.repository.GeneralRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.RepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.builder.RepoConfigModelBuilder;
import org.jfrog.access.client.project.ProjectResourceEntity;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.mockito.stubbing.Answer;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.List;
import java.util.Set;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.verifyNoMoreInteractions;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Yevdo Abramov
 * Created on 03/11/2020
 */
public class GetRepositoryConfigServiceTest {
    private static final String REPO_KEY = "prj-myrepo";
    private static final String PROJECT_NAME = "prj";
    @Mock
    private ProjectsService projectsService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private RepoConfigModelBuilder configBuilder;

    private GetRepositoryConfigService service;
    private AutoCloseable mockitoAutoCloseable;

    @BeforeMethod
    public void setUp() {
        mockitoAutoCloseable = MockitoAnnotations.openMocks(this);
        service = new GetRepositoryConfigService();
        service.setConfigBuilder(configBuilder);
        service.setProjectsService(projectsService);
        service.setRepositoryService(repositoryService);
    }

    @AfterMethod
    public void tearDown() throws Exception {
        mockitoAutoCloseable.close();
    }

    @Test
    public void testExecuteWhenRepoNotFound() {
        // given
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(request.getPathParamByKey("repoType")).thenReturn("local");
        when(request.getPathParamByKey("repoKey")).thenReturn(REPO_KEY);
        when(repositoryService.getLocalRepoConfigsExcludingInfoRepos()).thenReturn(List.of());

        // when
        service.execute(request, response);

        // then
        assertEquals(response.getResponseCode(), HttpStatus.SC_NOT_FOUND);
        verify(repositoryService).getLocalRepoConfigsExcludingInfoRepos();
        verifyNoMoreInteractions(projectsService, repositoryService, configBuilder);
    }

    @Test
    public void testExecuteWhenRepoHasProjectMetadata() {
        // given
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(request.getPathParamByKey("repoType")).thenReturn("local");
        when(request.getPathParamByKey("repoKey")).thenReturn(REPO_KEY);

        LocalRepoConfig localConfig = LocalRepoConfig.create(REPO_KEY, RepoType.Maven);
        when(repositoryService.getLocalRepoConfigsExcludingInfoRepos()).thenReturn(List.of(localConfig));
        when(projectsService.getProjectKeyOfRepository(REPO_KEY)).thenReturn(PROJECT_NAME);
        when(projectsService.getProjectRepositoryEntity(PROJECT_NAME, REPO_KEY))
                .thenReturn(ProjectResourceEntity.builder()
                        .resourceName(REPO_KEY).environment("test1").environment("test2").build());
        doAnswer((Answer<Void>) invocation -> {
            Object[] args = invocation.getArguments();
            LocalRepositoryConfigModel configModel = (LocalRepositoryConfigModel) args[1];
            configModel.setGeneral(new GeneralRepositoryConfigModel());
            return null;
        }).when(configBuilder)
                .populateLocalDescriptorValuesToModel(eq(localConfig), any());

        // when
        service.execute(request, response);

        // then
        assertEquals(response.getResponseCode(), HttpStatus.SC_OK);
        RepositoryConfigModel repositoryConfigModel = (RepositoryConfigModel) response.getIModel();
        assertNotNull(repositoryConfigModel);
        assertEquals(repositoryConfigModel.getGeneral().getProjectKey(), PROJECT_NAME);
        assertTrue(repositoryConfigModel.getGeneral().getEnvironments().containsAll(Set.of("test1", "test2")));

        verify(repositoryService).getLocalRepoConfigsExcludingInfoRepos();
        verify(projectsService).getProjectKeyOfRepository(REPO_KEY);
        verify(projectsService).getProjectRepositoryEntity(PROJECT_NAME, REPO_KEY);
        verify(configBuilder).populateLocalDescriptorValuesToModel(eq(localConfig), any());
        verifyNoMoreInteractions(projectsService, repositoryService, configBuilder);
    }

    @Test
    public void testExecuteWhenRepoHasDefaultProjectMetadata() {
        // given
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(request.getPathParamByKey("repoType")).thenReturn("local");
        when(request.getPathParamByKey("repoKey")).thenReturn(REPO_KEY);

        LocalRepoConfig localRepoConfig = LocalRepoConfig.create(REPO_KEY, RepoType.Maven);
        when(repositoryService.getLocalRepoConfigsExcludingInfoRepos()).thenReturn(List.of(localRepoConfig));
        when(projectsService.getProjectKeyOfRepository(REPO_KEY)).thenReturn(ProjectsService.DEFAULT_PROJECT);
        when(projectsService.getProjectRepositoryEntity(ProjectsService.DEFAULT_PROJECT, REPO_KEY))
                .thenReturn(null);
        doAnswer((Answer<Void>) invocation -> {
            Object[] args = invocation.getArguments();
            LocalRepositoryConfigModel configModel = (LocalRepositoryConfigModel) args[1];
            configModel.setGeneral(new GeneralRepositoryConfigModel());
            return null;
        }).when(configBuilder)
                .populateLocalDescriptorValuesToModel(eq(localRepoConfig), any());

        // when
        service.execute(request, response);

        // then
        assertEquals(response.getResponseCode(), HttpStatus.SC_OK);
        RepositoryConfigModel repositoryConfigModel = (RepositoryConfigModel) response.getIModel();
        assertNotNull(repositoryConfigModel);
        assertEquals(repositoryConfigModel.getGeneral().getProjectKey(), ProjectsService.DEFAULT_PROJECT);
        assertTrue(repositoryConfigModel.getGeneral().getEnvironments().contains(ProjectsService.DEFAULT_ENVIRONMENT));

        verify(repositoryService).getLocalRepoConfigsExcludingInfoRepos();
        verify(projectsService).getProjectKeyOfRepository(REPO_KEY);
        verify(projectsService).getProjectRepositoryEntity(ProjectsService.DEFAULT_PROJECT, REPO_KEY);
        verify(configBuilder).populateLocalDescriptorValuesToModel(eq(localRepoConfig), any());
        verifyNoMoreInteractions(projectsService, repositoryService, configBuilder);
    }
}