package org.artifactory.ui.rest.service.builds.search;

import org.apache.commons.lang3.StringUtils;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.api.build.BuildService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.projects.ProjectsService;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.Arguments;
import org.junit.jupiter.params.provider.MethodSource;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import java.util.stream.Stream;

import static org.artifactory.repo.system.SystemRepos.BUILDINFO_REPO_KEY;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class SearchBuildIDsByNameServiceTest {

    private final static String PROJECT = "proj";
    private final static String PROJECT_BUILD_REPO = "proj" + BuildAddon.BUILD_INFO_REPO_EXTENSION;

    private SearchBuildIDsByNameService searchBuildIDsByNameService;
    @Mock
    private BuildService buildService;
    @Mock
    private AuthorizationService authService;

    private AutoCloseable autoCloseable;

    @BeforeEach
    public void setUp() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        searchBuildIDsByNameService = Mockito.spy(new SearchBuildIDsByNameService());
        searchBuildIDsByNameService.setBuildService(buildService);
        searchBuildIDsByNameService.setAuthService(authService);
        when(buildService.normaliseBuildRepoName(StringUtils.EMPTY)).thenReturn(BUILDINFO_REPO_KEY);
        when(buildService.normaliseBuildRepoName(PROJECT_BUILD_REPO)).thenReturn(PROJECT_BUILD_REPO);
        when(buildService.normaliseBuildRepoName(BUILDINFO_REPO_KEY))
                .thenReturn(BUILDINFO_REPO_KEY);
        when(buildService.getBuildInfoRepoForDefaultProject()).thenReturn(BUILDINFO_REPO_KEY);
        when(buildService.getBuildInfoRepoKeyByProject(PROJECT)).thenReturn(PROJECT_BUILD_REPO);
        when(buildService.getBuildInfoRepoKeyByProject(ProjectsService.DEFAULT_PROJECT))
                .thenReturn(BUILDINFO_REPO_KEY);
    }

    @ParameterizedTest
    @MethodSource("provideQueryParams")
    void testGetBuildRepoForSearch(String buildRepo, String projectKey, String expectedBuildRepo) {
        String resultRepo = searchBuildIDsByNameService.getBuildRepoForSearch(buildRepo, projectKey);
        Assertions.assertEquals(expectedBuildRepo, resultRepo);
    }

    private static Stream<Arguments> provideQueryParams() {
        return Stream.of(
                Arguments.of(StringUtils.EMPTY, StringUtils.EMPTY, null),
                Arguments.of(BUILDINFO_REPO_KEY, StringUtils.EMPTY, BUILDINFO_REPO_KEY),
                Arguments.of(StringUtils.EMPTY, ProjectsService.DEFAULT_PROJECT, BUILDINFO_REPO_KEY),
                Arguments.of(StringUtils.EMPTY, PROJECT, PROJECT_BUILD_REPO),
                Arguments.of(StringUtils.EMPTY, ProjectsService.UNASSIGNED, BUILDINFO_REPO_KEY),
                Arguments.of(PROJECT_BUILD_REPO, PROJECT, PROJECT_BUILD_REPO),
                Arguments.of(PROJECT_BUILD_REPO, ProjectsService.DEFAULT_PROJECT, PROJECT_BUILD_REPO)
        );
    }

    @ParameterizedTest
    @MethodSource("requestParamProvider")
    void testValidateQueryParams(String param, String value) {
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        ArtifactoryRestResponse response = mock(ArtifactoryRestResponse.class);
        when(request.getQueryParamByKey(param)).thenReturn(value);
        assertThrows(java.lang.NumberFormatException.class,
                () -> searchBuildIDsByNameService.execute(request, response));
    }

    private static Object[][] requestParamProvider() {
        return new Object[][] {
                {"before", "three days"},
                {"after", "one day"},
                {"num_of_rows", "two"},
        };
    }

    @AfterEach
    void tearDown() throws Exception {
        if (autoCloseable != null) {
            autoCloseable.close();
            autoCloseable = null;
        }
    }

}