package org.artifactory.ui.rest.service.utils.repositories;

import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RealRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalCacheRepoConfig;
import org.artifactory.model.LocalConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RealRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.utils.repositories.RepoKeyType;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import static org.apache.http.HttpStatus.SC_OK;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertNotNull;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

class GetAllRepositoriesPagedServiceTest {
    @Mock
    private RepositoryService repoService;
    @Mock
    private ArtifactoryRestRequest request;
    private ArtifactoryRestResponse response;
    private AutoCloseable autoCloseable;

    private GetAllRepositoriesPagedService service;

    @BeforeEach
    public void setUp() {
        autoCloseable = MockitoAnnotations.openMocks(this);
        service = new GetAllRepositoriesPagedService();
        service.setRepoService(repoService);
        response = new ArtifactoryRestResponse();
    }

    @AfterEach
    public void tearDown() throws Exception {
        autoCloseable.close();
    }

    @Test
    void testPagingForAllRepos() {
        List<RealRepoDescriptor> allRepos = List.of(
                generateRepositoryDscriptor(LocalRepoDescriptor.class, "localRepo_1"),
                generateRepositoryDscriptor(LocalRepoDescriptor.class, "localRepo_2"),
                generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo_3"),
                generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo_4"),
                generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo_5"),
                generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo_6")
        );
        when(repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(any())).thenReturn(allRepos);
        setMockingOfRequestParams(false, false, false, "", "3");

        List<RepoKeyType> repos = new ArrayList<>();
        String continueState = "0";
        executeAndAssertPaging(continueState, repos);

        assertEquals(repos.size(), 6);
        List<String> names = repos.stream().map(RepoKeyType::getRepoKey).toList();
        assertTrue(names.containsAll(Arrays.asList(
                "localRepo_1", "localRepo_2", "remoteRepo_3", "remoteRepo_4", "fedRepo_5", "fedRepo_6")));
    }

    @Test
    void testPagingForLocalRepos() {
        List<LocalConfig> localRepos = List.of(
                generateRepository(LocalRepoConfig.class, "localRepo_1"),
                generateRepository(LocalRepoConfig.class, "localRepo_2"),
                generateRepository(LocalRepoConfig.class, "localRepo_3")
        );
        when(repoService.getLocalRepoConfigsByProject(any())).thenReturn(localRepos);
        setMockingOfRequestParams(true, false, false, "", "2");

        List<RepoKeyType> repos = new ArrayList<>();
        String continueState = "0";
        executeAndAssertPaging(continueState, repos);

        assertEquals(repos.size(), 3);
        List<String> names = repos.stream().map(RepoKeyType::getRepoKey).toList();
        assertTrue(names.containsAll(Arrays.asList("localRepo_1", "localRepo_2", "localRepo_3")));
    }

    @Test
    void testPagingForRemoteRepos() {
        List<RemoteRepoDescriptor> remoteRepos = List.of(
                (RemoteRepoDescriptor) generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo_1"),
                (RemoteRepoDescriptor) generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo_2"),
                (RemoteRepoDescriptor) generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo_3")
        );
        when(repoService.getRemoteRepoDescriptorsByProject(any())).thenReturn(remoteRepos);
        setMockingOfRequestParams(false, true, false, "", "2");

        List<RepoKeyType> repos = new ArrayList<>();
        String continueState = "0";
        executeAndAssertPaging(continueState, repos);

        assertEquals(repos.size(), 3);
        List<String> names = repos.stream().map(RepoKeyType::getRepoKey).toList();
        assertTrue(names.containsAll(Arrays.asList("remoteRepo_1", "remoteRepo_2", "remoteRepo_3")));
    }

    @Test
    void testPagingForFederatedRepos() {
        List<FederatedRepoDescriptor> remoteRepos = List.of(
                (FederatedRepoDescriptor) generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo_1"),
                (FederatedRepoDescriptor) generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo_2"),
                (FederatedRepoDescriptor) generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo_3")
        );
        when(repoService.getFederatedRepoDescriptorsByProject(any())).thenReturn(remoteRepos);
        setMockingOfRequestParams(false, false, true, "", "2");

        List<RepoKeyType> repos = new ArrayList<>();
        String continueState = "0";
        executeAndAssertPaging(continueState, repos);

        assertEquals(repos.size(), 3);
        List<String> names = repos.stream().map(RepoKeyType::getRepoKey).toList();
        assertTrue(names.containsAll(Arrays.asList("fedRepo_1", "fedRepo_2", "fedRepo_3")));
    }

    @Test
    void testRepoFilteringForAllRepos() {
        List<RealRepoDescriptor> allRepos = List.of(
                generateRepositoryDscriptor(LocalRepoDescriptor.class, "localRepo_1"),
                generateRepositoryDscriptor(LocalRepoDescriptor.class, "repoToBeFilteredIn_1"),
                generateRepositoryDscriptor(RemoteRepoDescriptor.class, "repoToBeFilteredIn_2"),
                generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo_4"),
                generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo_5"),
                generateRepositoryDscriptor(FederatedRepoDescriptor.class, "repoToBeFilteredIn_3")
        );
        when(repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(any())).thenReturn(allRepos);
        setMockingOfRequestParams(false, false, false, "FilteredIn", "4");

        List<RepoKeyType> repos = new ArrayList<>();
        String continueState = "0";
        executeAndAssertPaging(continueState, repos);

        assertEquals(repos.size(), 3);
        List<String> names = repos.stream().map(RepoKeyType::getRepoKey).toList();
        assertTrue(names.containsAll(Arrays.asList(
                "repoToBeFilteredIn_1", "repoToBeFilteredIn_2", "repoToBeFilteredIn_3")));
    }

    private LocalConfig generateRepository(Class<? extends RealRepoConfig> clazz, String repoKey) {
        if (LocalCacheRepoConfig.class.equals(clazz) || RemoteRepoConfig.class.equals(clazz)) {
            return LocalCacheRepoConfig.create(RemoteRepoConfig.create(repoKey, RepoType.Generic));
        } else if (FederatedRepoConfig.class.equals(clazz)) {
            return FederatedRepoConfig.create(repoKey, RepoType.Generic);
        } else {
            return LocalRepoConfig.create(repoKey, RepoType.Generic);
        }
    }

    private RealRepoDescriptor generateRepositoryDscriptor(Class<? extends RealRepoDescriptor> clazz, String repoKey) {
        if (RemoteRepoDescriptor.class.equals(clazz)) {
            RemoteRepoDescriptor repo = new HttpRepoDescriptor();
            repo.setKey(repoKey);
            return repo;
        } else if (FederatedRepoDescriptor.class.equals(clazz)) {
            FederatedRepoDescriptor repo = new FederatedRepoDescriptor();
            repo.setKey(repoKey);
            return repo;
        } else {
            LocalRepoDescriptor repo = new LocalRepoDescriptor();
            repo.setKey(repoKey);
            return repo;
        }
    }

    private void setMockingOfRequestParams(boolean isAnyLocalRepo, boolean isAnyRemoteRepo, boolean isAnyFederatedRepo,
            String repoSearchStr, String limit) {
        when(request.getQueryParamByKey("isAnyLocalRepo")).thenReturn(String.valueOf(isAnyLocalRepo));
        when(request.getQueryParamByKey("isAnyRemoteRepo")).thenReturn(String.valueOf(isAnyRemoteRepo));
        when(request.getQueryParamByKey("isAnyFederatedRepo")).thenReturn(String.valueOf(isAnyFederatedRepo));
        when(request.getQueryParamByKey("repoSearchStr")).thenReturn(repoSearchStr);
        when(request.getQueryParamByKey("limit")).thenReturn(limit);
    }

    private void executeAndAssertPaging(String continueState, List<RepoKeyType> repos) {
        long totalRequiredRepos = 0;

        when(request.getQueryParamByKey("continueState")).thenReturn(continueState);
        while (continueState != null) {
            service.execute(request, response);

            assertEquals(response.buildResponse().getStatus(), SC_OK);
            assertTrue(response.getIModel() instanceof ContinueResult);

            ContinueResult<RepoKeyType> continueResult = (ContinueResult<RepoKeyType>) response.getIModel();
            Long currentExecutionTotalRequiredRepos = continueResult.getTotalRequiredDataAmount();
            if ("0".equals(continueState) || "".equals(continueState)) {
                assertNotNull(currentExecutionTotalRequiredRepos);
                totalRequiredRepos = currentExecutionTotalRequiredRepos;
            } else {
                assertNull(currentExecutionTotalRequiredRepos);
            }

            continueState = continueResult.getContinueState();
            when(request.getQueryParamByKey("continueState")).thenReturn(continueState);
            repos.addAll(continueResult.getData());
        }

        assertEquals(totalRequiredRepos, repos.size());
    }

    @Test
    void testPagingForAllReposWithBlankContinueState() {
        List<RealRepoDescriptor> allRepos = List.of(
                generateRepositoryDscriptor(LocalRepoDescriptor.class, "localRepo"),
                generateRepositoryDscriptor(RemoteRepoDescriptor.class, "remoteRepo"),
                generateRepositoryDscriptor(FederatedRepoDescriptor.class, "fedRepo")
        );
        when(repoService.getLocalFederatedAndRemoteRepoDescriptorsByProject(any())).thenReturn(allRepos);
        setMockingOfRequestParams(false, false, false, "", "2");

        List<RepoKeyType> repos = new ArrayList<>();
        String continueState = "";
        executeAndAssertPaging(continueState, repos);

        assertEquals(repos.size(), 3);
        List<String> names = repos.stream().map(RepoKeyType::getRepoKey).toList();
        assertTrue(names.containsAll(Arrays.asList("localRepo", "remoteRepo", "fedRepo")));
    }
}
