package org.artifactory.rest.resource.federation;

import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.rest.constant.HaRestConstants;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.federation.FederatedRemoteMembersUpdateException;
import org.artifactory.federation.FederatedRepoService;
import org.artifactory.repo.RepoDetails;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.filter.AllowedPermissionScope;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;
import org.springframework.web.bind.annotation.RequestBody;

import javax.annotation.security.RolesAllowed;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.Collection;
import java.util.List;

import static javax.ws.rs.core.Response.Status.CREATED;
import static javax.ws.rs.core.Response.Status.OK;
import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_OK;
import static org.artifactory.rest.resource.federation.PermissionScopes.INTERNAL_SCOPE_CONFIG_CONFLICTS;
import static org.artifactory.rest.resource.federation.PermissionScopes.INTERNAL_SCOPE_CONFIG_SYNC;
import static org.artifactory.rest.resource.federation.PermissionScopes.INTERNAL_SCOPE_FULL_SYNC;
import static org.artifactory.rest.resource.federation.PermissionScopes.INTERNAL_SCOPE_METADATA;
import static org.artifactory.rest.resource.federation.PermissionScopes.INTERNAL_SCOPE_REPOSITORY_LIST;
import static org.artifactory.rest.resource.federation.PermissionScopes.INTERNAL_SCOPE_VALIDATE_MIRROR;

/**
 * @author Alexei Vainshtein
 */

@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Path("federation")
@RolesAllowed({AuthorizationService.ROLE_ADMIN})
public class FederatedRepoResource {
    private static final Logger log = LoggerFactory.getLogger(FederatedRepoResource.class);

    @Context
    HttpServletRequest httpRequest;

    private FederatedRepoService federatedRepoService;

    @Autowired
    public void setService(FederatedRepoService federatedRepoService) {
        this.federatedRepoService = federatedRepoService;
    }

    @GET
    @Produces({MediaType.APPLICATION_JSON})
    @Path("validate/{repoKey}")
    @AllowedPermissionScope(INTERNAL_SCOPE_VALIDATE_MIRROR)
    public Response validate(
            @PathParam("repoKey") String repoKey,
            @QueryParam("time") long time,
            @QueryParam("source_repo_type") String repoType,
            @QueryParam("source_repo_key") String sourceRepoKey
    ) {
        try {
            federatedRepoService.validate(
                    repoKey,
                    httpRequest.getHeader(ArtifactoryRequest.ORIGINAL_CLIENT_URI),
                    time,
                    repoType,
                    sourceRepoKey
            );
            return Response.status(SC_OK).build();
        } catch (BadRequestException e) {
            return Response.status(SC_BAD_REQUEST)
                    .entity("Validation of federated repository failed: " + e.getMessage())
                    .build();
        }
    }

    @GET
    @Produces({MediaType.APPLICATION_JSON})
    @Path("repos")
    @AllowedPermissionScope(INTERNAL_SCOPE_REPOSITORY_LIST)
    public Response repos() {
        // Will return all the federation repos:
        try {
            List<RepoDetails> repos = federatedRepoService.getRepos();
            return Response.status(SC_OK).entity(repos).build();
        } catch (BadRequestException e) {
            return Response.status(SC_BAD_REQUEST)
                    .entity("Retrieval of federated repositories failed: " + e.getMessage())
                    .build();
        }
    }

    @POST
    @Path("migrate/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    public Response migrate(@PathParam("repoKey") String repoKey) {
        try {
            federatedRepoService.migrate(repoKey);
        } catch (BadRequestException e) {
            String errorMessage = String.format("Failed to migrate repository %s due to: %s", repoKey, e.getMessage());
            log.error(errorMessage, repoKey, e.getMessage());
            log.debug("Failed to migrate repository due to:", e);
            return Response.status(SC_BAD_REQUEST).entity(errorMessage).build();
        }
        return Response.status(SC_OK).entity("Migration finished successfully").build();
    }

    @POST
    @Path("fullSync/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    @AllowedPermissionScope(INTERNAL_SCOPE_FULL_SYNC)
    public Response fullSync(@PathParam("repoKey") String repoKey, @QueryParam("mirror") String mirrorUrl) {
        try {
            federatedRepoService.triggerFullSync(repoKey, mirrorUrl);
        } catch (BadRequestException e) {
            String errorMessage =
                    String.format("Failed trigger full sync for repository %s due to: %s", repoKey, e.getMessage());
            log.error(errorMessage, repoKey, e.getMessage());
            log.debug("Failed trigger full sync for repository due to:", e);
            return Response.status(SC_BAD_REQUEST).entity(errorMessage).build();
        }
        return Response.status(SC_OK).entity("Full sync triggered successfully").build();
    }

    @POST
    @Path("configSync/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    @AllowedPermissionScope(INTERNAL_SCOPE_CONFIG_SYNC)
    public Response configSync(@PathParam("repoKey") String repoKey) {
        try {
            federatedRepoService.configSync(repoKey);
        } catch (BadRequestException e) {
            String errorMessage =
                    String.format("Failed to execute configuration sync for repository %s due to: %s", repoKey,
                            e.getMessage());
            log.error(errorMessage, repoKey, e.getMessage());
            log.debug("Failed to execute configuration sync for repository due to:", e);
            return Response.status(SC_BAD_REQUEST).entity(errorMessage).build();
        } catch (FederatedRemoteMembersUpdateException e) {
            log.error("Failed to execute configuration sync for repository due to: {}", e.getMessage());
            log.debug("Failed to execute configuration sync for repository due to: ", e);
            return Response.status(SC_BAD_REQUEST).entity(e.getMessage()).build();
        }
        return Response.status(SC_OK).entity("Configuration sync executed successfully").build();
    }
    @GET
    @Path("conflicts/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    @AllowedPermissionScope(INTERNAL_SCOPE_CONFIG_CONFLICTS)
    public Response findConfigConflicts(@PathParam("repoKey") String repoKey) {
        try {
            Collection<String> conflicts = federatedRepoService.findConflicts(repoKey);
            return Response.status(OK).entity(conflicts).build();
        } catch (BadRequestException e) {
            return Response.status(SC_BAD_REQUEST).entity(e.getMessage()).build();
        } catch (Exception e) {
            String message = "Repository configuration conflict mismatch failed";
            log.error("{} for repository '{}'\n{}", message, repoKey, e.getMessage());
            return Response.serverError().entity(message).build();
        }
    }

    @GET
    @Path("ready/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    public Response isReady(@PathParam("repoKey") String repoKey) {
        try {
            return Response.status(SC_OK).entity(federatedRepoService.isActive(repoKey)).build();
        } catch (BadRequestException e) {
            String errorMessage =
                    String.format("Failed to execute readiness request for repository %s due to: %s", repoKey,
                            e.getMessage());
            log.error(errorMessage);
            log.debug(errorMessage, e);
            return Response.status(SC_OK).entity(false).build();
        }
    }

    @POST
    @Path("metadata")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @AllowedPermissionScope(INTERNAL_SCOPE_METADATA)
    public Response metadata(@RequestBody List<String> types) {
        try {
            String metadata = federatedRepoService.getMetadata(types);
            return Response.status(SC_OK).entity(metadata).build();
        } catch (BadRequestException e) {
            String errorMessage =
                    String.format("Failed to execute metadata request due to: %s", e.getMessage());
            log.error(errorMessage);
            log.debug(errorMessage, e);
            return Response.status(SC_BAD_REQUEST).entity(errorMessage).build();
        }
    }

    @PUT
    @Path("metadata")
    @Consumes(MediaType.APPLICATION_JSON)
    @Produces(MediaType.APPLICATION_JSON)
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    @AllowedPermissionScope(INTERNAL_SCOPE_METADATA)
    public Response updateMetadata(@QueryParam("originUrl") String mirrorUrl,
            @QueryParam("updateMembers") boolean updateMembers, @RequestBody String metadata) {
        federatedRepoService.putMirrorMetadata(metadata, mirrorUrl, updateMembers);
        return Response.status(CREATED).entity("Metadata received by the member").build();
    }

    @POST
    @Path("metadata/invalidate")
    @AllowedPermissionScope(INTERNAL_SCOPE_METADATA)
    @RolesAllowed({AuthorizationService.ROLE_ADMIN, HaRestConstants.ROLE_HA})
    public Response invalidateMetadata(@QueryParam("type") String packageType, @QueryParam("url") String memberUrl,
            @QueryParam("proxy") String proxyName) {
        log.debug("Received invalidate request for package type {} and member: {} ", packageType, memberUrl);
        federatedRepoService.invalidateMetadata(packageType, memberUrl, proxyName);
        return Response.status(OK).build();
    }

    @PUT
    @Path("negotiation/disable")
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    public Response disableNegotiation(@QueryParam("repoTypes") String repoTypes) {
        String userName = ContextHelper.get().getAuthorizationService().currentUsername();
        log.debug("Received request from user {} to disable repo of types: {}", userName, repoTypes);
        federatedRepoService.setDisabledRepoTypes(repoTypes);
        return Response.status(OK).entity(String.format("Repos of type: %s negotiation set to disabled",
                repoTypes.equals("*") ? "all types" : repoTypes)).build();
    }

    @POST
    @Path("repos/enable")
    @RolesAllowed({AuthorizationService.ROLE_ADMIN})
    public Response enableRepos(@RequestBody List<String> repoTypes) {
        String userName = ContextHelper.get().getAuthorizationService().currentUsername();
        log.debug("Received request from user {} to enable all repos of type {}.", userName, repoTypes);
        federatedRepoService.enableRepos(repoTypes);
        return Response.status(OK).entity(String.format("Repos of type %s enabled.", repoTypes)).build();
    }
}
