/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.rest.resource.system;


import org.apache.commons.lang.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.HaAddon;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.BaseUrlModel;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.config.MailServerModel;
import org.artifactory.api.config.ProxiesListModel;
import org.artifactory.api.config.ProxyModel;
import org.artifactory.api.config.RepositoryConfigService;
import org.artifactory.api.config.SecurityPlatformModel;
import org.artifactory.api.context.ContextHelper;
import org.artifactory.api.rest.constant.ConfigRestConstants;
import org.artifactory.api.rest.constant.HaRestConstants;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.common.ConstantValues;
import org.artifactory.descriptor.config.CentralConfigDescriptor;
import org.artifactory.descriptor.config.MutableCentralConfigDescriptor;
import org.artifactory.descriptor.repo.ProxyDescriptor;
import org.artifactory.exception.DocumentBuildException;
import org.artifactory.platform.PlatformConfigService;
import org.artifactory.rest.exception.BadRequestException;
import org.artifactory.rest.exception.NotFoundException;
import org.artifactory.rest.resource.BaseResource;
import org.artifactory.rest.util.PATCH;
import org.artifactory.security.access.AccessService;
import org.artifactory.util.AolUtils;
import org.artifactory.util.HttpUtils;
import org.artifactory.util.UrlValidator;
import org.jfrog.common.YamlUtils;
import org.jfrog.common.config.diff.DataDiff;
import org.jfrog.common.config.diff.MapToDiffConverter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.security.RolesAllowed;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.PUT;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.io.File;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.stream.Collectors;

import static org.artifactory.api.rest.constant.ConfigRestConstants.SERVICE_KEY;


/**
 * @author freds
 */

@RolesAllowed({AuthorizationService.ROLE_ADMIN, HaRestConstants.ROLE_HA})
public class ConfigResource {
    private static final Logger log = LoggerFactory.getLogger(ConfigResource.class);

    private final CentralConfigService centralConfigService;
    private final HttpServletRequest request;
    private final UrlValidator urlValidator;
    private final AolService aolService;
    private final AddonsManager addonsManager;
    private final RepositoryConfigService repositoryConfigService;
    private final AccessService accessService;
    private final PlatformConfigService platformConfigService;

    public ConfigResource(CentralConfigService centralConfigService, HttpServletRequest httpServletRequest,
            AolService aolService, AddonsManager addonsManager, AccessService accessService,
            RepositoryConfigService repositoryConfigService, PlatformConfigService platformConfigService) {
        this.centralConfigService = centralConfigService;
        this.request = httpServletRequest;
        this.aolService = aolService;
        this.addonsManager = addonsManager;
        this.repositoryConfigService = repositoryConfigService;
        this.urlValidator = new UrlValidator("http", "https");
        this.accessService = accessService;
        this.platformConfigService = platformConfigService;
    }

    @GET
    @Produces(MediaType.APPLICATION_XML)
    public CentralConfigDescriptor getConfig() {
        AolUtils.assertDashboardUser();
        return centralConfigService.getDescriptor();
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path(ConfigRestConstants.PLATFORM_URL_BASE_PATH)
    public BaseUrlModel getPlatformBaseUrl() {
        return centralConfigService.getPlatformBaseUrl();
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path(ConfigRestConstants.PLATFORM_SAAS_PATH)
    public SaasResponse isSaas() {
        SaasResponse response = new SaasResponse();
        response.setSaas(aolService.isAol());
        return response;
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path(ConfigRestConstants.MAIL_PATH)
    public MailServerModel getMailServer() {
        return new MailServerModel(centralConfigService.getDescriptor().getMailServer());
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path(ConfigRestConstants.PROXIES_PATH)
    public ProxiesListModel getPlatformProxies() {
        return new ProxiesListModel(centralConfigService.getDescriptor().getProxies());
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path(ConfigRestConstants.PROXY_PATH)
    public ProxyModel getServiceProxy(@PathParam(SERVICE_KEY) String serviceKey) {
        ProxyDescriptor proxyDescriptor = centralConfigService.getDescriptor().getProxies()
                .stream()
                .filter(proxy -> (proxy.isPlatformDefault() || proxy.getServicesList().contains(serviceKey)))
                .findFirst()
                .orElse(null);
        if (proxyDescriptor == null) {
            throw new NotFoundException();
        }
        return new ProxyModel(proxyDescriptor);
    }

    @GET
    @Produces(MediaType.APPLICATION_JSON)
    @Path(ConfigRestConstants.PLATFORM_SECURITY)
    public SecurityPlatformModel getPlatformSecurity() {
        return new SecurityPlatformModel(
                centralConfigService.getDescriptor().getSecurity().isHideUnauthorizedResources(),
                accessService.isAnonymousAccessEnabled());
    }

    @POST
    @Consumes(MediaType.APPLICATION_XML)
    @Produces(MediaType.TEXT_PLAIN)
    public String setConfig(String xmlContent) {
        isOperationAllowedForSassOrHa();
        log.debug("Received new configuration data.");
        try {
            centralConfigService.setConfigXml(xmlContent, true);
            return "Reload of new configuration succeeded";
        } catch (DocumentBuildException e) {
            // re-thrown exception not logged. Logging it by purpose.
            log.warn(e.getMessage(), e);
            throw new BadRequestException(e.getMessage(), e);
        }
    }

    @GET
    @Produces(MediaType.TEXT_PLAIN)
    @Path(ConfigRestConstants.LOGO_URL_PATH)
    public String logoUrl() {
        String logo = ContextHelper.get().beanForType(PlatformConfigService.class).getPlatformGeneralConfig().getLogo();
        if (StringUtils.isNotBlank(logo)) {
            return logo;
        }

        File logoFile = new File(ContextHelper.get().getArtifactoryHome().getLogoDir(), "logo");
        if (logoFile.exists()) {
            return HttpUtils.getServletContextUrl(request) + "/webapp/logo?" + logoFile.lastModified();
        }

        return null;
    }

    @PUT
    @Consumes({MediaType.TEXT_PLAIN})
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    @Path(ConfigRestConstants.URL_BASE_PATH)
    public Response setUrlBase(String urlBase) {
        log.debug("Updating URL base: {}", urlBase);
        final String messageFailed = "Updating URL base has failed.\n";
        final String messageOk = "URL base has been successfully updated to \"%s\".\n";

        validateUrl(urlBase);
        persistUrl(urlBase);

        if (!centralConfigService.getMutableDescriptor().getUrlBase().equals(urlBase)) {
            return Response.serverError().entity(messageFailed).build();
        }
        return Response.ok().entity(String.format(messageOk, urlBase)).build();
    }

    @PATCH
    @Consumes({BaseResource.MEDIA_TYPE_APPLICATION_YAML})
    @RolesAllowed(AuthorizationService.ROLE_ADMIN)
    public Response patchConfig(String patchYaml) {
        try {
            log.debug("Patch on current config with: {}", patchYaml);
            Map<String, Object> hashMap = YamlUtils.getInstance().readValue(patchYaml, HashMap.class);
            int platformConfigResources = platformConfigService.movePlatformResources(hashMap);
            int accessConfigResources = accessService.moveConfigResourcesToAccess(hashMap);
            int dataDiffsSize = 0;
            if (!hashMap.isEmpty()) {
                Set<DataDiff<?>> newDataDiffs = MapToDiffConverter.mapToListOfChanges(hashMap);
                dataDiffsSize = newDataDiffs.size();
                Set<DataDiff<?>> repoDiffs = extractRepoConfigs(newDataDiffs);
                if (!newDataDiffs.isEmpty()) {
                    centralConfigService.mergeAndSaveNewData(newDataDiffs);
                }
                if (!repoDiffs.isEmpty()) {
                    repositoryConfigService.repoConfigSaveNewData(repoDiffs);
                }
            }
            return Response.ok()
                    .entity(dataDiffsSize + platformConfigResources + accessConfigResources +
                            " changes to config merged successfully")
                    .build();
        } catch (IllegalArgumentException e) {
            throw new BadRequestException(e.getMessage());
        }
    }

    private Set<DataDiff<?>> extractRepoConfigs(Set<DataDiff<?>> newDataDiffs) {

        Set<DataDiff<?>> repoDiffs =
                newDataDiffs.stream().filter(e -> e.getFieldName().startsWith("localRepositories") ||
                        e.getFieldName().startsWith("remoteRepositories") ||
                        e.getFieldName().startsWith("federatedRepositories") ||
                        e.getFieldName().startsWith("virtualRepositories")).collect(Collectors.toSet());
        newDataDiffs.removeIf(e -> e.getFieldName().startsWith("localRepositories") ||
                e.getFieldName().startsWith("remoteRepositories") ||
                e.getFieldName().startsWith("federatedRepositories") ||
                e.getFieldName().startsWith("virtualRepositories"));
        return repoDiffs;
    }

    /**
     * Saves new urlBase in configuration.
     *
     * @param urlBase url to save.
     */
    private void persistUrl(String urlBase) {
        MutableCentralConfigDescriptor descriptor = centralConfigService.getMutableDescriptor();
        descriptor.setUrlBase(urlBase);
        centralConfigService.saveEditedDescriptorAndReload(descriptor);
    }

    /**
     * Validates URL using UrlValidator.
     *
     * @param urlBase url to validate.
     * @throws BadRequestException if URL is in illegal form.
     * @see org.artifactory.util.UrlValidator
     */
    private void validateUrl(String urlBase) {
        if (!StringUtils.isBlank(urlBase)) { // we allow removing custom urlBase
            try {
                urlValidator.validate(urlBase);
            } catch (UrlValidator.UrlValidationException ex) {
                throw new BadRequestException(ex.getMessage());
            }
        }
    }

    private void isOperationAllowedForSassOrHa() {
        HaAddon haAddon = addonsManager.addonByType(HaAddon.class);
        if (aolService.isAol() && !(aolService.isDashboardUser() || haAddon.isHaAuthentication() ||
                ConstantValues.allowAolConfigOverrideChanges.getBoolean())) {
            throw new BadRequestException(AolUtils.NOT_SUPPORTED_FOR_CLOUD);
        }
    }
}
