package org.artifactory.rest.resource.repositories;

import com.google.common.collect.Lists;
import com.google.gson.JsonParser;
import org.apache.http.HttpStatus;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.HttpRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.RepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.model.FederatedRepoConfig;
import org.artifactory.model.LocalRepoConfig;
import org.artifactory.model.RemoteRepoConfig;
import org.artifactory.model.VirtualRepoConfig;
import org.artifactory.repo.RepoDetails;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.spring.ArtifactoryApplicationContext;
import org.jfrog.common.JsonUtils;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.core.Response;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;

import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.RETURNS_DEEP_STUBS;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

/**
 * @author Mor Merhav
 */
class RepositoriesResourceTest {

    @Mock
    private HttpServletRequest httpRequest;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private AuthorizationService authorizationService;
    @Mock
    private ArtifactoryApplicationContext applicationContext;
    @InjectMocks
    private RepositoriesResource repositoriesResource;

    private AutoCloseable mockitoAutoCloseable;

    @BeforeEach
    void setUp() {
        mockitoAutoCloseable = MockitoAnnotations.openMocks(this);
        when(httpRequest.getHeader("x-forwarded-host")).thenReturn("localhost:8082");
        when(httpRequest.getHeader("x-forwarded-proto")).thenReturn("http");

        CentralConfigService centralConfigService = mock(CentralConfigService.class, RETURNS_DEEP_STUBS);
        when(centralConfigService.getDescriptor().getUrlBase()).thenReturn(null);
        when(applicationContext.getCentralConfig()).thenReturn(centralConfigService);

        when(httpRequest.getContextPath()).thenReturn("/artifactory");

        ArtifactoryContextThreadBinder.bind(applicationContext);
    }

    @AfterEach
    void tearDown() throws Exception {
        mockitoAutoCloseable.close();
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    void getAllRepoDetailsShouldBeSortedByTypeAndNameAlphabetically() {
        when(authorizationService.userHasPermissionsOnRepositoryRoot(any())).thenReturn(true);

        List<LocalRepoDescriptor> localDescriptors =
                List.of(
                        createLocalDescriptor("c-maven-local", RepoType.Maven),
                        createLocalDescriptor("c-generic-local", RepoType.Generic),
                        createLocalDescriptor("a-maven-local", RepoType.Maven),
                        createLocalDescriptor("a-generic-local", RepoType.Generic),
                        createLocalDescriptor("b-maven-local", RepoType.Maven),
                        createLocalDescriptor("b-generic-local", RepoType.Generic));
        List<RemoteRepoDescriptor> remoteDescriptors =
                List.of(createRemoteRepoDescriptor("c-maven-remote", RepoType.Maven),
                        createRemoteRepoDescriptor("c-generic-remote", RepoType.Generic),
                        createRemoteRepoDescriptor("a-maven-remote", RepoType.Maven),
                        createRemoteRepoDescriptor("a-generic-remote", RepoType.Generic),
                        createRemoteRepoDescriptor("b-maven-remote", RepoType.Maven),
                        createRemoteRepoDescriptor("b-generic-remote", RepoType.Generic));
        List<FederatedRepoDescriptor> federatedRepoDescriptors =
                List.of(createFederatedRepoDescriptor("c-maven-federated", RepoType.Maven),
                        createFederatedRepoDescriptor("c-generic-federated", RepoType.Generic),
                        createFederatedRepoDescriptor("a-maven-federated", RepoType.Maven),
                        createFederatedRepoDescriptor("a-generic-federated", RepoType.Generic),
                        createFederatedRepoDescriptor("b-maven-federated", RepoType.Maven),
                        createFederatedRepoDescriptor("b-generic-federated", RepoType.Generic));
        List<VirtualRepoDescriptor> virtualRepoDescriptors =
                List.of(createVirtualRepoDescriptor("c-maven-virtual", RepoType.Maven),
                        createVirtualRepoDescriptor("c-generic-virtual", RepoType.Generic),
                        createVirtualRepoDescriptor("a-maven-virtual", RepoType.Maven),
                        createVirtualRepoDescriptor("a-generic-virtual", RepoType.Generic),
                        createVirtualRepoDescriptor("b-maven-virtual", RepoType.Maven),
                        createVirtualRepoDescriptor("b-generic-virtual", RepoType.Generic));

        when(repositoryService.getLocalRepoDescriptorsExcludingInfoReposByProject(any())).thenReturn(
                localDescriptors);
        when(repositoryService.getRemoteRepoDescriptorsByProject(any())).thenReturn(remoteDescriptors);
        when(repositoryService.getFederatedRepoDescriptorsByProject(any())).thenReturn(federatedRepoDescriptors);
        when(repositoryService.getVirtualRepoDescriptorsByProject(any())).thenReturn(virtualRepoDescriptors);

        Response response = repositoriesResource.getAllRepoDetails(null, null, null);

        Assertions.assertEquals(HttpStatus.SC_OK, response.getStatus());
        List<RepoDetails> actualRepoDetails = (List<RepoDetails>) response.getEntity();
        Assertions.assertNotNull(actualRepoDetails);

        List<RepoDescriptor> allRepoDescriptors = Lists.newArrayList();
        allRepoDescriptors.addAll(localDescriptors);
        allRepoDescriptors.addAll(remoteDescriptors);
        allRepoDescriptors.addAll(federatedRepoDescriptors);
        allRepoDescriptors.addAll(virtualRepoDescriptors);

        List<RepoDetails> expectedRepoDetails = allRepoDescriptors.stream().map(this::toRepoDetails)
                .sorted(Comparator.comparing(RepoDetails::getType).thenComparing(RepoDetails::getKey)).toList();

        Assertions.assertEquals(expectedRepoDetails.size(), actualRepoDetails.size());

        // For this test only repo type, package type and key ordering is important
        normalizeActualRepoDetails(actualRepoDetails);

        String expectedRepoDetailsJson = JsonUtils.getInstance().writeValueAsString(expectedRepoDetails);
        String actualRepoDetailsJson = JsonUtils.getInstance().writeValueAsString(actualRepoDetails);

        Assertions.assertEquals(JsonParser.parseString(expectedRepoDetailsJson),
                JsonParser.parseString(actualRepoDetailsJson));
    }

    private void normalizeActualRepoDetails(List<RepoDetails> actualRepoDetails) {
        for (RepoDetails repoDetail : actualRepoDetails) {
            repoDetail.setUrl(null);
        }
    }

    private LocalRepoDescriptor createLocalDescriptor(String repoKey, RepoType repoType) {
        return new LocalRepoDescriptor(LocalRepoConfig.create(repoKey, repoType), List.of(), null, List.of());
    }

    private RemoteRepoDescriptor createRemoteRepoDescriptor(String repoKey, RepoType repoType) {
        return new HttpRepoDescriptor(RemoteRepoConfig.create(repoKey, repoType), List.of(), List.of(), null,
                List.of());
    }

    private FederatedRepoDescriptor createFederatedRepoDescriptor(String repoKey, RepoType repoType) {
        return new FederatedRepoDescriptor(FederatedRepoConfig.create(repoKey, repoType), List.of(), null, List.of(),
                List.of());
    }

    private VirtualRepoDescriptor createVirtualRepoDescriptor(String repoKey, RepoType repoType) {
        return new VirtualRepoDescriptor(VirtualRepoConfig.create(repoKey, repoType), List.of(), new HashMap<>(), null,
                new HashMap<>(), List.of(), null, null);
    }

    private RepoDetails toRepoDetails(RepoDescriptor repoDescriptor) {
        RepoDetails repoDetails = new RepoDetails();
        repoDetails.setKey(repoDescriptor.getKey());
        repoDetails.setType(extractType(repoDescriptor));
        repoDetails.setDescription(repoDescriptor.getDescription());
        repoDetails.setPackageType(repoDescriptor.getType());
        if (repoDescriptor instanceof HttpRepoDescriptor httpRepoDescriptor) {
            repoDetails.setUrl(httpRepoDescriptor.getUrl());
        }
        return repoDetails;
    }

    private RepoDetailsType extractType(RepoDescriptor repoDescriptor) {
        return switch (repoDescriptor.getClass().getSimpleName()) {
            case "LocalRepoDescriptor" -> RepoDetailsType.LOCAL;
            case "HttpRepoDescriptor" -> RepoDetailsType.REMOTE;
            case "FederatedRepoDescriptor" -> RepoDetailsType.FEDERATED;
            case "VirtualRepoDescriptor" -> RepoDetailsType.VIRTUAL;
            default -> null;
        };
    }
}